'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var _$1 = require('lodash');

function _interopNamespace(e) {
    if (e && e.__esModule) return e;
    var n = Object.create(null);
    if (e) {
        Object.keys(e).forEach(function (k) {
            if (k !== 'default') {
                var d = Object.getOwnPropertyDescriptor(e, k);
                Object.defineProperty(n, k, d.get ? d : {
                    enumerable: true,
                    get: function () { return e[k]; }
                });
            }
        });
    }
    n.default = e;
    return Object.freeze(n);
}

var ___namespace = /*#__PURE__*/_interopNamespace(_$1);

var Temperaments;
(function (Temperaments) {
    Temperaments["Passive"] = "passive";
    Temperaments["Aggressive"] = "aggressive";
})(Temperaments || (Temperaments = {}));
const friends = [];
const neutral = [];
const avoid = [];
const kos = [];
const shards = {
    'default': {
        buffer: 3,
        friends: friends,
        neutral: neutral,
        avoid: avoid,
        kos: kos,
        authorizedSieges: [],
        maxColonies: 10,
        autoExpand: true,
        autoAttack: true,
        temperament: Temperaments.Passive,
        explorers: true,
    },
    'shard2': {
        buffer: 3,
        friends: friends,
        neutral: neutral,
        avoid: avoid,
        kos: kos,
        authorizedSieges: [],
        maxColonies: 11,
        autoExpand: true,
        autoAttack: true,
        temperament: Temperaments.Passive,
        explorers: true,
    },
    'shard3': {
        buffer: 3,
        friends: friends,
        neutral: neutral,
        avoid: avoid,
        kos: kos,
        authorizedSieges: [],
        maxColonies: 7,
        autoExpand: false,
        autoAttack: true,
        temperament: Temperaments.Passive,
        explorers: false,
    },
};

class Metrics {
    constructor(labels = {}) {
        this.defaultLabels = {};
        this.metricsMap = {};
        this.defaultLabels = labels;
    }
    getKey(key, labels = {}) {
        return `${key}_${Object.values(labels).sort().join(',')}`;
    }
    gauge(key, value, labels = {}) {
        this.metricsMap[this.getKey(key, labels)] = {
            key,
            value,
            type: 'gauge',
            labels: Object.assign(labels, this.defaultLabels),
            created: Game.time,
            updated: Game.time,
        };
    }
    counter(key, amount, labels = {}) {
        let metric = this.metricsMap[this.getKey(key, labels)];
        if (!metric) {
            metric = {
                key,
                value: 0,
                type: 'counter',
                labels: Object.assign({}, labels, this.defaultLabels),
                created: Game.time,
                updated: Game.time,
            };
        }
        metric.value += amount;
        metric.updated = Game.time;
        this.metricsMap[this.getKey(key, labels)] = metric;
    }
    histogram(_key, _value, _labels_ = {}) {
    }
    summary(_key, _value, _labels = {}) {
    }
    // write the metrics to Memory
    write() {
        const fresh = Object.values(this.metricsMap).filter((metric) => {
            return metric.updated + 5 > Game.time;
        });
        Memory.metrics = fresh;
    }
}
let accMetricsActive = false;
let accMetrics = [];
const setActive = () => {
    accMetricsActive = true;
};
const setInactive = () => {
    accMetricsActive = false;
    reset();
};
const addTraceMetrics = (trace) => {
    accMetrics = accMetrics.concat(trace.getMetrics());
};
const reset = () => {
    accMetrics = [];
};
const reportMetrics = () => {
    if (!accMetricsActive) {
        return;
    }
    const summary = _.reduce(accMetrics, (acc, timing) => {
        const rollup = acc.get(timing.key) || {
            key: timing.key,
            total: 0,
            count: 0,
            max: 0,
        };
        rollup.count++;
        rollup.total += timing.value;
        if (timing.value > rollup.max) {
            rollup.max = timing.value;
        }
        acc.set(timing.key, rollup);
        return acc;
    }, new Map());
    const rollups = Array.from(summary.values());
    let summaryArray = _.reduce(rollups, (result, metric) => {
        result.push(metric);
        return result;
    }, []);
    summaryArray = _.sortBy(summaryArray, (metric) => {
        return metric.total;
        // return metric.total / metric.count;
    });
    console.log('------- CPU Usage report --------');
    console.log('= Time Avg, Count, Total time, Max Time');
    // slice to 75 so that we don't overflow the console in the game
    summaryArray.reverse().slice(0, 75).forEach((metric) => {
        console.log(`* ${(metric.total / metric.count).toFixed(2)}, ${metric.count.toFixed(0)}, ` +
            `${metric.total.toFixed(2)}, ${metric.max.toFixed(2)} - ${metric.key}`);
    });
};

class Tracer {
    constructor(name, kv, metricsCollector) {
        this.name = name;
        this.kv = kv;
        this.metricsCollector = metricsCollector;
        this.start = 0;
        this.children = [];
        this.logFilter = null;
        this.logColors = false;
        this.collect = false;
        this.collectFilter = null;
        this.collectMin = 0.5;
        this.metrics = [];
    }
    as(name) {
        const trace = this.clone();
        trace.name = `${this.name}.${name}`;
        return trace;
    }
    withFields(fields) {
        const child = this.clone();
        for (const [key, value] of fields.entries()) {
            child.kv.set(key, value);
        }
        return child;
    }
    getMetricsCollector() {
        return this.metricsCollector;
    }
    /**
     * @deprecated Use `info(...)` instead.
     */
    log(message, details = {}) {
        this.info(message, details);
    }
    info(message, details = {}) {
        if (!this.shouldLog()) {
            return;
        }
        if (this.logColors) {
            console.log(`<font color="#3FBF7F">[INFO]`, this.name, '::', message, JSON.stringify(details), JSON.stringify(this.kv), '</font>');
        }
        else {
            console.log(`[INFO]`, this.name, '::', message, JSON.stringify(details), JSON.stringify(this.kv));
        }
    }
    notice(message, details = {}) {
        if (this.logColors) {
            console.log(`<font color="#2B7FD3">[NOTICE]`, this.name, '::', message, JSON.stringify(details), JSON.stringify(this.kv), '</font>');
        }
        else {
            console.log(`[NOTICE]`, this.name, '::', message, JSON.stringify(details), JSON.stringify(this.kv));
        }
    }
    warn(message, details = {}) {
        if (this.logColors) {
            console.log(`<font color="#ffbb00">[WARN]`, this.name, '::', message, JSON.stringify(details), JSON.stringify(this.kv), '</font>');
        }
        else {
            console.log(`[WARN]`, this.name, '::', message, JSON.stringify(details), JSON.stringify(this.kv));
        }
    }
    error(message, details = {}) {
        if (this.logColors) {
            console.log(`<font color="#FF5555">[ERROR]`, this.name, '::', message, JSON.stringify(details), JSON.stringify(this.kv), '</font>');
        }
        else {
            console.log(`[ERROR]`, this.name, '::', message, JSON.stringify(details), JSON.stringify(this.kv));
        }
    }
    startTimer(metric) {
        const start = Game.cpu.getUsed();
        return () => {
            const stop = Game.cpu.getUsed();
            const cpuTime = stop - start;
            if (this.shouldCollectMetrics(cpuTime)) {
                this.pushMetric(cpuTime);
            }
            if (this.shouldLog()) {
                this.writeLog(metric, cpuTime);
            }
            return cpuTime;
        };
    }
    setLogFilter(filter) {
        this.logFilter = filter;
    }
    setLogColors(on) {
        this.logColors = on;
    }
    setCollectMetrics(active) {
        this.collect = active;
    }
    setMetricFilter(filter) {
        this.collectFilter = filter;
    }
    setMetricMin(min) {
        this.collectMin = min;
    }
    outputMetrics() {
        ___namespace.sortBy(this.getMetrics(), 'start').forEach((metric) => {
            console.log(`${metric.value.toFixed(2).padStart(5, ' ')}ms: ${metric.key} at ${metric.updated}`, JSON.stringify(metric.labels));
        });
    }
    getMetrics() {
        let metrics = [].concat(this.metrics);
        this.children.forEach((child) => {
            metrics = metrics.concat(child.getMetrics());
        });
        return metrics;
    }
    clone() {
        const child = new Tracer(this.name, this.kv, this.metricsCollector);
        child.kv = new Map(this.kv.entries());
        child.logFilter = this.logFilter;
        child.logColors = this.logColors;
        child.collect = this.collect;
        child.collectFilter = this.collectFilter;
        child.collectMin = this.collectMin;
        this.children.push(child);
        return child;
    }
    shouldCollectMetrics(cpuTime) {
        return this.collect && (!this.collectFilter || this.name.startsWith(this.collectFilter)) &&
            (!this.collectMin || (this.collectMin > 0 && this.collectMin < cpuTime));
    }
    shouldLog() {
        return this.logFilter === this.kv.get('pid');
    }
    pushMetric(cpuTime) {
        const labels = {};
        this.kv.forEach((value, key) => (labels[key] = value));
        const item = {
            key: this.name, type: 'gauge', value: cpuTime, labels: labels,
            created: Game.time, updated: Game.time
        };
        this.metrics.push(item);
    }
    writeLog(metric, cpuTime) {
        console.log(`${cpuTime.toFixed(2).padStart(5, ' ')}ms: ${this.name} ${metric || ''} at ${this.start}`, JSON.stringify(this.kv));
    }
    /**
     * @deprecated The method is being replaced with startTimer
     */
    begin(name) {
        const trace = this.clone().as(name);
        trace.start = Game.cpu.getUsed();
        return trace;
    }
    /**
     * @deprecated  The method is being replaced with startTimer
     */
    end() {
        // If tracing not active minimize the overhead of the tracer
        if (!this.start) {
            return 0;
        }
        const stop = Game.cpu.getUsed();
        const cpuTime = stop - this.start;
        if (this.shouldCollectMetrics(cpuTime)) {
            this.pushMetric(cpuTime);
        }
        if (this.shouldLog()) {
            this.writeLog(null, cpuTime);
        }
        return cpuTime;
    }
}

function newMultipliers() {
    return {
        'heal': 1,
        'attack': 1,
        'move': 1,
        'tough': 1,
        'rangedAttack': 1,
    };
}
function buildAttacker(requiredTanking, maxEnergy, multipliers, trace) {
    let healParts = 0;
    let moveParts = 0;
    let attackParts = 0;
    let toughParts = 0;
    let neededEnergy = 0;
    const healingTick = Math.floor(12 * multipliers.heal);
    const movingTick = Math.floor(2 * multipliers.move);
    // if we have energy for a healing part, include at least one healing part
    if (maxEnergy > 380) { // heal, attack, move
        requiredTanking = _.max([requiredTanking, 1]);
    }
    trace.info('buildAttacker', {
        requiredTanking,
        maxEnergy,
        multipliers,
    });
    while (true) { // eslint-disable-line no-constant-condition
        const numNonMoveParts = healParts + attackParts + toughParts;
        const numParts = numNonMoveParts + moveParts;
        // Cannot have more than 50 parts
        if (numParts >= 50) {
            // trace.info('cannot have more than 50 parts', {healParts, moveParts, attackParts, toughParts});
            break;
        }
        let ΔHealParts = 0;
        let ΔMoveParts = 0;
        let ΔAttackParts = 0;
        const ΔToughParts = 0;
        let ΔEnergy = 0;
        if (moveParts * movingTick <= numNonMoveParts) {
            ΔMoveParts += 1;
            // trace.info('adding move parts', {ΔMoveParts});
        }
        else if (healParts * healingTick < requiredTanking) {
            ΔHealParts += 1;
            // trace.info('adding heal parts', {
            //   ΔHealParts,
            //   healParts,
            //   healingTick,
            //   totalHealingTick: healParts * healingTick
            // });
        }
        else {
            ΔAttackParts += 1;
            // trace.info('adding attack parts', {ΔAttackParts});
        }
        const ΔHealPartsEnergy = ΔHealParts * BODYPART_COST[HEAL];
        const ΔMovePartsEnergy = ΔMoveParts * BODYPART_COST[MOVE];
        const ΔAttackPartsEnergy = ΔAttackParts * BODYPART_COST[ATTACK];
        const ΔToughPartsEnergy = ΔToughParts * BODYPART_COST[TOUGH];
        ΔEnergy += ΔHealPartsEnergy + ΔMovePartsEnergy + ΔAttackPartsEnergy + ΔToughPartsEnergy;
        // trace.info('pass', {ΔEnergy, ΔHealParts, ΔMoveParts, ΔAttackParts, ΔToughParts});
        // If the next pass requires too much energy, we are done
        if (neededEnergy + ΔEnergy > maxEnergy) {
            // trace.info('pass requires too much energy', {ΔEnergy});
            break;
        }
        healParts += ΔHealParts;
        attackParts += ΔAttackParts;
        moveParts += ΔMoveParts;
        toughParts += ΔToughParts;
        neededEnergy += ΔEnergy;
    }
    let numNonMoveParts = healParts + attackParts + toughParts;
    let numParts = numNonMoveParts + moveParts;
    let availableMove = moveParts * movingTick - numNonMoveParts;
    while (numParts + availableMove < 50 && availableMove > 0 && maxEnergy >= neededEnergy + BODYPART_COST[TOUGH]) {
        // trace.info('adding tough parts', {numParts, availableMove, neededEnergy, maxEnergy});
        toughParts += 1;
        neededEnergy += BODYPART_COST[TOUGH];
        numNonMoveParts = healParts + attackParts + toughParts;
        numParts = numNonMoveParts + moveParts;
        availableMove = moveParts * movingTick - numNonMoveParts;
    }
    let ok = false;
    if (healParts * healingTick >= requiredTanking) {
        ok = true;
    }
    trace.info('final parts', { ok, healParts, moveParts, attackParts, toughParts, neededEnergy, maxEnergy });
    const body = [];
    let partsCount = 0;
    // Add heal and move (2:1) parts to front
    while (healParts) {
        if (moveParts && partsCount % movingTick === 0) {
            body.unshift(MOVE);
            moveParts--;
            // partsCount++;
        }
        body.unshift(HEAL);
        healParts--;
        partsCount++;
    }
    // Add attack and move parts (2:1) to front
    while (attackParts) {
        if (moveParts && partsCount % movingTick === 0) {
            body.unshift(MOVE);
            moveParts--;
            // partsCount++;
        }
        body.unshift(ATTACK);
        attackParts--;
        partsCount++;
    }
    // Add remaining move parts to front
    while (moveParts) {
        body.unshift(MOVE);
        moveParts--;
        // partsCount++;
    }
    // Add tough parts to front
    while (toughParts) {
        body.unshift(TOUGH);
        toughParts--;
        partsCount++;
    }
    return [body, ok];
}

function buildDefender(requiredTanking, maxEnergy, multipliers, trace) {
    let healParts = 0;
    let moveParts = 0;
    let attackParts = 0;
    let rangedAttackParts = 0;
    let toughParts = 0;
    let neededEnergy = 0;
    const healingTick = Math.floor(12 * multipliers.heal);
    const movingTick = Math.floor(2 * multipliers.move);
    // if we have energy for a healing part, include at least one healing part
    if (maxEnergy > 380) { // heal, attack, move
        requiredTanking = _.max([requiredTanking, 1]);
    }
    trace.info('buildAttacker', {
        requiredTanking,
        maxEnergy,
        multipliers,
    });
    while (true) { // eslint-disable-line no-constant-condition
        const numNonMoveParts = healParts + attackParts + toughParts + rangedAttackParts;
        const numParts = numNonMoveParts + moveParts;
        // Cannot have more than 50 parts
        if (numParts >= 50) {
            // trace.info('cannot have more than 50 parts', {healParts, moveParts, attackParts, toughParts});
            break;
        }
        let ΔHealParts = 0;
        let ΔMoveParts = 0;
        const ΔAttackParts = 0;
        let ΔRangedAttackParts = 0;
        const ΔToughParts = 0;
        let ΔEnergy = 0;
        if (moveParts * movingTick <= numNonMoveParts) {
            ΔMoveParts += 1;
            // trace.info('adding move parts', {ΔMoveParts});
        }
        else if (healParts * healingTick < requiredTanking) {
            ΔHealParts += 1;
            // trace.info('adding heal parts', {
            //   ΔHealParts,
            //   healParts,
            //   healingTick,
            //   totalHealingTick: healParts * healingTick
            // });
        }
        else {
            // TODO generalize builder to work with attack and defenders, this is the only
            // part of this that actually varies
            ΔRangedAttackParts += 1;
            // trace.info('adding ranged attack parts', {ΔRangedAttackParts});
        }
        const ΔHealPartsEnergy = ΔHealParts * BODYPART_COST[HEAL];
        const ΔMovePartsEnergy = ΔMoveParts * BODYPART_COST[MOVE];
        const ΔAttackPartsEnergy = ΔAttackParts * BODYPART_COST[ATTACK];
        const ΔRangedAttackPartsEnergy = ΔRangedAttackParts * BODYPART_COST[RANGED_ATTACK];
        const ΔToughPartsEnergy = ΔToughParts * BODYPART_COST[TOUGH];
        ΔEnergy += ΔHealPartsEnergy + ΔMovePartsEnergy + ΔAttackPartsEnergy + ΔRangedAttackPartsEnergy +
            ΔToughPartsEnergy;
        // trace.info('pass', {ΔEnergy, ΔHealParts, ΔMoveParts, ΔAttackParts, ΔToughParts});
        // If the next pass requires too much energy, we are done
        if (neededEnergy + ΔEnergy > maxEnergy) {
            // trace.info('pass requires too much energy', {ΔEnergy});
            break;
        }
        healParts += ΔHealParts;
        attackParts += ΔAttackParts;
        rangedAttackParts += ΔRangedAttackParts;
        moveParts += ΔMoveParts;
        toughParts += ΔToughParts;
        neededEnergy += ΔEnergy;
    }
    let numNonMoveParts = healParts + attackParts + toughParts;
    let numParts = numNonMoveParts + moveParts;
    let availableMove = moveParts * movingTick - numNonMoveParts;
    while (numParts + availableMove < 50 && availableMove > 0 && maxEnergy >= neededEnergy + BODYPART_COST[TOUGH]) {
        // trace.info('adding tough parts', {numParts, availableMove, neededEnergy, maxEnergy});
        toughParts += 1;
        neededEnergy += BODYPART_COST[TOUGH];
        numNonMoveParts = healParts + attackParts + toughParts;
        numParts = numNonMoveParts + moveParts;
        availableMove = moveParts * movingTick - numNonMoveParts;
    }
    let ok = false;
    if (healParts * healingTick >= requiredTanking && rangedAttackParts >= 1) {
        ok = true;
    }
    trace.info('final parts', { ok, healParts, moveParts, attackParts, toughParts, neededEnergy, maxEnergy });
    const body = [];
    let partsCount = 0;
    // Add heal and move (2:1) parts to front
    while (healParts) {
        if (moveParts && partsCount % movingTick === 0) {
            body.unshift(MOVE);
            moveParts--;
            // partsCount++;
        }
        body.unshift(HEAL);
        healParts--;
        partsCount++;
    }
    // Add ranged attack and move parts (2:1) to front
    while (rangedAttackParts) {
        if (moveParts && partsCount % movingTick === 0) {
            body.unshift(MOVE);
            moveParts--;
            // partsCount++;
        }
        body.unshift(RANGED_ATTACK);
        rangedAttackParts--;
        partsCount++;
    }
    // Add attack and move parts (2:1) to front
    while (attackParts) {
        if (moveParts && partsCount % movingTick === 0) {
            body.unshift(MOVE);
            moveParts--;
            // partsCount++;
        }
        body.unshift(ATTACK);
        attackParts--;
        partsCount++;
    }
    // Add remaining move parts to front
    while (moveParts) {
        body.unshift(MOVE);
        moveParts--;
        // partsCount++;
    }
    // Add tough parts to front
    while (toughParts) {
        body.unshift(TOUGH);
        toughParts--;
        partsCount++;
    }
    return [body, ok];
}

// General role memory IDs
const MEMORY_ROLE = 'role';
// Old source memory ID (deprecated)
const MEMORY_SOURCE = 'source';
const MEMORY_SOURCE_POSITION = 'source_position';
const MEMORY_SOURCE_CONTAINER = 'source_container';
// General movement memory IDs (use this for nearly all movement)
const MEMORY_DESTINATION = 'destination';
const MEMORY_DESTINATION_ROOM = 'destination_room';
const MEMORY_DESTINATION_SHARD = 'destination_shard';
const MEMORY_DESTINATION_POS = 'destination_pos';
// Used for debugging, don't use for decision making, use MEMORY_BASE instead
const MEMORY_ORIGIN_SHARD = 'shard';
const MEMORY_ORIGIN = 'origin';
const MEMORY_BASE = 'base';
const MEMORY_START_TICK = 'start_tick';
const MEMORY_COMMUTE_DURATION = 'commute_duration';
// Assign
const MEMORY_ASSIGN_SHARD = 'assignment_shard';
const MEMORY_ASSIGN_ROOM = 'assignment_room';
const MEMORY_ASSIGN_ROOM_POS = 'assignment_room_pos';
const MEMORY_TASK_TYPE = 'task_type';
const TASK_ID = 'task_id';
const TASK_PHASE = 'task_phase';
const TASK_TTL$2 = 'task_ttl';
// haul task
const MEMORY_HAUL_PICKUP = 'haul_pickup';
const MEMORY_HAUL_RESOURCE = 'haul_resource';
const MEMORY_HAUL_AMOUNT = 'haul_amount';
const MEMORY_HAUL_DROPOFF = 'haul_dropoff';
const REACTOR_TTL = 'reactor_ttl';
// terminal
const TERMINAL_TASK = 'terminal_task';
const TERMINAL_TASK_TYPE = 'terminal_task_type';
// buy/sell task
const MEMORY_ORDER_TYPE = 'order_type';
const MEMORY_ORDER_RESOURCE = 'order_resource';
const MEMORY_ORDER_AMOUNT = 'order_amount';
// transfer task
const TRANSFER_BASE = 'transfer_base';
const TRANSFER_ROOM = 'transfer_room';
const TRANSFER_RESOURCE = 'transfer_resource';
const TRANSFER_AMOUNT = 'transfer_amount';
// boosts
const PREPARE_BOOSTS = 'prepare_boosts';
const DESIRED_BOOSTS = 'desired_boosts';
// Attacker
const MEMORY_ATTACK = 'attack';
const MEMORY_HEAL = 'heal';
const MEMORY_POSITION_X = 'position_x';
const MEMORY_POSITION_Y = 'position_y';
const MEMORY_POSITION_ROOM = 'position_room';
// Reaction status
const REACTION_STATUS_ROOM = 'reaction_room';
const REACTION_STATUS_LAB = 'reaction_lab';
const REACTION_STATUS_RESOURCE = 'reaction_resource';
const REACTION_STATUS_RESOURCE_AMOUNT = 'reaction_resource_amount';
const REACTION_STATUS_PHASE = 'reaction_status_phase';
// Room status
const BASE_STATUS_NAME = 'room_status_name';
const BASE_STATUS_LEVEL = 'room_status_level';
const BASE_STATUS_LEVEL_COMPLETED = 'room_status_level_completed';
const BASE_STATUS_ENERGY = 'room_status_energy';
const BASE_STATUS_ALERT_LEVEL = 'room_status_alert_level';
const BASE_STATUS_TERMINAL = 'room_status_terminal';
// PID Controller suffixes
const PID_SUFFIX_P = 'pid_p';
const PID_SUFFIX_I = 'pid_i';
const PID_SUFFIX_D = 'pid_d';
const PID_SUFFIX_INTEGRAL = 'pid_integral';
const PID_SUFFIX_TIME = 'pid_time';
const PID_SUFFIX_ERROR = 'pid_error';
const PID_SUFFIX_SETPOINT = 'pid_setpoint';
// Reactor tasks
const REACTOR_TASK_TYPE = 'reactor_task_type';
const REACTOR_INPUT_A = 'reactor_input_a';
const REACTOR_INPUT_B = 'reactor_input_b';
const REACTOR_OUTPUT = 'reactor_output';
const REACTOR_AMOUNT = 'reactor_amount';
// Defense parties
const MEMORY_DEFENSE_IN_POSITION = 'defense_in_position';
// Party
const MEMORY_PARTY_ID = 'party_id';
// Idle at parking lot
const MEMORY_IDLE = 'idle';

const PER_LEVEL_ENERGY = 100000;
var AlertLevel;
(function (AlertLevel) {
    AlertLevel["GREEN"] = "green";
    AlertLevel["YELLOW"] = "yellow";
    AlertLevel["RED"] = "red";
})(AlertLevel || (AlertLevel = {}));
function getBasePrimaryRoom(base) {
    return Game.rooms[base.primary];
}
function getCreepBase(kernel, creep) {
    const baseId = creep.memory[MEMORY_BASE];
    if (!baseId) {
        return null;
    }
    const base = kernel.getPlanner().getBaseById(baseId);
    if (!base) {
        return null;
    }
    return base;
}
function setParking(base, layout, origin) {
    base.parking = new RoomPosition(layout.parking.x + origin.x, layout.parking.y + origin.y, origin.roomName);
}
function getBoostPosition(base) {
    return base.boostPosition;
}
function setBoostPosition(base, pos) {
    base.boostPosition = pos;
}
function getLabsForAction(base, action) {
    if (!base.boosts.has(action)) {
        return [];
    }
    return base.boosts.get(action);
}
function setLabsByAction(base, labsByAction) {
    base.boosts = labsByAction;
}
function getStoredResources(base) {
    return getStorageStructures(base).reduce((acc, structure) => {
        Object.keys(structure.store).forEach((resource) => {
            const current = acc.get(resource) || 0;
            acc.set(resource, structure.store.getUsedCapacity(resource) + current);
        });
        return acc;
    }, new Map());
}
function getStoredResourceAmount(base, resource) {
    const resources = getStoredResources(base);
    return resources.get(resource) || 0;
}
function getEnergyFullness(base) {
    const structures = getStorageStructures(base);
    if (!structures.length) {
        return 0;
    }
    const stores = structures.reduce((acc, structure) => {
        acc.capacity += structure.store.getCapacity(RESOURCE_ENERGY);
        acc.used += structure.store.getUsedCapacity(RESOURCE_ENERGY);
        return acc;
    }, { capacity: 0, used: 0 });
    if (!stores.capacity) {
        return 0;
    }
    return stores.used / stores.capacity;
}
function getReserveBuffer(base) {
    var _a;
    const room = getBasePrimaryRoom(base);
    if (!room) {
        return 0;
    }
    if (!((_a = room.controller) === null || _a === void 0 ? void 0 : _a.my)) {
        return 0;
    }
    const roomLevel = getBaseLevel(base);
    if (roomLevel < 4) {
        return 2000;
    }
    return (roomLevel - 3) * PER_LEVEL_ENERGY;
}
function getStorageStructures(base) {
    var _a, _b;
    const structures = [];
    const room = getBasePrimaryRoom(base);
    if (!room) {
        return structures;
    }
    if ((_a = room.storage) === null || _a === void 0 ? void 0 : _a.isActive()) {
        structures.push(room.storage);
    }
    if ((_b = room.terminal) === null || _b === void 0 ? void 0 : _b.isActive()) {
        structures.push(room.terminal);
    }
    return structures;
}
function getStructureWithResource(base, resource) {
    const structures = getStructuresWithResource(base, resource);
    if (!structures.length) {
        return null;
    }
    return structures[0];
}
function getStructureForResource(base, resource) {
    const structures = getStructuresForResource(base, resource);
    if (!structures.length) {
        return null;
    }
    return structures[0];
}
function getStructuresWithResource(base, resource) {
    var _a, _b;
    const structures = [];
    const room = getBasePrimaryRoom(base);
    if (!room) {
        return null;
    }
    if (((_a = room.storage) === null || _a === void 0 ? void 0 : _a.store.getUsedCapacity(resource)) > 0) {
        structures.push(room.storage);
    }
    if (((_b = room.terminal) === null || _b === void 0 ? void 0 : _b.store.getUsedCapacity(resource)) > 0) {
        structures.push(room.terminal);
    }
    return structures;
}
function getStructuresForResource(base, resource) {
    var _a, _b;
    const structures = [];
    const room = getBasePrimaryRoom(base);
    if (!room) {
        return null;
    }
    if (((_a = room.storage) === null || _a === void 0 ? void 0 : _a.store.getFreeCapacity(resource)) > 0) {
        structures.push(room.storage);
    }
    if (((_b = room.terminal) === null || _b === void 0 ? void 0 : _b.store.getFreeCapacity(resource)) > 0) {
        structures.push(room.terminal);
    }
    return structures;
}
function getBaseLevel(base) {
    var _a;
    let level = 0;
    const room = getBasePrimaryRoom(base);
    if (!room) {
        return level;
    }
    if ((_a = room.controller) === null || _a === void 0 ? void 0 : _a.level) {
        level = room.controller.level;
    }
    return level;
}
function getBaseLevelCompleted(base) {
    var _a;
    const room = getBasePrimaryRoom(base);
    if (!room) {
        return 0;
    }
    if (!((_a = room.controller) === null || _a === void 0 ? void 0 : _a.my)) {
        return 0;
    }
    return room.controller.progress / room.controller.progressTotal;
}
function getDamagedStructures(base) {
    return base.damagedStructures.map((structureId) => {
        return Game.getObjectById(structureId);
    });
}
function getNextDamagedStructure(base) {
    const structures = getDamagedStructures(base);
    if (!structures.length) {
        return null;
    }
    return structures[0];
}
function baseEnergyStorageCapacity(base) {
    const room = getBasePrimaryRoom(base);
    if (!room) {
        return 0;
    }
    return room.energyCapacityAvailable || 0;
}
function getStoredEffects(base) {
    return base.storedEffects;
}
const threadBase = (name, ttl) => (action) => {
    let lastCall = 0;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const tick = function (trace, kernel, base, ...args) {
        if (lastCall + ttl <= Game.time) {
            lastCall = Game.time;
            const actionTrace = trace.begin(name);
            const result = action(actionTrace, kernel, base, ...args);
            actionTrace.end();
            return result;
        }
        else {
            trace.info(`thread ${name} sleeping for ${lastCall + ttl - Game.time}`);
        }
        return null;
    };
    tick.reset = () => {
        lastCall = 0;
    };
    return tick;
};
function resetRemotes(base, trace) {
    trace.info(`resetting remotes ${base.id}`);
    base.rooms = [base.primary];
}
function addRoom(base, roomName, trace) {
    trace.info('adding room', { baseId: base.id, roomName });
    if (base.rooms.indexOf(roomName) !== -1) {
        trace.error('room already exists', { roomName, baseId: base.id });
        return;
    }
    base.rooms.push(roomName);
}

/**
 * Logic for determining base remotes
 *
 * The goal is to create a dynamic list of rooms near the base that can be safely mined.
 *
 * Over a few passes, rooms adjacent to the base and rooms checked in previous passes are
 * added to a list if they are safe to mine. It's possible for a room to not be safe to mine,
 * but be passable.
 *
 * TODO: Move to remotes.ts
 */
const PASSES$1 = 3;
const findRemotes = (kernel, base, trace) => {
    trace.info('checking remote mining', { baseId: base.id });
    const scribe = kernel.getScribe();
    const candidates = new Set();
    const dismissed = new Map();
    const seen = new Set();
    // Add the primary room to the list of seen rooms
    seen.add(base.primary);
    const startRoomStatus = Game.map.getRoomStatus(base.primary).status;
    let nextPass = [base.primary];
    trace.info('starting remote selection pass', { nextPass, maxPasses: PASSES$1, startRoomStatus });
    for (let i = 0; i <= PASSES$1; i++) {
        const found = [];
        nextPass.forEach((currentRoom) => {
            const exits = Game.map.describeExits(currentRoom);
            if (!exits) {
                trace.error('no exits found', { currentRoom });
                return;
            }
            const adjacentRooms = Object.values(exits);
            adjacentRooms.forEach((adjacentRoom) => {
                // Already dismissed this room, don't check it again
                if (dismissed.has(adjacentRoom)) {
                    return;
                }
                // Already seen this room, don't check it again
                if (seen.has(adjacentRoom)) {
                    return;
                }
                seen.add(adjacentRoom);
                // Check the room, determine if passable and if room is viable for mining
                const [passable, dismiss] = checkCandidate(kernel, base, adjacentRoom, startRoomStatus, trace);
                trace.info('checked candidate room', {
                    currentRoom,
                    adjacentRoom,
                    passable,
                    dismiss,
                });
                // Room is passable, so add to list of rooms to check
                if (passable) {
                    found.push(adjacentRoom);
                }
                // Room should not be considered as a candidate for mining
                if (dismiss) {
                    dismissed.set(adjacentRoom, dismiss);
                    return;
                }
                trace.info('adding room to candidates', { adjacentRoom });
                candidates.add(adjacentRoom);
            });
        });
        nextPass = found;
    }
    if (candidates.size === 0) {
        trace.info('no candidates rooms found', { candidates: Array.from(candidates.keys()), base });
        return [[], dismissed];
    }
    trace.info('candidates rooms found', {
        candidates: Array.from(candidates.keys()),
        dismissed: Array.from(dismissed.entries()),
        base
    });
    const sortedCandidates = _.sortByOrder(Array.from(candidates.keys()), [
        (roomName) => {
            const route = Game.map.findRoute(base.primary, roomName);
            if (route === ERR_NO_PATH) {
                return 9999;
            }
            return route.length;
        },
        (roomName) => {
            const roomEntry = scribe.getRoomById(roomName);
            if (!roomEntry) {
                return 9999;
            }
            return roomEntry.numSources;
        }
    ], ['asc', 'desc']);
    trace.info('next remote mining rooms', { sortedCandidates });
    //return [sortedCandidates, debug];
    return [sortedCandidates, dismissed];
};
function checkCandidate(kernel, base, room, baseRoomStatus, trace) {
    var _a, _b, _c;
    // room should be same status as base primary
    const roomStatus = Game.map.getRoomStatus(room).status;
    if (roomStatus !== baseRoomStatus) {
        return [false, 'status mismatch'];
    }
    // filter rooms already belonging to a base
    const roomBase = kernel.getPlanner().getBaseByRoom(room);
    if (roomBase && base.id !== roomBase.id) {
        return [true, 'already assigned to a base'];
    }
    const roomEntry = kernel.getScribe().getRoomById(room);
    // filter out rooms we have not seen
    if (!roomEntry) {
        return [false, 'no entry in scribe'];
    }
    // filter out rooms that do not have a source
    if (roomEntry.numSources === 0) {
        return [true, 'no sources'];
    }
    // Filter our rooms without a controller
    if (!((_a = roomEntry.controller) === null || _a === void 0 ? void 0 : _a.pos)) {
        return [true, 'no controller'];
    }
    // Filter out rooms that are claimed by players
    if (((_b = roomEntry.controller) === null || _b === void 0 ? void 0 : _b.owner) &&
        ((_c = roomEntry.controller) === null || _c === void 0 ? void 0 : _c.owner) !== kernel.getPlanner().getUsername() &&
        roomEntry.controller.owner !== 'Invader') {
        return [false, 'is owned'];
    }
    // If enemies present that we cannot beat, do not consider this room
    const damage = _.max([roomEntry.hostilesDmg, roomEntry.keepersDmg]);
    if (damage > 0) {
        const basePrimary = getBasePrimaryRoom(base);
        if (!basePrimary) {
            return [false, 'no primary room'];
        }
        const multipliers = newMultipliers();
        const [_parts, ok] = buildDefender(damage, basePrimary.energyCapacityAvailable, multipliers, trace);
        if (!ok) {
            trace.warn('defender build failed, we cannot take the room', {
                roomEntry, basePrimary, damage,
                energyCapacityAvailable: basePrimary.energyCapacityAvailable
            });
            return [false, 'hostiles too strong'];
        }
        trace.warn('defender build succeeded, we can take the room', {
            roomEntry, basePrimary, damage,
            energyCapacityAvailable: basePrimary.energyCapacityAvailable
        });
        // TODO: When defense of remotes is sorted out, remove this
        return [false, 'hostiles present'];
    }
    return [true, ''];
}
// Calculate the max number of remotes based on level and number of spawns
// TODO collect spawner saturation metrics and use that to calculate max remotes
function desiredRemotes(base, level, spawnUtilization, trace) {
    const room = getBasePrimaryRoom(base);
    const spawns = room.find(FIND_STRUCTURES, {
        filter: (s) => s.structureType === STRUCTURE_SPAWN && s.isActive(),
    });
    // No spawns, no remotes
    if (!spawns.length) {
        return 0;
    }
    if (spawnUtilization > 0.9) {
        base.rooms.length - 2;
        trace.info(`${base.id} should drop a room`);
    }
    else if (spawnUtilization < 0.5) {
        trace.info(`${base.id} should add a room`);
    }
    else {
        trace.info(`${base.id} should keep the same number of rooms`);
    }
    let desiredRemotes = 0;
    switch (level) {
        case 0:
        case 1:
        case 2:
        case 3:
            desiredRemotes = 3;
            break;
        case 4:
            desiredRemotes = 3;
            break;
        case 5:
            desiredRemotes = 3;
            break;
        case 6:
            desiredRemotes = 2;
            break;
        case 7:
        case 8:
            if (spawns.length < 2) {
                desiredRemotes = 1;
            }
            else if (spawns.length < 3) {
                desiredRemotes = 2;
            }
            else {
                desiredRemotes = 6;
            }
            break;
        default:
            throw new Error('unexpected controller level');
    }
    if (room.storage) {
        const reserveEnergy = room.storage.store.getUsedCapacity(RESOURCE_ENERGY);
        if (reserveEnergy > 500000) {
            return _.min([desiredRemotes, 0]);
        }
        else if (reserveEnergy > 400000) {
            return _.min([desiredRemotes, 1]);
        }
    }
    return desiredRemotes;
}

const EMPTY = 'empty';
const ANY = 'any';
const buildingCodes = {
    'X': EMPTY,
    '.': ANY,
    'R': STRUCTURE_ROAD,
    'P': STRUCTURE_SPAWN,
    'E': STRUCTURE_EXTENSION,
    'C': STRUCTURE_CONTAINER,
    'T': STRUCTURE_TOWER,
    'S': STRUCTURE_STORAGE,
    'L': STRUCTURE_LINK,
    'M': STRUCTURE_TERMINAL,
    'B': STRUCTURE_LAB,
    'O': STRUCTURE_OBSERVER,
    'N': STRUCTURE_NUKER,
};
const getConstructionPosition = (pos, origin, layout) => {
    return new RoomPosition(pos.x + origin.x - layout.origin.x, pos.y + origin.y - layout.origin.y, origin.roomName);
};

const STATUS_RUNNING = 'running';
const STATUS_SLEEPING = 'sleeping';
const STATUS_TERMINATED = 'terminated';
const running = () => {
    return { status: STATUS_RUNNING };
};
const sleeping = (sleepFor) => {
    return { status: STATUS_SLEEPING, sleepFor };
};
const terminate = () => {
    return { status: STATUS_TERMINATED };
};
class Process {
    constructor(id, type, priority, runnable) {
        this.id = id;
        this.type = type;
        this.priority = priority;
        this.adjustedPriority = priority;
        this.runnable = runnable;
        this.skippable = true;
        this.status = STATUS_RUNNING;
        this.lastRun = 0;
        this.nextRun = 0;
    }
    isRunning() {
        return this.status === STATUS_RUNNING;
    }
    isSleeping() {
        return this.status === STATUS_SLEEPING;
    }
    isTerminated() {
        return this.status === STATUS_TERMINATED;
    }
    shouldWake() {
        if (this.nextRun <= Game.time) {
            return true;
        }
        return false;
    }
    setRunning() {
        this.status = STATUS_RUNNING;
        this.nextRun = 0;
    }
    setSleeping(duration) {
        this.status = STATUS_SLEEPING;
        this.nextRun = Game.time + duration;
    }
    setTerminated() {
        this.status = STATUS_TERMINATED;
    }
    setSkippable(skippable) {
        this.skippable = skippable;
    }
    canSkip() {
        return this.skippable;
    }
    skip() {
        this.adjustedPriority -= 1;
    }
    run(kernel, trace) {
        this.lastRun = Game.time;
        const result = this.runnable.run(kernel, trace);
        switch (result.status) {
            case STATUS_RUNNING:
                this.setRunning();
                break;
            case STATUS_SLEEPING:
                this.setSleeping(result.sleepFor);
                break;
            case STATUS_TERMINATED:
                this.setTerminated();
                break;
        }
        this.adjustedPriority = this.priority;
    }
}

const CONSTRUCTION_INTERVAL = 100;
const MAX_STRUCTURE_SITES = 5;
const baseLayouts = [
    {
        origin: { x: 0, y: 0 },
        parking: { x: 0, y: 0 },
        muster: [],
        buildings: [],
    },
    {
        origin: { x: 0, y: 4 },
        parking: { x: 0, y: 0 },
        muster: [
            { x: -3, y: -1, direction: TOP },
            { x: -3, y: 1, direction: BOTTOM },
            { x: 3, y: -1, direction: TOP },
            { x: 3, y: 1, direction: BOTTOM },
        ],
        buildings: [
            ['P'],
            ['.'],
            ['.'],
            ['.'],
            ['.'],
        ],
    },
    {
        origin: { x: 2, y: 5 },
        parking: { x: 3, y: 3 },
        muster: [
            { x: -4, y: -1, direction: TOP },
            { x: -4, y: 1, direction: BOTTOM },
            { x: 4, y: -1, direction: TOP },
            { x: 4, y: 1, direction: BOTTOM },
        ],
        buildings: [
            ['.', 'E', 'R', 'E', '.'],
            ['.', 'R', 'P', 'R', '.'],
            ['.', 'R', 'E', 'R', '.'],
            ['.', 'R', 'E', 'R', '.'],
            ['.', 'X', 'R', 'E', '.'],
            ['.', 'R', 'X', 'R', '.'],
            ['.', 'X', 'R', 'X', '.'],
        ],
    },
    {
        origin: { x: 2, y: 5 },
        parking: { x: 3, y: 3 },
        muster: [
            { x: -5, y: -2, direction: TOP },
            { x: -5, y: 2, direction: BOTTOM },
            { x: 5, y: -2, direction: TOP },
            { x: 5, y: 2, direction: BOTTOM },
            { x: -2, y: -5, direction: LEFT },
            { x: 2, y: -5, direction: RIGHT },
            { x: -2, y: 5, direction: LEFT },
            { x: 2, y: 5, direction: RIGHT },
        ],
        buildings: [
            ['.', 'E', 'R', 'E', '.'],
            ['E', 'R', 'P', 'R', 'E'],
            ['E', 'R', 'E', 'R', 'E'],
            ['.', 'R', 'E', 'R', '.'],
            ['.', 'T', 'R', 'E', '.'],
            ['.', 'R', 'X', 'R', '.'],
            ['.', '.', 'R', 'E', '.'],
        ],
    },
    {
        origin: { x: 3, y: 6 },
        parking: { x: 3, y: 3 },
        muster: [
            { x: -8, y: -2, direction: TOP },
            { x: -8, y: 2, direction: BOTTOM },
            { x: 8, y: -2, direction: TOP },
            { x: 8, y: 2, direction: BOTTOM },
            { x: -2, y: -8, direction: LEFT },
            { x: 2, y: -8, direction: RIGHT },
            { x: -2, y: 8, direction: LEFT },
            { x: 2, y: 8, direction: RIGHT },
        ],
        buildings: [
            ['.', '.', 'R', 'E', 'R', '.', '.', '.'],
            ['.', 'R', 'E', 'R', 'E', 'R', 'R', '.'],
            ['.', 'E', 'R', 'P', 'R', 'E', '.', 'R'],
            ['.', 'E', 'R', 'E', 'R', 'E', '.', '.'],
            ['.', 'X', 'R', 'E', 'R', 'E', '.', '.'],
            ['.', 'R', 'T', 'R', 'E', 'R', '.', 'R'],
            ['R', 'X', 'R', 'S', 'R', 'X', 'R', '.'],
            ['.', 'R', 'E', 'R', 'E', 'R', 'X', 'R'],
            ['.', 'X', 'R', 'X', 'R', 'R', 'R', 'R'],
            ['.', 'E', 'R', 'E', 'R', 'R', 'X', 'R'],
            ['.', 'E', 'R', 'X', 'R', 'R', 'R', 'R'],
            ['.', 'R', 'E', 'R', 'E', 'R', 'R', 'R'],
            ['.', '.', 'R', 'E', 'R', 'E', '.', '.'],
        ],
    },
    {
        origin: { x: 6, y: 7 },
        parking: { x: 3, y: 3 },
        muster: [
            { x: -8, y: -2, direction: TOP },
            { x: -8, y: 2, direction: BOTTOM },
            { x: 8, y: -2, direction: TOP },
            { x: 8, y: 2, direction: BOTTOM },
            { x: -2, y: -8, direction: LEFT },
            { x: 2, y: -8, direction: RIGHT },
            { x: -2, y: 8, direction: LEFT },
            { x: 2, y: 8, direction: RIGHT },
        ],
        buildings: [
            ['.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.'],
            ['.', '.', '.', 'E', 'E', 'R', 'E', 'R', 'E', 'E', '.', '.', '.'],
            ['.', '.', '.', 'R', 'R', 'E', 'R', 'E', 'R', 'R', '.', '.', '.'],
            ['.', '.', 'R', 'E', 'E', 'R', 'P', 'R', 'E', 'E', 'R', '.', '.'],
            ['.', 'R', '.', 'E', 'E', 'R', 'E', 'R', 'E', 'E', '.', 'R', '.'],
            ['.', 'R', '.', 'X', 'X', 'R', 'X', 'R', 'X', 'X', '.', 'R', '.'],
            ['.', '.', 'R', 'X', 'R', 'T', 'R', 'E', 'R', 'X', 'R', '.', '.'],
            ['.', '.', '.', 'R', 'X', 'R', 'S', 'R', 'L', 'R', '.', '.', '.'],
            ['.', '.', 'R', 'X', 'R', 'E', 'R', 'T', 'R', '.', 'R', '.', '.'],
            ['.', 'R', '.', 'X', 'X', 'R', 'X', 'R', 'R', 'R', 'R', 'R', '.'],
            ['.', 'R', '.', 'E', 'E', 'R', 'E', 'R', 'R', 'X', 'R', 'R', '.'],
            ['.', '.', 'R', 'E', 'E', 'R', 'X', 'R', 'R', 'R', 'R', 'R', '.'],
            ['.', '.', '.', 'R', 'R', 'E', 'R', 'E', 'R', 'R', 'R', '.', '.'],
            ['.', '.', '.', 'E', 'E', 'R', 'E', 'R', 'E', 'E', '.', '.', '.'],
            ['.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.'],
        ],
    },
    {
        origin: { x: 6, y: 7 },
        parking: { x: 3, y: 3 },
        muster: [
            { x: -8, y: -2, direction: TOP },
            { x: -8, y: 2, direction: BOTTOM },
            { x: 8, y: -2, direction: TOP },
            { x: 8, y: 2, direction: BOTTOM },
            { x: -2, y: -8, direction: LEFT },
            { x: 2, y: -8, direction: RIGHT },
            { x: -2, y: 8, direction: LEFT },
            { x: 2, y: 8, direction: RIGHT },
        ],
        buildings: [
            ['.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.'],
            ['.', '.', '.', 'E', 'E', 'R', 'E', 'R', 'E', 'E', '.', '.', '.'],
            ['.', '.', 'E', 'R', 'R', 'E', 'R', 'E', 'R', 'R', 'E', '.', '.'],
            ['.', '.', 'R', 'E', 'E', 'R', 'P', 'R', 'E', 'E', 'R', '.', '.'],
            ['.', 'R', 'E', 'E', 'E', 'R', 'E', 'R', 'E', 'E', 'E', 'R', '.'],
            ['.', 'R', 'E', 'B', 'B', 'R', 'X', 'R', 'X', 'X', 'E', 'R', '.'],
            ['.', '.', 'R', 'B', 'R', 'T', 'R', 'E', 'R', 'X', 'R', '.', '.'],
            ['.', '.', '.', 'R', 'M', 'R', 'S', 'R', 'L', 'R', '.', '.', '.'],
            ['.', '.', 'R', 'X', 'R', 'E', 'R', 'T', 'R', 'X', 'R', '.', '.'],
            ['.', 'R', 'E', 'X', 'X', 'R', 'X', 'R', 'R', 'R', 'R', 'R', '.'],
            ['.', 'R', 'E', 'E', 'E', 'R', 'E', 'R', 'R', 'X', 'R', 'R', '.'],
            ['.', '.', 'R', 'E', 'E', 'R', 'X', 'R', 'R', 'R', 'R', 'R', '.'],
            ['.', '.', 'E', 'R', 'R', 'E', 'R', 'E', 'R', 'R', 'R', '.', '.'],
            ['.', '.', '.', 'E', 'E', 'R', 'E', 'R', 'E', 'E', '.', '.', '.'],
            ['.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.'],
        ],
    },
    {
        origin: { x: 6, y: 7 },
        parking: { x: 3, y: 3 },
        muster: [
            { x: -8, y: -2, direction: TOP },
            { x: -8, y: 2, direction: BOTTOM },
            { x: 8, y: -2, direction: TOP },
            { x: 8, y: 2, direction: BOTTOM },
            { x: -2, y: -8, direction: LEFT },
            { x: 2, y: -8, direction: RIGHT },
            { x: -2, y: 8, direction: LEFT },
            { x: 2, y: 8, direction: RIGHT },
        ],
        buildings: [
            ['.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.'],
            ['.', '.', '.', 'E', 'E', 'R', 'E', 'R', 'E', 'E', '.', '.', '.'],
            ['.', '.', 'E', 'R', 'R', 'E', 'R', 'E', 'R', 'R', 'E', '.', '.'],
            ['.', 'E', 'R', 'E', 'E', 'R', 'P', 'R', 'E', 'E', 'R', 'E', '.'],
            ['.', 'R', 'E', 'E', 'E', 'R', 'E', 'R', 'E', 'E', 'E', 'R', '.'],
            ['.', 'R', 'E', 'B', 'B', 'R', 'X', 'R', 'B', 'B', 'E', 'R', '.'],
            ['R', 'E', 'R', 'B', 'R', 'T', 'R', 'X', 'R', 'B', 'R', 'E', 'R'],
            ['.', 'R', 'E', 'R', 'M', 'R', 'S', 'R', 'L', 'R', 'E', 'R', '.'],
            ['R', 'E', 'R', 'X', 'R', 'T', 'R', 'T', 'R', 'E', 'R', 'E', 'R'],
            ['.', 'R', 'E', 'X', 'X', 'R', 'X', 'R', 'R', 'R', 'R', 'R', '.'],
            ['.', 'R', 'E', 'E', 'E', 'R', 'E', 'R', 'R', 'X', 'R', 'R', '.'],
            ['.', 'E', 'R', 'E', 'E', 'R', 'P', 'R', 'R', 'R', 'R', 'R', '.'],
            ['.', '.', 'E', 'R', 'R', 'E', 'R', 'E', 'R', 'R', 'R', 'E', '.'],
            ['.', '.', '.', 'E', 'E', 'R', 'E', 'R', 'E', 'E', 'E', '.', '.'],
            ['.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.'],
        ],
    },
    {
        origin: { x: 6, y: 7 },
        parking: { x: 3, y: 3 },
        muster: [
            { x: -8, y: -2, direction: TOP },
            { x: -8, y: 2, direction: BOTTOM },
            { x: 8, y: -2, direction: TOP },
            { x: 8, y: 2, direction: BOTTOM },
            { x: -2, y: -8, direction: LEFT },
            { x: 2, y: -8, direction: RIGHT },
            { x: -2, y: 8, direction: LEFT },
            { x: 2, y: 8, direction: RIGHT },
        ],
        buildings: [
            ['.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.'],
            ['.', '.', '.', 'E', 'E', 'R', 'O', 'R', 'E', 'E', '.', '.', '.'],
            ['.', 'E', 'E', 'R', 'R', 'E', 'R', 'E', 'R', 'R', 'E', 'E', '.'],
            ['.', 'E', 'R', 'E', 'E', 'R', 'P', 'R', 'E', 'E', 'R', 'E', '.'],
            ['E', 'R', 'E', 'E', 'E', 'R', 'E', 'R', 'E', 'E', 'E', 'R', 'E'],
            ['E', 'R', 'E', 'B', 'B', 'R', 'T', 'R', 'B', 'B', 'E', 'R', 'E'],
            ['R', 'E', 'R', 'B', 'R', 'T', 'R', 'T', 'R', 'B', 'R', 'E', 'R'],
            ['E', 'R', 'E', 'R', 'M', 'R', 'S', 'R', 'L', 'R', 'P', 'R', 'E'],
            ['R', 'E', 'R', 'B', 'R', 'T', 'R', 'T', 'R', 'X', 'R', 'E', 'R'],
            ['E', 'R', 'E', 'B', 'B', 'R', 'T', 'R', 'R', 'R', 'R', 'R', 'E'],
            ['E', 'R', 'E', 'E', 'E', 'R', 'E', 'R', 'R', 'N', 'R', 'R', 'E'],
            ['.', 'E', 'R', 'E', 'E', 'R', 'P', 'R', 'R', 'R', 'R', 'R', 'E'],
            ['.', 'E', 'E', 'R', 'R', 'E', 'R', 'E', 'R', 'R', 'R', 'E', '.'],
            ['.', '.', '.', 'E', 'E', 'R', 'E', 'R', 'E', 'E', 'E', '.', '.'],
            ['.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.'],
        ],
    },
];
class BaseConstructionRunnable {
    constructor(baseId, id) {
        this.id = id;
        this.baseId = baseId;
    }
    run(kernel, trace) {
        trace = trace.begin('base_construction_run');
        trace.info('base construction run', { id: this.id, baseId: this.baseId });
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.error('no base config');
            trace.end();
            return terminate();
        }
        const origin = base.origin;
        if (!origin) {
            trace.error('no origin');
            trace.end();
            return sleeping(CONSTRUCTION_INTERVAL);
        }
        const baseLevel = getBaseLevel(base);
        // Update parking lot for room level
        const levelLayout = baseLayouts[baseLevel];
        if (levelLayout) {
            setParking(base, levelLayout, origin);
        }
        else {
            trace.error('no level layout', { roomLevel: baseLevel });
        }
        // We have check all the things, now it's time to work out what we are building and build it
        const unfinished = this.selectLayout(baseLevel, origin, trace);
        if (unfinished) {
            this.buildLayout(base, unfinished, origin, trace);
        }
        else {
            trace.info('no unfinished layout', { roomLevel: baseLevel });
        }
        trace.end();
        return sleeping(CONSTRUCTION_INTERVAL);
    }
    selectLayout(roomLevel, origin, trace) {
        // for (let i = 0; i <= roomLevel; i++) {
        //  const layout = baseLayouts[i];
        const layout = baseLayouts[roomLevel];
        if (!this.layoutComplete(layout, origin, trace)) {
            return layout;
        }
        // }
        return null;
    }
    buildLayout(base, layout, origin, trace) {
        trace.info('building layout', { layout });
        let toBuild = [];
        let numSites = 0;
        const room = getBasePrimaryRoom(base);
        if (!room) {
            trace.error('no primary room');
            return;
        }
        // const roomVisual = new RoomVisual(room.name);
        for (let y = 0; y < layout.buildings.length; y++) {
            const row = layout.buildings[y];
            for (let x = 0; x < row.length; x++) {
                const code = row[x];
                if (buildingCodes[code] === ANY) {
                    continue;
                }
                const pos = getConstructionPosition({ x, y }, origin, layout);
                const structure = pos.lookFor(LOOK_STRUCTURES)[0];
                if (structure) {
                    trace.info('structure present', { structure: structure.structureType });
                    if (structure.structureType !== buildingCodes[code]) {
                        trace.warn('wrong site, remove', { existing: structure.structureType, expected: buildingCodes[code] });
                        structure.destroy();
                    }
                    continue;
                }
                const site = pos.lookFor(LOOK_CONSTRUCTION_SITES)[0];
                if (site) {
                    numSites++;
                    if (site.structureType !== buildingCodes[code]) {
                        trace.warn('wrong site, remove', { existing: site.structureType, expected: buildingCodes[code] });
                        site.remove();
                    }
                    continue;
                }
                const structureType = buildingCodes[code];
                if (!structureType || structureType === EMPTY) {
                    continue;
                }
                toBuild.push({ x: pos.x, y: pos.y, structureType });
            }
        }
        if (numSites > MAX_STRUCTURE_SITES) {
            trace.info('too many sites', { numSites });
            return;
        }
        // Sort by type priority
        toBuild = prioritizeBySitesType(toBuild);
        // Build sites until we hit max sites
        while (toBuild.length > 0) {
            if (numSites >= MAX_STRUCTURE_SITES) {
                return;
            }
            // Remove first item from the list
            const site = toBuild.shift();
            if (!site) {
                continue;
            }
            const pos = new RoomPosition(site.x, site.y, room.name);
            const result = room.createConstructionSite(pos, site.structureType);
            if (result !== OK && result !== ERR_FULL) {
                trace.error('failed to build structure', { structureType: site.structureType, pos, result });
                return;
            }
            numSites++;
            trace.notice('building', { structureType: site.structureType, pos, result });
        }
    }
    layoutComplete(layout, origin, trace) {
        if (layout.buildings.length === 0) {
            return true;
        }
        for (let y = 0; y < layout.buildings.length; y++) {
            const row = layout.buildings[y];
            for (let x = 0; x < row.length; x++) {
                const pos = getConstructionPosition({ x, y }, origin, layout);
                const structures = pos.lookFor(LOOK_STRUCTURES);
                const code = row[x];
                // Anything can be there
                if (buildingCodes[code] === ANY) {
                    continue;
                }
                // No structures and empty
                if (buildingCodes[code] === EMPTY && !structures.length) {
                    continue;
                }
                if (!structures.length) {
                    trace.info('missing structures', { pos });
                    return false;
                }
                if (structures.length && structures[0].structureType !== buildingCodes[code]) {
                    trace.info('incorrect structure present', { pos, structures: structures.map((s) => s.structureType) });
                    return false;
                }
            }
        }
        return true;
    }
}
const getPrioritizedSites = function (room) {
    let sites = room.find(FIND_MY_CONSTRUCTION_SITES);
    if (!sites || !sites.length) {
        return [];
    }
    sites = _.sortBy(sites, (site) => {
        switch (site.structureType) {
            case STRUCTURE_SPAWN:
                return 0 - site.progress / site.progressTotal;
            case STRUCTURE_TOWER:
                return 1 - site.progress / site.progressTotal;
            case STRUCTURE_RAMPART:
                return 2 - site.progress / site.progressTotal;
            case STRUCTURE_WALL:
                return 3 - site.progress / site.progressTotal;
            case STRUCTURE_STORAGE:
                return 4 - site.progress / site.progressTotal;
            case STRUCTURE_LINK:
                return 5 - site.progress / site.progressTotal;
            case STRUCTURE_EXTRACTOR:
                return 6 - site.progress / site.progressTotal;
            case STRUCTURE_EXTENSION:
                return 7 - site.progress / site.progressTotal;
            case STRUCTURE_TERMINAL:
                return 8 - site.progress / site.progressTotal;
            case STRUCTURE_LAB:
                return 9 - site.progress / site.progressTotal;
            case STRUCTURE_CONTAINER:
                return 10 - site.progress / site.progressTotal;
            case STRUCTURE_ROAD:
                return 20 - site.progress / site.progressTotal;
            default:
                return 15 - site.progress / site.progressTotal;
        }
    });
    return sites;
};
const prioritizeBySitesType = (sites) => {
    return _.sortBy(sites, (site) => {
        switch (site.structureType) {
            case STRUCTURE_SPAWN:
                return 0;
            case STRUCTURE_TOWER:
                return 1;
            case STRUCTURE_RAMPART:
                return 2;
            case STRUCTURE_WALL:
                return 3;
            case STRUCTURE_STORAGE:
                return 4;
            case STRUCTURE_LINK:
                return 5;
            case STRUCTURE_EXTENSION:
                return 6;
            case STRUCTURE_TERMINAL:
                return 7;
            case STRUCTURE_EXTRACTOR:
                return 8;
            case STRUCTURE_LAB:
                return 9;
            case STRUCTURE_CONTAINER:
                return 10;
            case STRUCTURE_ROAD:
                return 20;
            default:
                return 15;
        }
    });
};
const getInfrastructureSites = function (room) {
    let sites = room.find(FIND_MY_CONSTRUCTION_SITES).filter((site) => {
        return site.structureType === STRUCTURE_SPAWN || site.structureType === STRUCTURE_TOWER ||
            site.structureType === STRUCTURE_STORAGE || site.structureType === STRUCTURE_EXTENSION;
    });
    if (!sites || !sites.length) {
        return [];
    }
    sites = _.sortBy(sites, (site) => {
        switch (site.structureType) {
            case STRUCTURE_SPAWN:
                return 0 - site.progress / site.progressTotal;
            case STRUCTURE_TOWER:
                return 1 - site.progress / site.progressTotal;
            case STRUCTURE_STORAGE:
                return 2 - site.progress / site.progressTotal;
            case STRUCTURE_EXTENSION:
                return 3 - site.progress / site.progressTotal;
            default:
                return 15 - site.progress / site.progressTotal;
        }
    });
    return sites;
};

const topEdge = 1;
const bottomEdge = 48;
const leftEdge = 1;
const rightEdge = 48;
const blockingObjects = new Map([
    [STRUCTURE_WALL, true],
    [STRUCTURE_RAMPART, true],
]);
function getRegion(room, starting) {
    const regionMap = new Map();
    if (isWallOrInRegion(room, regionMap, starting)) {
        return regionMap;
    }
    const stack = [{ x: starting.x, y: starting.y }];
    while (stack.length) {
        const pos = stack.pop();
        // Move position to top/edge
        while (pos.y - 1 >= topEdge && !isWallOrInRegion(room, regionMap, { x: pos.x, y: pos.y - 1 })) {
            pos.y--;
        }
        let reachLeft = false;
        let reachRight = false;
        while (pos.y <= bottomEdge && !isWallOrInRegion(room, regionMap, pos)) {
            regionMap.set([pos.x, pos.y].join(','), { x: pos.x, y: pos.y });
            if (pos.x >= leftEdge) {
                if (!isWallOrInRegion(room, regionMap, { x: pos.x - 1, y: pos.y })) {
                    if (!reachLeft) {
                        stack.push({ x: pos.x - 1, y: pos.y });
                        reachLeft = true;
                    }
                }
                else {
                    reachLeft = false;
                }
            }
            if (pos.x <= rightEdge) {
                if (!isWallOrInRegion(room, regionMap, { x: pos.x + 1, y: pos.y })) {
                    if (!reachRight) {
                        stack.push({ x: pos.x + 1, y: pos.y });
                        reachRight = true;
                    }
                }
                else {
                    reachRight = false;
                }
            }
            pos.y++;
        }
    }
    // Get ramparts and add to region map
    room.find(FIND_STRUCTURES, {
        filter: (structure) => structure.structureType === STRUCTURE_RAMPART,
    }).forEach((rampart) => {
        regionMap.set([rampart.pos.x, rampart.pos.y].join(','), { x: rampart.pos.x, y: rampart.pos.y });
    });
    return regionMap;
}
function isWallOrInRegion(room, regionMap, pos) {
    if (regionMap.has([pos.x, pos.y].join(','))) {
        return true;
    }
    const objects = room.lookAt(pos.x, pos.y);
    return !!_.find(objects, (object) => {
        if (object.type === LOOK_STRUCTURES) {
            return blockingObjects.has(object.structure.structureType);
        }
        if (object.type === LOOK_TERRAIN && object.terrain === 'wall') {
            return true;
        }
        return false;
    });
}

// used to calculate positions along a direction
const DIRECTION_OFFSET = {
    [TOP]: { x: 0, y: -1 },
    [TOP_RIGHT]: { x: 1, y: -1 },
    [RIGHT]: { x: 1, y: 0 },
    [BOTTOM_RIGHT]: { x: 1, y: 1 },
    [BOTTOM]: { x: 0, y: 1 },
    [BOTTOM_LEFT]: { x: -1, y: 1 },
    [LEFT]: { x: -1, y: 0 },
    [TOP_LEFT]: { x: -1, y: -1 },
};
const getNearbyPositions = (pos, range) => {
    const positions = [];
    for (let x = pos.x - range; x <= pos.x + range; x++) {
        for (let y = pos.y - range; y <= pos.y + range; y++) {
            if (x < 0 || y < 0 || x > 49 || y > 49) {
                continue;
            }
            try {
                positions.push(new RoomPosition(x, y, pos.roomName));
            }
            catch (e) {
                console.log(x, y, pos.roomName);
                throw e;
            }
        }
    }
    return positions;
};

/**
 * Cost Matrix
 *
 * Contains functions that create the various cost matrices used by
 * processes & creeps.
 *
 * TODO could be dried out a bit more
 * TODO add tests
 */
let costMatrix255 = null;
const createDefenderCostMatrix = (roomId, _trace) => {
    const costMatrix = new PathFinder.CostMatrix();
    const room = Game.rooms[roomId];
    if (!room) {
        return costMatrix;
    }
    const spawn = room.find(FIND_STRUCTURES, {
        filter: (structure) => structure.structureType === STRUCTURE_SPAWN,
    })[0];
    if (!spawn) {
        // No spawn, return a cost matrix with 0s
        return new PathFinder.CostMatrix();
    }
    const costs = get255CostMatrix();
    // Set every position in base to 0
    const regionValues = Array.from(getRegion(room, spawn.pos).values());
    regionValues.forEach((pos) => {
        costs.set(pos.x, pos.y, 0);
    });
    return costs;
};
const createCommonCostMatrix = (kernel, roomName, trace) => {
    const costMatrix = new PathFinder.CostMatrix();
    const room = Game.rooms[roomName];
    if (!room) {
        trace.info('room not visible', { roomName: roomName });
        return costMatrix;
    }
    const terrain = Game.map.getRoomTerrain(roomName);
    const structures = room.find(FIND_STRUCTURES);
    trace.info('found structures', { numStructures: structures.length });
    // Favor roads and avoid blocking structures
    structures.forEach(function (struct) {
        if (struct.structureType === STRUCTURE_ROAD) {
            // Favor roads over plain tiles
            costMatrix.set(struct.pos.x, struct.pos.y, 1);
        }
        if (struct.structureType === STRUCTURE_STORAGE) {
            for (let x = struct.pos.x - 1; x <= struct.pos.x + 1; x++) {
                for (let y = struct.pos.y - 1; y <= struct.pos.y + 1; y++) {
                    if (x < 0 || y < 0 || x > 49 || y > 49 || terrain.get(x, y) === TERRAIN_MASK_WALL) {
                        continue;
                    }
                    // Dont override roads
                    if (costMatrix.get(x, y) < 3) {
                        costMatrix.set(x, y, 3);
                    }
                }
            }
        }
        // TODO figure out how to not use "any"
        const isObstacle = OBSTACLE_OBJECT_TYPES.indexOf(struct.structureType) > -1;
        if (isObstacle) {
            // Can't walk through non-walkable buildings
            costMatrix.set(struct.pos.x, struct.pos.y, 255);
        }
    });
    // Add blocking structures to the cost matrix
    const sites = room.find(FIND_CONSTRUCTION_SITES);
    trace.info('found construction sites', { numSites: sites.length });
    sites.forEach((site) => {
        if (OBSTACLE_OBJECT_TYPES.indexOf(site.structureType) !== -1) {
            costMatrix.set(site.pos.x, site.pos.y, 255);
        }
    });
    // Avoid controllers
    structures.filter((structure) => structure.structureType === STRUCTURE_CONTROLLER).forEach((structure) => {
        getNearbyPositions(structure.pos, 3).forEach((pos) => {
            if (terrain.get(pos.x, pos.y) !== TERRAIN_MASK_WALL && costMatrix.get(pos.x, pos.y) < 3) {
                costMatrix.set(pos.x, pos.y, 3);
            }
        });
    });
    // avoid sources
    applySourceMineralBuffer(room, costMatrix, terrain, 5, trace);
    const base = kernel.getPlanner().getBaseById(roomName);
    if (base) {
        // Dont path through the parking lot
        applyParkingLotBuffer(base, costMatrix, terrain, 5);
    }
    return costMatrix;
};
const singleRoomCommonMatrix = (kernel, roomName, trace) => {
    const costMatrix = createCommonCostMatrix(kernel, roomName, trace);
    // Block all exits
    for (let x = 0; x < 50; x++) {
        costMatrix.set(x, 0, 255);
        costMatrix.set(x, 49, 255);
    }
    for (let y = 0; y < 50; y++) {
        costMatrix.set(0, y, 255);
        costMatrix.set(49, y, 255);
    }
    return costMatrix;
};
const haulerCostMatrixMatrix = (kernel, roomName, trace) => {
    const costMatrix = createCommonCostMatrix(kernel, roomName, trace);
    const terrain = Game.map.getRoomTerrain(roomName);
    // Add roads in base final base layout
    const base = kernel.getPlanner().getBaseById(roomName);
    if (base) {
        applyBaseRoads(base, costMatrix, terrain, 1);
        // Don't path through the parking lot
        applyParkingLotBuffer(base, costMatrix, terrain, 5);
    }
    return costMatrix;
};
const createSourceRoadMatrix = (kernel, roomName, trace) => {
    const costMatrix = new PathFinder.CostMatrix();
    // If room is not visible then use empty matrix
    const room = Game.rooms[roomName];
    if (!room) {
        trace.error('room not visible', { roomName: roomName });
        return costMatrix;
    }
    const terrain = Game.map.getRoomTerrain(roomName);
    const structures = room.find(FIND_STRUCTURES);
    const nonRoadStructures = structures.filter((s) => s.structureType !== STRUCTURE_ROAD);
    nonRoadStructures.forEach(function (struct) {
        var _a;
        if (struct.structureType !== STRUCTURE_CONTAINER &&
            (struct.structureType !== STRUCTURE_RAMPART || !struct.my)) {
            // Controller links dont count as obstacle, otherwise controller pad will shift around
            if (struct.structureType === STRUCTURE_LINK && room.controller.pos.inRangeTo(struct.pos, 3)) {
                return;
            }
            // Give walls a low weight otherwise we will minimize holes, but only if
            // Only for rooms we control, we can destroy the walls
            if (((_a = room.controller) === null || _a === void 0 ? void 0 : _a.my) && struct.structureType === STRUCTURE_WALL) {
                costMatrix.set(struct.pos.x, struct.pos.y, 10);
                return;
            }
            // Can't walk through non-walkable buildings
            costMatrix.set(struct.pos.x, struct.pos.y, 255);
        }
    });
    applySourceMineralBuffer(room, costMatrix, terrain, 5, trace);
    applyRoads(room, costMatrix, 1);
    applyRoadSites(room, costMatrix, 1);
    // Add roads in base final base layout
    const base = kernel.getPlanner().getBaseById(roomName);
    if (base) {
        applyBaseRoads(base, costMatrix, terrain, 1);
        applyControllerBuffer(costMatrix, terrain, room.controller, 4);
        applyParkingLotBuffer(base, costMatrix, terrain, 5);
    }
    return costMatrix;
};
const createPartyCostMatrix = (roomName, _trace) => {
    var _a;
    const costMatrix = new PathFinder.CostMatrix();
    const terrain = Game.map.getRoomTerrain(roomName);
    for (let x = 0; x <= 49; x++) {
        for (let y = 0; y <= 49; y++) {
            const mask = terrain.get(x, y);
            if (mask) {
                const maskValue = (mask === TERRAIN_MASK_WALL) ? 255 : 5;
                // The party origin is the bottom left corner
                // center
                if (costMatrix.get(x, y) < maskValue) {
                    costMatrix.set(x, y, maskValue);
                }
                // down
                if (x >= 0 && x <= 49 && y + 1 >= 0 && y + 1 <= 48 && costMatrix.get(x, y + 1) < maskValue) {
                    costMatrix.set(x, y + 1, maskValue);
                }
                // left
                if (x - 1 >= 0 && x - 1 <= 49 && y >= 0 && y <= 48 && costMatrix.get(x - 1, y) < maskValue) {
                    costMatrix.set(x - 1, y, maskValue);
                }
                // down left
                if (x - 1 >= 0 && x - 1 <= 49 && y + 1 >= 0 && y + 1 <= 48 && costMatrix.get(x - 1, y + 1) < maskValue) {
                    costMatrix.set(x - 1, y + 1, maskValue);
                }
                continue;
            }
            if (x <= 1 || y <= 1 || x >= 48 || y >= 48) {
                if (costMatrix.get(x, y) < 5) {
                    costMatrix.set(x, y, 5);
                }
            }
        }
    }
    const room = Game.rooms[roomName];
    if (room && ((_a = room.controller) === null || _a === void 0 ? void 0 : _a.my)) {
        room.find(FIND_STRUCTURES).forEach((s) => {
            if (OBSTACLE_OBJECT_TYPES.indexOf(s.structureType) !== -1) {
                costMatrix.set(s.pos.x, s.pos.y, 255);
            }
        });
    }
    return costMatrix;
};
const createOpenSpaceMatrix = (roomName, trace) => {
    trace = trace.begin('createOpenSpaceMatrix');
    const costMatrix = new PathFinder.CostMatrix();
    const seen = new Set();
    let pass = 0;
    const passes = [];
    passes[pass] = [];
    // Add walls and room positions near edges to initial pass
    // Mark each position as seen
    const terrain = Game.map.getRoomTerrain(roomName);
    for (let x = 0; x <= 49; x++) {
        for (let y = 0; y <= 49; y++) {
            if (x < 3 || y < 3 || x > 46 || y > 46 || terrain.get(x, y) === TERRAIN_MASK_WALL) {
                passes[pass].push(new RoomPosition(x, y, roomName));
                seen.add(`${x},${y}`);
                costMatrix.set(x, y, pass);
            }
        }
    }
    do {
        const currentPass = passes[pass];
        pass++;
        passes[pass] = [];
        currentPass.forEach((centerPos) => {
            getNearbyPositions(centerPos, 1).forEach((pos) => {
                if (seen.has(`${pos.x},${pos.y}`)) {
                    return;
                }
                passes[pass].push(pos);
                seen.add(`${pos.x},${pos.y}`);
                costMatrix.set(pos.x, pos.y, pass);
            });
        });
    } while (passes[pass].length);
    const position = passes[pass - 1][0];
    const cpuTime = trace.end();
    trace.info('results', { cpuTime, pass, position });
    return [costMatrix, pass, position];
};
const get255CostMatrix = () => {
    if (costMatrix255) {
        return costMatrix255.clone();
    }
    const costs = new PathFinder.CostMatrix();
    for (let x = 0; x <= 49; x++) {
        for (let y = 0; y <= 49; y++) {
            costs.set(x, y, 255);
        }
    }
    costMatrix255 = costs;
    return costs.clone();
};
const visualizeCostMatrix = (roomName, costMatrix, trace) => {
    if (typeof (costMatrix) === 'boolean') {
        trace.info('costmatrix is boolean', { roomName });
        return;
    }
    trace.info('show matrix', { roomName });
    const visual = new RoomVisual(roomName);
    for (let x = 0; x <= 49; x++) {
        for (let y = 0; y <= 49; y++) {
            const cost = costMatrix.get(x, y);
            visual.text((cost).toString(), x, y);
        }
    }
};
const applyControllerBuffer = (costMatrix, terrain, controller, cost, _trace) => {
    const controllerPos = controller.pos;
    for (let x = controllerPos.x - 3; x <= controllerPos.x + 3; x++) {
        for (let y = controllerPos.y - 3; y <= controllerPos.y + 3; y++) {
            if (x < 0 || y < 0 || x > 49 || y > 49 || terrain.get(x, y) === TERRAIN_MASK_WALL) {
                continue;
            }
            // If the cost is already higher then leave it
            if (costMatrix.get(x, y) > cost) {
                continue;
            }
            costMatrix.set(x, y, cost);
        }
    }
};
const applySourceMineralBuffer = (room, costMatrix, terrain, cost, trace) => {
    let sources = room.find(FIND_SOURCES);
    const minerals = room.find(FIND_MINERALS);
    sources = sources.concat(minerals);
    trace.info('found sources and minerals', { numStructures: sources.length });
    sources.forEach(function (source) {
        const pos = source.pos;
        for (let x = pos.x - 2; x < pos.x + 3; x++) {
            for (let y = pos.y - 2; y < pos.y + 3; y++) {
                if (x < 0 || y < 0 || x > 49 || y > 49 || terrain.get(x, y) === TERRAIN_MASK_WALL) {
                    continue;
                }
                // If the cost is already higher then leave it
                if (costMatrix.get(x, y) > cost) {
                    continue;
                }
                costMatrix.set(x, y, cost);
            }
        }
    });
};
const applyRoads = (room, costMatrix, cost, _trace) => {
    const structures = room.find(FIND_STRUCTURES);
    const roads = structures.filter((s) => s.structureType === STRUCTURE_ROAD);
    // Add existing roads
    roads.forEach(function (struct) {
        const currentCost = costMatrix.get(struct.pos.x, struct.pos.y);
        if (currentCost >= 5 && currentCost != 255) {
            costMatrix.set(struct.pos.x, struct.pos.y, currentCost - 2);
            return;
        }
        costMatrix.set(struct.pos.x, struct.pos.y, cost);
    });
};
const applyRoadSites = (room, costMatrix, cost, _trace) => {
    // add road construction sites
    room.find(FIND_MY_CONSTRUCTION_SITES).forEach((site) => {
        if (site.structureType === STRUCTURE_ROAD) {
            const currentCost = costMatrix.get(site.pos.x, site.pos.y);
            if (currentCost >= 5 && currentCost != 255) {
                costMatrix.set(site.pos.x, site.pos.y, currentCost - 2);
                return;
            }
            costMatrix.set(site.pos.x, site.pos.y, cost);
        }
    });
};
const applyBaseRoads = (base, costMatrix, terrain, cost, _trace) => {
    const layout = baseLayouts[getBaseLevel(base)];
    const buildings = layout.buildings;
    for (let i = 0; i < buildings.length; i++) {
        for (let j = 0; j < buildings[i].length; j++) {
            if (buildingCodes[buildings[i][j]] === STRUCTURE_ROAD) {
                const y = base.origin.y - layout.origin.y + i;
                const x = base.origin.x - layout.origin.x + j;
                costMatrix.set(x, y, cost);
            }
        }
    }
};
const applyParkingLotBuffer = (base, costMatrix, terrain, cost, _trace) => {
    for (let x = base.parking.x - 1; x <= base.parking.x + 1; x++) {
        for (let y = base.parking.y - 1; y <= base.parking.y + 1; y++) {
            if (x < 0 || y < 0 || x > 49 || y > 49 || terrain.get(x, y) === TERRAIN_MASK_WALL) {
                continue;
            }
            costMatrix.set(x, y, cost);
        }
    }
};

const PASSES = 6;
const MIN_DISTANCE_FOR_ORIGIN = 8;
const DismissedReasonNoRoomEntry = 'no_room_entry';
const DismissedReasonNoController = 'no_controller';
const DismissedReasonAdjacentClaimed = 'adjacent_claimed';
const DismissedReasonOwned = 'owned';
const DismissedReasonDifferentRoomStatus = 'different_room_status';
const pickExpansion = (kernel, trace) => {
    const candidates = new Set();
    const claimed = new Set();
    const dismissed = new Map();
    let seen = new Set();
    const bases = kernel.getPlanner().getBases();
    let baseRoomStatus = null;
    if (bases.length) {
        baseRoomStatus = Game.map.getRoomStatus(bases[0].primary).status;
    }
    trace.notice('base room status', { baseRoomStatus });
    // First pass through bases, find all the rooms that are assigned to a base already
    bases.forEach((base) => {
        // Build map of claimed rooms
        base.rooms.forEach((roomName) => {
            claimed.add(roomName);
        });
    });
    let nextPass = Array.from(claimed.keys());
    seen = new Set(claimed);
    trace.info('starting expansion pass', { nextPass, maxPasses: PASSES });
    for (let i = 0; i < PASSES; i++) {
        const found = [];
        nextPass.forEach((parentRoom) => {
            _.forEach(Game.map.describeExits(parentRoom), (roomName, _key) => {
                // Check room in next pass
                if (!seen.has(roomName)) {
                    found.push(roomName);
                }
                seen.add(roomName);
                if (dismissed.has(roomName)) {
                    return;
                }
                if (claimed.has(roomName)) {
                    return;
                }
                if (baseRoomStatus !== null) {
                    const roomStatus = Game.map.getRoomStatus(roomName).status;
                    if (roomStatus !== baseRoomStatus) {
                        trace.info('different room status', { roomName, roomStatus, baseRoomStatus });
                        dismissed.set(roomName, DismissedReasonDifferentRoomStatus);
                        return;
                    }
                }
                const roomEntry = kernel.getScribe().getRoomById(roomName);
                if (!roomEntry) {
                    trace.info('no room entry', { roomName });
                    dismissed.set(roomName, DismissedReasonNoRoomEntry);
                    return;
                }
                trace.info('room entry', { roomEntry });
                if (!roomEntry.controller || !roomEntry.controller.pos) {
                    trace.info('dismiss candidate, no controller', { parentRoom, roomEntry });
                    dismissed.set(roomName, DismissedReasonNoController);
                    return;
                }
                // It room is owned by another player, we can't expand there
                // Also add to claims room list to we don't pick adjacent rooms
                if (roomEntry.controller.owner) {
                    trace.info('dismissed room owned', { parentRoom, roomEntry });
                    dismissed.set(roomName, DismissedReasonOwned);
                    claimed.add(roomName);
                    return;
                }
                // If previous room was claimed, do not build as this room is too close to another base
                if (claimed.has(parentRoom)) {
                    dismissed.set(roomName, DismissedReasonAdjacentClaimed);
                    trace.info('dismissing room, parent claimed', { parentRoom, roomName });
                    return;
                }
                trace.info('adding room to candidates', { roomName });
                candidates.add(roomName);
            });
        });
        nextPass = found;
    }
    trace.info('candidates', {
        candidates: Array.from(candidates.keys()),
        dismissed: Array.from(dismissed.entries())
    });
    const filterCandidates = new Set();
    for (const candidate of candidates.keys()) {
        if (claimed.has(candidate) || dismissed.has(candidate)) {
            continue;
        }
        filterCandidates.add(candidate);
    }
    trace.info('filtered candidates', { filterCandidates: Array.from(filterCandidates.keys()) });
    // TODO factor in available remotes
    const sortedCandidates = _.sortByOrder([...filterCandidates.keys()], (roomName) => {
        const roomEntry = kernel.getScribe().getRoomById(roomName);
        if (!roomEntry) {
            trace.error('no room entry', { roomName });
            return 0;
        }
        trace.info('room source', { roomName, numSources: roomEntry.numSources, roomEntry });
        return roomEntry.numSources;
    }, ['desc']);
    trace.info('sorted candidates', { sortedCandidates });
    if (sortedCandidates.length < 3) {
        trace.notice('not enough candidates', { sortedCandidates });
        return { selected: null, distance: null, origin: null, candidates, claimed, dismissed, seen };
    }
    for (let i = 0; i < sortedCandidates.length; i++) {
        const roomName = sortedCandidates[i];
        const [_costMatrix, distance, origin] = createOpenSpaceMatrix(roomName, trace);
        trace.info('open space matrix', { roomName, distance, origin });
        if (distance >= MIN_DISTANCE_FOR_ORIGIN) {
            return { selected: roomName, distance, origin, candidates, claimed, dismissed, seen };
        }
    }
    return { selected: null, distance: null, origin: null, candidates, claimed, dismissed, seen };
};

class BasesDebugger {
    constructor(id, kernel) {
        this.results = null;
        this.display = false;
        this.id = id;
        this.kernel = kernel;
    }
    run(kernel, trace) {
        trace.info('expand debugger run', { results: this.results });
        if (this.display) {
            if (!this.results || Game.time % 10 === 0) {
                this.results = pickExpansion(this.kernel, trace);
            }
            trace.info('expand debugger', { results: this.results });
            const selected = this.results.selected;
            if (selected) {
                Game.map.visual.text('O', new RoomPosition(25, 25, selected), { color: '#00ff00', fontSize: 20 });
                new RoomVisual(selected).text('x', this.results.origin, { color: '#00ff00' });
            }
            for (const roomName of this.results.candidates.keys()) {
                Game.map.visual.text('0', new RoomPosition(25, 25, roomName), { color: '#0000ff', fontSize: 20 });
            }
            for (const roomName of this.results.claimed.keys()) {
                Game.map.visual.text('X', new RoomPosition(25, 25, roomName), { color: '#0000ff', fontSize: 20 });
            }
            for (const [roomName, reason] of this.results.dismissed) {
                let text = 'X';
                switch (reason) {
                    case DismissedReasonNoController:
                        text = 'NC';
                        break;
                    case DismissedReasonNoRoomEntry:
                        text = 'NE';
                        break;
                    case DismissedReasonAdjacentClaimed:
                        text = 'AC';
                        break;
                    case DismissedReasonOwned:
                        text = 'OW';
                        break;
                    case DismissedReasonDifferentRoomStatus:
                        text = 'RS';
                        break;
                    default:
                        text = '?';
                }
                Game.map.visual.text(text, new RoomPosition(25, 25, roomName), { color: '#ff0000', fontSize: 20 });
            }
        }
        return running();
    }
    debug() {
        this.display = true;
    }
    clear() {
        this.display = false;
    }
}

const COST_MATRIX_TTL = 1000;
var AllowedCostMatrixTypes;
(function (AllowedCostMatrixTypes) {
    AllowedCostMatrixTypes["PARTY"] = "party";
    AllowedCostMatrixTypes["COMMON"] = "common";
    AllowedCostMatrixTypes["SINGLE_ROOM_COMMON"] = "single_room_common";
    AllowedCostMatrixTypes["HAULER"] = "hauler";
    AllowedCostMatrixTypes["BASE_DEFENSE"] = "base";
    AllowedCostMatrixTypes["OPEN_SPACE"] = "open_space";
    AllowedCostMatrixTypes["SOURCE_ROAD"] = "source_road";
    AllowedCostMatrixTypes["CONTROLLER_ROAD"] = "controller_road";
    AllowedCostMatrixTypes["DAMAGE"] = "damage";
})(AllowedCostMatrixTypes || (AllowedCostMatrixTypes = {}));
class CostMatrixCacheItem {
    constructor(id, costMatrixType) {
        this.roomId = id;
        this.costMatrixType = costMatrixType;
        this.costMatrix = null;
        this.time = 0;
    }
    update(kernel, trace) {
        let costMatrix = new PathFinder.CostMatrix();
        trace.info('updating', { room: this.roomId, type: this.costMatrixType });
        switch (this.costMatrixType) {
            case AllowedCostMatrixTypes.PARTY:
                costMatrix = createPartyCostMatrix(this.roomId);
                break;
            case AllowedCostMatrixTypes.COMMON:
                costMatrix = createCommonCostMatrix(kernel, this.roomId, trace);
                break;
            case AllowedCostMatrixTypes.SINGLE_ROOM_COMMON:
                costMatrix = singleRoomCommonMatrix(kernel, this.roomId, trace);
                break;
            case AllowedCostMatrixTypes.HAULER:
                costMatrix = haulerCostMatrixMatrix(kernel, this.roomId, trace);
                break;
            case AllowedCostMatrixTypes.BASE_DEFENSE:
                costMatrix = createDefenderCostMatrix(this.roomId);
                break;
            case AllowedCostMatrixTypes.SOURCE_ROAD:
                costMatrix = createSourceRoadMatrix(kernel, this.roomId, trace);
                break;
            case AllowedCostMatrixTypes.OPEN_SPACE:
                [costMatrix] = createOpenSpaceMatrix(this.roomId, trace);
                break;
            default:
                trace.error('unknown cost matrix type', { type: this.costMatrixType });
        }
        this.costMatrix = costMatrix;
        this.time = Game.time;
    }
    getCostMatrix(kernel, trace) {
        if (!this.costMatrix || this.isExpired(Game.time)) {
            trace.info('cache miss/expired', {
                room: this.roomId,
                type: this.costMatrixType,
                expired: this.isExpired(Game.time),
            });
            this.update(kernel, trace);
        }
        return this.costMatrix;
    }
    isExpired(time) {
        return time - this.time > COST_MATRIX_TTL;
    }
}
class CostMatrixCache {
    constructor() {
        this.rooms = new Map();
    }
    getCostMatrix(kernel, roomId, costMatrixType, trace) {
        if (!this.rooms.get(roomId)) {
            trace.info('room not in cache', { roomId });
            this.rooms.set(roomId, new Map());
        }
        const room = this.rooms.get(roomId);
        let roomMatrix = room.get(costMatrixType);
        if (!roomMatrix) {
            trace.info('matrix type not in room cache', { roomId, costMatrixType });
            roomMatrix = new CostMatrixCacheItem(roomId, costMatrixType);
            room.set(costMatrixType, roomMatrix);
        }
        return roomMatrix.getCostMatrix(kernel, trace);
    }
    getStats() {
        return {
            roomCacheSize: this.rooms.size,
        };
    }
}

class CostMatrixDebugger {
    constructor(id) {
        this.id = id;
        this.costMatrix = null;
    }
    run(kernel, trace) {
        trace.info('costmatrix debugger', { path: this.costMatrix });
        if (this.costMatrix) {
            // Display on the map
            visualizeCostMatrix(this.roomId, this.costMatrix, trace);
        }
        return running();
    }
    debug(roomId, costMatrixType) {
        const kernel = global.AI;
        const trace = new Tracer('costmatrix_debugger_debug', new Map(), new Metrics());
        trace.info('debug matrix', { roomId, costMatrixType });
        let costMatrix = new PathFinder.CostMatrix();
        switch (costMatrixType) {
            case AllowedCostMatrixTypes.PARTY:
                costMatrix = createPartyCostMatrix(roomId);
                break;
            case AllowedCostMatrixTypes.COMMON:
                costMatrix = createCommonCostMatrix(kernel, roomId, trace);
                break;
            case AllowedCostMatrixTypes.BASE_DEFENSE:
                costMatrix = createDefenderCostMatrix(roomId);
                break;
            case AllowedCostMatrixTypes.SOURCE_ROAD:
                costMatrix = createSourceRoadMatrix(kernel, roomId, trace);
                break;
            case AllowedCostMatrixTypes.OPEN_SPACE:
                [costMatrix] = createOpenSpaceMatrix(roomId, trace);
                break;
            case AllowedCostMatrixTypes.DAMAGE:
                //costMatrix = createDamageMatrix(global.AI, roomId, trace);
                break;
            default:
                trace.error('unexpected matrix type', { matrixType: costMatrixType });
        }
        if (typeof (costMatrix) !== 'boolean') {
            this.costMatrix = costMatrix;
        }
        else {
            this.costMatrix = null;
        }
    }
    clear() {
        this.costMatrix = null;
    }
}

const threadKernel = (name, ttl) => (action) => {
    let lastCall = 0;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const tick = function (trace, kernel, ...args) {
        if (lastCall + ttl <= Game.time) {
            lastCall = Game.time;
            const actionTrace = trace.begin(name);
            const result = action(actionTrace, kernel, ...args);
            actionTrace.end();
            return result;
        }
        else {
            trace.info(`thread ${name} sleeping for ${lastCall + ttl - Game.time}`);
        }
        return null;
    };
    tick.reset = () => {
        lastCall = 0;
    };
    return tick;
};

const CONSUME_EVENTS_TTL$2 = 1;
const DASHBOARD_EVENTS_TTL = 1;
const RADIUS = 3;
function getLinesStream() {
    return `hud_lines`;
}
var HudIndicatorStatus;
(function (HudIndicatorStatus) {
    HudIndicatorStatus["Green"] = "green";
    HudIndicatorStatus["Yellow"] = "yellow";
    HudIndicatorStatus["Red"] = "red";
    HudIndicatorStatus["Stale"] = "stale";
})(HudIndicatorStatus || (HudIndicatorStatus = {}));
function getDashboardStream() {
    return `hud_dashboard`;
}
const HudEventSet = 'set';
class Dashboard {
    constructor(key) {
        this.key = key;
        this.indicators = new Map();
    }
    setIndicator(indicator) {
        this.indicators.set(indicator.key, indicator);
    }
    getIndicators() {
        return this.indicators;
    }
}
class HUDRunnable {
    constructor() {
        this.dashboards = new Map();
        this.lines = new Map();
        this.dashboardConsumer = null;
        this.threadDashboardEvents = threadKernel('dashboard_events', DASHBOARD_EVENTS_TTL)(this.consumeDashboardEvents.bind(this));
        this.hudLinesConsumer = null;
        this.threadLinesEvents = threadKernel('consume_events', CONSUME_EVENTS_TTL$2)(this.consumeLinesEvents.bind(this));
    }
    run(kernel, trace) {
        if (!this.hudLinesConsumer) {
            const linesStreamId = getLinesStream();
            const consumer = kernel.getBroker().getStream(linesStreamId).addConsumer('hud');
            this.hudLinesConsumer = consumer;
        }
        if (!this.dashboardConsumer) {
            const dashboardStreamId = getDashboardStream();
            const consumer = kernel.getBroker().getStream(dashboardStreamId).addConsumer('hud');
            this.dashboardConsumer = consumer;
        }
        this.threadLinesEvents(trace, kernel);
        this.threadDashboardEvents(trace, kernel);
        const lines = Array.from(this.lines.values());
        _.forEach(_.groupBy(lines, 'room'), (lines, room) => {
            let lineNum = 0;
            const roomVisual = new RoomVisual(room);
            lines.forEach((line) => {
                lineNum++;
                roomVisual.text(line.text, 49, lineNum, { font: 0.7, align: 'right' });
            });
        });
        trace.info('dashboards', { dashboards: this.dashboards });
        Array.from(this.dashboards.values()).forEach((dashboard) => {
            const indicators = dashboard.getIndicators();
            trace.info('indicators', { indicators });
            let indicatorNum = 0;
            Array.from(indicators.values()).forEach((indicator) => {
                let fill = '#00ff00';
                if (indicator.status === HudIndicatorStatus.Red) {
                    fill = '#ff0000';
                }
                else if (indicator.status === HudIndicatorStatus.Yellow) {
                    fill = '#ffff00';
                }
                const position = new RoomPosition(1 + RADIUS + (2 * RADIUS * indicatorNum) + indicatorNum, 4, indicator.room);
                Game.map.visual.circle(position, { radius: RADIUS, fill });
                Game.map.visual.text(indicator.display, position, { fontSize: 5 });
                indicatorNum++;
            });
        });
        return running();
    }
    consumeLinesEvents(_trace) {
        this.hudLinesConsumer.getEvents().forEach((event) => {
            const line = event.data;
            this.lines.set(line.key, line);
        });
    }
    consumeDashboardEvents(trace) {
        this.dashboardConsumer.getEvents().forEach((event) => {
            trace.info('event', { event });
            const indicator = event.data;
            const room = indicator.room;
            if (!this.dashboards.has(room)) {
                this.dashboards.set(room, new Dashboard(room));
            }
            this.dashboards.get(room).setIndicator(indicator);
        });
    }
}

const ENTIRE_ROOM_BOUNDS = { x1: 0, y1: 0, x2: 49, y2: 49 };
const UNWALKABLE = -1;
const NORMAL = 0;
const PROTECTED = 1;
const NO_BUILD = 2;
const TO_EXIT = 3;
const EXIT = 4;
class RoomMatrix {
    constructor(roomName, protect, bounds) {
        this.roomName = roomName;
        this.bounds = bounds;
        this.matrix = this.createRoomMatrix(roomName, protect, bounds);
    }
    get(x, y) {
        return this.matrix[x][y];
    }
    set(x, y, value) {
        this.matrix[x][y] = value;
    }
    // create a 50x50 matrix, fill with unwalkable, set as normal if not at edge of room or walls
    createRoomMatrix(roomName, protect, bounds) {
        const matrix = Array(50).fill(0).map((_x) => Array(50).fill(UNWALKABLE));
        const terrain = Game.map.getRoomTerrain(roomName);
        for (let i = bounds.x1; i <= bounds.x2; i++) {
            for (let j = bounds.y1; j <= bounds.y2; j++) {
                // If wall, mark unwalkable
                if (terrain.get(i, j) === TERRAIN_MASK_WALL) {
                    matrix[i][j] = UNWALKABLE;
                    continue;
                }
                // If edge of boundary, mark as to exit
                if (i <= bounds.x1 || j <= bounds.y1 || i >= bounds.x2 || j >= bounds.y2) {
                    matrix[i][j] = EXIT; // Sink Tiles mark from given bounds
                    continue;
                }
                matrix[i][j] = NORMAL; // mark normal
            }
        }
        // Set tiles in matrix to PROTECTED if they are in the given rects
        const jmax = protect.length;
        for (let j = 0; j < jmax; j++) {
            const r = protect[j];
            for (let x = r.x1; x <= r.x2; x++) {
                for (let y = r.y1; y <= r.y2; y++) {
                    if (x <= 2 || x >= 48 || y <= 2 || y >= 48) {
                        continue;
                    }
                    if (terrain.get(x, y) !== TERRAIN_MASK_WALL) {
                        matrix[x][y] = PROTECTED;
                    }
                }
            }
        }
        // Marks tiles Near Exits for sink- where you cannot build wall/rampart
        const max = 49;
        for (let y = 1; y < 49; y++) {
            if (matrix[0][y - 1] === EXIT)
                matrix[1][y] = TO_EXIT;
            if (matrix[0][y] === EXIT)
                matrix[1][y] = TO_EXIT;
            if (matrix[0][y + 1] === EXIT)
                matrix[1][y] = TO_EXIT;
            if (matrix[49][y - 1] === EXIT)
                matrix[48][y] = TO_EXIT;
            if (matrix[49][y] === EXIT)
                matrix[48][y] = TO_EXIT;
            if (matrix[49][y + 1] === EXIT)
                matrix[48][y] = TO_EXIT;
        }
        for (let x = 1; x < 49; x++) {
            if (matrix[x - 1][0] === EXIT)
                matrix[x][1] = TO_EXIT;
            if (matrix[x][0] === EXIT)
                matrix[x][1] = TO_EXIT;
            if (matrix[x + 1][0] === EXIT)
                matrix[x][1] = TO_EXIT;
            if (matrix[x - 1][49] === EXIT)
                matrix[x][48] = TO_EXIT;
            if (matrix[x][49] === EXIT)
                matrix[x][48] = TO_EXIT;
            if (matrix[x + 1][49] === EXIT)
                matrix[x][48] = TO_EXIT;
        }
        // mark Border Tiles as not usable
        for (let y = 1; y < max; y++) {
            matrix[0][y] == UNWALKABLE;
            matrix[49][y] == UNWALKABLE;
        }
        for (let x = 1; x < max; x++) {
            matrix[x][0] == UNWALKABLE;
            matrix[x][49] == UNWALKABLE;
        }
        const room = Game.rooms[roomName];
        if (!room) {
            throw new Error(`RoomMatrix: Room ${roomName} not found`);
        }
        // mark area around sources as unbuildable
        room.find(FIND_SOURCES).forEach((source) => {
            getNearbyPositions(source.pos, 2).forEach((pos) => {
                if (terrain.get(pos.x, pos.y) !== TERRAIN_MASK_WALL) {
                    matrix[pos.x][pos.y] = NO_BUILD;
                }
            });
        });
        // mark area around minerals as unbuildable
        room.find(FIND_MINERALS).forEach((mineral) => {
            getNearbyPositions(mineral.pos, 2).forEach((pos) => {
                if (terrain.get(pos.x, pos.y) !== TERRAIN_MASK_WALL) {
                    matrix[pos.x][pos.y] = NO_BUILD;
                }
            });
        });
        // mark area around controller as unbuildable
        const controller = room.controller;
        getNearbyPositions(controller.pos, 2).forEach((pos) => {
            if (terrain.get(pos.x, pos.y) !== TERRAIN_MASK_WALL) {
                matrix[pos.x][pos.y] = NO_BUILD;
            }
        });
        return matrix;
    }
}
class Graph {
    constructor(menge_v) {
        this.numVertices = menge_v;
        this.level = Array(menge_v);
        // Array: for every vertex an edge Array mit {v,r,c,f} vertex_to,res_edge,capacity,flow
        this.edges = Array(menge_v).fill(0).map((_x) => []);
    }
    getTopEdges(x, y) {
        const pos = y * 50 + x;
        return this.edges[pos];
    }
    getBottomEdges(x, y) {
        const pos = y * 50 + x + 2500;
        return this.edges[pos];
    }
    newEdge(u, v, capacity) {
        // Normal forward Edge
        this.edges[u].push({ v: v, r: this.edges[v].length, capacity, flow: 0 });
        // reverse Edge for Residal Graph
        this.edges[v].push({ v: u, r: this.edges[u].length - 1, capacity: 0, flow: 0 });
    }
    // this.Bfs = function (s, t) { // calculates Level Graph and if theres a path from s to t
    bfs(s, t) {
        if (t >= this.numVertices) {
            return false;
        }
        this.level.fill(-1); // reset old levels
        this.level[s] = 0;
        const q = []; // queue with s as starting point
        q.push(s);
        let u = 0;
        let edge = null;
        while (q.length) {
            u = q.shift();
            const imax = this.edges[u].length;
            for (let i = 0; i < imax; i++) {
                edge = this.edges[u][i];
                if (this.level[edge.v] < 0 && edge.flow < edge.capacity) {
                    this.level[edge.v] = this.level[u] + 1;
                    q.push(edge.v);
                }
            }
        }
        return this.level[t] >= 0; // return if theres a path to t -> no level, no path!
    }
    // DFS like: send flow at along path from s->t recursivly while increasing the level of the visited vertices by one
    // u vertex, f flow on path, t =Sink , c Array, c[i] saves the count of edges explored from vertex i
    // this.Dfsflow = function (u, f, t, c) {
    dfsFlow(u, f, t, c) {
        if (u === t) { // Sink reached , aboard recursion
            return f;
        }
        let edge = null;
        let flow_till_here = 0;
        let flow_to_t = 0;
        while (c[u] < this.edges[u].length) { // Visit all edges of the vertex  one after the other
            edge = this.edges[u][c[u]];
            if (this.level[edge.v] === this.level[u] + 1 && edge.flow < edge.capacity) { // Edge leads to Vertex with a level one higher, and has flow left
                flow_till_here = Math.min(f, edge.capacity - edge.flow);
                flow_to_t = this.dfsFlow(edge.v, flow_till_here, t, c);
                if (flow_to_t > 0) {
                    edge.flow += flow_to_t; // Add Flow to current edge
                    this.edges[edge.v][edge.r].flow -= flow_to_t; // subtract from reverse Edge -> Residual Graph neg. Flow to use backward direction of BFS/DFS
                    return flow_to_t;
                }
            }
            c[u]++;
        }
        return 0;
    }
    // breadth-first-search which uses the level array to mark the vertices reachable from s
    // this.Bfsthecut = function (s) {
    bfsTheCut(s) {
        const e_in_cut = [];
        this.level.fill(-1);
        this.level[s] = 1;
        const q = [];
        q.push(s);
        let u = 0;
        let edge = null;
        while (q.length) {
            u = q.shift();
            let i = 0;
            const imax = this.edges[u].length;
            for (; i < imax; i++) {
                edge = this.edges[u][i];
                if (edge.flow < edge.capacity) {
                    if (this.level[edge.v] < 1) {
                        this.level[edge.v] = 1;
                        q.push(edge.v);
                    }
                }
                if (edge.flow === edge.capacity && edge.capacity > 0) { // blocking edge -> could be in min cut
                    edge.u = u;
                    e_in_cut.push(edge);
                }
            }
        }
        const min_cut = [];
        let i = 0;
        const imax = e_in_cut.length;
        for (; i < imax; i++) {
            if (this.level[e_in_cut[i].v] === -1) // Only edges which are blocking and lead to from s unreachable vertices are in the min cut
             {
                min_cut.push(e_in_cut[i].u);
            }
        }
        return min_cut;
    }
    // calculates min-cut graph (Dinic Algorithm)
    // this.Calcmincut = function (s, t) { // calculates min-cut graph (Dinic Algorithm)
    calcMinCut(s, t) {
        if (s == t) {
            return -1;
        }
        let returnvalue = 0;
        let count = [];
        let flow = 0;
        while (this.bfs(s, t) === true) {
            count = Array(this.numVertices + 1).fill(0);
            flow = 0;
            do {
                flow = this.dfsFlow(s, Number.MAX_VALUE, t, count);
                if (flow > 0) {
                    returnvalue += flow;
                }
            } while (flow);
        }
        return returnvalue;
    }
}
// Removes unneccary cut-tiles if bounds are set to include some 	dead ends
function delete_tiles_to_dead_ends(roomName, cut_tiles_array) {
    // Get Terrain and set all cut-tiles as unwalkable
    const room_array = new RoomMatrix(roomName, [], ENTIRE_ROOM_BOUNDS);
    for (let i = cut_tiles_array.length - 1; i >= 0; i--) {
        room_array[cut_tiles_array[i].x][cut_tiles_array[i].y] = UNWALKABLE;
    }
    // Floodfill from exits: save exit tiles in array and do a bfs-like search
    const unvisited_pos = [];
    let y = 0;
    const max = 49;
    for (; y < max; y++) {
        if (room_array[1][y] === TO_EXIT)
            unvisited_pos.push(50 * y + 1);
        if (room_array[48][y] === TO_EXIT)
            unvisited_pos.push(50 * y + 48);
    }
    let x = 0;
    for (; x < max; x++) {
        if (room_array[x][1] === TO_EXIT)
            unvisited_pos.push(50 + x);
        if (room_array[x][48] === TO_EXIT)
            unvisited_pos.push(2400 + x); // 50*48=2400
    }
    // Iterate over all unvisited TO_EXIT- Tiles and mark neigbours as TO_EXIT tiles, if walkable (NORMAL), and add to unvisited
    const surr = [[0, -1], [-1, -1], [-1, 0], [-1, 1], [0, 1], [1, 1], [1, 0], [1, -1]];
    let index;
    let dx;
    let dy;
    while (unvisited_pos.length > 0) {
        index = unvisited_pos.pop();
        x = index % 50;
        y = Math.floor(index / 50);
        for (let i = 0; i < 8; i++) {
            dx = x + surr[i][0];
            dy = y + surr[i][1];
            if (room_array[dx][dy] === NORMAL) {
                unvisited_pos.push(50 * dy + dx);
                room_array[dx][dy] = TO_EXIT;
            }
        }
    }
    // Remove min-Cut-Tile if there is no TO-EXIT  surrounding it
    let leads_to_exit = false;
    for (let i = cut_tiles_array.length - 1; i >= 0; i--) {
        leads_to_exit = false;
        x = cut_tiles_array[i].x;
        y = cut_tiles_array[i].y;
        for (let i = 0; i < 8; i++) {
            dx = x + surr[i][0];
            dy = y + surr[i][1];
            if (room_array[dx][dy] === TO_EXIT) {
                leads_to_exit = true;
            }
        }
        if (!leads_to_exit) {
            cut_tiles_array.splice(i, 1);
        }
    }
}
const infini = Number.MAX_VALUE;
const surr = [[0, -1], [-1, -1], [-1, 0], [-1, 1], [0, 1], [1, 1], [1, 0], [1, -1]];
// Function to create Source, Sink, Tiles arrays: takes a rectangle-Array as input for Tiles that are to Protect
// rects have top-left/bot_right Coordinates {x1,y1,x2,y2}
function create_graph(roomName, protect, bounds) {
    // An Array with Terrain information: -1 not usable, 2 Sink (Leads to Exit)
    const roomMatrix = new RoomMatrix(roomName, protect, bounds);
    // initialise graph
    // possible 2*50*50 +2 (st) Vertices (Walls etc set to unused later)
    const g = new Graph(2 * 50 * 50 + 2);
    // per Tile (0 in Array) top + bot with edge of c=1 from top to bott  (use every tile once!)
    // infini edge from bot to top vertices of adjacent tiles if they not protected (array =1) (no reverse edges in normal graph)
    // per prot. Tile (1 in array) Edge from source to this tile with infini cap.
    // per exit Tile (2in array) Edge to sink with infini cap.
    // source is at  pos 2*50*50, sink at 2*50*50+1 as first tile is 0,0 => pos 0
    // top vertices <-> x,y : v=y*50+x   and x= v % 50  y=v/50 (math.floor?)
    // bot vertices <-> top + 2500
    const source = 2 * 50 * 50;
    const sink = 2 * 50 * 50 + 1;
    const max = 49;
    for (let x = 1; x < max; x++) {
        for (let y = 1; y < max; y++) {
            const top = y * 50 + x;
            const bot = top + 2500;
            const vertex = roomMatrix.get(x, y);
            if (vertex === NORMAL) { // normal Tile
                g.newEdge(top, bot, 1);
                for (let i = 0; i < 8; i++) {
                    const dx = x + surr[i][0];
                    const dy = y + surr[i][1];
                    const surroundingVertex = roomMatrix.get(dx, dy);
                    if (surroundingVertex === NORMAL || surroundingVertex === TO_EXIT || surroundingVertex === NO_BUILD) {
                        g.newEdge(bot, dy * 50 + dx, infini);
                    }
                }
            }
            else if (vertex === PROTECTED) { // protected Tile
                g.newEdge(source, top, infini);
                g.newEdge(top, bot, 1);
                for (let i = 0; i < 8; i++) {
                    const dx = x + surr[i][0];
                    const dy = y + surr[i][1];
                    const surroundingVertex = roomMatrix.get(dx, dy);
                    if (surroundingVertex === NORMAL || surroundingVertex === TO_EXIT || surroundingVertex === NO_BUILD) {
                        g.newEdge(bot, dy * 50 + dx, infini);
                    }
                }
            }
            else if (vertex === NO_BUILD) {
                g.newEdge(top, bot, infini);
                for (let i = 0; i < 8; i++) {
                    const dx = x + surr[i][0];
                    const dy = y + surr[i][1];
                    const surroundingVertex = roomMatrix.get(dx, dy);
                    if (surroundingVertex === NORMAL || surroundingVertex === TO_EXIT || surroundingVertex === NO_BUILD) {
                        g.newEdge(bot, dy * 50 + dx, infini);
                    }
                }
            }
            else if (vertex === TO_EXIT) { // near Exit
                g.newEdge(top, sink, infini);
            }
        }
    } // graph finished
    return [g, roomMatrix];
}
// Function for user: calculate min cut tiles from room, rect[]
function getCutTiles(roomName, rect, bounds) {
    const [graph, matrix] = create_graph(roomName, rect, bounds);
    console.log('graph', JSON.stringify(graph));
    const source = 2 * 50 * 50; // Position Source / Sink in Room-Graph
    const sink = 2 * 50 * 50 + 1;
    const count = graph.calcMinCut(source, sink);
    console.log('NUmber of Tiles in Cut:', count);
    const positions = [];
    if (count > 0) {
        const cut_edges = graph.bfsTheCut(source);
        // Get Positions from Edge
        let u;
        let x;
        let y;
        const imax = cut_edges.length;
        for (let i = 0; i < imax; i++) {
            u = cut_edges[i]; // x= v % 50  y=v/50 (math.floor?)
            x = u % 50;
            y = Math.floor(u / 50);
            positions.push({ 'x': x, 'y': y });
        }
    }
    // if bounds are given,
    // try to dectect islands of walkable tiles, which are not conntected to the exits, and delete them from the cut-tiles
    const whole_room = (bounds.x1 == 0 && bounds.y1 == 0 && bounds.x2 == 49 && bounds.y2 == 49);
    if (positions.length > 0 && !whole_room) {
        delete_tiles_to_dead_ends(roomName, positions);
    }
    return [positions, matrix, graph];
}

class MinCutDebugger {
    constructor(id) {
        this.id = id;
        this.graph = null;
        this.matrix = null;
        this.cut = null;
    }
    run(kernel, trace) {
        trace.info('mincut debugger', {});
        if (this.matrix && Game.time % 2 === 0) {
            const visual = new RoomVisual(this.matrix.roomName);
            // Visualize the room
            for (let x = 0; x < 50; x++) {
                for (let y = 0; y < 50; y++) {
                    const vertex = this.matrix.get(x, y);
                    if (vertex === UNWALKABLE) {
                        visual.circle(x, y, { radius: 0.5, fill: '#111166', opacity: 0.3 });
                    }
                    else if (vertex === NORMAL) {
                        visual.circle(x, y, { radius: 0.5, fill: '#e8e863', opacity: 0.3 });
                    }
                    else if (vertex === PROTECTED) {
                        visual.circle(x, y, { radius: 0.5, fill: '#75e863', opacity: 0.3 });
                    }
                    else if (vertex === NO_BUILD) {
                        visual.circle(x, y, { radius: 0.5, fill: '#7878eb', opacity: 0.3 });
                    }
                    else if (vertex === TO_EXIT) {
                        visual.circle(x, y, { radius: 0.5, fill: '#b063e8', opacity: 0.3 });
                    }
                }
            }
            // Visualize the top edge count
            /*
            for (let x = 0; x < 49; x++) {
              for (let y = 0; y < 49; y++) {
                let edges = this.graph.getTopEdges(x, y);
                visual.text(`T${edges.length.toString()}`, x, y, {align: 'left', font: 0.2});
                edges.forEach(edge => {
                  const x2 = edge.x();
                  const y2 = edge.y();
                  visual.line(x, y, x2, edge.y, {color: '#ffffff', width: 0.1});
                });
      
                edges = this.graph.getBottomEdges(x, y);
                visual.text(`B${edges.length.toString()}`, x, y, {align: 'right', font: 0.2});
              }
            }
            */
        }
        if (this.cut) {
            const roomVisual = new RoomVisual(this.matrix.roomName);
            for (const pos of this.cut) {
                roomVisual.text('X', pos.x, pos.y);
            }
        }
        return running();
    }
    debug(kernel, roomName) {
        const trace = new Tracer('mincut_deugger', new Map([['pid', 'mincut_debugger']]), new Metrics());
        const base = kernel.getPlanner().getBaseById(roomName);
        trace.notice('base', { origin: base === null || base === void 0 ? void 0 : base.origin });
        const baseOrigin = base.origin;
        const baseBounds = {
            x1: baseOrigin.x - 9, y1: baseOrigin.y - 9,
            x2: baseOrigin.x + 9, y2: baseOrigin.y + 9,
        };
        const protect = [baseBounds];
        /* In the future we will build walls around key assets
        Game.rooms[roomName].find(FIND_SOURCES).forEach(source => {
          protect.push({
            x1: source.pos.x - 2, y1: source.pos.y - 2,
            x2: source.pos.x + 2, y2: source.pos.y + 2,
          });
        });
    
        Game.rooms[roomName].find(FIND_STRUCTURES, {
          filter: (structure) => {
            return structure.structureType === STRUCTURE_EXTRACTOR ||
              structure.structureType === STRUCTURE_LINK ||
              structure.structureType === STRUCTURE_CONTAINER;
          }
        }).forEach(structure => {
          protect.push({
            x1: structure.pos.x - 2, y1: structure.pos.y - 2,
            x2: structure.pos.x + 2, y2: structure.pos.y + 2,
          });
        })
    
    
        Game.rooms[roomName].find(FIND_STRUCTURES, {
          filter: (structure) => {
            return structure.structureType === STRUCTURE_CONTROLLER;
          }
        }).forEach(structure => {
          protect.push({
            x1: structure.pos.x - 1, y1: structure.pos.y - 1,
            x2: structure.pos.x + 1, y2: structure.pos.y + 1,
          });
        });
        */
        let cpu = Game.cpu.getUsed();
        const [cut, matrix, graph] = getCutTiles(roomName, protect, ENTIRE_ROOM_BOUNDS);
        this.cut = cut;
        this.matrix = matrix;
        this.graph = graph;
        cpu = Game.cpu.getUsed() - cpu;
        trace.notice('path', { roomName, cpu, cut, matrix });
    }
    clear() {
        this.matrix = null;
        this.cut = null;
    }
}

const Priorities = {
    CRITICAL: 0,
    CORE_LOGISTICS: 1,
    DEFENCE: 2,
    RESOURCES: 3,
    OFFENSE: 4,
    ATTACK: 5,
    LOGISTICS: 6,
    MAINTENANCE: 7,
    EXPLORATION: 8,
    DEBUG: 9,
};
const LOW_BUCKET_MIN_PRIORITY = Priorities.RESOURCES;
class Scheduler {
    constructor() {
        this.processTable = [];
        this.processMap = new Map();
        this.ranOutOfTime = 0;
        this.cpuThrottle = 0;
        this.slowProcessThreshold = 7.5;
        this.updateTimeLimit();
    }
    setCPUThrottle(throttle) {
        this.cpuThrottle = throttle;
    }
    setSlowProcessThreshold(threshold) {
        this.slowProcessThreshold = threshold;
    }
    registerProcess(process) {
        this.processTable.push(process);
        this.processMap.set(process.id, process);
    }
    unregisterProcess(process) {
        this.processTable = ___namespace.pull(this.processTable, process);
        this.processMap.delete(process.id);
    }
    outOfTime(process) {
        process.skip();
        this.ranOutOfTime++;
    }
    isOutOfTime() {
        const timeUsed = Game.cpu.getUsed();
        if (timeUsed >= this.timeLimit) {
            return true;
        }
        return false;
    }
    getOutOfTimeCount() {
        return this.ranOutOfTime;
    }
    hasProcess(id) {
        return this.processMap.has(id);
    }
    getProcess(id) {
        return this.processMap.get(id);
    }
    getProcesses() {
        return this.processTable;
    }
    updateTimeLimit() {
        const limit = Game.cpu.limit;
        const bucket = Game.cpu.bucket;
        this.timeLimit = limit * ___namespace.max([0.5, 1 - 10000 / bucket * 0.05]);
        if (this.cpuThrottle > 0 && this.timeLimit > this.cpuThrottle) {
            this.timeLimit = this.cpuThrottle;
        }
    }
    tick(kernel, trace) {
        trace = trace.begin('scheduler_tick');
        // Time limit change between ticks
        this.updateTimeLimit();
        // Sort process table priority
        // -1 should maintain the same order
        this.processTable = ___namespace.sortByAll(this.processTable, ['adjustedPriority', 'skippable', 'lastRun']);
        const toRemove = [];
        const processCpu = new Map();
        if (Game.cpu.bucket < 1000) {
            trace.notice('low bucket, running only critical processes', { bucket: Game.cpu.bucket });
        }
        // Iterate processes and act on their status
        this.processTable.forEach((process) => {
            const processTrace = trace.as(process.type).withFields(new Map([['pid', process.id]]));
            // If bucket is low only run the most critical processes, should keep the bucket away from 0
            if (Game.cpu.bucket < 1000 && process.priority >= LOW_BUCKET_MIN_PRIORITY) {
                this.outOfTime(process);
                trace.getMetricsCollector().counter('scheduler_out_of_time_total', 1);
                return;
            }
            // If tick out of time, skip if process can be skipped
            if (this.isOutOfTime() && process.canSkip()) {
                this.outOfTime(process);
                trace.getMetricsCollector().counter('scheduler_out_of_time_total', 1);
                return;
            }
            if (process.isRunning()) {
                trace.getMetricsCollector().counter('scheduler_processes_running_total', 1, { process_type: process.type });
                const startProcessCpu = Game.cpu.getUsed();
                try {
                    process.run(kernel, processTrace);
                }
                catch (e) {
                    processTrace.error('process error', { id: process.id, error: e.stack });
                    trace.getMetricsCollector().counter('scheduler_process_error_total', 1, { process_type: process.type });
                }
                const processTime = Game.cpu.getUsed() - startProcessCpu;
                trace.getMetricsCollector().counter('scheduler_cpu_time_total', processTime, { process_type: process.type });
                // We want to report slow processes
                if (processTime > this.slowProcessThreshold) {
                    processTrace.warn(`slow process - ${processTrace.name}`, { id: process.id, type: process.type, time: processTime });
                    trace.getMetricsCollector().counter('scheduler_slow_processes_total', 1, { process_type: process.type });
                }
                // Track time spent on each process by type
                if (!processCpu.has(process.type)) {
                    processCpu.set(process.type, 0);
                }
                processCpu.set(process.type, processCpu.get(process.type) + processTime);
            }
            else if (process.isSleeping()) {
                if (process.shouldWake()) {
                    process.setRunning();
                }
                trace.getMetricsCollector().counter('scheduler_processes_sleeping_total', 1, { process_type: process.type });
            }
            else if (process.isTerminated()) {
                toRemove.push(process);
                trace.getMetricsCollector().counter('scheduler_processes_terminated_total', 1, { process_type: process.type });
            }
        });
        toRemove.forEach((remove) => {
            this.unregisterProcess(remove);
        });
        trace.end();
    }
}

const RUNNING = 'running';
const SUCCESS = 'success';
const FAILURE = 'failure';
const rootNode = (id, behavior) => {
    return function (creep, trace, kernel) {
        const rootTrace = trace.begin(id);
        const result = behavior.tick(creep, rootTrace, kernel);
        trace.info('root result', { id });
        if (result === FAILURE) {
            trace.info('root failure', { id });
        }
        rootTrace.end();
    };
};
const selectorNode = (id, children) => {
    return {
        id,
        children,
        tickChildren: function (creep, trace, kernel) {
            let i = getState(creep, this.id);
            setState(creep, this.id, 0);
            for (; i < this.children.length; i++) {
                const result = this.children[i].tick(creep, trace, kernel);
                trace.log('result', { result });
                switch (result) {
                    case RUNNING:
                        setState(creep, this.id, i);
                        return RUNNING;
                    case FAILURE:
                        continue;
                    case SUCCESS:
                        return SUCCESS;
                }
            }
            return FAILURE;
        },
        tick: function (creep, trace, kernel) {
            trace = trace.begin(this.id);
            trace.info('tick', { id: this.id });
            const result = this.tickChildren(creep, trace, kernel);
            trace.info('result', result);
            trace.end();
            return result;
        },
        clear: function (creep, trace) {
            clearState(creep, this.id);
            this.children.forEach((child) => {
                child.clear(creep, trace);
            });
        },
    };
};
const sequenceNode = (id, children) => {
    return {
        id,
        children,
        tickChildren: function (creep, trace, kernel) {
            let i = getState(creep, this.id);
            for (; i < this.children.length; i++) {
                const result = this.children[i].tick(creep, trace, kernel);
                switch (result) {
                    case RUNNING:
                        setState(creep, this.id, i);
                        return RUNNING;
                    case FAILURE:
                        return FAILURE;
                    case SUCCESS:
                        continue;
                }
            }
            return SUCCESS;
        },
        tick: function (creep, trace, kernel) {
            trace = trace.begin(this.id);
            const result = this.tickChildren(creep, trace, kernel);
            trace.info('result', result);
            trace.end();
            return result;
        },
        clear: function (creep, trace) {
            clearState(creep, this.id);
            this.children.forEach((child) => {
                child.clear(creep, trace);
            });
        },
    };
};
const sequenceAlwaysNode = (id, children) => {
    return {
        id,
        children,
        tickChildren: function (creep, trace, kernel) {
            for (let i = 0; i < this.children.length; i++) {
                const result = this.children[i].tick(creep, trace, kernel);
                switch (result) {
                    case RUNNING:
                        return RUNNING;
                    case FAILURE:
                        return FAILURE;
                    case SUCCESS:
                        continue;
                }
            }
            return SUCCESS;
        },
        tick: function (creep, trace, kernel) {
            trace = trace.begin(this.id);
            const result = this.tickChildren(creep, trace, kernel);
            trace.info('result', result);
            trace.end();
            return result;
        },
        clear: function (creep, trace) {
            clearState(creep, this.id);
            this.children.forEach((child) => {
                child.clear(creep, trace);
            });
        },
    };
};
const repeatUntilFailure = (id, node) => {
    return {
        id,
        node,
        tick: function (creep, trace, kernel) {
            trace = trace.begin(this.id);
            const result = this.node.tick(creep, trace, kernel);
            trace.info('result', result);
            trace.end();
            if (result === FAILURE) {
                return FAILURE;
            }
            return RUNNING;
        },
        clear: function (creep, trace) {
            clearState(creep, this.id);
            this.node.clear(creep, trace);
        },
    };
};
const repeatUntilSuccess = (id, node) => {
    return {
        id,
        node,
        tick: function (creep, trace, kernel) {
            trace = trace.begin(this.id);
            const result = this.node.tick(creep, trace, kernel);
            trace.info('result', result);
            trace.end();
            if (result === SUCCESS) {
                return SUCCESS;
            }
            return RUNNING;
        },
        clear: function (creep, trace) {
            clearState(creep, this.id);
            this.node.clear(creep, trace);
        },
    };
};
const runUntilConditionMet = (id, condition, node) => {
    return {
        id,
        node,
        condition,
        tick: function (creep, trace, kernel) {
            trace = trace.begin(this.id);
            // if previous tick set that it was running, then skip condition check
            const isRunning = getState(creep, this.id);
            // clear running state, so we can set it again if needed
            clearState(creep, this.id);
            if (isRunning) {
                trace.info('previous tick returned running, so dont check condition and tick node');
            }
            else {
                trace.info('checking condition', { id: this.id });
                const conditionResult = this.condition(creep, trace, kernel);
                if (conditionResult) {
                    // We made it through the condition, clear that branches state so that
                    // next time around we start from scratch
                    this.clear(creep, trace);
                    trace.info('condition met', { id: this.id });
                    trace.end();
                    // We are done, move on
                    return SUCCESS;
                }
                trace.info('condition not met, tick node', { id: this.id });
            }
            const result = this.node.tick(creep, trace, kernel);
            trace.info('result', { result });
            if (result === FAILURE) {
                return FAILURE;
            }
            // if result is running then set state to running
            if (result === RUNNING) {
                setState(creep, this.id, true);
            }
            // if the condition is not met and it didnt fail, we return running
            return RUNNING;
        },
        clear: function (creep, trace) {
            clearState(creep, this.id);
            this.node.clear(creep, trace);
        },
    };
};
const repeatUntilConditionMet = (id, condition, node) => {
    return {
        id,
        node,
        condition,
        tick: function (creep, trace, kernel) {
            trace = trace.begin(this.id);
            trace.info('checking condition', { id: this.id });
            const conditionResult = this.condition(creep, trace, kernel);
            if (!conditionResult) {
                trace.info('condition not met', { id: this.id });
                const result = this.node.tick(creep, trace, kernel);
                trace.info('result', { result });
                trace.end();
                if (result === FAILURE) {
                    return FAILURE;
                }
                return RUNNING;
            }
            // We made it through the condition, clear that branches state so that
            // next time around we start from scratch
            this.clear(creep, trace);
            trace.info('condition met', { id: this.id });
            trace.end();
            return SUCCESS;
        },
        clear: function (creep, trace) {
            clearState(creep, this.id);
            this.node.clear(creep, trace);
        },
    };
};
const leafNode = (id, behavior) => {
    return {
        id,
        behavior,
        tick: function (creep, trace, kernel) {
            trace = trace.begin(this.id);
            const result = this.behavior(creep, trace, kernel);
            trace.info('result', result);
            trace.end();
            return result;
        },
        clear: function (creep, trace) {
            clearState(creep, this.id);
        },
    };
};
function getState(creep, id, trace) {
    let state = 0;
    if (creep.memory[id]) {
        state = creep.memory[id];
    }
    // Avoids state getting stuck on previous value, next
    // time we enter the node it will be 0 unless override it
    clearState(creep, id);
    // trace.log('get state (clears existing state)', {id, state});
    return state;
}
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function setState(creep, id, state, _trace) {
    creep.memory[id] = state;
    // trace.log('set state', {id, state});
}
function clearState(creep, id, _trace) {
    // trace.log('clear state', {id, state: creep.memory[id]});
    delete creep.memory[id];
}

const commonPolicy = {
    room: {
        avoidHostileRooms: true,
        avoidFriendlyRooms: false,
        avoidRoomsWithKeepers: true,
        avoidRoomsWithTowers: true,
        avoidUnloggedRooms: false,
        sameRoomStatus: true,
        costMatrixType: AllowedCostMatrixTypes.COMMON,
    },
    destination: {
        range: 1,
    },
    path: {
        allowIncomplete: true,
        maxSearchRooms: 12,
        maxOps: 6000,
        maxPathRooms: 6,
        ignoreCreeps: true,
        sourceKeeperBuffer: 3,
    },
};
const singleRoomPolicy = {
    room: {
        avoidHostileRooms: true,
        avoidFriendlyRooms: false,
        avoidRoomsWithKeepers: true,
        avoidRoomsWithTowers: false,
        avoidUnloggedRooms: false,
        sameRoomStatus: true,
        costMatrixType: AllowedCostMatrixTypes.SINGLE_ROOM_COMMON,
    },
    destination: {
        range: 1,
    },
    path: {
        allowIncomplete: true,
        maxSearchRooms: 1,
        maxOps: 1000,
        maxPathRooms: 1,
        ignoreCreeps: true,
    },
};
const roadPolicy = {
    room: {
        avoidHostileRooms: true,
        avoidFriendlyRooms: false,
        avoidRoomsWithKeepers: true,
        avoidRoomsWithTowers: false,
        avoidUnloggedRooms: false,
        sameRoomStatus: true,
        costMatrixType: AllowedCostMatrixTypes.SOURCE_ROAD,
    },
    destination: {
        range: 1,
    },
    path: {
        allowIncomplete: true,
        maxSearchRooms: 12,
        maxOps: 5000,
        maxPathRooms: 6,
        ignoreCreeps: true,
        controllerBuffer: 3,
    },
};
const controllerRoadPolicy = {
    room: {
        avoidHostileRooms: true,
        avoidFriendlyRooms: false,
        avoidRoomsWithKeepers: true,
        avoidRoomsWithTowers: false,
        avoidUnloggedRooms: false,
        sameRoomStatus: true,
        costMatrixType: AllowedCostMatrixTypes.SOURCE_ROAD,
    },
    destination: {
        range: 2,
    },
    path: {
        allowIncomplete: true,
        maxSearchRooms: 12,
        maxOps: 5000,
        maxPathRooms: 6,
        ignoreCreeps: true,
    },
};

const getPath = (kernel, origin, destination, policy, trace) => {
    const pathDetails = {
        tries: 1,
        passes: 0,
        searchedRooms: new Set(),
        rejectedRooms: new Map(),
        blockedRooms: new Set(),
        incompletePaths: [],
    };
    for (; pathDetails.passes < pathDetails.tries; pathDetails.passes++) {
        // Get list of rooms on the way to destination
        const roomRoute = Game.map.findRoute(origin.roomName, destination.roomName, {
            routeCallback: getRoomRouteCallback(kernel, origin.roomName, destination.roomName, policy.room, pathDetails),
        });
        // If we have no route, return null
        if (roomRoute === ERR_NO_PATH) {
            trace.info('no room path', { origin: origin.roomName, destination: destination.roomName });
            return [null, pathDetails];
        }
        // Map findRoute results to map of names for fast lookup
        const rooms = _.values(roomRoute);
        const allowedRooms = _.reduce(rooms, (acc, entry) => {
            acc.set(entry.room, true);
            return acc;
        }, new Map());
        // Add origin room to list of allowed room
        allowedRooms.set(origin.roomName, true);
        const goal = {
            pos: destination,
            range: policy.destination.range,
        };
        const opts = {
            maxRooms: policy.path.maxSearchRooms,
            roomCallback: getRoomCallback(kernel, origin.roomName, destination.roomName, policy.path, policy.room, allowedRooms, pathDetails, trace),
            maxOps: policy.path.maxOps,
            plainCost: policy.path.plainCost || 2,
            swampCost: policy.path.swampCost || 5,
        };
        trace.info('findPath', { origin, goal, opts });
        const result = PathFinder.search(origin, goal, opts);
        // If route is complete or we don't care, go with it
        if (result.incomplete === false || policy.path.allowIncomplete) {
            trace.info('path found', { path: result.path });
            return [result, pathDetails];
        }
        // If route has no where to go, fail
        if (result.path.length <= 1) {
            trace.info('path length less than 1', { origin, goal, opts });
            return [null, pathDetails];
        }
        const lastRoom = result.path[result.path.length - 1].roomName;
        if (lastRoom === destination.roomName) {
            trace.info('last room is destination', { origin, goal, opts });
            return [null, pathDetails];
        }
        // Add last room to blocked and try again if allowed
        if (lastRoom != destination.roomName) {
            trace.info('last room not destination', { origin, goal, opts });
            pathDetails.blockedRooms.add(lastRoom);
            pathDetails.incompletePaths.push(result);
        }
    }
    trace.info('passes exhausted', { origin });
    return [null, pathDetails];
};
const getClosestBaseByPath = (kernel, destination, policy, trace) => {
    const roomEntry = kernel.getScribe().getRoomById(destination.roomName);
    let selectedBase = null;
    let selectedPathLength = 99999;
    // Get colonies and filter by the policy
    let bases = kernel.getPlanner().getBases();
    bases = applyAllowedBasePolicy(bases, roomEntry, policy.base, trace);
    // Iterate colonies and find the closest one within the policies
    bases.forEach((config) => {
        // Get the origin position from the base by apply the base policy
        const originPosition = getOriginPosition(kernel, config, policy.base);
        if (!originPosition) {
            trace.error('no origin position', { config, policy });
            return;
        }
        // Find the path from the origin to the destination
        const [result, debug] = getPath(kernel, originPosition, destination, policy, trace);
        if (!result) {
            trace.info('no path', { config, policy, debug });
            return;
        }
        // If the path is longer then the current selection, skip
        if (result.path.length > selectedPathLength) {
            trace.info('path too long', { config, policy, debug });
            return;
        }
        // If path has more rooms then allowed, skip
        const roomsInPath = _.uniq(result.path.map((pos) => pos.roomName));
        if (roomsInPath.length > policy.path.maxPathRooms) {
            trace.info('path has too many rooms long', { config, policy, debug });
            return false;
        }
        // Update the selected base and path
        selectedBase = config;
        selectedPathLength = result.path.length;
    });
    return selectedBase;
};
const applyAllowedBasePolicy = (bases, destRoomEntry, policy, trace) => {
    // Do not search colonies below the minimum level
    if (policy.minRoomLevel) {
        bases = bases.filter((config) => {
            const room = Game.rooms[config.primary];
            if (!room) {
                trace.warn('room not found', { room: config.primary });
                return false;
            }
            return _.get(config, 'controller.level', 0) >= policy.minRoomLevel;
        });
    }
    // Do not search colonies further than the max linear distance
    if (policy.maxLinearDistance) {
        bases = bases.filter((config) => {
            return Game.map.getRoomLinearDistance(destRoomEntry.id, config.primary) <= policy.maxLinearDistance;
        });
    }
    trace.info('filtered colonies', { colonies: bases.map((base) => base.id) });
    return bases;
};
const getOriginPosition = (_kernel, base, policy, _trace) => {
    if (policy.start === 'spawn') {
        return base.origin;
    }
    return null;
};
const getRoomRouteCallback = (kernel, originRoom, destRoom, policy, searchDetails, trace) => {
    return (toRoom, _fromRoom) => {
        searchDetails.searchedRooms.add(toRoom);
        // Always allow entry to destination room
        if (destRoom === toRoom) {
            return 1;
        }
        // Always allow movement in origin room
        if (originRoom === toRoom) {
            return 1;
        }
        if (searchDetails.blockedRooms.has(toRoom)) {
            return Infinity;
        }
        const roomEntry = kernel.getScribe().getRoomById(toRoom);
        // If we have not scanned the room, dont enter it
        if (!roomEntry && policy.avoidUnloggedRooms) {
            return Infinity;
        }
        if (roomEntry) {
            const [allow, reason] = applyRoomCallbackPolicy(kernel, roomEntry, policy);
            if (!allow) {
                searchDetails.rejectedRooms.set(toRoom, reason);
                return Infinity;
            }
        }
        return 1;
    };
};
const getRoomCallback = (kernel, originRoom, destRoom, pathPolicy, roomPolicy, allowedRooms, pathDetails, trace) => {
    return (roomName) => {
        // If this room is not destination, check if we should avoid it
        if (originRoom != roomName && destRoom !== roomName) {
            if (!allowedRooms.has(roomName)) {
                pathDetails.rejectedRooms.set(roomName, 'not in allowed rooms');
                return false;
            }
            const roomEntry = kernel.getScribe().getRoomById(roomName);
            // If we have not scanned the room, dont enter it
            if (!roomEntry && roomPolicy.avoidUnloggedRooms) {
                pathDetails.rejectedRooms.set(roomName, 'unlogged');
                return false;
            }
            if (roomEntry) {
                const [allow, reason] = applyRoomCallbackPolicy(kernel, roomEntry, roomPolicy);
                if (!allow) {
                    pathDetails.rejectedRooms.set(roomName, reason);
                    return false;
                }
            }
        }
        // Fetch cached cost matrix for the room
        let costMatrix = kernel.getCostMatrixCache().getCostMatrix(kernel, roomName, roomPolicy.costMatrixType, trace);
        costMatrix = costMatrix.clone();
        const room = Game.rooms[roomName];
        if (room) {
            // Mark creeps as not passible
            if (!pathPolicy.ignoreCreeps) {
                room.find(FIND_CREEPS).forEach((creep) => {
                    costMatrix.set(creep.pos.x, creep.pos.y, 255);
                });
            }
            // add road construction sites
            if (pathPolicy.preferRoadSites) {
                room.find(FIND_MY_CONSTRUCTION_SITES).forEach((site) => {
                    if (site.structureType === STRUCTURE_ROAD) {
                        costMatrix.set(site.pos.x, site.pos.y, 1);
                    }
                });
            }
            // Add a buffer around source keepers
            if (pathPolicy.sourceKeeperBuffer > 0) {
                room.find(FIND_HOSTILE_CREEPS, {
                    filter: (creep) => {
                        return creep.owner.username === 'Source Keeper';
                    },
                }).forEach((sourceKeeper) => {
                    getNearbyPositions(sourceKeeper.pos, pathPolicy.sourceKeeperBuffer).forEach((pos) => {
                        costMatrix.set(pos.x, pos.y, 10);
                    });
                });
            }
            // Add a buffer around hostile creeps
            if (pathPolicy.hostileCreepBuffer > 0) {
                room.find(FIND_HOSTILE_CREEPS, {
                    filter: (creep) => {
                        return creep.owner.username !== 'Source Keeper';
                    },
                }).forEach((hostileCreep) => {
                    getNearbyPositions(hostileCreep.pos, pathPolicy.hostileCreepBuffer).forEach((pos) => {
                        costMatrix.set(pos.x, pos.y, 10);
                    });
                });
            }
            // Add a buffer around the controller
            if (pathPolicy.controllerBuffer > 0) {
                const controller = room.controller;
                if (controller && controller.my) {
                    getNearbyPositions(controller.pos, pathPolicy.controllerBuffer).forEach((pos) => {
                        costMatrix.set(pos.x, pos.y, 10);
                    });
                }
            }
        }
        return costMatrix;
    };
};
const applyRoomCallbackPolicy = (kernel, roomEntry, policy, _trace) => {
    var _a, _b;
    const owner = (_a = roomEntry.controller) === null || _a === void 0 ? void 0 : _a.owner;
    const claimedByMe = owner === kernel.getPlanner().getUsername() && ((_b = roomEntry.controller) === null || _b === void 0 ? void 0 : _b.level) > 0;
    const ownerIsNotMe = owner !== kernel.getPlanner().getUsername();
    const isFriendly = kernel.getFriends().includes(owner);
    // If owner is not me, and we want to avoid friendly rooms and is friendly, dont enter
    if (owner && !claimedByMe && policy.avoidFriendlyRooms) {
        return [false, 'friendly'];
    }
    // if not claimed by me and hostiles are present, dont enter
    if (policy.avoidHostileRooms && !claimedByMe &&
        ((roomEntry.hasHostiles && roomEntry.hostilesDmg > 0) || roomEntry.hasKeepers)) {
        return [false, 'hostile'];
    }
    // If owner is not me and has hostile creeps, dont enter
    if (policy.avoidHostileRooms && owner && ownerIsNotMe && !isFriendly &&
        (roomEntry.hasHostiles && roomEntry.hostilesDmg > 0)) {
        return [false, 'hostile'];
    }
    // if room has towers and the room is not mine, dont enter
    if (policy.avoidRoomsWithTowers && roomEntry.numTowers && owner && ownerIsNotMe) {
        return [false, 'towers'];
    }
    // if has Keepers, dont enter
    if (policy.avoidRoomsWithKeepers && roomEntry.hasKeepers) {
        return [false, 'keepers'];
    }
    if (policy.sameRoomStatus) ;
    // trace.info('room allowed', {roomName: roomEntry.id});
    return [true, 'good'];
};
const visualizePath = (path, _trace) => {
    const pathByRooms = path.reduce((acc, pos) => {
        if (!acc.has(pos.roomName)) {
            acc.set(pos.roomName, []);
        }
        acc.get(pos.roomName).push(pos);
        return acc;
    }, new Map());
    // Display in the rooms
    Array.from(pathByRooms.entries()).forEach(([key, value]) => {
        new RoomVisual(key).poly(value);
    });
};

const MAX_POSITION_TTL = 5;
const MEMORY_MOVE_POS_TTL = 'move_pos_ttl';
const MEMORY_MOVE_PREV_POS = 'move_prev_pos';
const MEMORY_MOVE_STUCK_COUNT = 'move_stuck_count';
const PATH_ORIGIN_KEY = 'path_origin_id';
const PATH_DESTINATION_KEY = 'path_dest_key';
const getMoveOpts = (ignoreCreeps = false, reusePath = 100, maxOps = 2000, range = 0) => {
    return { reusePath, maxOps, ignoreCreeps, range };
};
const moveToMemory = (creep, memoryId, range, ignoreCreeps, reusePath, maxOps) => {
    const destination = Game.getObjectById(creep.memory[memoryId]);
    if (!destination) {
        return FAILURE;
    }
    return moveTo(creep, destination.pos, range, ignoreCreeps, reusePath, maxOps);
};
const moveTo = (creep, destination, range, ignoreCreeps, reusePath, maxOps) => {
    if (creep.pos.inRangeTo(destination, range)) {
        return SUCCESS;
    }
    const moveOpts = getMoveOpts(ignoreCreeps, reusePath, maxOps, range);
    const result = creep.moveTo(destination, moveOpts);
    if (result === ERR_NO_PATH) {
        // Clear existing path so we build a new one
        delete creep.memory['_move'];
        return RUNNING;
    }
    if (result !== OK && result !== ERR_TIRED) {
        return FAILURE;
    }
    return RUNNING;
};
const setDestination = (creep, destinationId, roomId = null, shardName = null) => {
    creep.memory[MEMORY_DESTINATION] = destinationId;
    if (roomId) {
        creep.memory[MEMORY_DESTINATION_ROOM] = roomId;
    }
    if (!shardName) {
        shardName = Game.shard.name;
    }
    creep.memory[MEMORY_DESTINATION_SHARD] = shardName;
};
const isStuck = (creep) => {
    const wasStuck = creep.memory[MEMORY_MOVE_STUCK_COUNT] || 0;
    if (wasStuck > 0) {
        creep.memory[MEMORY_MOVE_STUCK_COUNT] = wasStuck - 1;
        return true;
    }
    let prevPos = creep.memory[MEMORY_MOVE_PREV_POS] || null;
    if (prevPos != null) {
        prevPos = new RoomPosition(prevPos.x, prevPos.y, prevPos.roomName);
    }
    // Compare current and last and increase stuck ttl
    if (prevPos && creep.pos.isEqualTo(prevPos)) {
        if (!creep.memory[MEMORY_MOVE_POS_TTL]) {
            creep.memory[MEMORY_MOVE_POS_TTL] = 0;
        }
        creep.memory[MEMORY_MOVE_POS_TTL] += 1;
        // Creep is stuck, clear cached path, and get new path factoring in creeps
        if (creep.memory[MEMORY_MOVE_POS_TTL] > MAX_POSITION_TTL) {
            clearMovementCache(creep);
            creep.memory[MEMORY_MOVE_STUCK_COUNT] = 5;
            return true;
        }
    }
    else {
        // Creep is not stuck, update previous position and clear ttl
        creep.memory[MEMORY_MOVE_PREV_POS] = creep.pos;
        creep.memory[MEMORY_MOVE_POS_TTL] = 0;
    }
    return false;
};
const getDestinationFromPosInMemory = (creep, memoryId) => {
    const positionString = creep.memory[memoryId];
    if (!positionString) {
        return null;
    }
    const posArray = positionString.split(',');
    if (!posArray || posArray.length !== 3) {
        return null;
    }
    return new RoomPosition(posArray[0], posArray[1], posArray[2]);
};
const getDestinationFromMemory = (creep, memoryId) => {
    const destId = creep.memory[memoryId];
    if (!destId) {
        return null;
    }
    //
    const dest = Game.getObjectById(destId);
    if (!dest) {
        return null;
    }
    return dest.pos;
};
const getAndSetCreepPath = (kernel, pathCache, creep, destination, range, policy, trace) => {
    const path = pathCache.getPath(kernel, creep.pos, destination, range, policy, trace);
    const originKey = pathCache.getKey(creep.pos, 0);
    const destKey = pathCache.getKey(destination, range);
    return [path, originKey, destKey];
};
const clearMovementCache = (creep) => {
    delete creep.memory['_move'];
    delete creep.memory[PATH_ORIGIN_KEY];
    delete creep.memory[PATH_DESTINATION_KEY];
};
const updateCreepCachedPath = (kernel, creep, destination, range, policy, trace) => {
    const pathCache = kernel.getPathCache();
    let path = null;
    let originKey = creep.memory[PATH_ORIGIN_KEY] || null;
    let destKey = creep.memory[PATH_DESTINATION_KEY] || null;
    trace.info('keys', { originKey, destKey });
    if (originKey && destKey) {
        path = pathCache.getCachedPath(originKey, destKey, trace);
    }
    if (path) {
        trace.info('heap cache hit', { originKey, destKey });
        return path.value;
    }
    if (!path) {
        trace.info('heap cache miss', { originKey, destKey });
        const getSetResult = getAndSetCreepPath(kernel, pathCache, creep, destination, range, policy, trace);
        originKey = getSetResult[1];
        destKey = getSetResult[2];
        trace.info('setting keys', { originKey, destKey });
        creep.memory[PATH_ORIGIN_KEY] = originKey;
        creep.memory[PATH_DESTINATION_KEY] = destKey;
        return getSetResult[0];
    }
    trace.info('missing path', { originKey, destKey });
    clearMovementCache(creep);
    return null;
};
const cachedMoveToMemoryPos = (memoryId, range = 1, policy) => {
    return leafNode('cached_move_to_position', (creep, trace, kingdom) => {
        const destination = getDestinationFromPosInMemory(creep, memoryId);
        if (!destination) {
            clearMovementCache(creep);
            creep.say('📍❓');
            trace.error('missing creep destination', { memoryId });
            return FAILURE;
        }
        return cachedMoveToPosition(kingdom, creep, destination, range, policy, trace);
    });
};
const cachedMoveToRoom = (memoryId, policy) => {
    return leafNode('cached_move_to_position', (creep, trace, kingdom) => {
        const destinationRoom = creep.memory[memoryId];
        if (!destinationRoom) {
            clearMovementCache(creep);
            creep.say('📍❓');
            trace.error('missing creep destination', { memoryId });
            return FAILURE;
        }
        if (creep.pos.roomName === destinationRoom) {
            return SUCCESS;
        }
        const destination = new RoomPosition(25, 25, destinationRoom);
        return cachedMoveToPosition(kingdom, creep, destination, 25, policy, trace);
    });
};
const cachedMoveToMemoryObjectId = (memoryId, range = 1, policy) => {
    return leafNode('cached_move_to_object_id', (creep, trace, kingdom) => {
        const destination = getDestinationFromMemory(creep, memoryId);
        if (!destination) {
            clearMovementCache(creep);
            creep.say('📍❓');
            trace.error('missing creep destination', { id: creep.memory[memoryId] });
            return FAILURE;
        }
        return cachedMoveToPosition(kingdom, creep, destination, range, policy, trace);
    });
};
const cachedMoveToPosition = (kernel, creep, destination, range = 1, policy, trace) => {
    // Check if creep has arrived
    if (creep.pos.inRangeTo(destination, range)) {
        clearMovementCache(creep);
        trace.info('reached destination', { destination, range });
        return SUCCESS;
    }
    const stuck = isStuck(creep);
    let result = null;
    if (!stuck) {
        const pathfinderResult = updateCreepCachedPath(kernel, creep, destination, range, policy, trace);
        trace.info('pathfinder result', { result: pathfinderResult, creepName: creep.name, destination, range, policy });
        if (!pathfinderResult) {
            creep.say('🚧');
            trace.error('no path found', { destination, range });
            return FAILURE;
        }
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        if (global.LOG_WHEN_PID === creep.name) {
            visualizePath(pathfinderResult.path);
        }
        result = creep.moveByPath(pathfinderResult.path);
        trace.info('move by path result', { result, path: pathfinderResult.path });
    }
    else {
        const moveOpts = getMoveOpts(false, 50, policy.path.maxOps);
        result = creep.moveTo(destination, moveOpts);
        trace.info('stuck move result', { result, origin: creep.pos, destination });
    }
    if (result === ERR_NO_PATH) {
        // Clear existing path so we build a new one
        clearMovementCache(creep);
        trace.info('move by result no path', { result });
        return RUNNING;
    }
    if (result !== OK && result !== ERR_TIRED) {
        clearMovementCache(creep);
        creep.say('⛔️');
        trace.error('move by result not OK', { result });
        return FAILURE;
    }
    if (result === ERR_TIRED) {
        creep.memory[MEMORY_MOVE_POS_TTL] -= 1;
    }
    return RUNNING;
};
const moveToCreepMemory = (memoryID, range = 1, ignoreCreeps, reusePath, maxOps) => {
    return leafNode('bt.movement.moveToCreepMemory', (creep) => {
        return moveToMemory(creep, memoryID, range, ignoreCreeps, reusePath, maxOps);
    });
};
const fillCreepFromDestination = (creep, trace) => {
    const destinationMemory = creep.memory[MEMORY_DESTINATION];
    const destination = Game.getObjectById(creep.memory[MEMORY_DESTINATION]);
    if (!destination) {
        trace.log('could not find destination', { destinationMemory });
        return FAILURE;
    }
    const resource = creep.memory[MEMORY_HAUL_RESOURCE] || RESOURCE_ENERGY;
    let amount = creep.memory[MEMORY_HAUL_AMOUNT] || undefined;
    if (amount > creep.store.getFreeCapacity(resource)) {
        amount = creep.store.getFreeCapacity(resource);
    }
    if (!destination.store) {
        const result = creep.pickup(destination);
        if (result !== OK) {
            trace.error('failed to pick up resource', { result });
            return FAILURE;
        }
        return FAILURE;
    }
    if (amount > destination.store.getUsedCapacity(resource)) {
        amount = destination.store.getUsedCapacity(resource);
    }
    if (amount === 0) {
        trace.log('amount is 0');
        return FAILURE;
    }
    // If we are seeing a specific amount, we are done when we have that amount in the hold
    if (amount && creep.store.getUsedCapacity(resource) >= amount) {
        trace.log('success: have amount we are looking for', { amount, resource });
        return SUCCESS;
    }
    const result = creep.withdraw(destination, resource, amount);
    trace.log('widthdrawl result', { result, destinationId: destination.id, resource, amount });
    if (result === OK) {
        return RUNNING;
    }
    if (result === ERR_FULL) {
        return SUCCESS;
    }
    if (result === ERR_NOT_ENOUGH_RESOURCES) {
        return SUCCESS;
    }
    return FAILURE;
};
const moveToShard = (shardMemoryKey) => {
    return repeatUntilConditionMet('moveToShard', (creep, _trace, _kingdom) => {
        const destinationShardName = creep.memory[shardMemoryKey];
        // If creep doesn't have a harvest room assigned, we are done
        if (!destinationShardName) {
            return true;
        }
        // If the creep reaches the room we are done
        if (Game.shard.name === destinationShardName) {
            return true;
        }
        return false;
    }, sequenceNode('select_and_enter_portal', [
        leafNode('select_portal', (creep, trace, kingdom) => {
            const destinationShardName = creep.memory[shardMemoryKey];
            // Lookup closest portal, use path cache
            let portals = kingdom.getScribe().getPortals(destinationShardName);
            if (!portals || !portals.length) {
                trace.info('unable to find portal', { shardName: destinationShardName });
                return FAILURE;
            }
            portals = _.filter(portals, (portal) => {
                return Game.map.getRoomLinearDistance(portal.pos.roomName, creep.room.name) < 4;
            });
            const portal = portals[0] || null;
            if (!portal) {
                trace.info('unable to find nearby portal', { shardName: destinationShardName });
                return FAILURE;
            }
            // Store creep and set portal as destination
            kingdom.getScribe().setCreepBackup(creep);
            creep.memory[MEMORY_DESTINATION_POS] = [portal.pos.x, portal.pos.y, portal.pos.roomName].join(',');
            return SUCCESS;
        }),
        cachedMoveToMemoryPos(MEMORY_DESTINATION_POS, 0, commonPolicy),
    ]));
};

const BOOST_PHASE = 'boost_phase';
const BOOST_PHASE_START = 'boosting_start';
const BOOST_PHASE_MOVE = 'boosting_move';
const BOOST_PHASE_READY = 'boosting_ready';
const BOOST_PHASE_DONE = 'boosting_done';
const behaviorBoosts = (behaviorNode) => {
    return sequenceAlwaysNode('boosting', [
        leafNode('get_boosted', (creep, trace, kernel) => {
            const phase = creep.memory[BOOST_PHASE] || BOOST_PHASE_START;
            if (phase === BOOST_PHASE_DONE) {
                trace.info('boosting complete');
                return SUCCESS;
            }
            // Mark done of no requested boosts
            const desiredActions = creep.memory[DESIRED_BOOSTS] || [];
            if (!desiredActions.length) {
                trace.info('no requested action boosts');
                creep.memory[BOOST_PHASE] = BOOST_PHASE_DONE;
                return SUCCESS;
            }
            const base = getCreepBase(kernel, creep);
            if (!base) {
                trace.error('no base config', { creep: creep.name });
                creep.memory[BOOST_PHASE] = BOOST_PHASE_DONE;
                return SUCCESS;
            }
            const boosterPos = getBoostPosition(base);
            if (!boosterPos) {
                trace.info('no booster position');
                creep.memory[BOOST_PHASE] = BOOST_PHASE_DONE;
                return SUCCESS;
            }
            switch (phase) {
                case BOOST_PHASE_START:
                    // eslint-disable-next-line no-fallthrough
                    creep.memory[BOOST_PHASE] = BOOST_PHASE_MOVE;
                // eslint-disable-next-line no-fallthrough
                case BOOST_PHASE_MOVE: {
                    // Move to booster location
                    const result = moveTo(creep, boosterPos, 0, false, 50, 1000);
                    if (result === SUCCESS) {
                        creep.memory[BOOST_PHASE] = BOOST_PHASE_READY;
                        return RUNNING;
                    }
                    return result;
                }
                case BOOST_PHASE_READY:
                    desiredActions.forEach((desiredAction) => {
                        const labs = getLabsForAction(base, desiredAction);
                        if (!labs.length) {
                            trace.info('no labs for action', { action: desiredAction });
                            return;
                        }
                        const lab = labs[0];
                        const result = lab.boostCreep(creep);
                        trace.info('boosted', {
                            labId: lab.id,
                            compound: lab.mineralType,
                            result,
                        });
                    });
                    creep.memory[BOOST_PHASE] = BOOST_PHASE_DONE;
                    return RUNNING;
                default:
                    throw new Error(`Unknown boost phase: ${phase}`);
            }
        }),
        behaviorNode,
    ]);
};

const ROLE_HARASSER = 'harasser';
const MEMORY_HARASS_BASE = 'harasser.target_base';
const MEMORY_HARASS_ROOMS = 'harasser.target_rooms';
const MEMORY_HARASS_VISITED = 'harasser.visited';
const MEMORY_HARASS_CURRENT = 'harasser.current';
const MEMORY_HARASS_CREEP_TARGET = 'harasser.creep_target';
const harasserPolicy = {
    room: {
        avoidHostileRooms: false,
        avoidFriendlyRooms: false,
        avoidRoomsWithKeepers: false,
        avoidRoomsWithTowers: true,
        avoidUnloggedRooms: false,
        sameRoomStatus: true,
        costMatrixType: AllowedCostMatrixTypes.COMMON,
    },
    destination: {
        range: 1,
    },
    path: {
        allowIncomplete: true,
        maxSearchRooms: 12,
        maxOps: 5000,
        maxPathRooms: 6,
        ignoreCreeps: true,
        sourceKeeperBuffer: 3,
    },
};
const HarasserDefinition = {
    base: [MOVE, RANGED_ATTACK],
    parts: [MOVE, HEAL, MOVE, RANGED_ATTACK, MOVE, RANGED_ATTACK, MOVE, TOUGH, MOVE, RANGED_ATTACK],
    boosts: ['rangedAttack'],
    energyLimit: 0,
    energyMinimum: 550,
    ignoreSpawnEnergyLimit: false,
    processPriority: Priorities.ATTACK,
    skippable: false,
};
const behavior$d = sequenceNode('harasser_root', [
    leafNode('pick_next_room', (creep, _trace, _kernel) => {
        let targetRooms = creep.memory[MEMORY_HARASS_ROOMS];
        if (!targetRooms) {
            const targetBase = creep.memory[MEMORY_HARASS_BASE];
            targetRooms = _.values(Game.map.describeExits(targetBase));
            creep.memory[MEMORY_HARASS_ROOMS] = _.shuffle(targetRooms);
        }
        // Depending on the number of target rooms we have visited, rotate through the list
        const roomCount = creep.memory[MEMORY_HARASS_VISITED] || 0;
        creep.memory[MEMORY_HARASS_CURRENT] = targetRooms[roomCount % targetRooms.length];
        return SUCCESS;
    }),
    repeatUntilConditionMet('stop_and_attack', (creep, trace, _kernel) => {
        // if we reach the target room, move to attack phase
        if (creep.memory[MEMORY_HARASS_CURRENT] === creep.room.name) {
            return true;
        }
        const creepHeal = scoreHealing(creep, true);
        // Check for creeps weaker than us
        const hostiles = creep.room.find(FIND_HOSTILE_CREEPS, {
            filter: (hostile) => {
                return hostile.owner.username !== 'Source Keeper' &&
                    scoreAttacking(hostile) < creepHeal;
            },
        });
        trace.info(`Harasser ${creep.name} has ${hostiles.length} hostiles`);
        // If there are weaker creeps, move to attack phase
        return hostiles.length > 0;
    }, cachedMoveToRoom(MEMORY_HARASS_CURRENT, harasserPolicy)),
    leafNode('attack_phase', (creep, trace, kernel) => {
        trace.info('harass room', { room: creep.memory[MEMORY_HARASS_CURRENT] });
        // Heal if damaged and have heal parts
        if (creep.hits < creep.hitsMax && creep.getActiveBodyparts(HEAL) > 0) {
            creep.heal(creep);
        }
        const creepHeal = scoreHealing(creep, true);
        const dontAttack = kernel.getDontAttack();
        let hostiles = creep.room.find(FIND_HOSTILE_CREEPS, {
            filter: (c) => {
                return c.owner.username !== 'Source Keeper';
            },
        });
        // filter out friendlies and neutrals
        hostiles = hostiles.filter((c) => {
            var _a;
            return dontAttack.indexOf((_a = c.owner) === null || _a === void 0 ? void 0 : _a.username) === -1;
        });
        // if strong hostiles our job is done, move to the next room
        const strongHostiles = _.filter(hostiles, (c) => {
            return scoreAttacking(c) > creepHeal;
        });
        if (strongHostiles.length > 0) {
            trace.warn('defenders present', { num: strongHostiles.length });
            return SUCCESS; // GO TO THE NEXT ROOM
        }
        let weakCreep = null;
        const targetId = creep.memory[MEMORY_HARASS_CREEP_TARGET];
        if (targetId) {
            const target = Game.getObjectById(targetId);
            if (target) {
                trace.info('have target', { target: targetId });
                weakCreep = target;
            }
            else {
                delete creep.memory[MEMORY_HARASS_CREEP_TARGET];
            }
        }
        // TODO, attack when getting attacked - don't focus on the weakest creep
        if (!weakCreep) {
            // if we have weak hostiles, attack them
            const weakHostiles = _.difference(hostiles, strongHostiles);
            if (weakHostiles.length > 0) {
                trace.info('picking creep', { num: weakHostiles.length });
                weakCreep = creep.pos.findClosestByRange(weakHostiles);
                creep.memory[MEMORY_HARASS_CREEP_TARGET] = weakCreep.id;
            }
        }
        if (weakCreep) {
            trace.info('attacking weak creep', { id: weakCreep.id });
            const range = creep.pos.getRangeTo(weakCreep);
            if (range <= 3) {
                trace.info('attack closest', { target: weakCreep.id });
                creep.rangedAttack(weakCreep);
            }
            if (range > 3) {
                creep.moveTo(weakCreep);
            }
            return RUNNING;
        }
        // attack containers if not one of my bases
        const baseRoom = kernel.getPlanner().getBaseByRoom(creep.room.name);
        if (!baseRoom) {
            // if there are containers, destroy them
            const containers = creep.room.find(FIND_STRUCTURES, {
                filter: (s) => {
                    return s.structureType === STRUCTURE_CONTAINER;
                },
            });
            if (containers.length > 0) {
                const container = creep.pos.findClosestByRange(containers);
                const range = creep.pos.getRangeTo(container);
                if (range <= 3) {
                    trace.info('attack container', { target: container.id });
                    creep.rangedAttack(container);
                }
                if (range > 3) {
                    creep.moveTo(container);
                }
                return RUNNING;
            }
        }
        return SUCCESS;
    }),
    leafNode('harass_room_done', (creep, _trace, _kernel) => {
        // if we are not in target room, don't switch rooms (we probably picked
        // on someone along the way)
        if (creep.room.name !== creep.memory[MEMORY_HARASS_CURRENT]) {
            return SUCCESS;
        }
        // Increment visited count, moving us to the next room in the list
        const roomCount = creep.memory[MEMORY_HARASS_VISITED] || 0;
        creep.memory[MEMORY_HARASS_VISITED] = roomCount + 1;
        return SUCCESS;
    }),
]);
const roleHarasser = {
    run: rootNode('harasser', behaviorBoosts(behavior$d)),
};
function scoreHealing(creep, healSelf = false) {
    let toughBonus = 1;
    return _.reduce(creep.body, (healing, part) => {
        // Don't count damaged parts
        if (part.hits < 1) {
            return healing;
        }
        if (part.type === HEAL) {
            let multiplier = 1;
            if (part.boost) {
                multiplier = BOOSTS[HEAL][part.boost].heal;
            }
            return healing + (healSelf ? HEAL_POWER : RANGED_HEAL_POWER) * multiplier;
        }
        // TODO this need to be solved with a dynamic programming problem that
        // looks at each part and sums a healing bonus for teach toughness part
        // up to whatever the damage per tick may be
        if (healSelf && part.type === TOUGH && part.boost) {
            toughBonus = _.min([BOOSTS[TOUGH][part.boost].damage, toughBonus]);
        }
        return healing;
    }, 0);
}
function scoreAttacking(creep) {
    return _.reduce(creep.body, (dmg, part) => {
        if (part.type === ATTACK) {
            let multiplier = 1;
            if (part.boost) {
                multiplier = BOOSTS[ATTACK][part.boost].attack;
            }
            return dmg + ATTACK_POWER * multiplier;
        }
        if (part.type === RANGED_ATTACK) {
            let multiplier = 1;
            if (part.boost) {
                multiplier = BOOSTS[RANGED_ATTACK][part.boost].rangedAttack;
            }
            return dmg + RANGED_ATTACK_POWER * multiplier;
        }
        return dmg;
    }, 0);
}

/**
 * Creep definitions
 *
 * TODO move to role files
 */
const WORKER_BUILDER = 'builder';
const WORKER_HARVESTER = 'harvester';
const WORKER_MINER = 'miner';
const WORKER_UPGRADER = 'upgrader';
const WORKER_DEFENDER = 'defender';
const WORKER_DEFENDER_BOOSTED = 'defender_boosted';
const WORKER_DEFENDER_DRONE = 'defender_drone';
const WORKER_ATTACKER = 'attacker';
const WORKER_REPAIRER = 'repairer';
const WORKER_DISTRIBUTOR = 'distributor';
const WORKER_RESERVER = 'reserver';
const WORKER_HAULER = 'hauler';
const WORKER_EXPLORER = 'explorer';
const ROLE_WORKER = 'worker';
// The 'base' should at most 300 energy as it will form the base of the creep
// The 'parts' are additional parts that will be used to fill up to the 'energyLimit'
const DEFINITIONS = new Map();
DEFINITIONS.set(ROLE_HARASSER, HarasserDefinition);
DEFINITIONS.set(WORKER_UPGRADER, {
    base: [CARRY, WORK, MOVE],
    parts: [MOVE, WORK, WORK, MOVE, WORK, CARRY],
    boosts: ['upgradeController'],
    energyLimit: 0,
    energyMinimum: 200,
    softEnergyMinimum: 1000,
    ignoreSpawnEnergyLimit: false,
    processPriority: Priorities.CORE_LOGISTICS,
    skippable: true,
});
DEFINITIONS.set(WORKER_DISTRIBUTOR, {
    base: [MOVE, WORK, CARRY],
    parts: [MOVE, CARRY, CARRY],
    boosts: [],
    energyLimit: 0,
    energyMinimum: 200,
    softEnergyMinimum: 300,
    ignoreSpawnEnergyLimit: true,
    processPriority: Priorities.CORE_LOGISTICS,
    skippable: true,
});
DEFINITIONS.set(WORKER_DEFENDER, {
    base: [MOVE, RANGED_ATTACK],
    parts: [MOVE, HEAL, MOVE, RANGED_ATTACK, MOVE, RANGED_ATTACK, MOVE, TOUGH, MOVE, RANGED_ATTACK],
    boosts: ['rangedAttack'],
    energyLimit: 0,
    energyMinimum: 550,
    ignoreSpawnEnergyLimit: false,
    processPriority: Priorities.DEFENCE,
    skippable: false,
});
DEFINITIONS.set(WORKER_DEFENDER_BOOSTED, {
    base: [MOVE, RANGED_ATTACK],
    parts: [MOVE, HEAL, MOVE, TOUGH, MOVE, RANGED_ATTACK, MOVE, TOUGH, MOVE, RANGED_ATTACK],
    boosts: ['heal', 'damage', 'rangedAttack'],
    energyLimit: 0,
    energyMinimum: 1000,
    ignoreSpawnEnergyLimit: false,
    processPriority: Priorities.DEFENCE,
    skippable: false,
});
DEFINITIONS.set(WORKER_DEFENDER_DRONE, {
    base: [MOVE, RANGED_ATTACK],
    parts: [MOVE, HEAL, MOVE, TOUGH, MOVE, RANGED_ATTACK, MOVE, TOUGH, MOVE, RANGED_ATTACK],
    boosts: [],
    energyLimit: 0,
    energyMinimum: 300,
    ignoreSpawnEnergyLimit: false,
    processPriority: Priorities.DEFENCE,
    skippable: true,
});
DEFINITIONS.set(WORKER_HARVESTER, {
    base: [MOVE, CARRY, WORK],
    parts: [MOVE, WORK, WORK, MOVE, WORK, CARRY],
    boosts: [],
    energyLimit: 3200,
    energyMinimum: 200,
    softEnergyMinimum: 1000,
    ignoreSpawnEnergyLimit: false,
    processPriority: Priorities.RESOURCES,
    skippable: true,
});
DEFINITIONS.set(ROLE_WORKER, {
    base: [MOVE, WORK, CARRY],
    parts: [MOVE, CARRY, WORK, MOVE, CARRY, CARRY],
    boosts: [],
    energyLimit: 0,
    energyMinimum: 300,
    ignoreSpawnEnergyLimit: false,
    processPriority: Priorities.RESOURCES,
    skippable: true,
});
DEFINITIONS.set(WORKER_MINER, {
    base: [MOVE, WORK, WORK],
    parts: [MOVE, CARRY, WORK, MOVE, WORK, WORK],
    boosts: [],
    energyLimit: 1750,
    energyMinimum: 250,
    softEnergyMinimum: 300,
    ignoreSpawnEnergyLimit: true,
    processPriority: Priorities.RESOURCES,
    skippable: true,
});
DEFINITIONS.set(WORKER_HAULER, {
    base: [MOVE, WORK, CARRY],
    parts: [MOVE, CARRY, CARRY],
    boosts: [],
    energyLimit: 0,
    energyMinimum: 200,
    softEnergyMinimum: 300,
    ignoreSpawnEnergyLimit: true,
    processPriority: Priorities.LOGISTICS,
    skippable: true,
});
DEFINITIONS.set(WORKER_ATTACKER, {
    base: [TOUGH, MOVE, MOVE, MOVE, MOVE, MOVE, HEAL, HEAL, HEAL, HEAL, HEAL, RANGED_ATTACK, WORK, WORK],
    parts: [WORK, TOUGH, MOVE, HEAL, MOVE, HEAL, MOVE, RANGED_ATTACK, WORK],
    boosts: ['tough', 'heal', 'damage'],
    energyLimit: 0,
    energyMinimum: 150,
    ignoreSpawnEnergyLimit: true,
    processPriority: Priorities.ATTACK,
    skippable: false,
});
DEFINITIONS.set(WORKER_BUILDER, {
    base: [MOVE, CARRY, WORK],
    parts: [MOVE, WORK, WORK, MOVE, WORK, CARRY],
    boosts: [],
    energyLimit: 1500,
    energyMinimum: 200,
    softEnergyMinimum: 800,
    ignoreSpawnEnergyLimit: true,
    processPriority: Priorities.MAINTENANCE,
    skippable: true,
});
DEFINITIONS.set(WORKER_REPAIRER, {
    base: [CARRY, MOVE, WORK, MOVE],
    parts: [CARRY, MOVE, WORK, MOVE],
    boosts: [],
    energyLimit: 900,
    energyMinimum: 250,
    softEnergyMinimum: 800,
    ignoreSpawnEnergyLimit: false,
    processPriority: Priorities.MAINTENANCE,
    skippable: true,
});
DEFINITIONS.set(WORKER_RESERVER, {
    base: [MOVE, CLAIM],
    parts: [CLAIM, MOVE],
    boosts: [],
    energyLimit: 2600,
    energyMinimum: 650,
    ignoreSpawnEnergyLimit: false,
    processPriority: Priorities.RESOURCES,
    skippable: true,
});
DEFINITIONS.set(WORKER_EXPLORER, {
    base: [MOVE],
    parts: [],
    boosts: [],
    energyLimit: 50,
    energyMinimum: 50,
    ignoreSpawnEnergyLimit: false,
    processPriority: Priorities.EXPLORATION,
    skippable: true,
});

const TOPIC_TERMINAL_TASK = 'task_terminal';
const TASK_REACTION = 'task_reaction';
const BOOST_PREP = 'boost_prep';
const NUKER_TARGETS = 'nuker_targets';
const ACTIVE_REACTIONS = 'active_reactions';
const ROOM_STATUES = 'room_status';
const ATTACK_ROOM = 'attack_room';

const NotAttachedError = new Error('Consumer not attached to stream');
class Event {
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    constructor(key, time, type, data) {
        this.key = key;
        this.time = time;
        this.type = type;
        this.data = data;
    }
}
class Consumer {
    constructor(id, stream) {
        this.id = id;
        this.stream = stream;
        this.offset = stream.getLength();
    }
    getEvents() {
        if (this.stream == null) {
            throw NotAttachedError;
        }
        const events = this.stream.getEvents(this.offset);
        this.offset = this.stream.getLength();
        return events;
    }
    detach() {
        this.stream = null;
    }
    getOffset() {
        return this.offset;
    }
    // When topic compacted, we need the ability to shift each consumers offset
    shiftOffet(shift) {
        this.offset -= shift;
    }
}
class Stream {
    constructor() {
        this.consumers = new Map();
        this.events = [];
    }
    publish(event) {
        this.events.push(event);
    }
    removeConsumed() {
        let minOffset = -1;
        for (const consumer of this.consumers.values()) {
            if (minOffset === -1) {
                minOffset = consumer.getOffset();
            }
            minOffset = Math.min(minOffset, consumer.getOffset());
        }
        if (minOffset > 0) {
            this.events = this.events.slice(minOffset);
            for (const consumer of this.consumers.values()) {
                consumer.shiftOffet(minOffset);
            }
        }
    }
    getLength() {
        return this.events.length;
    }
    getEvents(offset) {
        const events = this.events.slice(offset);
        return events;
    }
    addConsumer(consumerId) {
        const consumer = new Consumer(consumerId, this);
        this.consumers.set(consumerId, consumer);
        return consumer;
    }
    removeConsumer(consumer) {
        this.consumers.delete(consumer.id);
        consumer.detach();
    }
}
class EventBroker {
    constructor() {
        this.streams = new Map();
    }
    getStream(streamId) {
        if (!this.streams.get(streamId)) {
            this.streams.set(streamId, new Stream());
        }
        return this.streams.get(streamId);
    }
    removeConsumed() {
        for (const stream of this.streams.values()) {
            stream.removeConsumed();
        }
    }
}

const threadBaseRoom = (name, ttl) => (action) => {
    let lastCall = 0;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const tick = function (trace, kernel, base, room, ...args) {
        if (lastCall + ttl <= Game.time) {
            lastCall = Game.time;
            const actionTrace = trace.begin(name);
            const result = action(actionTrace, kernel, base, room, ...args);
            actionTrace.end();
            return result;
        }
        else {
            trace.info(`thread ${name} sleeping for ${lastCall + ttl - Game.time}`);
        }
        return null;
    };
    tick.reset = () => {
        lastCall = 0;
    };
    return tick;
};

/**
 * Base Creep Spawning
 *
 * Tracks the spawns in a base, pull events from the spawn topic, and spawns the requested creeps.
 *
 * TODO - Move to topic with base id in the name - IN PROGRESS
 */
const SPAWN_TTL = 5;
const REQUEST_BOOSTS_TTL = 5;
const PROCESS_EVENTS_TTL = 20;
const INITIAL_TOPIC_LENGTH = 9999;
const RED_TOPIC_LENGTH = 10;
const YELLOW_TOPIC_LENGTH = 5;
const SPAWN_REQUEST_ROLE = 'role';
const SPAWN_REQUEST_SPAWN_MIN_ENERGY = 'spawn_min_energy';
const SPAWN_REQUEST_PARTS = 'parts';
const UTILIZATION_CARDINALITY = 150;
const MIN_UTILIZATION_SAMPLES = 30;
function createSpawnRequest(priority, ttl, role, 
// eslint-disable-next-line @typescript-eslint/no-explicit-any
memory, parts, energyLimit) {
    return {
        priority,
        ttl,
        details: {
            role,
            memory,
            energyLimit,
            parts,
        },
    };
}
function getShardSpawnTopic() {
    return 'shard_spawn';
}
function getBaseSpawnTopic(baseId) {
    return `base_${baseId}_spawn`;
}
function createUtilizationUpdate(utilization) {
    return {
        utilization,
    };
}
function getBaseSpawnUtilizationTopic(baseId) {
    return `base_${baseId}_spawn_utilization`;
}
class SpawnManager {
    constructor(id, baseId) {
        this.utilization = [];
        this.id = id;
        this.baseId = baseId;
        this.spawnThread = threadBaseRoom('spawn_thread', SPAWN_TTL)(this.spawning.bind(this));
        this.eventsThread = threadBase('produce_events_thread', PROCESS_EVENTS_TTL)((trace, kernel, base) => {
            this.processEvents(trace, kernel, base);
        });
    }
    run(kernel, trace) {
        trace = trace.begin('spawn_manager_run');
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.error('no base, terminating', { baseId: this.baseId });
            trace.end();
            return terminate();
        }
        const room = getBasePrimaryRoom(base);
        if (!room) {
            trace.error('no room object,', { base });
            trace.end();
            return running();
        }
        trace.info('Spawn manager run', { id: this.id, baseId: this.baseId });
        this.spawnThread(trace, kernel, base, room);
        this.eventsThread(trace, kernel, base);
        trace.end();
        return running();
    }
    spawning(trace, kernel, base, room) {
        const spawns = room.find(FIND_MY_STRUCTURES, {
            filter: (s) => s.structureType === STRUCTURE_SPAWN,
        });
        // If there are no spawns then we should request another base in the kernel produce the creep
        if (spawns.length === 0) {
            let request = null;
            // eslint-disable-next-line no-cond-assign
            while (request = kernel.getTopics().getNextRequest(getBaseSpawnTopic(base.id))) {
                trace.notice('sending spawn request to shard/neighbors', { request: request });
                kernel.getTopics().addRequestV2(getShardSpawnTopic(), request);
            }
            return;
        }
        // iterate spawns and fetch next request if idle
        spawns.forEach((spawn) => {
            const isIdle = !spawn.spawning;
            const spawnEnergy = spawn.room.energyAvailable;
            const energyCapacity = spawn.room.energyCapacityAvailable;
            const energyPercentage = spawnEnergy / energyCapacity;
            trace.info('spawn status', {
                id: spawn.id,
                isIdle,
                spawnEnergy,
                energyCapacity,
                energyPercentage,
            });
            // track utilization
            this.utilization.push(isIdle ? 0 : 1);
            if (this.utilization.length > UTILIZATION_CARDINALITY) {
                this.utilization.shift();
            }
            // currently spawning something
            if (!isIdle) {
                trace.getMetricsCollector().counter('spawn_busy', 1, { spawn: spawn.id, base: base.id });
                handleActiveSpawning(kernel, base, spawn, trace);
                return;
            }
            // check if spawner has energy
            if (!isSpawnerEnergyReady(kernel, base, spawn, trace)) {
                trace.getMetricsCollector().counter('spawn_not_enough_energy', 1, { spawn: spawn.id, base: base.id });
                trace.info('spawner not ready, skipping', { spawn: spawn.id, spawnEnergy, energyCapacity });
                return;
            }
            const request = getNextRequest(kernel, base, spawn, trace);
            if (request) {
                // determine topic pressure
                const spawnTopicSize = kernel.getTopics().getLength(getBaseSpawnTopic(base.id));
                const spawnTopicBackPressure = Math.floor(energyCapacity * (1 - (0.09 * spawnTopicSize)));
                let energyLimit = _.max([300, spawnTopicBackPressure]);
                // Allow request to override energy limit
                if (request.details.energyLimit) {
                    energyLimit = request.details.energyLimit;
                }
                trace.notice('spawning', { id: this.id, spawnEnergy, energyLimit, request });
                trace.getMetricsCollector().counter('spawn_spawning', 1, { spawn: spawn.id, base: base.id });
                createCreep(base, room.name, spawn, request, spawnEnergy, energyLimit, trace);
                return;
            }
            trace.getMetricsCollector().counter('spawn_no_request', 1, { spawn: spawn.id, base: base.id });
            handleIdle(trace);
        });
    }
    processEvents(trace, kernel, base) {
        const baseRoom = getBasePrimaryRoom(base);
        if (!baseRoom) {
            trace.error('no primary room for base', { base: base.id });
            return;
        }
        const roomName = baseRoom.name;
        const baseTopic = kernel.getTopics().getTopic(getBaseSpawnTopic(base.id));
        let utilization = 0;
        if (this.utilization.length) {
            utilization = _.sum(this.utilization) / this.utilization.length;
        }
        if (this.utilization.length > MIN_UTILIZATION_SAMPLES) {
            const utilizationStream = kernel.getBroker().getStream(getBaseSpawnUtilizationTopic(base.id));
            utilizationStream.publish(new Event(this.id, Game.time, "set", createUtilizationUpdate(utilization)));
        }
        let creeps = [];
        let topicLength = 9999;
        if (baseTopic) {
            topicLength = baseTopic.length;
            creeps = baseTopic.map((message) => {
                return `${message.details[SPAWN_REQUEST_ROLE]}(${message.priority},${message.ttl - Game.time})`;
            });
        }
        const line = {
            key: `${this.id}`,
            room: baseRoom.name,
            order: 5,
            text: `Next spawn: ${creeps.join(',')}, Utilization: ${utilization.toFixed(2)}, Topic length: ${topicLength}`,
            time: Game.time,
        };
        const event = new Event(this.id, Game.time, HudEventSet, line);
        trace.info('produce_events', { event });
        kernel.getBroker().getStream(getLinesStream()).publish(event);
        const indicatorStream = kernel.getBroker().getStream(getDashboardStream());
        // Processes
        let processStatus = HudIndicatorStatus.Green;
        if (topicLength === INITIAL_TOPIC_LENGTH) {
            processStatus = HudIndicatorStatus.Stale;
        }
        else if (topicLength > RED_TOPIC_LENGTH) {
            processStatus = HudIndicatorStatus.Red;
        }
        else if (topicLength > YELLOW_TOPIC_LENGTH) {
            processStatus = HudIndicatorStatus.Yellow;
        }
        const spawnLengthIndicator = {
            room: roomName,
            key: 'spawn_length',
            display: 'S',
            status: processStatus,
        };
        indicatorStream.publish(new Event(this.id, Game.time, HudEventSet, spawnLengthIndicator));
    }
}
function peekNextNeighborRequest(kernel, base, trace) {
    trace.info("peeking neighbor request", { baseRoom: base.primary, baseNeighbors: base.neighbors });
    const topic = kernel.getTopics();
    const requests = topic.getTopic(getShardSpawnTopic());
    if (!requests) {
        return null;
    }
    // Reverse message so we get higher priority first
    return getNextNeighborRequestFilter(kernel, base, requests.reverse(), trace);
}
function getNextNeighborRequest(kernel, base, trace) {
    trace.info("getting neighbor request", { baseRoom: base.primary, baseNeighbors: base.neighbors });
    const topic = kernel.getTopics();
    return topic.getMessageOfMyChoice(getShardSpawnTopic(), (messages) => {
        // Reverse message so we get higher priority first
        return getNextNeighborRequestFilter(kernel, base, messages.reverse(), trace);
    });
}
function requestBoosts(kernel, base, spawn, boosts, priority) {
    const request = {
        priority: priority,
        ttl: REQUEST_BOOSTS_TTL + Game.time,
        details: {
            [TASK_ID]: `bp-${spawn.id}-${Game.time}`,
            [PREPARE_BOOSTS]: boosts,
        }
    };
    kernel.getTopics().addRequestV2(BOOST_PREP, request);
}
function isSpawnerEnergyReady(kernel, base, spawn, trace) {
    const spawnEnergy = spawn.room.energyAvailable;
    // check minimum energy
    let minEnergy = 300;
    // TODO factor number of crepes in base for minimum energy
    //const numCreeps = kernel.getCreepsManager().getCreepsByBase(base.id).length;
    minEnergy = _.max([300, minEnergy]);
    if (spawnEnergy < minEnergy) {
        trace.info('low energy, not spawning', { id: spawn.id, spawnEnergy, minEnergy });
        return false;
    }
    return true;
}
function getNextRequest(kernel, base, spawn, trace) {
    const spawnEnergy = spawn.room.energyAvailable;
    const energyCapacity = spawn.room.energyCapacityAvailable;
    // check local requests
    let localRequest = kernel.getTopics().peekNextRequest(getBaseSpawnTopic(base.id));
    if (localRequest && !isRequestEnergyReady(kernel, base, spawn, localRequest, trace)) {
        trace.info('local request not ready, skipping', { spawn: spawn.id, spawnEnergy, energyCapacity });
        localRequest = null;
    }
    // Check if we have a reserve of energy and can help neighbors
    let neighborRequest = peekNextNeighborRequest(kernel, base, trace);
    if (neighborRequest && !isRequestEnergyReady(kernel, base, spawn, neighborRequest, trace)) {
        trace.info('neighbor request not ready, skipping', { spawn: spawn.id, spawnEnergy, energyCapacity });
        neighborRequest = null;
    }
    trace.info('spawn requests', { localRequest, neighborRequest });
    let requestSource = null;
    // Select local request if available
    if (localRequest) {
        trace.info('found local request', { localRequest });
        requestSource = 'local';
    }
    // If no request selected and neighbor request available, select neighbor request
    if (!localRequest && neighborRequest) {
        trace.warn('found neighbor request', { neighborRequest });
        requestSource = 'neighbor';
    }
    // If local priority w/ bonus is less than neighbor priority, select neighbor request
    if (((localRequest === null || localRequest === void 0 ? void 0 : localRequest.priority) + 1) < (neighborRequest === null || neighborRequest === void 0 ? void 0 : neighborRequest.priority)) {
        trace.warn('neighbor request has higher priority', { neighborRequest, localRequest });
        requestSource = 'neighbor';
    }
    let request = null;
    // Take the request off the right queue
    switch (requestSource) {
        case 'local':
            trace.info('spawning from local request', { localRequest });
            request = kernel.getTopics().getNextRequest(getBaseSpawnTopic(base.id));
            break;
        case 'neighbor':
            trace.info('spawning from neighbor request', { neighborRequest });
            request = getNextNeighborRequest(kernel, base, trace);
            break;
        default:
            trace.info('no request found, skipping', { request });
            break;
    }
    return request;
}
function isRequestEnergyReady(kernel, base, spawn, request, trace) {
    const spawnEnergy = spawn.room.energyAvailable;
    const energyCapacity = spawn.room.energyCapacityAvailable;
    const requestMinEnergy = request.details[SPAWN_REQUEST_SPAWN_MIN_ENERGY] || 0;
    if (spawnEnergy < requestMinEnergy) {
        trace.warn('base does not have energy', { requestMinEnergy, spawnEnergy, request });
        return false;
    }
    // get role from definition
    const role = request.details[SPAWN_REQUEST_ROLE];
    const definition = DEFINITIONS.get(role);
    if (!definition) {
        trace.error('unknown role', { role });
        return false;
    }
    // the soft min is only enforced if the base has enough capacity to meet it
    if (definition.softEnergyMinimum && definition.softEnergyMinimum < energyCapacity &&
        spawnEnergy < definition.softEnergyMinimum) {
        trace.info('no enough energy (soft)', { spawnEnergy, energyCapacity, definition });
        return false;
    }
    // if definition has a minimum energy requirement, check if we have enough energy
    if (definition.energyMinimum && spawnEnergy < definition.energyMinimum) {
        trace.warn('not enough energy (hard)', { spawnEnergy, definition });
        return false;
    }
    return true;
}
function handleActiveSpawning(kernel, base, spawn, trace) {
    const creep = Game.creeps[spawn.spawning.name];
    spawn.room.visual.text(spawn.spawning.name + '🛠️', spawn.pos.x - 1, spawn.pos.y, { align: 'right', opacity: 0.8 });
    const role = creep.memory[MEMORY_ROLE];
    const definition = DEFINITIONS.get(role);
    if (!definition) {
        trace.error('unknown role', { creepName: creep.name, role });
        return;
    }
    const boosts = definition.boosts;
    const priority = definition.processPriority;
    trace.info('spawning', { creepName: creep.name, role, boosts, priority });
    trace.getMetricsCollector().gauge('spawn_spawning_role', spawn.spawning.remainingTime, { spawn: spawn.id, role, base: base.id });
    if (boosts) {
        requestBoosts(kernel, base, spawn, boosts, priority);
    }
    return;
}
function handleIdle(trace) {
    trace.info('no request');
}
function createCreep(base, room, spawn, request, energy, energyLimit, trace) {
    const role = request.details[SPAWN_REQUEST_ROLE];
    let parts = request.details[SPAWN_REQUEST_PARTS] || null;
    const memory = request.details.memory || {};
    const definition = DEFINITIONS.get(role);
    if (!definition) {
        trace.error('no definition for role', { role });
        return;
    }
    const ignoreSpawnEnergyLimit = definition.ignoreSpawnEnergyLimit || false;
    if (energy > energyLimit && !ignoreSpawnEnergyLimit) {
        energy = energyLimit;
    }
    const roleEnergyLimit = definition.energyLimit;
    if (roleEnergyLimit && energy > roleEnergyLimit) {
        energy = roleEnergyLimit;
    }
    trace.info("parts from request", { parts });
    // if parts not provided, work them out from the definition
    if (!parts || !parts.length) {
        parts = getBodyParts(definition, energy);
        trace.info("parts decided", { parts, energy, definition });
    }
    const name = [role, Game.shard.name, Game.time].join('_');
    // Requests to the kernel should include the destination base, don't overwrite it
    if (!memory[MEMORY_BASE]) {
        memory[MEMORY_BASE] = base.id;
    }
    // Used for debugging, don't use for decision making, use MEMORY_BASE instead
    memory[MEMORY_ORIGIN_SHARD] = Game.shard.name;
    memory[MEMORY_ORIGIN] = room;
    memory[MEMORY_ROLE] = role;
    memory[DESIRED_BOOSTS] = definition.boosts;
    const result = spawn.spawnCreep(parts, name, { memory });
    if (result !== OK) {
        trace.error('spawn error', { result, spawn: spawn.id, name, parts });
    }
}
function getBodyParts(definition, maxEnergy) {
    let base = definition.base.slice(0);
    let i = 0;
    while (true) { // eslint-disable-line no-constant-condition
        const nextPart = definition.parts[i % definition.parts.length];
        const estimate = base.concat([nextPart]).reduce((acc, part) => {
            return acc + BODYPART_COST[part];
        }, 0);
        if (estimate <= maxEnergy && base.length < 50) {
            base.push(nextPart);
        }
        else {
            break;
        }
        i++;
    }
    base = _.sortBy(base, (part) => {
        switch (part) {
            case TOUGH:
                return 0;
            case WORK:
            case CARRY:
                return 1;
            case MOVE:
                return 2;
            case ATTACK:
                return 8;
            case RANGED_ATTACK:
                return 9;
            case HEAL:
                return 10;
            default:
                return 1;
        }
    });
    return base;
}
function getNextNeighborRequestFilter(kernel, base, messages, trace) {
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    return _.find(messages, (message) => {
        // Select message if portal nearby
        // RAKE check distance on other side of the portal too
        const assignedShard = message.details.memory[MEMORY_ASSIGN_SHARD] || null;
        if (assignedShard && assignedShard != Game.shard.name) {
            trace.warn('request in another shard', { assignedShard, shard: Game.shard.name });
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            const portals = kernel.getScribe().
                getPortals(assignedShard).filter((portal) => {
                const distance = Game.map.getRoomLinearDistance(baseRoom.name, portal.pos.roomName);
                return distance < 2;
            });
            if (!portals.length) {
                return false;
            }
            return true;
        }
        // Determine destination room
        let destinationRoom = null;
        const baseRoom = message.details.memory[MEMORY_BASE];
        if (baseRoom) {
            destinationRoom = baseRoom;
        }
        const assignedRoom = message.details.memory[MEMORY_ASSIGN_ROOM];
        if (assignedRoom) {
            destinationRoom = assignedRoom;
        }
        const positionRoom = message.details.memory[MEMORY_POSITION_ROOM];
        if (positionRoom) {
            destinationRoom = positionRoom;
        }
        // If no destination room, can be produced by anyone
        if (!destinationRoom) {
            trace.warn('no destination room, can be produced by anyone', { message });
            return true;
        }
        // If the room is part of a base, check if the base is a neighbor
        const destinationBase = kernel.getPlanner().getBaseByRoom(destinationRoom);
        if (destinationBase) {
            const isNeighbor = base.neighbors.some((neighborId) => {
                return neighborId == destinationBase.id;
            });
            if (isNeighbor) {
                return true;
            }
        }
        return false;
    });
}

/**
 * Formation/party logic
 *
 * Handles logic related to filling the group, relative positioning, moving in
 * formation, setting heals, and targets for a group of creeps. Attack and defense
 * logic use this to abstract away creep handling.
 *
 */
const REQUEST_PARTY_MEMBER_TTL$1 = 25;
const MAX_PARTY_SIZE = 4;
const MAX_PREVIOUS_POSITIONS = 6;
const FORMATION_QUAD = 'quad';
const FORMATION_SINGLE_FILE = 'single_file';
const DIRECTION_2BY2_FORMATION = {
    [TOP]: [
        { x: 0, y: -1 },
        { x: 1, y: -1 },
        { x: 0, y: 0 },
        { x: 1, y: 0 }, // BR
    ],
    [TOP_RIGHT]: [
        { x: 1, y: -1 },
        { x: 0, y: -1 },
        { x: 1, y: 0 },
        { x: 0, y: 0 }, // BL
    ],
    [RIGHT]: [
        { x: 1, y: -1 },
        { x: 1, y: 0 },
        { x: 0, y: -1 },
        { x: 0, y: 0 }, // BL
    ],
    [BOTTOM_RIGHT]: [
        { x: 1, y: 0 },
        { x: 1, y: -1 },
        { x: 0, y: -1 },
        { x: 0, y: 0 }, // BL
    ],
    [BOTTOM]: [
        { x: 0, y: 0 },
        { x: 1, y: 0 },
        { x: 0, y: -1 },
        { x: 1, y: -1 }, // TR
    ],
    [BOTTOM_LEFT]: [
        { x: 0, y: 0 },
        { x: 1, y: 0 },
        { x: 0, y: -1 },
        { x: 1, y: -1 }, // TR
    ],
    [LEFT]: [
        { x: 0, y: -1 },
        { x: 0, y: 0 },
        { x: 1, y: -1 },
        { x: 1, y: 0 }, // BR
    ],
    [TOP_LEFT]: [
        { x: 0, y: -1 },
        { x: 1, y: -1 },
        { x: 0, y: 0 },
        { x: 1, y: 0 }, // BR
    ],
};
class PartyRunnable {
    constructor(id, baseId, previousPositions, role, parts, minEnergy, priority, ttl) {
        this.id = id;
        this.baseId = baseId;
        this.role = role;
        this.parts = parts;
        this.minEnergy = minEnergy;
        this.priority = priority;
        this.requestCreepTTL = ttl;
        this.isDone = false;
        this.deployTicks = 0;
        this.setPreviousPositions(previousPositions);
        this.setFormation(FORMATION_SINGLE_FILE);
        this.setDirection(TOP);
        this.threadRequestCreeps = threadBase('request_creeps', REQUEST_PARTY_MEMBER_TTL$1)(this.requestCreeps.bind(this));
    }
    run(kernel, trace) {
        trace = trace.begin('party_run');
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.error('base not found', { baseId: this.baseId });
            trace.end();
            return terminate();
        }
        // TODO possible race condition with outer layer and this
        const creeps = this.getAssignedCreeps();
        if (this.isDone) {
            trace.info('party done, cleaning up');
            // TODO recycle
            creeps.forEach((creep) => {
                creep.suicide();
            });
            trace.end();
            return terminate();
        }
        trace.info('party run', {
            id: this.id,
            position: this.position,
            deployTicks: this.deployTicks,
            creeps: creeps.map((creep) => creep.name),
            previousPositions: this.previousPositions,
        });
        if (!Game.flags['debug']) {
            this.threadRequestCreeps(trace, kernel, base);
        }
        else {
            trace.info('in debug mode, not spawning');
        }
        this.updateCreepPositions(creeps, trace);
        trace.end();
        return running();
    }
    updateCreepPositions(creeps, trace) {
        if (this.formation === FORMATION_SINGLE_FILE) {
            creeps.forEach((creep, idx) => {
                let position = this.position;
                if (this.previousPositions[idx]) {
                    position = this.previousPositions[idx];
                }
                trace.info('setting creep position in single file', { name: creep.name, idx, position });
                creep.memory[MEMORY_POSITION_X] = position.x;
                creep.memory[MEMORY_POSITION_Y] = position.y;
                creep.memory[MEMORY_POSITION_ROOM] = position.roomName;
            });
        }
        else if (this.formation === FORMATION_QUAD) {
            const quadPositions = this.getCreepOffsets();
            // TODO first four should be in formation, remaining should be behind
            // Can possibly use 3rd and grater previous positions
            creeps.forEach((creep, idx) => {
                idx = idx % 4;
                const x = ___namespace.max([___namespace.min([this.position.x + quadPositions[idx].x, 49]), 0]);
                const y = ___namespace.max([___namespace.min([this.position.y + quadPositions[idx].y, 49]), 0]);
                const roomName = this.position.roomName;
                trace.info('setting creep position in quad', { name: creep.name, x, y, roomName });
                creep.memory[MEMORY_POSITION_X] = x;
                creep.memory[MEMORY_POSITION_Y] = y;
                creep.memory[MEMORY_POSITION_ROOM] = roomName;
            });
        }
        else {
            trace.error('unknown formation', this.formation);
        }
    }
    done() {
        this.isDone = true;
    }
    getPosition() {
        return this.position;
    }
    getCreepPosition(idx, trace) {
        // Set to current position
        let x = this.position.x;
        let y = this.position.y;
        let roomName = this.position.roomName;
        if (this.formation === FORMATION_SINGLE_FILE) {
            // only update position if we have a previous position
            if (this.previousPositions[idx]) {
                x = this.previousPositions[idx].x;
                y = this.previousPositions[idx].y;
                roomName = this.previousPositions[idx].roomName;
            }
        }
        else if (this.formation === FORMATION_QUAD) {
            const quadPositions = this.getCreepOffsets();
            x = ___namespace.max([___namespace.min([this.position.x + quadPositions[idx].x, 49]), 0]);
            y = ___namespace.max([___namespace.min([this.position.y + quadPositions[idx].y, 49]), 0]);
            roomName = this.position.roomName;
        }
        else {
            trace.warn('unknown formation', this.formation);
        }
        return new RoomPosition(x, y, roomName);
    }
    inPosition(position, trace) {
        let inPosition = true;
        const visual = new RoomVisual();
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const showVisuals = global.LOG_WHEN_PID === this.id;
        if (showVisuals) {
            if (this.formation === FORMATION_QUAD) {
                visual.rect(position.x - 0.5, position.y - 1.5, 2, 2);
            }
            if (this.formation === FORMATION_SINGLE_FILE) {
                visual.rect(position.x - 0.5, position.y - 3.5, 1, 4);
            }
        }
        let positionBlocked = false;
        this.getAssignedCreeps().slice(0, 4).forEach((creep, idx) => {
            // if creep is on an edge, report it as ready
            if (creep.pos.x === 49 || creep.pos.x === 0 || creep.pos.y === 49 || creep.pos.y === 0) {
                trace.info('creep is at edge', {
                    name: creep.name,
                    pos: creep.pos,
                });
                if (showVisuals) {
                    visual.text('E', creep.pos.x, creep.pos.y + 0.5, {
                        color: (idx) ? '#0000FF' : '#FFFFFF',
                    });
                }
                return;
            }
            if (creep.fatigue > 0) {
                trace.info('not ready: fatigued', { creepName: creep.name, fatigue: creep.fatigue });
                inPosition = false;
                if (showVisuals) {
                    visual.text('F', creep.pos.x, creep.pos.y + 0.5, {
                        color: (idx) ? '#0000FF' : '#FFFFFF',
                    });
                }
                return;
            }
            const position = this.getCreepPosition(idx, trace);
            const terrain = position.lookFor(LOOK_TERRAIN);
            if (terrain.length > 0 && terrain[0] === 'wall') {
                if (showVisuals) {
                    visual.text('W', position.x, position.y + 0.5);
                }
                trace.info('not ready: terrain', {
                    name: creep.name,
                    terrain: terrain,
                    position: position,
                    idx,
                });
                positionBlocked = true;
                return;
            }
            if (creep.pos.x !== position.x || creep.pos.y !== position.y ||
                creep.pos.roomName !== position.roomName) {
                trace.info('not ready: out of position', {
                    creepName: creep.name,
                    creepPos: creep.pos,
                    desired: position,
                    formation: this.formation,
                    idx,
                });
                inPosition = false;
                /*
                if (creep.pos.getRangeTo(this.position) < 5) {
                  trace.log("not ready: out of position", {
                    creepName: creep.name,
                    creepPos: creep.pos,
                    desired: {x, y, roomName},
                  });
        
                  inPosition = false;
                } else {
                  trace.log("far out of position, not halting", {
                    creepName: creep.name,
                    creepPos: creep.pos,
                    desired: {x, y, roomName},
                  });
                }
                */
                if (showVisuals) {
                    visual.text('O', creep.pos.x, creep.pos.y + 0.5);
                }
                return;
            }
            trace.info('in position', {
                creepName: creep.name,
                creepPos: creep.pos,
                desired: position,
                formation: this.formation,
                idx,
            });
            if (showVisuals) {
                visual.text('R', creep.pos.x, creep.pos.y + 0.5);
            }
        });
        return positionBlocked || inPosition;
    }
    onEdge(position) {
        return position.x < 1 || position.x > 48 || position.y < 1 || position.y > 48;
    }
    shiftPosition(position, direction) {
        const x = position.x + DIRECTION_OFFSET[direction].x;
        const y = position.y + DIRECTION_OFFSET[direction].y;
        return new RoomPosition(x, y, position.roomName);
    }
    isBlocked(position, trace) {
        let structures = [];
        let positions = [];
        if (this.formation === FORMATION_SINGLE_FILE) {
            positions = [
                RoomPosition(position.x, position.y, position.roomName),
            ];
        }
        else {
            positions = this.getCreepOffsets().map((offset) => {
                const x = ___namespace.min([___namespace.max([position.x + offset.x, 0]), 49]);
                const y = ___namespace.min([___namespace.max([position.y + offset.y, 0]), 49]);
                return new RoomPosition(x, y, position.roomName);
            });
        }
        positions.forEach((position) => {
            structures = structures.concat(position.lookFor(LOOK_STRUCTURES))
                .filter((structure) => {
                return structure.structureType !== STRUCTURE_ROAD;
            });
            structures = structures.concat(position.lookFor(LOOK_TERRAIN))
                .filter((terrain) => {
                return terrain === 'wall';
            });
        });
        trace.info('blocking structures', { structures });
        return structures.length > 0;
    }
    // Returns objects that are blocking the party
    getBlockingObjects(position, _direction, _trace) {
        let objects = [];
        let positions = [];
        if (this.formation === FORMATION_SINGLE_FILE) {
            positions = [
                new RoomPosition(___namespace.min([___namespace.max([position.x, 0]), 49]), ___namespace.min([___namespace.max([position.y, 0]), 49]), position.roomName),
            ];
        }
        else if (this.formation === FORMATION_QUAD) {
            positions = this.getCreepOffsets().map((offset) => {
                const x = ___namespace.min([___namespace.max([position.x + offset.x, 0]), 49]);
                const y = ___namespace.min([___namespace.max([position.y + offset.y, 0]), 49]);
                return new RoomPosition(x, y, position.roomName);
            });
        }
        // We cannot lookFor in a room that is not loaded
        if (Game.rooms[position.roomName]) {
            positions.forEach((position) => {
                // Structures can block
                objects = objects.concat(position.lookFor(LOOK_STRUCTURES))
                    .filter((structure) => {
                    var _a, _b;
                    return structure.structureType !== STRUCTURE_ROAD && !(structure.structureType === STRUCTURE_RAMPART &&
                        ((_b = (_a = structure.room) === null || _a === void 0 ? void 0 : _a.controller) === null || _b === void 0 ? void 0 : _b.my));
                });
            });
        }
        return objects;
    }
    setRole(role) {
        this.role = role;
    }
    setParts(parts) {
        this.parts = parts;
    }
    setFormation(formation) {
        this.formation = formation;
    }
    getFormation() {
        return this.formation;
    }
    getCreepOffsets() {
        if (this.formation === FORMATION_SINGLE_FILE) {
            return this.previousPositions;
        }
        else if (this.formation === FORMATION_QUAD) {
            return DIRECTION_2BY2_FORMATION[this.direction];
        }
        else {
            throw new Error('Unknown formation');
        }
    }
    getCreepPositions() {
        return this.getCreepOffsets().map((offset) => {
            const x = ___namespace.max([___namespace.min([this.position.x + offset.x, 49]), 0]);
            const y = ___namespace.max([___namespace.min([this.position.y + offset.y, 49]), 0]);
            const roomName = this.position.roomName;
            return new RoomPosition(x, y, roomName);
        });
    }
    getDirection() {
        return this.direction;
    }
    setDirection(direction) {
        this.direction = direction;
    }
    setPreviousPositions(previousPositions) {
        this.previousPositions = previousPositions;
        this.position = previousPositions[0];
    }
    setMinEnergy(minEnergy) {
        this.minEnergy = minEnergy;
    }
    setPosition(position) {
        this.position = position;
        if (!this.position.isEqualTo(this.previousPositions[0])) {
            this.previousPositions.unshift(this.position);
            this.previousPositions.splice(MAX_PREVIOUS_POSITIONS);
        }
        this.previousPositions = ___namespace.unique(this.previousPositions, false, (position) => {
            return position.x + ',' + position.y + ',' + position.roomName;
        });
    }
    getPreviousPositions() {
        return this.previousPositions;
    }
    setTarget(targets, trace) {
        const target = ___namespace.find(targets, (target) => {
            return this.position.getRangeTo(target) <= 3;
        });
        if (target) {
            Game.map.visual.line(this.position, target.pos, { color: '#FF0000' });
        }
        else {
            trace.info('no targets in range');
            return null;
        }
        trace.info('setting targets', { target });
        const creeps = this.getAssignedCreeps();
        creeps.forEach((creep) => {
            creep.memory[MEMORY_ATTACK] = target === null || target === void 0 ? void 0 : target.id;
        });
        return target;
    }
    setHeal(_trace) {
        const creeps = this.getAssignedCreeps();
        if (!creeps.length) {
            return;
        }
        // Locate most damaged creep and heal
        const healOrder = ___namespace.sortBy(creeps, (creep) => {
            return creep.hits / creep.hitsMax;
        });
        let targetId = null;
        if (healOrder.length && healOrder[0].hits < healOrder[0].hitsMax) {
            targetId = healOrder[0].id;
        }
        const damagePercent = healOrder[0].hits / healOrder[0].hitsMax;
        creeps.forEach((creep, idx) => {
            creep.memory[MEMORY_HEAL] = null;
            if ((idx === 0 && damagePercent < 0.5) || (idx === 1 && damagePercent < 0.75)) {
                creep.memory[MEMORY_HEAL] = targetId;
            }
            else if (idx >= 2 && damagePercent < 1) {
                creep.memory[MEMORY_HEAL] = targetId;
            }
        });
    }
    getAssignedCreeps() {
        const creeps = Object.values(Game.creeps).filter((creep) => {
            return creep.memory[MEMORY_PARTY_ID] === this.id;
        });
        const quad = ___namespace.sortBy(creeps, 'ttl').slice(0, 4);
        /*
        quad = quad.filter(creep => !creep.spawning);
        quad = _.sortBy(quad, (creep) => {
          return creep.body.filter(part => part.type === HEAL).reduce((acc, part) => {
            if (part.boost) {
              acc += HEAL_POWER * BOOSTS[HEAL][part.boost][HEAL];
            } else {
              acc += HEAL_POWER;
            }
    
            return acc;
          }, 0)
        })
        */
        return [].concat(quad, creeps.slice(4));
    }
    requestCreeps(trace, kernel, base) {
        const baseRoomName = base.primary;
        const baseRoom = Game.rooms[baseRoomName];
        if (!baseRoom) {
            trace.info('base room not visible', { baseRoomName });
            return;
        }
        const spawns = baseRoom.find(FIND_MY_STRUCTURES, {
            filter: (structure) => structure.structureType === STRUCTURE_SPAWN,
        });
        if (!spawns || !spawns.length) {
            trace.info('no spawns found');
            return;
        }
        const path = PathFinder.search(this.position, spawns[0].pos, {
            maxOps: 4000,
        });
        // TODO need to calculate plains cost
        const plainsCost = 2;
        // path length * move cost/ticks
        const travelTime = path.path.length * plainsCost;
        // 150 for the 50 parts
        const buildTime = 150;
        this.deployTicks = travelTime + buildTime;
        if (path.incomplete) {
            this.deployTicks += 200;
        }
        trace.info('deploy time', {
            pathLength: path.path.length,
            pathIncomplete: path.incomplete,
            plainsCost,
        });
        const creeps = this.getAssignedCreeps();
        const freshCreeps = creeps.filter((creep) => {
            return creep.spawning || (creep.ticksToLive > this.deployTicks);
        });
        if (freshCreeps.length >= MAX_PARTY_SIZE) {
            trace.info('we have enough creeps', {
                freshCreeps: freshCreeps.map((creep) => creep.name),
                deployTicks: this.deployTicks,
                lowestTTL: ___namespace.min(creeps.filter((creep) => creep.ticksToLive)
                    .map((creep) => creep.ticksToLive)),
            });
            return;
        }
        const idx = ___namespace.max([creeps.length - 1, 0]) % 4;
        const position = this.getCreepPosition(idx, trace);
        const priority = this.priority;
        const ttl = this.requestCreepTTL + Game.time;
        const role = this.role;
        const memory = {
            [MEMORY_PARTY_ID]: this.id,
            // Tell creep to move to it's rally point position
            [MEMORY_POSITION_X]: position.x,
            [MEMORY_POSITION_Y]: position.y,
            [MEMORY_POSITION_ROOM]: position.roomName,
            [MEMORY_ASSIGN_ROOM]: position.roomName,
            [MEMORY_BASE]: base.id,
        };
        const request = createSpawnRequest(priority, ttl, role, memory, this.parts, 0);
        trace.notice('requesting creep', {
            deployTicks: this.deployTicks,
            request,
        });
        kernel.getTopics().addRequestV2(getBaseSpawnTopic(base.id), request);
    }
}

// get muster point for a party to form up on
function getMusterVector(base) {
    const level = getBaseLevel(base);
    const musterPoints = baseLayouts[level].muster;
    // check muster points until we find one that is not obstructed
    for (let i = 0; i < musterPoints.length; i++) {
        const point = musterPoints[i];
        for (let j = 0; j < MAX_PARTY_SIZE; j++) {
            const origin = base.origin;
            const x = origin.x + point.x + DIRECTION_OFFSET[point.direction].x * j;
            const y = origin.y + point.y + DIRECTION_OFFSET[point.direction].y * j;
            const pos = new RoomPosition(x, y, base.primary);
            // if structure is present, skip this muster point
            if (pos.lookFor(LOOK_STRUCTURES).length) {
                continue;
            }
            // if construction site is present, skip this muster point
            if (pos.lookFor(LOOK_CONSTRUCTION_SITES).length) {
                continue;
            }
            // if wall is present, skip this muster point
            const roomTerrain = Game.rooms[base.primary].getTerrain();
            if (roomTerrain.get(x, y) === TERRAIN_MASK_WALL) {
                continue;
            }
        }
        const x = base.origin.x + point.x;
        const y = base.origin.y + point.y;
        return {
            pos: new RoomPosition(x, y, base.primary),
            direction: point.direction,
            magnitude: MAX_PARTY_SIZE,
        };
    }
    return null;
}

class MusterDebugger {
    constructor(id) {
        this.id = id;
    }
    run(kernel, trace) {
        trace.info('muster debugger', {});
        if (this.vector) {
            visualizeVector(this.vector, trace);
        }
        return running();
    }
    debug(roomName) {
        const kernel = global.AI;
        const base = kernel.getPlanner().getBaseByRoom(roomName);
        this.vector = getMusterVector(base);
    }
    clear() {
        this.vector = null;
    }
}
function visualizeVector(vector, trace) {
    const roomVisual = Game.rooms[vector.pos.roomName].visual;
    if (!roomVisual) {
        trace.warn('no room visual', { room: vector.pos.roomName });
        return;
    }
    for (let i = 0; i < vector.magnitude; i++) {
        let char = '0';
        if (i === 0) {
            char = 'X';
        }
        const x = vector.pos.x + i * DIRECTION_OFFSET[vector.direction].x;
        const y = vector.pos.y + i * DIRECTION_OFFSET[vector.direction].y;
        const pos = new RoomPosition(x, y, vector.pos.roomName);
        roomVisual.text(char, pos);
    }
}

const ATTACK_ROOM_TTL = 500;
var AttackStatus;
(function (AttackStatus) {
    AttackStatus["REQUESTED"] = "requested";
    AttackStatus["COMPLETED"] = "completed";
})(AttackStatus || (AttackStatus = {}));
var Phase;
(function (Phase) {
    Phase["PHASE_MARSHAL"] = "marshal";
    Phase["PHASE_EN_ROUTE"] = "en_route";
    Phase["PHASE_ATTACK"] = "attack";
})(Phase || (Phase = {}));

// Spawn requests
const PRIORITY_DISTRIBUTOR = 19; // if zero distributes priority +10
const PRIORITY_MINER_PRIMARY = 17; // primary room priority
const PRIORITY_HAULER = 15;
const PRIORITY_DEFENDER = 15; // hostile presence in base rooms
const PRIORITY_BUFFER_PATROL = 15;
const PRIORITY_UPGRADER = 14; // logic reduces priority: n * 2
const PRIORITY_MINER_REMOTE = 13; // remote rooms should have lower priority then base ops
const PRIORITY_RESERVER = 13; // many things request reservers/claimers
const PRIORITY_REPAIRER_URGENT = 12;
const PRIORITY_BUILDER = 11;
const EXPLORER = 10;
const DISTRIBUTOR_NO_RESERVE = 9; // no energy stored, don't use regular distributor
const PRIORITY_ATTACKER = 8;
const PRIORITY_HARASSER = 8;
const PRIORITY_REPAIRER = 7;
const TERMINAL_BUY = 2;
const TERMINAL_TRANSFER = 3;
const TERMINAL_ENERGY_BALANCE = 4;
// Dropped/Source hauling
const DUMP_NEXT_TO_STORAGE = -10;
const HAUL_DROPPED = 5;
const HAUL_CONTAINER = 0;
const HAUL_BASE_ROOM = 5;
const LOAD_FACTOR = 0.2;
// Core Hauling
const UNLOAD_LINK = 2;
const LOAD_LINK = 1.9;
const UNLOAD_BOOST = 1.7;
const HAUL_BOOST = 1.6;
const HAUL_TOWER_HOSTILES = 1.5;
const HAUL_TOWER = 1.5;
const HAUL_EXTENSION = 1.4;
const HAUL_TERMINAL = 1.1;
const HAUL_REACTION = 1.0;
const HAUL_NUKER = 0.8;
// Reactions
const REACTION_PRIORITIES = {
    G: 12,
    OH: 13,
    ZK: 9,
    UL: 9,
    LH: 1,
    ZH: 6,
    KH: 4,
    UH: 6,
    GH: 14,
    LO: 9,
    ZO: 1,
    KO: 6,
    UO: 1,
    GO: 5,
    LH2O: 1,
    KH2O: 5,
    ZH2O: 6,
    UH2O: 7,
    GH2O: 15,
    LHO2: 10,
    UHO2: 1,
    KHO2: 6,
    ZHO2: 1,
    GHO2: 6,
    XLH2O: 1,
    XKH2O: 6,
    XZH2O: 8,
    XUH2O: 8,
    XGH2O: 16,
    XLHO2: 11,
    XUHO2: 1,
    XKHO2: 7,
    XZHO2: 1,
    XGHO2: 7, // damage
};

const MEMORY_CLEANUP_TTL = 1000;
const MEMORY_OBJECT_TTL = 5000;
class PersistentMemory {
    constructor(id) {
        this.memoryId = id;
    }
    getMemory(trace) {
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const memoryObject = Memory.proc[this.memoryId];
        if (!memoryObject) {
            return {};
        }
        if (typeof memoryObject.stales === 'undefined') {
            trace.warn('stale not set', { memoryObject });
            memoryObject.stales = true;
        }
        // Don't return exceptionally old memory objects
        if (memoryObject.stales && memoryObject.time < Game.time - MEMORY_OBJECT_TTL) {
            trace.warn('memory object is stale', { memoryObject });
            return {};
        }
        return memoryObject.value;
    }
    setMemory(value, stales = true) {
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        Memory.proc[this.memoryId] = {
            id: this.memoryId,
            time: Game.time,
            stales: stales,
            value: value,
        };
    }
}
class MemoryManager {
    constructor() {
        // Ensure Memory storage is ready
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        if (!Memory.proc) {
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            Memory.proc = {};
        }
    }
    run(_kernel, _trace) {
        const now = Game.time;
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const memory = Memory;
        const proc = memory.proc;
        for (const key in proc) {
            const obj = proc[key];
            if (typeof obj.stales === 'undefined') {
                obj.stales = true;
            }
            if (obj.stales && obj.time < now - MEMORY_OBJECT_TTL) {
                delete proc[key];
            }
        }
        return sleeping(MEMORY_CLEANUP_TTL);
    }
}

const MIN_HEALING_BOOST_AMOUNT = 2000;
function scoreRoomDamage(room) {
    let towerDamage = room.numTowers * 600;
    // if no storage or low energy, try to run the room out of energy
    if (!room.storage || room.storage.energy < 10000) {
        towerDamage = room.numTowers * 150;
    }
    return towerDamage + room.hostilesDmg;
}
function scoreStorageHealing(storage) {
    const healing2x = storage.store.getUsedCapacity(RESOURCE_LEMERGIUM_OXIDE) > MIN_HEALING_BOOST_AMOUNT;
    const healing3x = storage.store.getUsedCapacity(RESOURCE_LEMERGIUM_ALKALIDE) > MIN_HEALING_BOOST_AMOUNT;
    const healing4x = storage.store.getUsedCapacity(RESOURCE_CATALYZED_LEMERGIUM_ALKALIDE) > MIN_HEALING_BOOST_AMOUNT;
    if (healing4x) {
        return 4;
    }
    if (healing3x) {
        return 3;
    }
    if (healing2x) {
        return 2;
    }
    return 1;
}

const REQUEST_ATTACKER_TTL = 30;
const UPDATE_PARTS_INTERVAL = 50;
const WAR_PARTY_ROLE = WORKER_ATTACKER;
const CORNERS = new Map([
    [TOP, null],
    [RIGHT, null],
    [BOTTOM, null],
    [LEFT, null],
    [TOP_LEFT, { x: -1, y: -2 }],
    [TOP_RIGHT, { x: 2, y: -2 }],
    [BOTTOM_LEFT, { x: -1, y: 1 }],
    [BOTTOM_RIGHT, { x: 2, y: 1 }], // BR
]);
const ADJACENT_SIDES = new Map([
    [TOP, []],
    [RIGHT, []],
    [BOTTOM, []],
    [LEFT, []],
    [TOP_RIGHT, [TOP, RIGHT]],
    [BOTTOM_RIGHT, [BOTTOM, RIGHT]],
    [BOTTOM_LEFT, [BOTTOM, LEFT]],
    [TOP_LEFT, [TOP, LEFT]],
]);
const warPartyQuadPolicy = {
    room: {
        avoidFriendlyRooms: true,
        avoidHostileRooms: false,
        avoidRoomsWithKeepers: false,
        avoidRoomsWithTowers: true,
        avoidUnloggedRooms: false,
        sameRoomStatus: true,
        costMatrixType: AllowedCostMatrixTypes.PARTY,
    },
    destination: {
        range: 1,
    },
    path: {
        allowIncomplete: false,
        maxSearchRooms: 16,
        maxOps: 6000,
        maxPathRooms: 5,
        ignoreCreeps: false,
    },
};
const warPartySingleFilePolicy = {
    room: {
        avoidFriendlyRooms: true,
        avoidHostileRooms: false,
        avoidRoomsWithKeepers: false,
        avoidRoomsWithTowers: true,
        avoidUnloggedRooms: false,
        sameRoomStatus: true,
        costMatrixType: AllowedCostMatrixTypes.COMMON,
    },
    destination: {
        range: 1,
    },
    path: {
        allowIncomplete: false,
        maxSearchRooms: 16,
        maxOps: 6000,
        maxPathRooms: 5,
        ignoreCreeps: false,
    },
};
class WarPartyRunnable extends PersistentMemory {
    constructor(id, baseId, targetRoom, muster, createdAt) {
        var _a;
        super(id);
        // Minimum fields
        this.id = id;
        this.baseId = baseId;
        this.targetRoom = targetRoom;
        this.createdAt = createdAt;
        this.muster = muster;
        this.phase = Phase.PHASE_MARSHAL;
        this.position = this.muster.pos;
        this.destination = new RoomPosition(25, 25, targetRoom);
        this.range = 3;
        this.direction = BOTTOM;
        this.formation = FORMATION_QUAD;
        // set previous positions to muster line
        const previousPositions = [];
        for (let i = 0; i < MAX_PARTY_SIZE; i++) {
            const x = this.position.x + DIRECTION_OFFSET[muster.direction].x * i;
            const y = this.position.y + DIRECTION_OFFSET[muster.direction].y * i;
            previousPositions.push(new RoomPosition(x, y, this.position.roomName));
        }
        this.previousPositions = previousPositions;
        this.parts = null;
        this.roomDamage = null;
        this.minEnergy = ((_a = DEFINITIONS.get(WAR_PARTY_ROLE)) === null || _a === void 0 ? void 0 : _a.energyMinimum) || 0;
        this.pathDestination = null;
        this.path = [];
        this.pathTime = 0;
        this.threadUpdateParts = threadBaseRoom('update_parts', UPDATE_PARTS_INTERVAL)(this.updateParts.bind(this));
    }
    run(kernel, trace) {
        var _a, _b, _c;
        trace = trace.begin('warparty_run');
        this.restoreMemory(trace);
        this.party = new PartyRunnable(this.id, this.baseId, this.previousPositions, WAR_PARTY_ROLE, [], this.minEnergy, PRIORITY_ATTACKER, REQUEST_ATTACKER_TTL);
        this.party.setDirection(this.direction);
        this.party.setFormation(this.formation);
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.end();
        }
        const baseAggresiveFlag = Game.flags[`aggressive_${this.baseId}`];
        if (kernel.getConfig().temperament === Temperaments.Passive && !baseAggresiveFlag) {
            trace.error('No aggressive flag');
            // Party will terminate itself and cause War Party to terminate
            this.party.done();
        }
        // TODO use a war party specific topic for notifying of target change
        const targetRoom = this.targetRoom;
        const creeps = this.getAssignedCreeps();
        const targetRoomObject = Game.rooms[targetRoom];
        const positionRoomObject = Game.rooms[this.position.roomName];
        const targetRoomEntry = kernel.getScribe().getRoomById(targetRoom);
        if (!targetRoomEntry) {
            this.updateMemory(trace);
            trace.end();
            trace.error('no target room entry', { targetRoom });
            return running();
        }
        const baseRoom = getBasePrimaryRoom(base);
        if (!baseRoom) {
            this.updateMemory(trace);
            trace.end();
            trace.error('no base room', { base: this.baseId });
            return running();
        }
        // If no parts, update parts
        if (!this.parts) {
            this.updateParts(trace, kernel, base, baseRoom, targetRoomEntry);
        }
        else {
            this.threadUpdateParts(trace, kernel, base, baseRoom, targetRoomEntry);
        }
        if (!targetRoom) {
            trace.error('no target room, terminating war party');
            // Party will terminate itself and cause War Party to terminate
            this.party.done();
        }
        trace.info('war party run', {
            id: this.id,
            baseId: base.id,
            primaryRoomId: base.primary,
            targetRoom,
            muster: this.muster,
            phase: this.phase,
            position: this.position,
            roomDamage: this.roomDamage,
            formation: this.getFormation(),
            previousPositions: this.getPreviousPositions(),
            creeps: creeps.length,
        });
        this.setHeal(trace);
        let targetPosition = new RoomPosition(25, 25, this.targetRoom);
        const roomEntry = kernel.getScribe().getRoomById(this.targetRoom);
        if (!roomEntry) {
            trace.info(`no room entry for ${this.targetRoom}, using center of room`);
            // TODO should probably delay until we have a room entry
        }
        else if (roomEntry.spawnLocation) {
            trace.info('setting spawn as target position', { pos: roomEntry.spawnLocation });
            // TODO fix issue with restored from memory room positions not having functions
            targetPosition = new RoomPosition(roomEntry.spawnLocation.x, roomEntry.spawnLocation.y, roomEntry.spawnLocation.roomName);
        }
        else if ((_a = roomEntry.controller) === null || _a === void 0 ? void 0 : _a.pos) {
            trace.info('setting controller as target position', { pos: roomEntry.controller.pos });
            // TODO fix issue with restored from memory room positions not having functions
            targetPosition = new RoomPosition(roomEntry.controller.pos.x, roomEntry.controller.pos.y, roomEntry.controller.pos.roomName);
        }
        // marshal
        if (this.phase === Phase.PHASE_MARSHAL) {
            this.setFormation(FORMATION_SINGLE_FILE);
            // If we have at least 4 creeps and they are in position, begin deployment
            if (this.inPosition(this.position, trace) && creeps.length >= 4) {
                this.phase = Phase.PHASE_EN_ROUTE;
                trace.info('moving to en route phase', { phase: this.phase });
            }
            else {
                this.setDestination(targetPosition, 3);
                this.position = this.muster.pos;
                this.marshal(this.position, creeps, trace);
            }
        }
        // deploy
        if (this.phase === Phase.PHASE_EN_ROUTE) {
            // If we are out of creep, remarshal
            if (!creeps.length || (((_b = this.position.findClosestByRange(creeps)) === null || _b === void 0 ? void 0 : _b.pos.getRangeTo(this.position)) > 5)) {
                this.phase = Phase.PHASE_MARSHAL;
                trace.info('moving to marshal phase', { phase: this.phase });
            }
            else if (targetRoom === this.position.roomName) {
                this.phase = Phase.PHASE_ATTACK;
                trace.info('moving to attack phase', { phase: this.phase });
            }
            else {
                this.setDestination(targetPosition, 3);
                this.deploy(kernel, positionRoomObject, targetRoom, creeps, trace);
            }
        }
        // attack
        if (this.phase === Phase.PHASE_ATTACK) {
            this.setFormation(FORMATION_QUAD);
            const numPartyInTargetRoom = this.getAssignedCreeps().
                filter((creep) => creep.room.name === this.targetRoom).length;
            if (!numPartyInTargetRoom || !targetRoomObject || !creeps.length ||
                ((_c = this.position.findClosestByRange(creeps)) === null || _c === void 0 ? void 0 : _c.pos.getRangeTo(this.position)) > 5) {
                this.phase = Phase.PHASE_MARSHAL;
                const roomName = base.primary;
                const roomObject = Game.rooms[roomName];
                if (!roomObject) {
                    trace.error(`no room object for ${roomName}`);
                }
                const energyCapacityAvailable = roomObject.energyCapacityAvailable;
                this.minEnergy = ___namespace.min([this.minEnergy + 1000, energyCapacityAvailable]);
                this.party.setMinEnergy(this.minEnergy);
                trace.info('moving to marshal phase', { phase: this.phase });
            }
            else {
                const done = this.engage(kernel, targetRoomObject, creeps, trace);
                if (done) {
                    trace.notice('done, notify war manager that room is cleared', { targetRoom: this.targetRoom });
                    // Inform that attack is completed
                    const attackUpdate = {
                        status: AttackStatus.COMPLETED,
                        roomId: targetRoom,
                    };
                    kernel.getTopics().addRequest(ATTACK_ROOM, 1, attackUpdate, ATTACK_ROOM_TTL + Game.time);
                    // TODO go into waiting for orders phase
                    // Done, terminate
                    this.party.done();
                }
            }
        }
        // Tick the party along
        const partyResult = this.party.run(kernel, trace);
        if (partyResult.status === STATUS_TERMINATED) {
            this.updateMemory(trace);
            trace.info('party terminated');
            trace.end();
            return partyResult;
        }
        if (global.LOG_WHEN_PID === this.id) {
            new RoomVisual(this.position.roomName).text('x', this.position.x, this.position.y);
            this.visualizePathToTarget(kernel, this.position, this.destination, this.range, trace);
        }
        this.updateMemory(trace);
        trace.end();
        return running();
    }
    restoreMemory(trace) {
        const memory = this.getMemory(trace);
        if (memory.phase) {
            this.phase = memory.phase;
        }
        if (memory.position) {
            this.position = new RoomPosition(memory.position.x, memory.position.y, memory.position.roomName);
        }
        if (memory.destination) {
            this.destination = new RoomPosition(memory.destination.x, memory.destination.y, memory.destination.roomName);
        }
        if (memory.range) {
            this.range = memory.range;
        }
        if (memory.previousPositions) {
            this.previousPositions = memory.previousPositions.map((pos) => {
                return new RoomPosition(pos.x, pos.y, pos.roomName);
            });
        }
        if (memory.direction) {
            this.direction = memory.direction;
        }
        if (memory.formation) {
            this.formation = memory.formation;
        }
    }
    updateMemory(_trace) {
        this.setMemory({
            phase: this.phase,
            position: this.position,
            destination: this.destination,
            range: this.range,
            direction: this.party.getDirection(),
            formation: this.party.getFormation(),
            previousPositions: this.party.getPreviousPositions(),
        });
    }
    updateParts(trace, kernel, base, baseRoom, targetRoomEntry) {
        const boosts = newMultipliers();
        const baseStorage = baseRoom.storage;
        if (baseStorage) {
            const availableHealingBoost = scoreStorageHealing(baseStorage);
            boosts[HEAL] = availableHealingBoost;
        }
        const availableEnergyCapacity = baseRoom.energyCapacityAvailable;
        // We want at least one creep attacking per quad, so ensure tank works with 3 creeps
        const roomDamage = scoreRoomDamage(targetRoomEntry) / 4;
        const [parts, ok] = buildAttacker(roomDamage, availableEnergyCapacity, boosts, trace);
        if (!ok) {
            // TODO suicide?
            trace.error('failed to build parts', { roomDamage, availableEnergyCapacity, boosts });
            return;
        }
        trace.info('updating parts', { parts });
        this.setParts(parts);
        this.roomDamage = roomDamage;
    }
    marshal(position, creeps, trace) {
        if (!creeps.length) {
            return;
        }
        this.setPosition(this.muster.pos, trace);
    }
    deploy(kernel, room, targetRoom, creeps, trace) {
        trace.info('deploy', {
            targetRoom,
            position: this.position,
            destination: this.destination,
        });
        const [nextPosition, direction, blockers] = this.getNextPosition(kernel, this.position, this.destination, this.range, trace);
        trace.info('next position', { targetRoom, nextPosition, blockers: blockers.map((blocker) => blocker.id) });
        if (nextPosition) {
            trace.info('setting next position', { nextPosition });
            this.setPosition(nextPosition, trace);
        }
        else {
            trace.info('no next position');
        }
        // Update direction
        this.direction = direction;
        let targets = [];
        const dontAttack = kernel.getConfig().friends.concat(kernel.getConfig().neutral);
        if (room) {
            // determine target (hostile creeps, towers, spawns, nukes, all other structures)
            targets = targets.concat(room.find(FIND_HOSTILE_CREEPS, {
                filter: (creep) => dontAttack.indexOf(creep.owner.username) === -1,
            }));
        }
        if (blockers.length) {
            trace.info('blockers', { blocked: blockers.map((structure) => structure.id) });
            targets = targets.concat(blockers);
        }
        if (targets.length) {
            targets = ___namespace.sortBy(targets, (target) => {
                return creeps[0].pos.getRangeTo(target);
            });
            trace.info('targets', { targetsLength: targets.length });
            const target = this.party.setTarget(targets, trace);
            if (target) {
                this.alignWithTarget(target, nextPosition, trace);
            }
        }
        else {
            trace.info('no targets');
        }
    }
    engage(kernel, room, creeps, trace) {
        var _a;
        let destination = new RoomPosition(25, 25, this.targetRoom);
        let range = 3;
        if (room && room.controller) {
            destination = room.controller.pos;
        }
        let targets = [];
        // If we have visibility into the room, get targets and choose first as destination
        if (room) {
            if ((((_a = room.controller) === null || _a === void 0 ? void 0 : _a.safeMode) || 0) > 500) {
                trace.notice('room is in safe mode, ending party');
                return true;
            }
            targets = this.getTargets(kernel, room);
            if (targets.length) {
                trace.info('target', targets[0]);
                destination = targets[0].pos;
                range = 1;
                // } else if (room.controller?.my) {
                //   trace.info('room is owned by me, ending party');
                //   return true;
                // } else if (room.controller?.level > 0) {
                //   trace.info('no targets, but room level is > 0, not ending party');
            }
            else {
                trace.info('no targets, done');
                return true;
            }
        }
        this.setDestination(destination, range);
        const [nextPosition, direction, blockers] = this.getNextPosition(kernel, this.position, this.destination, this.range, trace);
        trace.info('next position', { nextPosition, blockers: blockers.map((blocker) => blocker.id) });
        // Commented this out until direction actually matters
        // const directionChanged = direction != this.direction;
        // if (directionChanged) {
        //  trace.info("changing direction", {direction});
        //  this.setDirection(direction);
        // } else
        if (nextPosition) {
            trace.info('setting next position', { nextPosition });
            this.setPosition(nextPosition, trace);
        }
        else {
            trace.info('no next position');
        }
        // Update direction
        this.direction = direction;
        let nearbyTargets = [];
        if (room) {
            const friends = kernel.getConfig().friends;
            // determine target (hostile creeps, towers, spawns, nukes, all other structures)
            nearbyTargets = nearbyTargets.concat(this.position.findInRange(FIND_HOSTILE_CREEPS, 2, {
                filter: (creep) => friends.indexOf(creep.owner.username) === -1,
            }));
        }
        if (blockers.length) {
            trace.info('blockers', { blocked: blockers.map((structure) => structure.id) });
            nearbyTargets = nearbyTargets.concat(blockers);
        }
        // Add other targets
        if (targets.length) {
            nearbyTargets = nearbyTargets.concat(targets);
        }
        if (nearbyTargets.length) {
            nearbyTargets = ___namespace.sortBy(nearbyTargets, (target) => {
                return this.position.getRangeTo(target);
            });
            trace.info('nearby targets', { nearByTargetsLength: nearbyTargets.length });
            const target = this.party.setTarget(nearbyTargets, trace);
            if (target) {
                this.alignWithTarget(target, nextPosition, trace);
            }
        }
        else {
            trace.info('no targets');
            return false;
        }
        return false;
    }
    alignWithTarget(target, position, trace) {
        let inCorner = null;
        for (const [direction, corner] of CORNERS) {
            if (!corner) {
                return;
            }
            trace.info('corner', { corner, direction });
            const x = ___namespace.min([___namespace.max([this.position.x + corner.x, 0]), 49]);
            const y = ___namespace.min([___namespace.max([this.position.y + corner.y, 0]), 49]);
            const cornerPosition = new RoomPosition(x, y, this.position.roomName);
            trace.info('cornerPosition', { cornerPosition });
            if (target.pos.isEqualTo(cornerPosition)) {
                inCorner = direction;
            }
        }
        if (inCorner) {
            trace.info('in corner', { inCorner });
            const sides = ADJACENT_SIDES.get(inCorner);
            if (sides.length) {
                trace.info('sides', { sides });
                const side = ___namespace.find(sides, (side) => {
                    const shiftedPosition = this.party.shiftPosition(position, side);
                    return !this.isBlocked(shiftedPosition, trace);
                });
                if (side) {
                    trace.info('side', { side });
                    const shiftPosition = this.party.shiftPosition(position, side);
                    if (shiftPosition) {
                        trace.info('shifting position', { shiftPosition });
                        this.setPosition(shiftPosition, trace);
                    }
                }
            }
        }
    }
    getTargets(kernel, room) {
        const friends = kernel.getConfig().friends;
        let targets = [];
        // determine target (hostile creeps, towers, spawns, nukes, all other structures)
        targets = targets.concat(room.find(FIND_HOSTILE_STRUCTURES, {
            filter: (structure) => structure.structureType === STRUCTURE_TOWER &&
                friends.indexOf(structure.owner.username) === -1,
        }));
        targets = targets.concat(room.find(FIND_HOSTILE_STRUCTURES, {
            filter: (structure) => structure.structureType === STRUCTURE_SPAWN &&
                friends.indexOf(structure.owner.username) === -1,
        }));
        targets = targets.concat(room.find(FIND_HOSTILE_STRUCTURES, {
            filter: (structure) => structure.structureType === STRUCTURE_NUKER &&
                friends.indexOf(structure.owner.username) === -1,
        }));
        // Hostile creeps
        targets = targets.concat(room.find(FIND_HOSTILE_CREEPS, {
            filter: (creep) => friends.indexOf(creep.owner.username) === -1,
        }));
        // Hostile structures
        targets = targets.concat(room.find(FIND_HOSTILE_STRUCTURES, {
            filter: (structure) => {
                if (structure.structureType === STRUCTURE_CONTROLLER ||
                    structure.structureType === STRUCTURE_RAMPART) {
                    return false;
                }
                return friends.indexOf(structure.owner.username) === -1;
            },
        }));
        // Remove walls by controller so our reservers can block upgrading
        if (room.controller) {
            const wallsNearController = room.controller.pos.findInRange(FIND_STRUCTURES, 1, {
                filter: (structure) => {
                    return structure.structureType === STRUCTURE_WALL ||
                        structure.structureType === STRUCTURE_RAMPART;
                },
            });
            targets = targets.concat(wallsNearController);
        }
        // NOTE: testing room clearing behavior - 08/06/22
        // targets = targets.concat(room.find(FIND_STRUCTURES, {
        //   filter: (structure) => {
        //     if (structure.structureType === STRUCTURE_CONTROLLER) {
        //       return false;
        //     }
        //     if (structure instanceof OwnedStructure && structure.owner) {
        //       const structureOwner = structure.owner.username;
        //       if (structureOwner && kernel.getConfig().friends.indexOf(structureOwner) !== -1) {
        //         return false;
        //       }
        //       return true;
        //     }
        //     const roomOwner = structure.room.controller?.owner?.username;
        //     if (roomOwner && kernel.getConfig().friends.indexOf(roomOwner) !== -1) {
        //       return false;
        //     }
        //     return true;
        //   },
        // }));
        return targets;
    }
    setParts(parts) {
        this.parts = parts;
        this.party.setParts(parts);
    }
    getPreviousPositions() {
        return this.party.getPreviousPositions();
    }
    getAssignedCreeps() {
        return this.party.getAssignedCreeps();
    }
    isCreepLowHealth(creeps) {
        let lowHealth = false;
        creeps.forEach((creep) => {
            if (creep.hits === 0) {
                return lowHealth = true;
            }
            if (creep.hitsMax / creep.hits < 0.8) {
                return lowHealth = true;
            }
        });
        return lowHealth;
    }
    visualizePathToTarget(kernel, origin, destination, range, trace) {
        const path = this.getPath(kernel, origin, destination, range, trace);
        if (!path) {
            trace.log('no path to visualize');
            return;
        }
        visualizePath(path);
    }
    getPath(kernel, origin, destination, range, trace) {
        trace.info('get path', { path: this.path, pathDestination: this.pathDestination, destination });
        if (this.path && this.pathDestination && this.pathDestination.isEqualTo(destination) &&
            Game.time - this.pathTime < 50) {
            trace.info('path cache hit', { pathLength: this.path.length, ttl: Game.time - this.pathTime, origin, destination });
            return this.path;
        }
        trace.info('war party path cache miss', { origin, destination });
        this.pathDestination = destination;
        this.pathTime = Game.time;
        warPartyQuadPolicy.destination.range = range;
        let [result, debug] = getPath(kernel, origin, destination, warPartyQuadPolicy, trace);
        trace.info('quad search', {
            origin: origin,
            destination: destination,
            result,
            debug,
        });
        if (result && !result.incomplete) {
            this.setFormation(FORMATION_QUAD);
        }
        else {
            warPartySingleFilePolicy.destination.range = range;
            [result, debug] = getPath(kernel, origin, destination, warPartySingleFilePolicy, trace);
            trace.info('single file search', {
                origin: origin,
                destination: destination,
                result,
                debug,
            });
            // set single file formation
            this.setFormation(FORMATION_SINGLE_FILE);
        }
        if (!result) {
            trace.info('no path found', { origin, destination });
            this.path = [];
        }
        else {
            this.path = result.path;
            this.path = [origin].concat(this.path);
        }
        return this.path;
    }
    getNextPosition(kernel, currentPosition, destination, range, trace) {
        // Figure out where we are going
        const path = this.getPath(kernel, currentPosition, destination, range, trace);
        if (!path) {
            // Cant find where we are going, freeze
            // TODO maybe suicide
            trace.warn('warparty stuck', { id: this.id });
            return [currentPosition, this.direction, []];
        }
        if (path.length === 0) {
            trace.error('no path', { id: this.id, currentPosition, destination, path });
            return [currentPosition, this.direction, []];
        }
        // We know where we are going and the path
        trace.info('path found', { pathLength: path.length, currentPosition, destination });
        // Work out the closest position along the path and it's distance
        // Scan path and find closest position, use that as as position on path
        const currentIndex = ___namespace.findIndex(path, (position) => {
            return position.isEqualTo(currentPosition);
        });
        // Log a message if we could not find an index, should not happen
        if (currentIndex < 0) {
            trace.warn('could not find origin/creep index', { currentIndex, currentPosition, path });
        }
        // Assume we are off path
        let nextIndex = currentIndex;
        // Advance the position by one if creeps are ready, on an edge, start of path
        if (this.inPosition(this.position, trace) || this.onEdge() || currentIndex === -1) {
            nextIndex++;
        }
        // Get the next position (may be same as current, if creeps are not in position)
        let nextPosition = path[nextIndex];
        if (!nextPosition) {
            trace.info('no next position', { nextIndex, path });
            return [currentPosition, this.direction, []];
        }
        let direction = this.direction;
        // Determine if we plan to move
        const positionChanged = !currentPosition.isEqualTo(nextPosition);
        if (positionChanged) {
            direction = currentPosition.getDirectionTo(nextPosition);
        }
        // Check if our path is blocked
        let blockers = [];
        if (direction && currentPosition.roomName === nextPosition.roomName) {
            blockers = this.getBlockingObjects(nextPosition, direction, trace);
        }
        // We are blocked, don't move and instead clear blockage
        if (blockers.length) {
            trace.info('blocked', { blockers, nextPosition, direction });
            nextPosition = currentPosition;
        }
        trace.info('get next position', {
            positionChanged,
            currentIndex,
            currentPosition,
            direction,
            blockers: blockers.map((blocker) => blocker.id),
            nextIndex,
            nextPosition,
            pathLength: path.length,
            destination,
        });
        return [nextPosition, direction, blockers];
    }
    setFormation(formation) {
        this.party.setFormation(formation);
    }
    getFormation() {
        return this.party.getFormation();
    }
    setDestination(destination, range) {
        this.destination = destination;
        this.range = range;
    }
    setDirection(direction) {
        this.party.setDirection(direction);
    }
    getPosition() {
        return this.party.getPosition();
    }
    inPosition(position, trace) {
        return this.party.inPosition(position, trace);
    }
    isBlocked(position, trace) {
        return this.party.isBlocked(position, trace);
    }
    getBlockingObjects(position, direction, trace) {
        return this.party.getBlockingObjects(position, direction, trace);
    }
    onEdge() {
        return this.position.x === 1 || this.position.x >= 48 || this.position.y <= 1 ||
            this.position.y === 49;
    }
    setPosition(position, _trace) {
        this.position = position;
        this.party.setPosition(position);
    }
    setTarget(targetRequests, trace) {
        return this.party.setTarget(targetRequests, trace);
    }
    setHeal(trace) {
        this.party.setHeal(trace);
    }
}

const BufferPathPolicy = {
    base: {
        start: 'spawn',
        maxLinearDistance: 5,
        minRoomLevel: 0,
        hasSpawn: true,
    },
    room: {
        avoidHostileRooms: true,
        avoidFriendlyRooms: true,
        avoidRoomsWithKeepers: false,
        avoidRoomsWithTowers: false,
        avoidUnloggedRooms: false,
        sameRoomStatus: true,
        costMatrixType: AllowedCostMatrixTypes.PARTY,
    },
    destination: {
        range: 2,
    },
    path: {
        allowIncomplete: false,
        maxSearchRooms: 16,
        maxOps: 5000,
        maxPathRooms: 6,
        ignoreCreeps: true,
    },
};
const RUN_TTL$8 = 10;
class BufferManager {
    constructor(id, _trace) {
        this.id = id;
    }
    run(kernel, trace) {
        trace = trace.begin('buffer_manager_run');
        const hostileRoomsByBase = getHostileRoomsByBase(kernel, trace);
        hostileRoomsByBase.forEach((rooms, baseId) => {
            if (rooms.length < 1) {
                trace.info('no hostiles rooms', { baseId, rooms });
                return;
            }
            const base = kernel.getPlanner().getBaseById(baseId);
            if (!base) {
                trace.info('no base', { baseId });
                return;
            }
            const room = ___namespace.sortByAll(rooms, ['level', 'id']).shift();
            if (!room) {
                return;
            }
            trace.info('attack hostile room', { baseId, room });
            const attackRequest = {
                status: AttackStatus.REQUESTED,
                roomId: room.id,
            };
            kernel.getTopics().addRequest(ATTACK_ROOM, 1, attackRequest, ATTACK_ROOM_TTL + Game.time);
        });
        // TODO add HUD line and attack lines on map
        trace.end();
        return sleeping(RUN_TTL$8);
    }
}
function getHostileRoomsByBase(kernel, trace) {
    const hostileRooms = kernel.getScribe().getHostileRooms(kernel, trace);
    trace.info('hostile rooms', { hostileRooms });
    const config = kernel.getConfig();
    const dontAttack = config.friends.concat(config.neutral);
    const candidateRooms = hostileRooms.filter((room) => {
        return dontAttack.indexOf(room.owner) === -1 && room.numKeyStructures > 0;
    });
    trace.info('candidate rooms', { config, dontAttack, candidateRooms });
    const hostileRoomsByBase = new Map();
    // TODO fix this
    BufferPathPolicy.base.maxLinearDistance = config.buffer;
    candidateRooms.forEach((room) => {
        const base = getClosestBaseByPath(kernel, room.controllerPos, BufferPathPolicy, trace);
        if (!base) {
            trace.info('no base', { room });
            return;
        }
        if (!hostileRoomsByBase.has(base.id)) {
            hostileRoomsByBase.set(base.id, []);
        }
        trace.info('attack room from', { base, room });
        hostileRoomsByBase.set(base.id, hostileRoomsByBase.get(base.id).concat(room));
    });
    trace.info('hostile rooms by base', { hostileRoomsByBase: Array.from(hostileRoomsByBase.values()) });
    return hostileRoomsByBase;
}

class PathDebugger {
    constructor(id, kernel) {
        this.id = id;
        this.results = null;
        this.resultsDebug = null;
        this.kernel = kernel;
    }
    run(kernel, trace) {
        trace.info('path debugger', { path: this.results });
        if (this.results) {
            visualizePath(this.results.path);
            for (const roomName of this.resultsDebug.searchedRooms) {
                if (this.resultsDebug.blockedRooms.has(roomName)) {
                    Game.map.visual.text('X', new RoomPosition(25, 25, roomName), { color: '#ff0000' });
                    return;
                }
                Game.map.visual.text('0', new RoomPosition(25, 25, roomName), { color: '#ff0000' });
            }
            _.each(this.resultsDebug.incompletePaths, (path) => {
                displayRoomPaths(path.path, { stroke: '#ff0000' });
                Game.map.visual.poly(path.path, { stroke: '#ff0000' });
            });
        }
        return running();
    }
    debug(origin, goal, range, policyName) {
        const trace = new Tracer('path_debugger_debug', new Map([['pid', 'path_debugger']]), new Metrics());
        trace.setLogFilter(global.LOG_WHEN_PID);
        let policy = null;
        switch (policyName) {
            case 'warparty_quad':
                policy = warPartyQuadPolicy;
                break;
            case 'warparty_single_file':
                policy = warPartySingleFilePolicy;
                break;
            case 'buffer_path':
                policy = BufferPathPolicy;
                break;
            case 'common':
                policy = commonPolicy;
                break;
            default:
                trace.error('policy not supported', { policyName });
                return;
        }
        const [path, debug] = getPath(this.kernel, origin, goal, policy, trace);
        trace.notice('path', { origin, goal, range, policy, path, debug });
        this.results = path;
        this.resultsDebug = debug;
    }
    clear() {
        this.results = null;
    }
}
const displayRoomPaths = (path, style) => {
    const pathByRooms = path.reduce((acc, pos) => {
        if (!acc.has(pos.roomName)) {
            acc.set(pos.roomName, []);
        }
        acc.get(pos.roomName).push(pos);
        return acc;
    }, new Map());
    // Display in the rooms
    Array.from(pathByRooms.entries()).forEach(([key, value]) => {
        new RoomVisual(key).poly(value, style);
    });
};

const CACHE_ITEM_TTL = 1000;
class PathCacheItem {
    constructor(originId, goalId, path, time) {
        this.originId = originId;
        this.goalId = goalId;
        this.value = path;
        this.time = time;
        this.hits = 0;
        this.next = null;
        this.prev = null;
    }
    add(item) {
        item.next = this;
        if (this.prev) {
            item.prev = this.prev;
            this.prev.next = item;
        }
        this.prev = item;
    }
    remove() {
        this.next.prev = this.prev;
        this.prev.next = this.next;
        this.next = null;
        this.prev = null;
    }
    isExpired(time) {
        return time - this.time > CACHE_ITEM_TTL;
    }
}
class PathCache {
    constructor(maxSize, pathProvider) {
        this.pathProvider = pathProvider;
        this.maxSize = maxSize;
        this.listCount = 0;
        this.originGoalToPathMap = {};
        this.head = new PathCacheItem(null, null, null, Game.time);
        this.tail = new PathCacheItem(null, null, null, Game.time);
        this.head.add(this.tail);
        this.hits = 0;
        this.misses = 0;
        this.expired = 0;
    }
    getPath(kernel, origin, goal, range, policy, trace) {
        const originId = this.getKey(origin, 0);
        const goalId = this.getKey(goal, range);
        let item = this.getCachedPath(originId, goalId, trace);
        // If no item, calculate path; otherwise, move item to top of LRU cache
        if (!item) {
            // trace.log('cache miss: calculating path', {origin, goal});
            const getPolicy = _.cloneDeep(policy);
            getPolicy.destination.range = range;
            const [result, _debug] = this.pathProvider(kernel, origin, goal, getPolicy, trace);
            if (!result) {
                return null;
            }
            item = this.setCachedPath(originId, goalId, result, Game.time, trace);
        }
        return item.value;
    }
    setCachedPath(originKey, destKey, value, time, trace) {
        const item = new PathCacheItem(originKey, destKey, value, time);
        // TODO move to add logic
        this.head.add(item);
        this.listCount += 1;
        // Add path to cache
        const origins = this.originGoalToPathMap[originKey];
        if (!origins) {
            this.originGoalToPathMap[originKey] = {};
        }
        this.originGoalToPathMap[originKey][destKey] = item;
        trace.info('set cache path', { originKey, destKey });
        if (this.listCount > this.maxSize) {
            const toRemove = this.tail.next;
            trace.info('over max', {
                count: this.listCount,
                max: this.maxSize,
                originKey: toRemove.originId,
                destKey: toRemove.goalId,
            });
            // TODO use remove logic
            if (toRemove && this.originGoalToPathMap[toRemove.originId]) {
                delete this.originGoalToPathMap[toRemove.originId][toRemove.goalId];
                toRemove.remove();
                this.listCount -= 1;
            }
        }
        return item;
    }
    getCachedPath(originKey, destKey, _trace) {
        // TODO move to hash lookup method
        const destinations = this.originGoalToPathMap[originKey];
        if (!destinations) {
            this.misses += 1;
            return null;
        }
        const item = destinations[destKey];
        if (!item) {
            this.misses += 1;
            return null;
        }
        this.hits += 1;
        item.hits += 1;
        // TOOO move to remove logic
        // Remove from linked list
        item.remove();
        // Remove from hash map
        if (this.originGoalToPathMap[item.originId] && this.originGoalToPathMap[item.originId][item.goalId]) {
            delete this.originGoalToPathMap[item.originId][item.goalId];
        }
        this.listCount -= 1;
        if (item.isExpired(Game.time)) {
            this.expired += 1;
            return null;
        }
        // TODO move to add logic
        // Add to linked list
        this.head.add(item);
        // Add to hash map
        if (!this.originGoalToPathMap[item.originId]) {
            this.originGoalToPathMap[item.originId] = {};
        }
        this.originGoalToPathMap[item.originId][item.goalId] = item;
        this.listCount += 1;
        return item;
    }
    getKey(pos, range = 0) {
        return `${pos.roomName}_${pos.x}_${pos.y}_${range}`;
    }
    getSize(trace) {
        let count = 0;
        let node = this.head;
        while (node.prev) {
            if (node.prev === node) {
                trace.error('aborting, hit self referencing node');
                break;
            }
            if (node.prev.prev === node) {
                trace.error('aborting, hit cyclical (3) referencing node');
                break;
            }
            if (count > this.maxSize * 1.1) {
                trace.error('aborting count, too large');
                break;
            }
            if (node.value) {
                count++;
            }
            node = node.prev;
        }
        return count;
    }
    loadFromMemory(trace) {
        trace = trace.begin('load_from_memory');
        const memory = Memory['path_cache'] || {};
        const paths = memory.paths || [];
        paths.forEach((path) => {
            path.value.path = path.value.path.map((position) => {
                return new RoomPosition(position.x, position.y, position.roomName);
            });
            this.setCachedPath(path.originId, path.goalId, path.value, path.time, trace);
        });
        trace.end();
    }
    saveToMemory(trace) {
        trace = trace.begin('save_to_memory');
        const paths = [];
        let path = this.tail;
        // eslint-disable-next-line no-cond-assign
        while (path = path.next) {
            if (!path.value) {
                continue;
            }
            paths.push({
                originId: path.originId,
                goalId: path.goalId,
                value: {
                    path: path.value.path.map((pos) => {
                        return {
                            x: pos.x,
                            y: pos.y,
                            roomName: pos.roomName,
                        };
                    }),
                },
                time: path.time,
            });
        }
        const rooms = [];
        Memory['path_cache'] = {
            paths,
            rooms,
        };
        trace.end();
    }
    getStats(trace) {
        return {
            cacheHits: this.hits,
            cacheMisses: this.misses,
            cacheExpired: this.expired,
            listCount: this.listCount,
            size: this.getSize(trace),
            max: this.maxSize,
        };
    }
    debug() {
        var _a;
        let node = this.head;
        let count = 0;
        while (node.prev) {
            console.log(`O:${node.originId}\tG:${node.goalId}\tH:${node.hits}\tA:${Game.time - node.time}\tL:${((_a = node === null || node === void 0 ? void 0 : node.value) === null || _a === void 0 ? void 0 : _a.path.length) || 0}`);
            node = node.prev;
            count++;
        }
        console.log(this.listCount, count);
    }
}

const DEFAULT_TTL = 500;
class Topics {
    constructor() {
        this.topics = new Map();
        this.lastCleanup = 0;
    }
    getTopics() {
        return this.topics;
    }
    getTopic(key) {
        if (this.lastCleanup < Game.time) {
            this.removeStale();
        }
        if (!this.topics.has(key)) {
            return null;
        }
        return this.topics.get(key);
    }
    setTopic(key, value) {
        if (!this.topics.has(key)) {
            return null;
        }
        this.topics.set(key, value);
    }
    reset() {
        this.topics = new Map();
    }
    removeStale() {
        for (const [key, _topic] of this.topics) {
            this.topics.set(key, this.topics.get(key).filter((request) => {
                return request.ttl >= Game.time;
            }));
        }
        this.lastCleanup = Game.time;
    }
    createTopic(key) {
        this.topics.set(key, []);
        return this.topics.get(key);
    }
    /**
     * @deprecated Use addRequestV2 instead.
     */
    addRequest(key, priority, details, ttl = DEFAULT_TTL) {
        let topic = this.getTopic(key);
        if (!topic) {
            topic = this.createTopic(key);
        }
        const request = {
            priority,
            details,
            ttl,
        };
        topic.push(request);
        this.topics.set(key, _.sortBy(topic, 'priority'));
    }
    addRequestV2(key, request) {
        let topic = this.getTopic(key);
        if (!topic) {
            topic = this.createTopic(key);
        }
        topic.push(request);
        // TODO doing this each message we push is a bit slow
        this.topics.set(key, _.sortBy(topic, 'priority'));
    }
    peekNextRequest(key) {
        const topic = this.getTopic(key);
        if (!topic) {
            return null;
        }
        if (!topic.length) {
            return null;
        }
        return topic[topic.length - 1];
    }
    getNextRequest(key) {
        const topic = this.getTopic(key);
        if (!topic) {
            return null;
        }
        let request = null;
        // eslint-disable-next-line no-cond-assign
        while (request = topic.pop()) {
            if (request.ttl < Game.time) {
                continue;
            }
            break;
        }
        this.setTopic(key, topic);
        return request;
    }
    getFilteredRequests(key, filter) {
        const topic = this.getTopic(key);
        if (!topic) {
            return [];
        }
        return topic.filter(filter);
    }
    getMessageOfMyChoice(key, chooser) {
        const messages = this.getTopic(key);
        if (!messages) {
            return null;
        }
        const choice = chooser(messages);
        if (choice) {
            _.remove(messages, choice);
        }
        this.setTopic(key, messages);
        return choice;
    }
    getLength(key) {
        const topic = this.topics.get(key);
        if (!topic) {
            return 0;
        }
        return topic.length;
    }
    reportMetrics(metrics) {
        // Report topic counts
        for (const [key, value] of this.getTopics()) {
            metrics.gauge(`topic_length`, value.length, { topic: key });
        }
    }
}

const TASK_HAUL = 'haul';
const TASK_MARKET_ORDER = 'order';
const TASK_TRANSFER = 'transfer';
const REACTION = 'reaction';

const setCommuteDuration = leafNode('bt.harvest.setCommuteDuration', (creep) => {
    const startTick = creep.memory[MEMORY_START_TICK];
    // Only set commute if we have a start tick and if we have not already
    // set the commute time
    if (startTick && !creep.memory[MEMORY_COMMUTE_DURATION]) {
        const commuteTime = Game.time - startTick + creep.body.length;
        creep.memory[MEMORY_COMMUTE_DURATION] = commuteTime;
    }
    return SUCCESS;
});
const creepIsFresh = (creep, offset = 0) => {
    if (creep.spawning) {
        return true;
    }
    const commute = creep.memory[MEMORY_COMMUTE_DURATION] || 150;
    const buildTime = creep.body.length * 3;
    return creep.ticksToLive > (commute + buildTime + offset);
};

// TODO replace with PDI class and method to reset
// eslint-disable-next-line @typescript-eslint/no-explicit-any
const globalAny = global;
globalAny.RESET_PIDS = false;
const setup = (memory, setPoint, p, i, d) => {
    if (!p) {
        throw new Error('missing p');
    }
    memory.set(PID_SUFFIX_SETPOINT, setPoint);
    memory.set(PID_SUFFIX_P, p);
    memory.set(PID_SUFFIX_I, i || 0);
    memory.set(PID_SUFFIX_D, d || 0);
};
const update = (memory, value, time, _trace) => {
    if (globalAny.RESET_PIDS) {
        memory.set(PID_SUFFIX_ERROR, 0);
        memory.set(PID_SUFFIX_TIME, time);
        memory.set(PID_SUFFIX_INTEGRAL, 0);
    }
    const setPoint = memory.get(PID_SUFFIX_SETPOINT);
    const p = memory.get(PID_SUFFIX_P) || 0.4;
    const i = memory.get(PID_SUFFIX_I) || 0.001;
    const d = memory.get(PID_SUFFIX_D) || 0;
    if (!p) {
        throw new Error('update: missing p');
    }
    const err = value - setPoint;
    const prevTime = memory.get(PID_SUFFIX_TIME) || time;
    const dt = time - prevTime;
    const prevIntegral = memory.get(PID_SUFFIX_INTEGRAL) || 0;
    let integral = prevIntegral + (err * dt * i);
    // Bootstrapping can require a lot of workers/haulers. 10 was too few (Jan 2022)
    // 50 seems too much
    if (integral > 40) {
        integral = 40;
    }
    else if (integral < -10) {
        integral = -10;
    }
    const prevErr = memory.get(PID_SUFFIX_ERROR) || err;
    let det = 0;
    if (dt > 0) {
        det = -((err - prevErr) / dt);
    }
    memory.set(PID_SUFFIX_ERROR, err);
    memory.set(PID_SUFFIX_TIME, time);
    memory.set(PID_SUFFIX_INTEGRAL, integral);
    const result = p * err + integral + d * det;
    // TODO move to HUD
    // const roomVisual = new RoomVisual(roomId);
    // roomVisual.text(`PID: ${result} = ${p} * ${err} + ${integral} + ${d} * ${det}`, 0, 1,
    //  {align: 'left'});
    return result;
};

/**
 * Base Logistics
 *
 * - Using a PID, try to ensure queue of jobs is empty (plenty of workers/haulers)
 * - Ensure enough distributors are filling extensions so that we are not waiting for energy
 *
 */
function getBaseDistributorTopic(baseId) {
    return `base_${baseId}_distributor`;
}
const RUN_TTL$7 = 5;
const CALCULATE_LEG_TTL = 20;
const BUILD_SHORTEST_LEG_TTL = 100;
const CONSUME_EVENTS_TTL$1 = 50;
const PRODUCE_EVENTS_TTL$2 = 100;
const REQUEST_HAULER_TTL = 50;
const UPDATE_HAULERS_TTL = 50;
const UPDATE_PID_TTL = 5;
const MIN_DISTRIBUTORS = 1;
const REQUEST_DISTRIBUTOR_TTL = 10;
const HAUL_EXTENSION_TTL = 10;
const REQUEST_HAUL_DROPPED_RESOURCES_TTL = 20;
const LEG_CALCULATE_INTERVAL = 500;
// More sites means more spent per load on road construction & maintenance
const MAX_ROAD_SITES = 5;
const MIN_LOAD_SIZE = 100;
const getLogisticsTopic = (baseId) => `${baseId}_logistics`;
function getBaseHaulerTopic(baseId) {
    return `base_${baseId}_hauler`;
}
var LogisticsEventType;
(function (LogisticsEventType) {
    LogisticsEventType["RequestRoad"] = "request_road";
})(LogisticsEventType || (LogisticsEventType = {}));
class LogisticsRunnable extends PersistentMemory {
    constructor(baseId) {
        super(baseId);
        this.numHaulers = 0;
        this.numActiveHaulers = 0;
        this.numIdleHaulers = 0;
        this.avgHaulerCapacity = 200;
        this.baseId = baseId;
        this.legs = null;
        this.selectedLeg = null;
        this.passes = 0;
        this.desiredHaulers = 0;
        this.pidHaulersMemory = null;
        this.logisticsStreamConsumer = null;
        this.threadConsumeEvents = threadBase('consume_events', CONSUME_EVENTS_TTL$1)(this.consumeEvents.bind(this));
        // Iterate through all destinations and calculate the remaining roads to build
        this.calculateLegIterator = this.calculateLegGenerator();
        this.threadCalculateLeg = threadBase('calculate_leg', CALCULATE_LEG_TTL)((trace, kernel, _base) => {
            this.calculateLegIterator.next({ trace, kernel });
        });
        this.threadUpdateHaulers = threadBase('update_haulers_thread', UPDATE_HAULERS_TTL)(this.updateHaulers.bind(this));
        this.threadHaulerPID = threadBase('hauler_pid', UPDATE_PID_TTL)(this.updatePID.bind(this));
        this.threadRequestHaulers = threadBase('request_haulers_thread', REQUEST_HAULER_TTL)(this.requestHaulers.bind(this));
        this.threadRequestDistributor = threadBaseRoom('request_distributer_thread', REQUEST_DISTRIBUTOR_TTL)(this.requestDistributor.bind(this));
        this.threadRequestExtensionFilling = threadBaseRoom('request_extension_filling_thread', HAUL_EXTENSION_TTL)(this.requestExtensionFilling.bind(this));
        this.threadRequestHaulDroppedResources = threadBase('request_haul_dropped_thread', REQUEST_HAUL_DROPPED_RESOURCES_TTL)(this.requestHaulDroppedResources.bind(this));
        this.threadRequestHaulTombstonesAndRuins = threadBase('request_haul_tombstone_thread', REQUEST_HAUL_DROPPED_RESOURCES_TTL)(this.requestHaulTombstonesAndRuins.bind(this));
        // From the calculated legs, select shortest to build and build it
        this.threadBuildShortestLeg = threadBase('select_leg', BUILD_SHORTEST_LEG_TTL)(this.buildShortestLeg.bind(this));
        // Walls may be built that block access to sources, check and remove any walls along the path and replace with road
        this.threadEnsureWallPassage = threadBase('ensure_wall_passage', BUILD_SHORTEST_LEG_TTL)(this.ensureWallPassage.bind(this));
        this.threadProduceEvents = threadBase('produce_events', PRODUCE_EVENTS_TTL$2)(this.produceEvents.bind(this));
    }
    run(kernel, trace) {
        // Setup the stream consumer
        if (this.logisticsStreamConsumer === null) {
            const streamId = getLogisticsTopic(this.baseId);
            this.logisticsStreamConsumer = kernel.getBroker().getStream(streamId).
                addConsumer('logistics');
        }
        if (this.legs === null) {
            const memory = this.getMemory(trace);
            if (memory.legs) {
                this.legs = new Map(memory.legs);
                // Hydrate the room positions
                for (const leg of this.legs.values()) {
                    if (leg.path) {
                        leg.path = leg.path.map((pos => new RoomPosition(pos.x, pos.y, pos.roomName)));
                    }
                    if (leg.remaining) {
                        leg.remaining = leg.remaining.map((pos => new RoomPosition(pos.x, pos.y, pos.roomName)));
                    }
                }
            }
            else {
                this.legs = new Map();
            }
            memory.legs = Array.from(this.legs.entries());
            this.setMemory(memory);
        }
        if (this.pidHaulersMemory === null) {
            const memory = this.getMemory(trace);
            if (memory.pid) {
                this.pidHaulersMemory = new Map(memory.pid);
            }
            else {
                this.pidHaulersMemory = new Map();
            }
            memory.pid = Array.from(this.pidHaulersMemory.entries());
            this.setMemory(memory);
            setup(this.pidHaulersMemory, 0, 0.08, 0.0015, 0);
        }
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.error('missing base', { id: this.baseId });
            return terminate();
        }
        const room = getBasePrimaryRoom(base);
        if (!room) {
            trace.error('missing primary room', { id: this.baseId });
            return sleeping(20);
        }
        trace.info('logistics', {
            id: this.baseId, legs: this.legs.size, passes: this.passes,
            selectedLeg: this.selectedLeg
        });
        this.threadConsumeEvents(trace, kernel, base);
        this.filterOldLegs(trace, kernel, base);
        // If red alert, don't do anything
        if (base.alertLevel === AlertLevel.GREEN) {
            this.threadCalculateLeg(trace, kernel, base);
            this.threadBuildShortestLeg(trace, kernel, base);
            this.threadEnsureWallPassage(trace, kernel, base);
        }
        this.threadUpdateHaulers(trace, kernel, base);
        this.threadRequestExtensionFilling(trace, kernel, base, room);
        this.threadRequestHaulDroppedResources(trace, kernel, base);
        this.threadRequestHaulTombstonesAndRuins(trace, kernel, base);
        this.threadHaulerPID(trace, kernel, base);
        this.threadRequestHaulers(trace, kernel, base);
        this.threadRequestDistributor(trace, kernel, base, room);
        this.threadProduceEvents(trace, kernel, base);
        // CLEANUP add LOG_WHEN_PID_CHECK
        if (this.selectedLeg) {
            visualizePath(this.selectedLeg.path);
        }
        visualizeLegNodes(Array.from(this.legs.values()), trace);
        return sleeping(RUN_TTL$7);
    }
    consumeEvents(trace, kernel, _base) {
        this.logisticsStreamConsumer.getEvents().forEach((event) => {
            switch (event.type) {
                case LogisticsEventType.RequestRoad:
                    this.requestRoad(kernel, event.data.id, event.data.position, event.time, trace);
                    break;
            }
        });
        // Storing legs
        trace.info('storing legs after consuming events', { legs: Array.from(this.legs.entries()) });
        const memory = this.getMemory(trace);
        memory.legs = Array.from(this.legs.entries());
        this.setMemory(memory);
    }
    filterOldLegs(trace, _kernel, base) {
        for (const [id, leg] of this.legs) {
            if (base.rooms.indexOf(leg.destination.roomName) === -1) {
                trace.info('removing leg', { id });
                this.legs.delete(id);
            }
        }
    }
    updateHaulers(_trace, kernel, _base) {
        // Get list of haulers and workers
        const haulers = kernel.getCreepsManager().getCreepsByBaseAndRole(this.baseId, WORKER_HAULER);
        const workers = kernel.getCreepsManager().getCreepsByBaseAndRole(this.baseId, ROLE_WORKER);
        this.haulers = haulers.concat(workers);
        this.numHaulers = this.haulers.length;
        this.numActiveHaulers = this.haulers.filter((creep) => {
            const task = creep.memory[MEMORY_TASK_TYPE];
            return task === TASK_HAUL || creep.store.getUsedCapacity() > 0;
        }).length;
        this.numIdleHaulers = this.numHaulers - this.numActiveHaulers;
        // Updating the avg when there are no haulers causes some undesirable
        // situations (task explosion)
        if (this.numHaulers) {
            this.avgHaulerCapacity = this.haulers.reduce((total, hauler) => {
                return total + hauler.store.getCapacity();
            }, 0) / this.haulers.length;
            if (this.avgHaulerCapacity < MIN_LOAD_SIZE) {
                this.avgHaulerCapacity = MIN_LOAD_SIZE;
            }
        }
        else {
            this.avgHaulerCapacity = MIN_LOAD_SIZE;
        }
    }
    updatePID(trace, kernel, base) {
        const numHaulTasks = kernel.getTopics().getLength(getBaseHaulerTopic(base.id));
        const demand = numHaulTasks - this.numIdleHaulers;
        this.desiredHaulers = update(this.pidHaulersMemory, demand, Game.time);
        trace.info('desired haulers', { desired: this.desiredHaulers });
        trace.getMetricsCollector().gauge('base_haulers_desired', this.desiredHaulers, { base: base.id });
        trace.getMetricsCollector().gauge('base_haulers_active', this.numActiveHaulers, { base: base.id });
        trace.getMetricsCollector().gauge('base_haulers_idle', this.numIdleHaulers, { base: base.id });
        trace.getMetricsCollector().gauge('base_haulers_tasks', numHaulTasks, { base: base.id });
        // Update PID memory
        trace.info('pid memory', { pid: Array.from(this.pidHaulersMemory.entries()) });
        const memory = this.getMemory(trace);
        memory.pid = Array.from(this.pidHaulersMemory.entries());
        this.setMemory(memory);
        const hudLine = {
            key: `pid_${this.baseId}`,
            room: base.primary,
            text: `Hauler PID: ${this.desiredHaulers.toFixed(2)}, ` +
                `Haul Tasks: ${numHaulTasks}, ` +
                `Num Haulers: ${this.numHaulers}, ` +
                `Active Haulers: ${this.numActiveHaulers}, ` +
                `Idle Haulers: ${this.numIdleHaulers}`,
            time: Game.time,
            order: 10,
        };
        kernel.getBroker().getStream(getLinesStream()).publish(new Event(this.baseId, Game.time, HudEventSet, hudLine));
    }
    requestHaulers(trace, kernel, base) {
        if (Game.cpu.bucket < 2000) {
            trace.warn('bucket is low, not requesting haulers', { bucket: Game.cpu.bucket });
            return;
        }
        const room = Game.rooms[base.primary];
        if (!room) {
            trace.warn('room not found', { room: base.primary });
            return;
        }
        let role = WORKER_HAULER;
        // if the room does not have storage, then request general workers instead of haulers
        if (!(room === null || room === void 0 ? void 0 : room.storage)) {
            role = ROLE_WORKER;
        }
        trace.info('request haulers/workers', {
            numHaulers: this.numHaulers, desiredHaulers: this.desiredHaulers,
            baseId: base.id
        });
        const neededHaulers = this.desiredHaulers - this.numHaulers;
        for (let i = 0; i < Math.min(3, neededHaulers); i++) {
            let priority = PRIORITY_HAULER;
            if (this.numHaulers === 0 && i === 0) {
                priority += 5;
            }
            // Reduce priority as we have more haulers
            priority -= (this.numHaulers + i) * 0.1;
            const ttl = REQUEST_HAULER_TTL + Game.time;
            const memory = {
                [MEMORY_BASE]: base.id,
            };
            const request = createSpawnRequest(priority, ttl, role, memory, null, 0);
            trace.info('requesting hauler/worker', { role, priority, request });
            kernel.getTopics().addRequestV2(getBaseSpawnTopic(base.id), request);
        }
    }
    requestDistributor(trace, kernel, base, room) {
        let distributors = kernel.getCreepsManager().getCreepsByBaseAndRole(this.baseId, WORKER_DISTRIBUTOR);
        distributors = distributors.filter((c) => {
            // Making stale 150 ticks sooner to avoid not having enough distributors
            return creepIsFresh(c, 150);
        });
        const numDistributors = distributors.length;
        // If low on bucket base not under threat
        if (Game.cpu.bucket < 1000 && numDistributors > 0 && base.alertLevel === AlertLevel.GREEN) {
            trace.info('low CPU, limit distributors');
            return;
        }
        let desiredDistributors = MIN_DISTRIBUTORS;
        if (room.controller.level < 3) {
            desiredDistributors = 1;
        }
        const fullness = room.energyAvailable / room.energyCapacityAvailable;
        if (room.controller.level >= 3 && fullness < 0.5) {
            desiredDistributors = 3;
        }
        if (base.alertLevel !== AlertLevel.GREEN) {
            const numTowers = room.find(FIND_MY_STRUCTURES, {
                filter: (structure) => {
                    return structure.structureType === STRUCTURE_TOWER;
                },
            }).length;
            trace.info('hostiles in room and we need more distributors', { numTowers, desiredDistributors });
            desiredDistributors = (numTowers / 2) + desiredDistributors;
        }
        if (!room.storage || numDistributors >= desiredDistributors) {
            trace.info('do not request distributors', {
                hasStorage: !!room.storage,
                numDistributors,
                desiredDistributors,
                roomLevel: room.controller.level,
                fullness,
            });
            return;
        }
        let distributorPriority = PRIORITY_DISTRIBUTOR;
        if (getStoredResourceAmount(base, RESOURCE_ENERGY) === 0) {
            distributorPriority = DISTRIBUTOR_NO_RESERVE;
        }
        const ttl = REQUEST_DISTRIBUTOR_TTL + Game.time;
        const role = WORKER_DISTRIBUTOR;
        const memory = {
            [MEMORY_ASSIGN_ROOM]: room.name,
            [MEMORY_BASE]: base.id,
        };
        for (let i = 0; i < desiredDistributors - numDistributors; i++) {
            const request = createSpawnRequest(distributorPriority, ttl, role, memory, null, 0);
            trace.info('request distributor', { desiredDistributors, fullness, request });
            kernel.getTopics().addRequestV2(getBaseSpawnTopic(base.id), request);
        }
    }
    produceEvents(trace, kernel, base) {
        var _a, _b, _c, _d;
        const hudLine = {
            key: `${this.baseId}`,
            room: base.primary,
            text: `Logistics - passes: ${this.passes}, legs: ${this.legs.size}, ` +
                `selected: ${(_a = this.selectedLeg) === null || _a === void 0 ? void 0 : _a.id}, numRemaining: ${(_b = this.selectedLeg) === null || _b === void 0 ? void 0 : _b.remaining.length}, ` +
                `end: ${(_d = (_c = this.selectedLeg) === null || _c === void 0 ? void 0 : _c.remaining) === null || _d === void 0 ? void 0 : _d.slice(-3, 0)}`,
            time: Game.time,
            order: 5,
        };
        kernel.getBroker().getStream(getLinesStream()).publish(new Event(this.baseId, Game.time, HudEventSet, hudLine));
    }
    requestExtensionFilling(trace, kernel, base, room) {
        const pickup = getStructureWithResource(base, RESOURCE_ENERGY);
        if (!pickup) {
            trace.info('no energy available for extensions', { resource: RESOURCE_ENERGY });
            return;
        }
        const nonFullExtensions = room.find(FIND_STRUCTURES, {
            filter: (structure) => {
                return (structure.structureType === STRUCTURE_EXTENSION ||
                    structure.structureType === STRUCTURE_SPAWN) &&
                    structure.store.getFreeCapacity(RESOURCE_ENERGY) > 0 &&
                    structure.isActive();
            },
        });
        nonFullExtensions.forEach((extension) => {
            const details = {
                [TASK_ID]: `ext-${extension.id}-${Game.time}`,
                [MEMORY_TASK_TYPE]: TASK_HAUL,
                [MEMORY_HAUL_PICKUP]: pickup.id,
                [MEMORY_HAUL_RESOURCE]: RESOURCE_ENERGY,
                [MEMORY_HAUL_DROPOFF]: extension.id,
                [MEMORY_HAUL_AMOUNT]: extension.store.getFreeCapacity(RESOURCE_ENERGY),
            };
            kernel.getTopics().addRequest(getBaseDistributorTopic(this.baseId), HAUL_EXTENSION, details, HAUL_EXTENSION_TTL + Game.time);
        });
        trace.info('haul extensions', { numHaulTasks: nonFullExtensions.length });
    }
    requestHaulDroppedResources(trace, kernel, base) {
        // iterate rooms and request haulers for any tombstones
        base.rooms.forEach((roomName) => {
            if (roomName === base.primary) {
                if (base.alertLevel === AlertLevel.RED) {
                    trace.warn('base under attack, do not haul dropped resources', { alertLevel: base.alertLevel });
                    return;
                }
            }
            else {
                if (base.alertLevel !== AlertLevel.GREEN) {
                    trace.warn('base threatened do not haul dropped resources in remotes', { alertLevel: base.alertLevel });
                    return;
                }
            }
            const room = Game.rooms[roomName];
            if (!room) {
                return;
            }
            // Get resources to haul
            const droppedResourcesToHaul = room.find(FIND_DROPPED_RESOURCES);
            // No resources to haul, we are done
            if (!droppedResourcesToHaul.length) {
                return;
            }
            trace.info('avg hauler capacity', { numHaulers: this.haulers.length, avgCapacity: this.avgHaulerCapacity });
            droppedResourcesToHaul.forEach((resource) => {
                var _a;
                const topic = getBaseHaulerTopic(base.id);
                let priority = HAUL_DROPPED;
                // Increase priority if primary room
                // TODO factor distance (or number of rooms from base)
                if (base.primary === resource.room.name) {
                    priority += HAUL_BASE_ROOM;
                }
                if ((_a = room.storage) === null || _a === void 0 ? void 0 : _a.pos.isNearTo(resource.pos)) {
                    priority += DUMP_NEXT_TO_STORAGE;
                }
                const dropoff = getStructureForResource(base, resource.resourceType);
                if (!dropoff) {
                    trace.info('no dropoff for resource', { resource: resource.resourceType, baseId: base.id });
                }
                const haulersWithTask = this.haulers.filter((creep) => {
                    const task = creep.memory[MEMORY_TASK_TYPE];
                    const pickup = creep.memory[MEMORY_HAUL_PICKUP];
                    return task === TASK_HAUL && pickup === resource.id;
                });
                const haulerCapacity = haulersWithTask.reduce((total, hauler) => {
                    return total += hauler.store.getFreeCapacity();
                }, 0);
                const untaskedUsedCapacity = resource.amount - haulerCapacity;
                const loadsToHaul = Math.floor(untaskedUsedCapacity / this.avgHaulerCapacity);
                trace.info('loads', {
                    avgHaulerCapacity: this.avgHaulerCapacity, haulerCapacity,
                    untaskedUsedCapacity, loadsToHaul,
                });
                for (let i = 0; i < loadsToHaul; i++) {
                    // Reduce priority for each load after first
                    const loadPriority = priority - LOAD_FACTOR * i;
                    const details = {
                        [TASK_ID]: `pickup-${this.baseId}-${Game.time}`,
                        [MEMORY_TASK_TYPE]: TASK_HAUL,
                        [MEMORY_HAUL_PICKUP]: resource.id,
                        [MEMORY_HAUL_DROPOFF]: (dropoff === null || dropoff === void 0 ? void 0 : dropoff.id) || undefined,
                        [MEMORY_HAUL_RESOURCE]: resource.resourceType,
                        [MEMORY_HAUL_AMOUNT]: resource.amount,
                    };
                    trace.info('haul dropped', { room: roomName, topic, i, loadPriority, details });
                    kernel.getTopics().addRequest(topic, loadPriority, details, REQUEST_HAUL_DROPPED_RESOURCES_TTL + Game.time);
                }
            });
        });
    }
    requestHaulTombstonesAndRuins(trace, kernel, base) {
        trace.info('base requesting haul tombstones and ruins', { baseId: base.id, rooms: base.rooms });
        base.rooms.forEach((roomName) => {
            if (roomName === base.primary) {
                if (base.alertLevel === AlertLevel.RED) {
                    trace.warn('base under attack, do not haul tombstones', { alertLevel: base.alertLevel });
                    return;
                }
            }
            else {
                if (base.alertLevel !== AlertLevel.GREEN) {
                    trace.warn('base threatened do not haul tombstones in remotes', { alertLevel: base.alertLevel });
                    return;
                }
            }
            const room = Game.rooms[roomName];
            if (!room) {
                trace.warn('room not found', { roomName });
                return;
            }
            const tombstones = room.find(FIND_TOMBSTONES, {
                filter: (tombstone) => {
                    if (tombstone.store.getUsedCapacity() === 0) {
                        return false;
                    }
                    const numAssigned = this.haulers.filter((hauler) => {
                        return hauler.memory[MEMORY_HAUL_PICKUP] === tombstone.id;
                    }).length;
                    return numAssigned === 0;
                },
            });
            const ruins = room.find(FIND_RUINS, {
                filter: (ruin) => {
                    if (ruin.store.getUsedCapacity() === 0) {
                        return false;
                    }
                    const numAssigned = this.haulers.filter((hauler) => {
                        return hauler.memory[MEMORY_HAUL_PICKUP] === ruin.id;
                    }).length;
                    return numAssigned === 0;
                },
            });
            let toHaul = [];
            toHaul = toHaul.concat(tombstones).concat(ruins);
            trace.info('found ruins and tombstones', { room: roomName, toHaul: toHaul.length, ruins: ruins.length, tombstones: tombstones.length });
            toHaul.forEach((haul) => {
                Object.keys(haul.store).forEach((resourceType) => {
                    trace.info('tombstone or ruin', { id: haul.id, resource: resourceType, amount: haul.store[resourceType] });
                    const dropoff = getStructureForResource(base, resourceType);
                    if (!dropoff) {
                        trace.info('no dropoff for resource', { resource: resourceType });
                    }
                    const details = {
                        [TASK_ID]: `pickup-${this.baseId}-${Game.time}`,
                        [MEMORY_TASK_TYPE]: TASK_HAUL,
                        [MEMORY_HAUL_PICKUP]: haul.id,
                        [MEMORY_HAUL_DROPOFF]: dropoff === null || dropoff === void 0 ? void 0 : dropoff.id,
                        [MEMORY_HAUL_RESOURCE]: resourceType,
                        [MEMORY_HAUL_AMOUNT]: haul.store.getUsedCapacity(resourceType),
                    };
                    const topic = getBaseHaulerTopic(base.id);
                    const priority = HAUL_DROPPED;
                    trace.info('haul tombstone or ruin', { topic, priority, details });
                    kernel.getTopics().addRequest(topic, priority, details, REQUEST_HAUL_DROPPED_RESOURCES_TTL + Game.time);
                });
            });
        });
    }
    requestRoad(kernel, id, destination, time, trace) {
        trace.info('request road', { id, destination, time });
        if (this.legs.has(id)) {
            const leg = this.legs.get(id);
            leg.destination = destination;
            leg.requestedAt = time;
            trace.info('leg already exists, updating', { leg });
            this.legs.set(id, leg);
        }
        else {
            const leg = {
                id,
                destination,
                path: null,
                remaining: [],
                requestedAt: time,
                updatedAt: null,
            };
            trace.info('does not exist, creating', { leg });
            this.legs.set(id, leg);
        }
    }
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    *calculateLegGenerator() {
        let legs = [];
        while (true) {
            const details = yield;
            const kernel = details.kernel;
            const trace = details.trace;
            trace.info('calculate legs pass', { num: legs.length, legs });
            if (!legs.length) {
                trace.info('updating legs to calculate');
                legs = this.getLegsToCalculate(trace);
            }
            trace.info('legs to update', {
                legs: legs.map((l) => {
                    return { id: l.id, updatedAt: l.updatedAt, remaining: l.remaining.length };
                }),
            });
            const leg = legs.shift();
            if (leg) {
                const [path, remaining] = this.calculateLeg(kernel, leg, trace);
                if (path) {
                    leg.path = path || [];
                    leg.remaining = remaining || [];
                    leg.updatedAt = Game.time;
                    this.legs.set(leg.id, leg);
                    // Storing legs
                    trace.info('storing legs after calculating', { legs: Array.from(this.legs.entries()) });
                    const memory = this.getMemory(trace);
                    memory.legs = Array.from(this.legs.entries());
                    this.setMemory(memory);
                }
            }
            if (!legs.length && this.legs.size > 0) {
                this.passes += 1;
                trace.info('pass completed', { passes: this.passes });
            }
        }
    }
    getLegsToCalculate(_trace) {
        // Filter out legs that need to be updated
        return Array.from(this.legs.values()).filter((leg) => {
            return leg.updatedAt < Game.time - LEG_CALCULATE_INTERVAL;
        });
    }
    calculateLeg(kernel, leg, trace) {
        trace.info('updating leg', { leg });
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.error('missing base', { id: this.baseId });
            return [null, null];
        }
        const [pathResult, details] = getPath(kernel, base.origin, leg.destination, roadPolicy, trace);
        if (!pathResult) {
            trace.error('path not found', { origin: base.origin, dest: leg.destination, details });
            return [null, null];
        }
        const remaining = pathResult.path.filter((pos) => {
            // Do not count edges as we cannot build on them
            if (pos.x === 0 || pos.y === 0 || pos.x === 49 || pos.y === 49) {
                return false;
            }
            // assume road is built if room not visible, assuming otherwise has problems
            if (!Game.rooms[pos.roomName]) {
                return true;
            }
            const road = pos.lookFor(LOOK_STRUCTURES).find((s) => {
                return s.structureType === STRUCTURE_ROAD;
            });
            if (road) {
                trace.info('road found', { road });
                return false;
            }
            const site = pos.lookFor(LOOK_CONSTRUCTION_SITES).find((s) => {
                return s.structureType === STRUCTURE_ROAD;
            });
            if (site) {
                trace.info('site found', { road });
                return false;
            }
            return true;
        });
        trace.info('remaining', { remaining, leg });
        return [pathResult.path, remaining];
    }
    ensureWallPassage(trace, kernel, base) {
        const legs = Array.from(this.legs.values());
        const unfinishedLegs = legs.filter((leg) => {
            return leg.remaining.length > 0;
        });
        trace.info('unfinished legs', { unfinishedLegs });
        unfinishedLegs.forEach((leg) => {
            for (let i = 0; i < leg.path.length; i++) {
                const pos = leg.path[i];
                if (!Game.rooms[pos.roomName]) {
                    return;
                }
                // Check if wall is present and remove
                const wall = pos.lookFor(LOOK_STRUCTURES).find((s) => {
                    return s.structureType === STRUCTURE_WALL;
                });
                if (wall) {
                    trace.warn('remove wall', { pos });
                    wall.destroy();
                }
                // Check if wall site is to be built and remove
                const wallSite = pos.lookFor(LOOK_CONSTRUCTION_SITES).find((s) => {
                    return s.structureType === STRUCTURE_WALL;
                });
                if (wallSite) {
                    trace.warn('remove wall site', { pos });
                    wallSite.remove();
                }
                // If there was a wall site or if we have not built too many road sites, set as a passage
                if (wall || wallSite) {
                    if (!_.find(base.passages, { x: pos.x, y: pos.y })) {
                        trace.warn('set as passage', { pos });
                        base.passages.push({ x: pos.x, y: pos.y });
                    }
                    else {
                        trace.info('already a passage', { pos });
                    }
                }
            }
        });
    }
    buildShortestLeg(trace, kernel, base) {
        if (this.passes < 1) {
            trace.info('calculate all legs at least once before building shortest leg', { passes: this.passes });
            return;
        }
        trace.info('legs', { legs: this.legs });
        // Find shortest unfinished leg
        const legs = Array.from(this.legs.values());
        const unfinishedLegs = legs.filter((leg) => {
            return leg.remaining.length > 0;
        });
        const sortedLegs = _.sortByAll(unfinishedLegs, (leg) => {
            // Sort sources in the primary room first
            if (leg.destination.roomName === base.primary) {
                return 0;
            }
            return 1;
        }, (leg) => {
            return leg.remaining.length;
        });
        let roadSites = 0;
        let leg = null;
        for (let s = 0; s < sortedLegs.length; s++) {
            if (roadSites >= MAX_ROAD_SITES) {
                break;
            }
            leg = sortedLegs[s];
            trace.info("building next shortest leg", { leg });
            for (let i = 0; i < leg.path.length; i++) {
                if (roadSites >= MAX_ROAD_SITES) {
                    break;
                }
                const pos = leg.path[i];
                if (!Game.rooms[pos.roomName]) {
                    continue;
                }
                // Do not build on edges
                if (pos.x === 0 || pos.x === 49 || pos.y === 0 || pos.y === 49) {
                    trace.info('skip border site', { pos });
                    continue;
                }
                // Check if road is already present
                const road = pos.lookFor(LOOK_STRUCTURES).find((s) => {
                    return s.structureType === STRUCTURE_ROAD;
                });
                if (road) {
                    continue;
                }
                // Check if road is already planned
                const roadSite = pos.lookFor(LOOK_CONSTRUCTION_SITES).find((s) => {
                    return s.structureType === STRUCTURE_ROAD;
                });
                if (roadSite) {
                    roadSites += 1;
                    continue;
                }
                // Check if wall is present and remove
                const wall = pos.lookFor(LOOK_STRUCTURES).find((s) => {
                    return s.structureType === STRUCTURE_WALL;
                });
                if (wall) {
                    trace.warn('remove wall', { pos });
                    wall.destroy();
                }
                // Check if wall site is to be built and remove
                const wallSite = pos.lookFor(LOOK_CONSTRUCTION_SITES).find((s) => {
                    return s.structureType === STRUCTURE_WALL;
                });
                if (wallSite) {
                    trace.warn('remove wall site', { pos });
                    wallSite.remove();
                }
                // If there was a wall site or if we have not built too many road sites, build a road site
                if (wall || wallSite || roadSites <= MAX_ROAD_SITES) {
                    const result = pos.createConstructionSite(STRUCTURE_ROAD);
                    if (result !== OK) {
                        trace.error('failed to build road', { pos, result });
                        continue;
                    }
                    trace.warn('build road site', { pos });
                    roadSites += 1;
                }
            }
            this.selectedLeg = leg;
        }
    }
}
const visualizeLegNodes = (legs, trace) => {
    for (let i = 0; i < legs.length; i++) {
        const leg = legs[i];
        if (!leg.destination) {
            trace.error('missing leg destination', { leg });
            continue;
        }
        new RoomVisual(leg.destination.roomName).text('X', leg.destination.x, leg.destination.y);
    }
};

const MIN_COMPOUND = 500;
const MAX_COMPOUND = 2000;
const MIN_ENERGY$1 = 1000;
const MAX_ENERGY = 2000;
const REQUEST_UNLOAD_TTL$1 = 5;
const REQUEST_LOAD_TTL$1 = 5;
const REQUEST_ENERGY_TTL$1 = 5;
const REQUEST_REBALANCE_TTL = 10;
const UPDATE_ROOM_BOOSTER_INTERVAL = 5;
function getBaseBoostTopic(base) {
    return `base_${base.id}_room_boosts`;
}
class Compound {
    constructor(name, effect, bonus) {
        this.name = name;
        this.effect = effect;
        this.bonus = bonus;
    }
}
class Effect {
    constructor(name, part, compounds = []) {
        this.name = name;
        this.part = part;
        this.compounds = compounds;
    }
}
class BoosterRunnable {
    constructor(id, baseId, labIds) {
        this.id = id;
        this.baseId = baseId;
        this.labIds = labIds;
        this.allEffects = null;
        this.boostPosition = this.getCreepBoostPosition();
        this.threadUpdateBoosters = threadBase('update_room_booster', UPDATE_ROOM_BOOSTER_INTERVAL)(this.updateBoosters.bind(this));
    }
    run(kernel, trace) {
        trace = trace.begin('booster_run');
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.error('Base not found', { baseId: this.baseId });
            trace.end();
            return terminate();
        }
        const labs = this.getLabs();
        if (labs.length !== 3) {
            trace.info('not right number of labs - terminating', { num: labs.length });
            trace.end();
            return terminate();
        }
        trace.info('booster run', { labId: labs.map((lab) => lab.id) });
        const resourceEnd = trace.startTimer('resource');
        const loadedEffects = this.getLoadedEffects();
        const desiredEffects = this.getDesiredEffects(kernel);
        resourceEnd();
        const [needToLoad, couldUnload, emptyLabs] = this.getNeeds(loadedEffects, desiredEffects, trace);
        trace.info('room status', {
            loaded: loadedEffects,
            desired: desiredEffects,
        });
        trace.info('booster needs', {
            labIds: this.labIds,
            needToLoad,
            couldUnload,
            emptyLabIds: emptyLabs.map((lab) => lab.id),
        });
        let sleepFor = REQUEST_ENERGY_TTL$1;
        this.requestEnergyForLabs(kernel, base, trace);
        this.threadUpdateBoosters(trace, kernel, base);
        if (desiredEffects.size) {
            sleepFor = REQUEST_LOAD_TTL$1;
            this.sendHaulRequests(kernel, base, loadedEffects, needToLoad, emptyLabs, couldUnload, trace);
        }
        else {
            sleepFor = REQUEST_UNLOAD_TTL$1;
            this.rebalanceLabs(kernel, base, trace);
        }
        trace.end();
        return sleeping(sleepFor);
    }
    getLabs() {
        return this.labIds.map((labId) => Game.getObjectById(labId)).filter(((lab) => lab));
    }
    updateBoosters(trace, kernel, base) {
        const resourceEnd = trace.startTimer('resources');
        const allEffects = this.getCompoundByEffects();
        const storedEffects = this.getAvailableEffects(kernel, base);
        const labsByResource = this.getLabsByResource();
        const labsByAction = this.getLabsByAction();
        resourceEnd();
        const details = {
            baseId: this.baseId,
            position: this.boostPosition,
            allEffects,
            storedEffects,
            labsByResource,
            labsByAction: labsByAction,
        };
        trace.info('publishing room boosts', {
            baseId: this.baseId,
            position: this.boostPosition,
            labsByResource: _.reduce(Array.from(labsByResource.values()), (labs, lab) => {
                labs.set(lab.id, lab.mineralType);
                return labs;
            }, new Map()),
            availableEffects: storedEffects,
        });
        const request = {
            priority: 1,
            ttl: UPDATE_ROOM_BOOSTER_INTERVAL + Game.time,
            details,
        };
        kernel.getTopics().addRequestV2(getBaseBoostTopic(base), request);
    }
    getBoostPosition() {
        return this.boostPosition;
    }
    getCreepBoostPosition() {
        const labs = this.getLabs();
        if (!labs.length) {
            return null;
        }
        const roomName = labs[0].room.name;
        const topLeft = labs.reduce((acc, lab) => {
            if (lab.pos.x < acc.x) {
                acc.x = lab.pos.x;
            }
            if (lab.pos.y < acc.y) {
                acc.y = lab.pos.y;
            }
            return acc;
        }, { x: 50, y: 50 });
        let position = null;
        position = new RoomPosition(topLeft.x, topLeft.y, roomName);
        if (position.lookFor(LOOK_STRUCTURES).filter((structure) => {
            return structure.structureType !== STRUCTURE_ROAD;
        }).length === 0) {
            return position;
        }
        position = new RoomPosition(topLeft.x, topLeft.y + 1, roomName);
        if (position.lookFor(LOOK_STRUCTURES).filter((structure) => {
            return structure.structureType !== STRUCTURE_ROAD;
        }).length === 0) {
            return position;
        }
        position = new RoomPosition(topLeft.x + 1, topLeft.y, roomName);
        if (position.lookFor(LOOK_STRUCTURES).filter((structure) => {
            return structure.structureType !== STRUCTURE_ROAD;
        }).length === 0) {
            return position;
        }
        position = new RoomPosition(topLeft.x + 1, topLeft.y + 1, roomName);
        if (position.lookFor(LOOK_STRUCTURES).filter((structure) => {
            return structure.structureType !== STRUCTURE_ROAD;
        }).length === 0) {
            return position;
        }
        return labs[0].pos;
    }
    getLabsByAction() {
        return this.getLabs().reduce((acc, lab) => {
            if (lab.mineralType) {
                const action = this.allEffects.get(lab.mineralType);
                if (!acc.has(action.name)) {
                    acc.set(action.name, []);
                }
                acc.get(action.name).push(lab);
            }
            return acc;
        }, new Map());
    }
    // @deprecated - use getLabsByAction()
    getLabsByResource() {
        const labs = this.getLabs();
        return labs.reduce((acc, lab) => {
            if (lab.mineralType) {
                acc.set(lab.mineralType, lab);
            }
            return acc;
        }, new Map());
    }
    // @deprecated - use getLabsByAction()
    getLabByResource(resource) {
        const labs = this.getLabs();
        for (let i = 0; i < labs.length; i++) {
            if (labs[i].mineralType === resource) {
                return labs[i];
            }
        }
        return null;
    }
    getLabResources() {
        const labs = this.getLabs();
        return labs.reduce((acc, lab) => {
            if (lab.mineralType) {
                acc.set(lab.mineralType, lab.store.getUsedCapacity(lab.mineralType));
            }
            return acc;
        }, new Map());
    }
    getEmptyLabs() {
        const labs = this.getLabs();
        return labs.filter((lab) => {
            return !lab.mineralType;
        });
    }
    // Get compounds by effects
    getCompoundByEffects(availableResources = null) {
        // If we are after all events and they are already cached, then return the cache
        if (!availableResources && this.allEffects) {
            return this.allEffects;
        }
        const allEffects = new Map();
        Object.keys(BOOSTS).forEach((part) => {
            const resources = BOOSTS[part];
            Object.keys(resources).forEach((resource) => {
                if (availableResources && !availableResources.get(resource)) {
                    return;
                }
                const effects = resources[resource];
                Object.keys(effects).forEach((effect) => {
                    const bonus = effects[effect];
                    if (!allEffects.has(effect)) {
                        allEffects.set(effect, new Effect(effect, part));
                    }
                    allEffects.get(effect).compounds.push(new Compound(resource, effect, bonus));
                });
            });
        });
        // If all effects are not already cached, cache them.
        // Do not cache if we are checking against available resources.
        if (!this.allEffects && !availableResources) {
            this.allEffects = allEffects;
        }
        return allEffects;
    }
    getLoadedEffects() {
        const resources = this.getLabResources();
        return this.getCompoundByEffects(resources);
    }
    getAvailableEffects(kernel, base) {
        const availableResources = getStoredResources(base);
        const loadedResource = this.getLabResources();
        // TODO the merge does not sum values
        // @REACTOR double check this
        const combined = new Map(Array.from(availableResources.entries()).concat(Array.from(loadedResource.entries())));
        return this.getCompoundByEffects(combined);
    }
    getDesiredEffects(kernel) {
        const desiredEffects = new Map();
        const allEffects = this.getCompoundByEffects();
        let request = null;
        // eslint-disable-next-line no-cond-assign
        while (request = kernel.getTopics().getNextRequest(BOOST_PREP)) {
            const requestedEffects = request.details[PREPARE_BOOSTS];
            if (!requestedEffects) {
                continue;
            }
            requestedEffects.forEach((requested) => {
                desiredEffects.set(requested, allEffects.get(requested));
            });
        }
        return desiredEffects;
    }
    getNeeds(loadedEffects, desiredEffects, trace) {
        const loadedNames = [...loadedEffects.keys()];
        const desiredNames = [...desiredEffects.keys()];
        let preparedNames = _.intersection(loadedNames, desiredNames);
        preparedNames = preparedNames.filter((effectName) => {
            const effect = loadedEffects.get(effectName);
            const compound = effect.compounds[0].name;
            const lab = this.getLabByResource(compound);
            const currentAmount = lab.store.getUsedCapacity(compound);
            if (currentAmount < MIN_COMPOUND) {
                trace.info('not enough of compound', { compound, effectName, currentAmount });
                return false;
            }
            return true;
        });
        const needToLoad = _.difference(desiredNames, preparedNames);
        const couldUnload = _.difference(loadedNames, preparedNames);
        const emptyLabs = this.getEmptyLabs();
        return [needToLoad, couldUnload, emptyLabs];
    }
    sendHaulRequests(kernel, base, loadedEffects, needToLoad, emptyLabs, couldUnload, trace) {
        const numToLoad = needToLoad.length;
        const numEmpty = emptyLabs.length;
        if (numEmpty && numToLoad) {
            const numReadyToLoad = _.min([numEmpty, numToLoad]);
            const load = needToLoad.slice(0, numReadyToLoad);
            this.requestMaterialsForLabs(kernel, base, load, trace);
        }
        if (numToLoad > numEmpty) {
            const numToUnload = numToLoad - numEmpty;
            const unload = couldUnload.slice(0, numToUnload);
            this.requestUnloadOfLabs(kernel, base, loadedEffects, unload, trace);
        }
    }
    rebalanceLabs(kernel, base, trace) {
        const labs = this.getLabs();
        labs.forEach((lab) => {
            if (!lab.mineralType) {
                trace.info('lab has no mineral loaded', { labId: lab.id });
                return;
            }
            const currentAmount = lab.store.getUsedCapacity(lab.mineralType);
            let amount = 0;
            if (currentAmount < MIN_COMPOUND) {
                trace.info('lab is below min: unload it', { labId: lab.id, resource: lab.mineralType });
                amount = currentAmount;
            }
            else if (currentAmount > MAX_COMPOUND) {
                trace.info('lab is above min: return some', { labId: lab.id, resource: lab.mineralType });
                amount = currentAmount - MAX_COMPOUND;
            }
            if (amount) {
                const dropoff = getStructureForResource(base, lab.mineralType);
                if (!dropoff) {
                    trace.info('no dropoff for already loaded compound', { resource: lab.mineralType });
                    return;
                }
                const request = {
                    ttl: REQUEST_REBALANCE_TTL + Game.time,
                    priority: HAUL_BOOST,
                    details: {
                        [TASK_ID]: `bmc-${this.id}-${Game.time}`,
                        [MEMORY_TASK_TYPE]: TASK_HAUL,
                        [MEMORY_HAUL_PICKUP]: lab.id,
                        [MEMORY_HAUL_RESOURCE]: lab.mineralType,
                        [MEMORY_HAUL_DROPOFF]: dropoff.id,
                        [MEMORY_HAUL_AMOUNT]: amount,
                    },
                };
                trace.info('boost clear low', { priority: HAUL_BOOST, request });
                kernel.getTopics().addRequestV2(getBaseDistributorTopic(base.id), request);
            }
        });
    }
    requestUnloadOfLabs(kernel, base, loadedEffects, couldUnload, trace) {
        couldUnload.forEach((toUnload) => {
            const effect = loadedEffects.get(toUnload);
            const compound = effect.compounds[0];
            const pickup = this.getLabByResource(compound.name);
            if (!pickup) {
                trace.info('no pickup for already loaded compound', { resource: compound.name });
                return;
            }
            const baseCreeps = kernel.getCreepsManager().getCreepsByBase(this.baseId);
            const assignedCreeps = baseCreeps.filter((creep) => {
                const task = creep.memory[MEMORY_TASK_TYPE];
                const taskPickup = creep.memory[MEMORY_HAUL_PICKUP];
                const resource = creep.memory[MEMORY_HAUL_RESOURCE];
                return task === TASK_HAUL && taskPickup === pickup.id && resource == compound.name;
            });
            if (assignedCreeps.length) {
                trace.info('creep already unloading', { resource: compound.name });
                return;
            }
            const dropoff = getStructureForResource(base, compound.name);
            if (!dropoff) {
                trace.info('no dropoff for already loaded compound', { resource: compound.name });
                return;
            }
            const details = {
                [TASK_ID]: `bmu-${this.id}-${Game.time}`,
                [MEMORY_TASK_TYPE]: TASK_HAUL,
                [MEMORY_HAUL_PICKUP]: pickup.id,
                [MEMORY_HAUL_RESOURCE]: compound.name,
                [MEMORY_HAUL_DROPOFF]: dropoff.id,
                [MEMORY_HAUL_AMOUNT]: pickup.store.getUsedCapacity(compound.name),
            };
            const request = {
                ttl: REQUEST_UNLOAD_TTL$1 + Game.time,
                priority: UNLOAD_BOOST,
                details,
            };
            trace.info('boost unload', { priority: HAUL_BOOST, details });
            kernel.getTopics().addRequestV2(getBaseDistributorTopic(this.baseId), request);
        });
    }
    requestMaterialsForLabs(kernel, base, needToLoad, trace) {
        const allEffects = this.getCompoundByEffects();
        const storedResources = getStoredResources(base);
        needToLoad.forEach((toLoad) => {
            trace.info('need to load', { toLoad });
            const effect = allEffects.get(toLoad);
            if (!effect) {
                trace.info('not able to find desired effect', { toLoad });
                return;
            }
            const emptyLabs = this.getEmptyLabs();
            if (emptyLabs.length === 0) {
                trace.info('no destination for available compound', { toLoad });
                return;
            }
            const emptyLab = emptyLabs[0];
            const baseCreeps = kernel.getCreepsManager().getCreepsByBase(this.baseId);
            const assignedCreeps = baseCreeps.filter((creep) => {
                const task = creep.memory[MEMORY_TASK_TYPE];
                const taskDropoff = creep.memory[MEMORY_HAUL_DROPOFF];
                const resource = creep.memory[MEMORY_HAUL_RESOURCE];
                return task === TASK_HAUL && taskDropoff === emptyLab.id && resource !== RESOURCE_ENERGY;
            });
            if (assignedCreeps.length) {
                trace.info('creep already loading', { toLoad });
                return;
            }
            // Refactor this to a a function that further filters a set of effects
            const compound = effect.compounds.reduce((selected, compound) => {
                if (storedResources.get(compound.name) >= MIN_COMPOUND) {
                    if (!selected) {
                        selected = compound;
                    }
                    if (effect.name != 'damage') {
                        if (selected.bonus < compound.bonus) {
                            selected = compound;
                        }
                    }
                    else {
                        if (selected.bonus > compound.bonus) {
                            selected = compound;
                        }
                    }
                }
                return selected;
            }, null);
            if (!compound) {
                trace.info('no compound available for', { toLoad });
                return;
            }
            this.requestHaulingOfMaterial(kernel, base, compound, emptyLab, trace);
        });
    }
    requestHaulingOfMaterial(kernel, base, compound, lab, trace) {
        const pickup = getStructureWithResource(base, compound.name);
        if (!pickup) {
            trace.info('no pickup for available compound', { resource: compound.name });
            return;
        }
        const details = {
            [TASK_ID]: `brl-${this.id}-${Game.time}`,
            [MEMORY_TASK_TYPE]: TASK_HAUL,
            [MEMORY_HAUL_PICKUP]: pickup.id,
            [MEMORY_HAUL_RESOURCE]: compound.name,
            [MEMORY_HAUL_DROPOFF]: lab.id,
            [MEMORY_HAUL_AMOUNT]: MAX_COMPOUND,
        };
        const request = {
            ttl: REQUEST_LOAD_TTL$1 + Game.time,
            priority: HAUL_BOOST,
            details,
        };
        trace.info('boost load material', { priority: HAUL_BOOST, details });
        kernel.getTopics().addRequestV2(getBaseDistributorTopic(this.baseId), request);
    }
    requestEnergyForLabs(kernel, base, trace) {
        const labs = this.getLabs();
        labs.forEach((lab) => {
            // Only fill lab if needed
            if (lab.store.getUsedCapacity(RESOURCE_ENERGY) >= MIN_ENERGY$1) {
                trace.info('lab has energy', { labId: lab.id });
                return;
            }
            const pickup = getStructureWithResource(base, RESOURCE_ENERGY);
            const currentEnergy = lab.store.getUsedCapacity(RESOURCE_ENERGY);
            const details = {
                [TASK_ID]: `bel-${this.id}-${Game.time}`,
                [MEMORY_TASK_TYPE]: TASK_HAUL,
                [MEMORY_HAUL_PICKUP]: pickup.id,
                [MEMORY_HAUL_RESOURCE]: RESOURCE_ENERGY,
                [MEMORY_HAUL_DROPOFF]: lab.id,
                [MEMORY_HAUL_AMOUNT]: MAX_ENERGY - currentEnergy,
            };
            const request = {
                ttl: REQUEST_ENERGY_TTL$1 + Game.time,
                priority: HAUL_BOOST,
                details,
            };
            trace.info('boost load energy', { labId: lab.id, priority: HAUL_BOOST, details });
            kernel.getTopics().addRequestV2(getBaseDistributorTopic(this.baseId), request);
        });
    }
}

const RUN_TTL$6 = 50;
const BUILD_STRUCTURES_TTL = 200;
const PRODUCE_EVENTS_TTL$1 = 50;
const padLayout = new Map();
padLayout.set(TOP, {
    origin: { x: 1, y: 1 },
    buildings: [
        ['R', 'R', 'R'],
        ['.', 'L', '.'],
        ['.', '.', '.'],
    ],
});
padLayout.set(TOP_RIGHT, {
    origin: { x: 1, y: 1 },
    buildings: [
        ['.', 'R', 'R'],
        ['.', 'L', 'R'],
        ['.', '.', '.'],
    ],
});
padLayout.set(RIGHT, {
    origin: { x: 1, y: 1 },
    buildings: [
        ['.', '.', 'R'],
        ['.', 'L', 'R'],
        ['.', '.', 'R'],
    ],
});
padLayout.set(BOTTOM_RIGHT, {
    origin: { x: 1, y: 1 },
    buildings: [
        ['.', '.', '.'],
        ['.', 'L', 'R'],
        ['.', 'R', 'R'],
    ],
});
padLayout.set(BOTTOM, {
    origin: { x: 1, y: 1 },
    buildings: [
        ['.', '.', '.'],
        ['.', 'L', '.'],
        ['R', 'R', 'R'],
    ],
});
padLayout.set(BOTTOM_LEFT, {
    origin: { x: 1, y: 1 },
    buildings: [
        ['.', '.', '.'],
        ['R', 'L', '.'],
        ['R', 'R', '.'],
    ],
});
padLayout.set(LEFT, {
    origin: { x: 1, y: 1 },
    buildings: [
        ['R', '.', '.'],
        ['R', 'L', '.'],
        ['R', '.', '.'],
    ],
});
padLayout.set(TOP_LEFT, {
    origin: { x: 1, y: 1 },
    buildings: [
        ['R', 'R', '.'],
        ['R', 'L', '.'],
        ['.', '.', '.'],
    ],
});
class ControllerRunnable extends PersistentMemory {
    constructor(controllerId) {
        super(controllerId);
        this.controllerId = controllerId;
        this.threadBuildStructures = threadBase('check_structures', BUILD_STRUCTURES_TTL)(this.buildStructures.bind(this));
        this.threadProduceEvents = threadBase('consume_events', PRODUCE_EVENTS_TTL$1)(this.produceEvents.bind(this));
    }
    run(kernel, trace) {
        trace = trace.as('controller_run');
        trace.info('run', {
            controller: this.controllerId,
            nodePosition: this.nodePosition,
            roadPosition: this.roadPosition,
            nodeDirection: this.nodeDirection,
        });
        const controller = Game.getObjectById(this.controllerId);
        if (!controller) {
            trace.error('missing controller', { id: this.controllerId });
            return sleeping(RUN_TTL$6);
        }
        const base = kernel.getPlanner().getBaseByRoom(controller.room.name);
        if (!base) {
            trace.error('missing origin', { id: controller.room.name });
            return sleeping(RUN_TTL$6);
        }
        if (!this.nodePosition || !this.nodeDirection) {
            trace.info('node and road position not set, populating');
            this.populateNodePosition(kernel, controller, trace);
        }
        if (base.alertLevel === AlertLevel.GREEN) {
            this.threadBuildStructures(trace, kernel, base, controller.room);
            this.threadProduceEvents(trace, kernel, base, controller);
        }
        return sleeping(RUN_TTL$6);
    }
    populateNodePosition(kernel, controller, trace) {
        const memory = this.getMemory(trace);
        if ((memory === null || memory === void 0 ? void 0 : memory.nodePosition) && (memory === null || memory === void 0 ? void 0 : memory.roadPosition)) {
            this.nodePosition = new RoomPosition(memory.nodePosition.x, memory.nodePosition.y, memory.nodePosition.roomName);
            this.roadPosition = new RoomPosition(memory.roadPosition.x, memory.roadPosition.y, memory.roadPosition.roomName);
            this.nodeDirection = controller.pos.getDirectionTo(this.roadPosition);
            trace.info('populated node position from memory', {
                nodePosition: this.nodePosition,
                roadPosition: this.roadPosition,
                nodeDirection: this.nodeDirection,
            });
        }
        if (this.nodePosition && this.roadPosition) {
            trace.info('node and road positions are already set');
            return;
        }
        const base = kernel.getPlanner().getBaseByRoom(controller.pos.roomName);
        if (!base) {
            trace.error('missing base config', { room: controller.pos.roomName });
            return;
        }
        const [pathResult, details] = getPath(kernel, base.origin, controller.pos, controllerRoadPolicy, trace);
        if (!pathResult || !pathResult.path.length) {
            trace.error('no path found', { origin: base.origin, dest: controller.pos, pathResult, details });
            return;
        }
        // We are grabbing where we will put a link
        const pathLength = pathResult.path.length;
        this.nodePosition = pathResult.path[pathLength - 1];
        this.roadPosition = pathResult.path[pathLength - 2];
        this.nodeDirection = controller.pos.getDirectionTo(this.roadPosition);
        trace.warn('node and road position was not set: setting', {
            id: this.controllerId,
            position: this.nodePosition,
            direction: this.nodeDirection,
        });
        memory.nodePosition = this.nodePosition;
        memory.roadPosition = this.roadPosition;
        this.setMemory(memory, false);
    }
    produceEvents(trace, kernel, base, controller) {
        const position = this.roadPosition;
        if (!position) {
            trace.error('no road position', { room: controller.room.name });
            return;
        }
        // If there is not a link, build a road to the controller
        const link = controller.pos.findInRange(FIND_MY_STRUCTURES, 1, { filter: (s) => s.structureType === STRUCTURE_LINK })[0];
        if (!link) {
            const data = {
                id: controller.id,
                position: position,
            };
            kernel.getBroker().getStream(getLogisticsTopic(base.id)).
                publish(new Event(this.controllerId, Game.time, LogisticsEventType.RequestRoad, data));
        }
    }
    buildStructures(trace, kernel, base, room) {
        trace.info('building structures for controller', { controllerId: this.controllerId });
        if (!this.nodePosition || !this.nodeDirection) {
            trace.error('missing node position or direction', {
                id: this.controllerId,
                nodePosition: this.nodePosition,
                nodeDirection: this.nodeDirection,
            });
            return;
        }
        const layout = padLayout.get(this.nodeDirection);
        const terrain = room.getTerrain();
        trace.info('building structures', { layout });
        const roomVisual = new RoomVisual(room.name);
        for (let y = 0; y < layout.buildings.length; y++) {
            const row = layout.buildings[y];
            for (let x = 0; x < row.length; x++) {
                const code = row[x];
                const pos = getConstructionPosition({ x, y }, this.nodePosition, layout);
                trace.info('building structure', { code, pos });
                if (buildingCodes[code] === ANY) {
                    continue;
                }
                // We cant build links until RCL5
                if (code === 'L' && room.controller.level < 5) {
                    continue;
                }
                if (terrain.get(pos.x, pos.y) === TERRAIN_MASK_WALL) {
                    continue;
                }
                const structure = pos.lookFor(LOOK_STRUCTURES)[0];
                trace.info('structure', { structure, pos });
                if (structure) {
                    trace.info('structure present', { structure: structure.structureType });
                    if (structure.structureType !== buildingCodes[code]) {
                        trace.warn('wrong site, remove', { existing: structure.structureType, expected: buildingCodes[code] });
                        structure.destroy();
                    }
                    continue;
                }
                const site = pos.lookFor(LOOK_CONSTRUCTION_SITES)[0];
                trace.info('site', { site });
                if (site) {
                    if (site.structureType !== buildingCodes[code]) {
                        trace.warn('wrong site, remove', { existing: site.structureType, expected: buildingCodes[code] });
                        site.remove();
                    }
                    continue;
                }
                const structureType = buildingCodes[code];
                if (!structureType || structureType === EMPTY) {
                    trace.info('no structure type', { code, x, y });
                    continue;
                }
                roomVisual.text(code, pos.x, pos.y);
                const result = room.createConstructionSite(pos, structureType);
                if (result !== OK && result !== ERR_FULL) {
                    trace.error('failed to build structure', { structureType, pos, result });
                    return;
                }
                trace.notice('building structure', { pos, structureType });
            }
        }
    }
}

var DEFENSE_STATUS;
(function (DEFENSE_STATUS) {
    DEFENSE_STATUS["GREEN"] = "green";
    DEFENSE_STATUS["YELLOW"] = "yellow";
    DEFENSE_STATUS["RED"] = "red";
})(DEFENSE_STATUS || (DEFENSE_STATUS = {}));

const REQUEST_PARTY_MEMBER_TTL = 30;
const NO_TARGET_TTL = 20;
class DefensePartyRunnable {
    constructor(id, base, flagId, position) {
        this.id = id;
        this.base = base;
        this.flagId = flagId;
        this.noTargetTTL = 0;
        this.minEnergy = 0;
        this.party = new PartyRunnable(id, base.id, [position], WORKER_DEFENDER_DRONE, null, this.minEnergy, PRIORITY_BUFFER_PATROL, REQUEST_PARTY_MEMBER_TTL);
    }
    run(kernel, trace) {
        var _a, _b, _c;
        trace = trace.begin('defense_party_run');
        trace.info('defense party run top', { id: this.id });
        const prep = trace.begin('run_prep');
        // Check existence of flag
        const flag = this.getFlag();
        if (!flag) {
            trace.error('no flag with that id, terminating', { flagId: this.flagId });
            prep.end();
            trace.end();
            return terminate();
        }
        const creeps = this.getAssignedCreeps();
        const position = this.getPosition();
        trace.info('defense party run', {
            id: this.id,
            creeps: creeps.map((creep) => creep.name),
            flagId: this.flagId,
            noTargetTTL: this.noTargetTTL,
            position,
        });
        prep.end();
        const targetTrace = trace.begin('run_target');
        let targetCreeps = [];
        let targetStructures = [];
        // Get target requests and map to creeps
        if (flag.room) {
            const friends = kernel.getConfig().friends;
            targetCreeps = flag.room.find(FIND_HOSTILE_CREEPS).filter((creep) => {
                return friends.indexOf(creep.owner.username) === -1;
            });
            targetCreeps = ___namespace.sortBy(targetCreeps, (creep) => {
                return 50 - creep.getActiveBodyparts(ATTACK);
            });
            const username = kernel.getPlanner().getUsername();
            if (!((_a = flag.room.controller) === null || _a === void 0 ? void 0 : _a.my) && ((_c = (_b = flag.room.controller) === null || _b === void 0 ? void 0 : _b.reservation) === null || _c === void 0 ? void 0 : _c.username) !== username) {
                targetStructures = flag.room.find(FIND_STRUCTURES, {
                    filter: (structure) => {
                        return structure.structureType === STRUCTURE_CONTAINER;
                    },
                });
            }
        }
        let targets = [];
        targets = targets.concat(targetCreeps, targetStructures);
        trace.info('target requests', { targets: targets.map((target) => target.id) });
        targetTrace.end();
        const movementTrace = trace.begin('run_movement');
        // Select target from request; if no requests and its been a while, return to flag
        let destination = this.getPosition();
        if (targets.length) {
            this.noTargetTTL = NO_TARGET_TTL;
            destination = targets[0].pos;
        }
        else {
            this.noTargetTTL -= 1;
            if (this.noTargetTTL < 0 && flag) {
                destination = flag.pos;
            }
        }
        movementTrace.end();
        trace.info('goal', { destination, targets });
        const partyTrace = trace.begin('run_party');
        this.setPosition(destination, trace);
        this.setTarget(targets, trace);
        this.setHeal(trace);
        const partyResult = this.party.run(kernel, trace);
        partyTrace.end();
        if (partyResult.status === STATUS_TERMINATED) {
            trace.notice('party terminated, terminate this');
            trace.end();
            return partyResult;
        }
        trace.end();
        return running();
    }
    getAssignedCreeps() {
        return this.party.getAssignedCreeps();
    }
    inPosition(position, trace) {
        return this.party.inPosition(position, trace);
    }
    getFlag() {
        return Game.flags[this.flagId] || null;
    }
    getPosition() {
        return this.party.getPosition();
    }
    setPosition(position, _trace) {
        this.party.setPosition(position);
    }
    setTarget(targets, trace) {
        this.party.setTarget(targets, trace);
    }
    setHeal(trace) {
        this.party.setHeal(trace);
    }
}

/**
 * Base Defense Manager
 * Created: August 2022
 *
 * - Scan remotes for hostiles, if found set yellow alert and dispatch ranged defenders to room
 * - Scan primary room for hostiles, if found set red alert and dispatch melee defenders to ramparts
 * - Report base defense status to stream
 *
 * Base defense: 2 melee defenders that position themselves where enemy is trying to punch through.
 * create additional ranged defenders to the break through point.
 *
 * Room defense: Send enough defenders (singles or quads) to kill the hostiles. Frequently invaders.
 *
 * TODO: Request base defenders from neighbors
 */
const RUN_INTERVAL$3 = 5;
const TARGET_REQUEST_TTL = RUN_INTERVAL$3;
const REQUEST_DEFENDERS_TTL = 25;
// const REQUEST_DEFENDER_TTL = 5;
const MAX_DEFENDERS_PER_ROOM = 1;
function getBasePriorityTargetsTopic(baseId) {
    return `base_${baseId}_priority_targets`;
}
function getBaseDefenseTopic(baseId) {
    return `base_${baseId}_defense`;
}
class DefenseManager {
    // Legacy code
    // threadCheckBaseDefenses: KernelThreadFunc;
    // threadReturnDefendersToStation: KernelThreadFunc;
    // threadHandleDefenderRequest: KernelThreadFunc;
    constructor(kernel, base, trace) {
        this.baseId = base.id;
        this.scheduler = kernel.getScheduler();
        this.restoreFromMemory(kernel, trace);
        //this.threadCheckBaseDefenses = threadKernel('check_defense_thread', REQUEST_DEFENDERS_TTL)(checkBaseDefenses);
        //this.threadReturnDefendersToStation = threadKernel('recall_defenders_thread', REQUEST_DEFENDERS_TTL)(returnDefendersToStation);
        //this.threadHandleDefenderRequest = threadKernel('request_defenders_thread', REQUEST_DEFENDER_TTL)(this.requestDefenders.bind(this));
    }
    run(kernel, trace) {
        trace = trace.begin('defense_manager_run');
        trace.info('defense manager run');
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.error('base not found, terminating', { baseId: this.baseId });
            return terminate();
        }
        // Check base for attackers
        this.defendBase(kernel, base, trace);
        // Check remotes for attackers
        this.defendRemotes(kernel, base, trace);
        /* Legacy code
        const defenseStatusTrace = trace.begin('updateDefenseStatus');
        publishDefenseStatuses(kernel, hostilesByBase, defenseStatusTrace);
        defenseStatusTrace.end();
    
        this.handleDefendFlags(kernel, trace);
        this.threadCheckBaseDefenses(trace, kernel, hostilesByBase);
        this.threadReturnDefendersToStation(trace, kernel, hostilesByBase);
        this.threadHandleDefenderRequest(trace, kernel, hostilesByBase);
        */
        trace.end();
        return sleeping(RUN_INTERVAL$3);
    }
    restoreFromMemory(kernel, trace) {
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const memory = Memory;
        if (!memory.defense) {
            memory.defense = {
                parties: [],
            };
        }
        this.memory = memory.defense;
        this.defenseParties = this.memory.parties.map((party) => {
            if (!party.id || !party.position || !party.flagId) {
                trace.error('invalid defense party', { party });
                return null;
            }
            party.position = new RoomPosition(party.position.x, party.position.y, party.position.roomName);
            return this.createAndScheduleDefenseParty(kernel, party.id, party.flagId, party.position, trace);
        }).filter((party) => {
            return !!party;
        });
    }
    defendBase(kernel, base, trace) {
        // Ensure that towers have something to shoot at
        const targetTopicTrace = trace.begin('addHostilesToBaseTargetTopic');
        addHostilesToBaseTargetTopic(kernel, base, targetTopicTrace);
        targetTopicTrace.end();
    }
    defendRemotes(kernel, base, trace) {
        const roomsNeedingDefenders = new Map();
        const hostileAttackPowerByRoom = getHostileAttackPowerByBaseRoom(kernel, base);
        trace.info('hostile attack power by room', {
            baseId: this.baseId,
            rooms: Array.from(hostileAttackPowerByRoom.entries())
        });
        // check if hostiles in primary room
        const primaryRoomHostiles = hostileAttackPowerByRoom.get(base.primary);
        if (primaryRoomHostiles > 0) {
            // request rampart defense and regular defenders
            trace.warn('hostiles in primary room', { baseId: this.baseId, primaryRoomHostiles });
            // const memory = {
            //   [MEMORY.MEMORY_ASSIGN_ROOM]: base.primary,
            //   [MEMORY.MEMORY_BASE]: base.id,
            // };
            // const request = createSpawnRequest(PRIORITY_RAMPART_DEFENDER, REQUEST_DEFENDERS_TTL + Game.time,
            //   CREEPS.WORKER_DEFENDER, memory, null, 0);
            // trace.warn('requesting rampaert defenders', {roomName: base.primary, primaryRoomHostiles});
            // kernel.getTopics().addRequestV2(getBaseDefenseTopic(base.id), request);
        }
        for (const [roomName, attackPower] of hostileAttackPowerByRoom) {
            // if no hostiles, dont send defenders
            if (attackPower == 0) {
                continue;
            }
            // Don't send defenders to primary room
            if (roomName === base.primary) {
                continue;
            }
            // Request defenders from base
            trace.warn('requesting defenders', { roomName, attackPower });
        }
        // Check rooms for Invader Cores
        base.rooms.forEach((roomName) => {
            const roomEntry = kernel.getScribe().getRoomById(roomName);
            if (!roomEntry) {
                trace.warn('checking for invader cores, room entry not found', { roomName });
                return;
            }
            if (roomEntry.invaderCorePos && roomEntry.invaderCoreLevel === 0) {
                trace.warn('checking for invader cores, invader core found', { roomName });
                let hostileAttackPower = roomsNeedingDefenders.get(roomName) || -1;
                if (hostileAttackPower < 0) {
                    hostileAttackPower = 0;
                }
                roomsNeedingDefenders.set(roomName, hostileAttackPower);
            }
        });
        const defenders = kernel.getCreepsManager().getCreepsByBaseAndRole(this.baseId, WORKER_DEFENDER);
        // Handle defense needs
        for (const [roomName, attackPower] of roomsNeedingDefenders) {
            // Primary room also needs rampart defense
            // Check if rampart defenders exist
            // if not, build parts list and request spawning of rampart defenders
            // Check if defenders already exist (regular rooms and primary)
            const roomDefenders = defenders.filter((creep) => {
                return creep.memory[MEMORY_ASSIGN_ROOM] === roomName;
            });
            if (roomDefenders.length > MAX_DEFENDERS_PER_ROOM) {
                trace.warn('defenders already exist', { roomName, roomDefenders });
                continue;
            }
            const primaryRoom = Game.rooms[base.primary];
            if (!primaryRoom) {
                trace.warn('primary room not found', { baseId: this.baseId, primaryRoom });
                continue;
            }
            // Build parts list and request spawning of defenders
            const multipliers = newMultipliers();
            const [parts, ok] = buildDefender(attackPower, primaryRoom.energyCapacityAvailable, multipliers, trace);
            if (!ok) {
                trace.warn('defender build failed, we cannot take the room', {
                    roomName, attackPower,
                    energyCapacityAvailable: primaryRoom.energyCapacityAvailable
                });
                continue;
            }
            const memory = {
                [MEMORY_ASSIGN_ROOM]: roomName,
                [MEMORY_BASE]: base.id,
            };
            const request = createSpawnRequest(PRIORITY_DEFENDER, REQUEST_DEFENDERS_TTL + Game.time, WORKER_DEFENDER, memory, parts, 0);
            trace.warn('requesting defenders', { roomName, attackPower, request });
            kernel.getTopics().addRequestV2(getBaseDefenseTopic(base.id), request);
        }
    }
    createAndScheduleDefenseParty(kernel, id, flagId, position, trace) {
        const flag = Game.flags[flagId] || null;
        if (!flag) {
            trace.error('flag not found, not creating defense party', { flagId });
            return;
        }
        // TODO replace this with planner usage
        const base = kernel.getPlanner().getClosestBaseInRange(flag.pos.roomName, 5);
        if (!base) {
            trace.error('could not find base in range, not creating defense party', { roomName: flag.pos.roomName });
            return;
        }
        trace.notice('creating defense party', { id, position, flagId, baseId: base.id });
        const party = new DefensePartyRunnable(id, base, flagId, position);
        const process = new Process(id, 'defense_party', Priorities.DEFENCE, {
            run(kernel, trace) {
                return party.run(kernel, trace);
            },
        });
        this.scheduler.registerProcess(process);
        return party;
    }
    handleDefendFlags(kernel, trace) {
        // Removed old defense parties
        this.defenseParties = this.defenseParties.filter((party) => {
            return this.scheduler.hasProcess(party.id);
        });
        Object.values(Game.flags).forEach((flag) => {
            if (flag.name.startsWith('defend_')) {
                const flagDefensePartyId = `${flag.name}_party`;
                trace.info('defend flag', { flagDefensePartyId });
                if (!this.scheduler.hasProcess(flagDefensePartyId)) {
                    const party = this.createAndScheduleDefenseParty(kernel, flagDefensePartyId, flag.name, flag.pos, trace);
                    this.defenseParties.push(party);
                }
            }
        });
        // Update memory
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        Memory.defense = {
            parties: this.defenseParties.map((party) => {
                if (!Game.flags[party.flagId]) {
                    trace.error('missing flag', { flagId: party.flagId });
                    return null;
                }
                return {
                    id: party.id,
                    flagId: party.flagId,
                    position: party.getPosition(),
                };
            }).filter((party) => !!party),
        };
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        trace.info('defense memory', { memory: Memory.defense });
    }
    requestDefenders(trace, kernel, hostilesByBase) {
        // Check intra-base requests for defenders
        Array.from(hostilesByBase.keys()).forEach((baseId) => {
            const base = kernel.getPlanner().getBaseById(baseId);
            if (!base) {
                trace.error('cannot find base config', { baseId });
                return;
            }
            const request = kernel.getTopics().getNextRequest(getBaseDefenseTopic(base.id));
            if (request) {
                trace.log('got defender request', { request });
                this.handleDefenderRequest(kernel, base, request, trace);
            }
        });
    }
    handleDefenderRequest(kernel, base, request, trace) {
        const room = getBasePrimaryRoom(base);
        if (!room) {
            trace.error('cannot find primary room', { base });
            return;
        }
        trace.log('request details', {
            controllerLevel: (room === null || room === void 0 ? void 0 : room.controller) || null,
            request,
        });
        if (request.details.spawn) {
            const spawnRequest = createSpawnRequest(request.priority, request.ttl, request.details.role, request.details.memory, null, 0);
            trace.log('requesting spawning of defenders', { request });
            kernel.getTopics().addRequestV2(getBaseSpawnTopic(base.id), spawnRequest);
        }
        trace.info('requesting existing defense response', { request });
        // TODO replace with base defense topic
        // Order existing defenders to the room
        kernel.getCreepsManager().getCreepsByBaseAndRole(base.id, WORKER_DEFENDER).forEach((defender) => {
            defender.memory[MEMORY_ASSIGN_ROOM] = request.details.memory[MEMORY_ASSIGN_ROOM];
            defender.memory[MEMORY_ASSIGN_ROOM_POS] = request.details.memory[MEMORY_ASSIGN_ROOM_POS];
        });
    }
}
// function addHealTargetsToBaseTopic(kernal: Kernel, damagedCreeps) {
// }
function addHostilesToBaseTargetTopic(kernel, base, trace) {
    // Add targets to base target topic
    const primaryRoom = getBasePrimaryRoom(base);
    if (!primaryRoom) {
        trace.error('cannot find primary room', { base });
        return;
    }
    const hostiles = primaryRoom.find(FIND_HOSTILE_CREEPS);
    hostiles.forEach((hostile) => {
        let healingPower = scoreHealing(hostile, true);
        // Get adjacent creeps
        const adjacentCreeps = hostile.pos.findInRange(FIND_HOSTILE_CREEPS, 1);
        // Sum up total healing (healing of target + healing of adjacent creeps)
        healingPower = adjacentCreeps.reduce((total, creep) => {
            return total + scoreHealing(creep, true);
        }, healingPower);
        const details = {
            id: hostile.id,
            roomName: hostile.room.name,
            healingPower: healingPower,
            // TODO include rangedAttackPower
            // TODO include meleeAttackPower
        };
        const request = {
            priority: healingPower,
            ttl: TARGET_REQUEST_TTL + Game.time,
            details,
        };
        trace.info('requesting target', { details, healingPower });
        kernel.getTopics().addRequestV2(getBasePriorityTargetsTopic(base.id), request);
    });
    /*
    Array.from(hostilesByBases.entries()).forEach(([baseId, hostiles]) => {
      const base = kernel.getPlanner().getBaseById(baseId);
      if (!base) {
        return;
      }
  
      hostiles.forEach((hostile) => {
        let healingPower = scoreHealing(hostile as Creep, true);
  
        // Get adjacent creeps
        const adjacentCreeps = hostile.pos.findInRange(FIND_HOSTILE_CREEPS, 1);
  
        // Sum up total healing (healing of target + healing of adjacent creeps)
        healingPower = adjacentCreeps.reduce((total, creep) => {
          return total + scoreHealing(creep, true);
        }, healingPower);
  
        const details = {
          id: hostile.id,
          roomName: hostile.room.name,
          healingPower: healingPower,
          // TODO include rangedAttackPower
          // TODO include meleeAttackPower
        };
  
        trace.info('requesting target', {details, healingPower});
  
        kernel.getTopics().addRequest(getBasePriorityTargetsTopic(baseId), healingPower,
          details, TARGET_REQUEST_TTL + Game.time);
      });
    });
    */
}
function getHostileAttackPowerByBaseRoom(kernel, base, _trace) {
    const hostileAttackPowerByRoom = new Map();
    base.rooms.forEach((roomName) => {
        const roomEntry = kernel.getScribe().getRoomById(roomName);
        if (!roomEntry) {
            return;
        }
        hostileAttackPowerByRoom.set(roomName, roomEntry.hostilesDmg);
    });
    return hostileAttackPowerByRoom;
}

const TASK_PHASE_START = 'phase_start';
const TASK_PHASE_LOAD = 'phase_transfer_resources';
const TASK_PHASE_REACT = 'phase_react';
const TASK_PHASE_UNLOAD = 'phase_unload';
const TASK_TTL$1 = 25;
const REQUEST_LOAD_TTL = 20;
const REQUEST_UNLOAD_TTL = 20;
const REACTION_TTL = 0;
const NO_SLEEP = 0;
const NEW_TASK_SLEEP = 10;
const PRODUCE_STATUS_TTL$1 = 25;
const REACTION_STATUS_STREAM = 'stream_reaction_status';
const REACTION_STATUS_START = 'reaction_status_start';
const REACTION_STATUS_UPDATE = 'reaction_status_update';
const REACTION_STATUS_STOP = 'reaction_status_stop';
class ReactorRunnable extends PersistentMemory {
    constructor(id, baseId, labIds) {
        super(id);
        this.id = id;
        this.baseId = baseId;
        this.labIds = labIds;
        this.prevTime = Game.time;
        this.threadProduceStatus = threadBaseRoom('produce_status_thread', PRODUCE_STATUS_TTL$1)(this.produceUpdateStatus.bind(this));
    }
    run(kernel, trace) {
        trace = trace.begin('reactor_run');
        const ticks = Game.time - this.prevTime;
        this.prevTime = Game.time;
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.info('base not found');
            trace.end();
            return terminate();
        }
        const room = getBasePrimaryRoom(base);
        if (!room) {
            trace.info('room not found - terminating', {});
            trace.end();
            return sleeping(50);
        }
        const labs = this.labIds.map((labId) => Game.getObjectById(labId));
        if (_.filter(labs, (lab) => !lab).length) {
            trace.info('lab missing - terminating', { labIds: this.labIds });
            trace.end();
            return terminate();
        }
        let task = this.getTask(trace);
        if (!task) {
            trace.info('no current task', {});
            const request = kernel.getTopics().getNextRequest(TASK_REACTION);
            if (!request) {
                trace.info('no available tasks', {});
                trace.end();
                return sleeping(NEW_TASK_SLEEP);
            }
            task = {
                [TASK_ID]: `unload-${this.id}-${Game.time}`,
                [MEMORY_TASK_TYPE]: TASK_HAUL,
                [MEMORY_HAUL_PICKUP]: request.details[MEMORY_HAUL_PICKUP],
                [MEMORY_HAUL_RESOURCE]: request.details[MEMORY_HAUL_RESOURCE],
                [MEMORY_HAUL_AMOUNT]: request.details[MEMORY_HAUL_AMOUNT],
                [MEMORY_HAUL_DROPOFF]: request.details[MEMORY_HAUL_DROPOFF],
                [REACTOR_TTL]: TASK_TTL$1,
            };
            this.setTask(task);
            trace.info('got new task', { task });
        }
        this.threadProduceStatus(trace, kernel, base, room, labs, task);
        trace.info('reactor run', {
            labIds: this.labIds,
            ticks,
            task,
        });
        if (task) {
            let sleepFor = 0;
            [task, sleepFor] = this.processTask(kernel, base, room, labs, task, ticks, trace);
            this.setTask(task);
            if (sleepFor) {
                trace.end();
                return sleeping(sleepFor);
            }
        }
        trace.end();
        return running();
    }
    getOutput(trace) {
        if (this.isIdle(trace)) {
            return null;
        }
        return this.getTask(trace)[REACTOR_OUTPUT];
    }
    getTask(trace) {
        const memory = this.getMemory(trace);
        if (!memory) {
            return null;
        }
        return memory['task'] || null;
    }
    setTask(task) {
        this.setMemory({
            task: task,
        });
    }
    isIdle(trace) {
        return !this.getTask(trace);
    }
    clearTask(trace) {
        const memory = this.getMemory(trace);
        if (!memory) {
            return;
        }
        delete memory['task'];
        this.setMemory(memory);
    }
    // TODO create type for task
    processTask(kernel, base, room, labs, task, ticks, trace) {
        const inputA = task[REACTOR_INPUT_A];
        const amount = task[REACTOR_AMOUNT];
        const inputB = task[REACTOR_INPUT_B];
        const phase = task[TASK_PHASE] || TASK_PHASE_START;
        switch (phase) {
            case TASK_PHASE_START:
                trace.info('starting task', { task });
                task[TASK_PHASE] = TASK_PHASE_LOAD;
            // eslint-disable-next-line no-fallthrough
            case TASK_PHASE_LOAD: {
                // Maintain task TTL. We want to abort hard to perform tasks
                let ttl = task[REACTOR_TTL];
                // Check if we have lowered the TTL and use the new one
                if (ttl > TASK_TTL$1) {
                    ttl = TASK_TTL$1;
                }
                if (ttl <= 0) {
                    trace.info('ttl exceeded, clearing task', {});
                    this.clearTask(trace);
                    return [task, NO_SLEEP];
                }
                else {
                    task[TASK_PHASE] = TASK_PHASE_LOAD;
                    task[REACTOR_TTL] = ttl - ticks;
                }
                const readyA = this.prepareInput(kernel, base, labs[1], inputA, amount, trace);
                const readyB = this.prepareInput(kernel, base, labs[2], inputB, amount, trace);
                if (readyA && readyB) {
                    trace.info('loaded, moving to react phase', {});
                    task[TASK_PHASE] = TASK_PHASE_REACT;
                    return [task, NO_SLEEP];
                }
                return [task, REQUEST_LOAD_TTL];
            }
            case TASK_PHASE_REACT: {
                if (labs[0].cooldown) {
                    trace.info('reacting cooldown', { cooldown: labs[0].cooldown });
                    return [task, labs[0].cooldown];
                }
                const result = labs[0].runReaction(labs[1], labs[2]);
                if (result !== OK) {
                    trace.info('reacted, moving to unload phase', {});
                    task[TASK_PHASE] = TASK_PHASE_UNLOAD;
                    return [task, NO_SLEEP];
                }
                return [task, REACTION_TTL];
            }
            case TASK_PHASE_UNLOAD: {
                const lab = labs[0];
                if (!lab.mineralType || lab.store.getUsedCapacity(lab.mineralType) === 0) {
                    trace.info('unloaded, task complete', {});
                    this.clearTask(trace);
                    return [task, NO_SLEEP];
                }
                this.unloadLab(kernel, base, labs[0], trace);
                return [task, REQUEST_UNLOAD_TTL];
            }
            default:
                trace.error('BROKEN REACTION LOGIC', phase);
                this.clearTask(trace);
                return [task, NO_SLEEP];
        }
    }
    unloadLabs(kernel, base, labs, trace) {
        labs.forEach((lab) => {
            if (lab.mineralType) {
                this.unloadLab(kernel, base, lab, trace);
            }
        });
    }
    prepareInput(kernel, base, lab, resource, desiredAmount, trace) {
        let currentAmount = 0;
        if (lab.mineralType) {
            currentAmount = lab.store.getUsedCapacity(lab.mineralType);
        }
        // Unload the lab if it's not the right mineral
        if (lab.mineralType && lab.mineralType !== resource && lab.store.getUsedCapacity(lab.mineralType) > 0) {
            this.unloadLab(kernel, base, lab, trace);
            return false;
        }
        // Load the lab with the right mineral
        if (currentAmount < desiredAmount) {
            const pickup = getStructureWithResource(base, resource);
            const missingAmount = desiredAmount - currentAmount;
            if (!pickup) {
                this.requestResource(kernel, base, resource, missingAmount, trace);
            }
            else {
                this.loadLab(kernel, lab, pickup, resource, missingAmount, trace);
            }
            return false;
        }
        return true;
    }
    requestResource(kernel, base, resource, amount, trace) {
        // TODO this really should use topics/IPC
        trace.info('requesting resource from governor', { resource, amount });
        const resourceManager = kernel.getResourceManager();
        const requested = resourceManager.requestResource(base, resource, amount, REQUEST_LOAD_TTL, trace);
        if (!requested) {
            resourceManager.buyResource(base, resource, amount, REQUEST_LOAD_TTL, trace);
        }
    }
    loadLab(kernel, lab, pickup, resource, amount, trace) {
        trace.info('requesting load', {
            lab: lab.id,
            resource: resource,
            amount: amount,
            pickup: pickup.id,
            ttl: REQUEST_LOAD_TTL,
        });
        kernel.getTopics().addRequest(getBaseDistributorTopic(this.baseId), HAUL_REACTION, {
            [TASK_ID]: `load-${this.id}-${Game.time}`,
            [MEMORY_TASK_TYPE]: TASK_HAUL,
            [MEMORY_HAUL_PICKUP]: pickup.id,
            [MEMORY_HAUL_RESOURCE]: resource,
            [MEMORY_HAUL_AMOUNT]: amount,
            [MEMORY_HAUL_DROPOFF]: lab.id,
        }, REQUEST_LOAD_TTL + Game.time);
    }
    unloadLab(kernel, base, lab, trace) {
        const currentAmount = lab.store.getUsedCapacity(lab.mineralType);
        const dropoff = getStructureForResource(base, lab.mineralType);
        trace.info('requesting unload', {
            lab: lab.id,
            resource: lab.mineralType,
            amount: currentAmount,
            dropoff: dropoff.id,
            ttl: REQUEST_LOAD_TTL,
        });
        kernel.getTopics().addRequest(getBaseDistributorTopic(this.baseId), HAUL_REACTION, {
            [TASK_ID]: `unload-${this.id}-${Game.time}`,
            [MEMORY_TASK_TYPE]: TASK_HAUL,
            [MEMORY_HAUL_PICKUP]: lab.id,
            [MEMORY_HAUL_RESOURCE]: lab.mineralType,
            [MEMORY_HAUL_AMOUNT]: currentAmount,
            [MEMORY_HAUL_DROPOFF]: dropoff.id,
        }, REQUEST_LOAD_TTL + Game.time);
    }
    produceUpdateStatus(trace, kernel, base, room, labs, task) {
        var _a;
        if (!task) {
            trace.info('no task', { baseId: this.baseId });
            return;
        }
        const resource = task[REACTOR_OUTPUT] || null;
        const phase = task[TASK_PHASE] || null;
        const amount = ((_a = labs[0]) === null || _a === void 0 ? void 0 : _a.store.getUsedCapacity(task[REACTOR_OUTPUT])) || 0;
        // TODO add reactor status and reactor task
        const status = {
            [REACTION_STATUS_ROOM]: room.name,
            [REACTION_STATUS_LAB]: labs[0].id,
            [REACTION_STATUS_RESOURCE]: resource,
            [REACTION_STATUS_RESOURCE_AMOUNT]: amount,
            [REACTION_STATUS_PHASE]: phase,
        };
        trace.info('producing reactor status', { status });
        kernel.getTopics().addRequest(ACTIVE_REACTIONS, 1, status, PRODUCE_STATUS_TTL$1 + Game.time);
        // @REFACTOR there *may* be a bug here. I don't think it's calling stop
        if (resource) {
            kernel.getBroker().getStream(REACTION_STATUS_STREAM).
                publish(new Event(this.id, Game.time, REACTION_STATUS_UPDATE, status));
        }
        else {
            kernel.getBroker().getStream(REACTION_STATUS_STREAM).
                publish(new Event(this.id, Game.time, REACTION_STATUS_STOP, {}));
        }
    }
}

const reactorPositions = [
    [[3, -2], [3, -1], [2, -2]],
    [[-3, 2], [-3, 1], [-2, 2]], // Bottom left
];
// TODO I flipped TR and TL so that boosting was available at RCL6, think this through at a later time
const boosterPositions = [[-3, -2], [-3, -1], [-2, -2]]; // Top left
const RUN_TTL$5 = 20;
const ASSIGN_LABS_TTL = 20;
class LabsManager {
    constructor(baseId, id, scheduler, _trace) {
        this.id = id;
        this.baseId = baseId;
        this.scheduler = scheduler;
        this.reactorsIds = [];
        this.boosterIds = [];
        this.threadAssignLabs = threadBaseRoom('assign_labs', ASSIGN_LABS_TTL)(this.assignLabs.bind(this));
    }
    run(kernel, trace) {
        trace = trace.begin('labs_manager_run');
        trace.info('labs manager run', {
            reactorsIds: this.reactorsIds,
            boosterIds: this.boosterIds,
        });
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.info('no base config for room', { baseId: this.baseId });
            return terminate();
        }
        this.threadAssignLabs(trace, kernel, base);
        trace.end();
        return sleeping(RUN_TTL$5);
    }
    assignLabs(trace, kernel, base, room) {
        this.assignBasedOnPosition(kernel, base, room, trace);
        trace.info('assigned labs', { reactors: this.reactorsIds, booster: this.boosterIds });
        // Check that we have processes for reactors
        this.reactorsIds.forEach((reactorIds) => {
            const reactorId = `${reactorIds[0]}`;
            const hasProcess = this.scheduler.hasProcess(reactorId);
            if (!hasProcess) {
                trace.info('creating process for reactor', { reactorId });
                this.scheduler.registerProcess(new Process(reactorId, 'reactors', Priorities.RESOURCES, new ReactorRunnable(reactorId, base.id, reactorIds)));
            }
        });
        // Check that we have processes for boosters
        if (this.boosterIds.length) {
            const boosterId = `${this.boosterIds[0]}`;
            const hasProcess = this.scheduler.hasProcess(boosterId);
            if (!hasProcess) {
                trace.info('creating process for booster', { boosterId });
                const booster = new BoosterRunnable(boosterId, base.id, this.boosterIds);
                this.scheduler.registerProcess(new Process(boosterId, 'boosters', Priorities.RESOURCES, booster));
            }
        }
    }
    // Automated assignment of labs based on position
    assignBasedOnPosition(kernel, base, orgRoom, trace) {
        trace = trace.begin('assign_labs');
        const room = getBasePrimaryRoom(base);
        if (!room) {
            trace.end();
            return;
        }
        this.reactorsIds = [];
        const origin = base.origin;
        reactorPositions.forEach((offsets) => {
            let reactorIds = offsets.map(([x, y]) => {
                var _a;
                return (_a = room.lookForAt(LOOK_STRUCTURES, origin.x + x, origin.y + y).
                    find((s) => s.structureType === STRUCTURE_LAB)) === null || _a === void 0 ? void 0 : _a.id;
            });
            reactorIds = reactorIds.filter((id) => id);
            reactorIds = ___namespace.sortBy(reactorIds, 'id');
            if (reactorIds.length === 3) {
                this.reactorsIds.push(reactorIds);
            }
        });
        let boosterIds = boosterPositions.map(([x, y]) => {
            var _a;
            return (_a = room.lookForAt(LOOK_STRUCTURES, origin.x + x, origin.y + y).
                find((s) => s.structureType === STRUCTURE_LAB)) === null || _a === void 0 ? void 0 : _a.id;
        });
        boosterIds = boosterIds.filter((id) => id);
        boosterIds = ___namespace.sortBy(boosterIds, 'id');
        if (boosterIds.length === 3) {
            this.boosterIds = boosterIds;
        }
        trace.end();
    }
}

const TICK_STEP = 2;
const PROCESS_TTL = 250;
const HAUL_TTL = 10;
const ENERGY_READY_AMOUNT = 600;
const UPDATE_STRUCTURES_TTL = 100;
class LinkManager {
    constructor(id, baseId) {
        this.id = id;
        this.baseId = baseId;
        this.storageLink = null;
        this.sourceLinks = [];
        this.sinkLinks = null;
        this.ttl = PROCESS_TTL;
        this.haulTTL = 0;
        this.prevTime = Game.time;
        this.threadUpdateStructures = threadBaseRoom('update_structures', UPDATE_STRUCTURES_TTL)(this.updateStructures.bind(this));
    }
    run(kernel, trace) {
        trace = trace.begin('link_manager_run');
        const ticks = Game.time - this.prevTime;
        this.prevTime = Game.time;
        this.haulTTL -= ticks;
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.info('no base config for room', { baseId: this.baseId });
            return terminate();
        }
        const room = getBasePrimaryRoom(base);
        if (!room) {
            trace.error('base primary room not visible', { baseId: this.baseId });
            trace.end();
            return;
        }
        this.threadUpdateStructures(trace, kernel, base, room);
        trace.info('running', {
            storageId: this.storageId,
            storageLink: this.storageLink,
            sourceLinks: this.sourceLinks,
            sinkLinks: this.sinkLinks,
            haulTTL: this.haulTTL,
            ttl: this.ttl,
        });
        if (!this.storageLink) {
            trace.info('sleeping due to not having a storage link', {});
            trace.end();
            return sleeping(PROCESS_TTL);
        }
        const storageLink = Game.getObjectById(this.storageLink);
        if (!this.storageId || !storageLink) {
            trace.info('sleeping due to missing storage or storage link', {});
            trace.end();
            return sleeping(PROCESS_TTL); // Removed +1 from this 4/25/22
        }
        let performedTransfer = false;
        // If our link Ids are old we should terminate
        let shouldTerminate = false;
        this.ttl -= ticks;
        if (this.ttl < 0) {
            shouldTerminate = true;
        }
        // Create list of links with energy
        const hasEnergy = this.sourceLinks.map((linkId) => {
            const link = Game.getObjectById(linkId);
            if (!link) {
                trace.info('should terminate due to missing source link', { linkId });
                shouldTerminate = true;
                trace.end();
                return null;
            }
            return link;
        }).filter((link) => canTransfer(link));
        // Include storage link if it has energy and it isn't being hauled
        if (notEmpty(storageLink) && this.haulTTL <= 0) {
            hasEnergy.push(storageLink);
        }
        // Get links that need energy and transfer if we have source links with energy
        this.sinkLinks.map((linkId) => {
            const link = Game.getObjectById(linkId);
            if (!link) {
                trace.info('should terminate due to missing sink link', { linkId });
                shouldTerminate = true;
                return null;
            }
            return link;
        }).forEach((sink) => {
            if (!canReceive(sink)) {
                return;
            }
            const source = hasEnergy.pop();
            if (!source) {
                return;
            }
            performedTransfer = true;
            transferEnergy(source, sink, trace);
        });
        // If storage link needs energy try to get some from a source
        if (canReceive(storageLink)) {
            let source = hasEnergy.pop();
            // Don't transfer to itself
            if (source && source.id === storageLink.id) {
                source = hasEnergy.pop();
            }
            if (source) {
                performedTransfer = true;
                transferEnergy(source, storageLink, trace);
            }
        }
        trace.info('has energy', { hasEnergy, haulTTL: this.haulTTL, performedTransfer });
        if (!performedTransfer && this.haulTTL < 0) {
            // If we have no source ready with energy for sinks, then load energey into storage
            if (!hasEnergy.length) {
                const amount = storageLink.store.getFreeCapacity(RESOURCE_ENERGY);
                if (amount) {
                    const pickup = getStructureWithResource(base, RESOURCE_ENERGY);
                    if (pickup) {
                        this.fillLink(kernel, storageLink, pickup, _.min([amount, ENERGY_READY_AMOUNT]), trace);
                    }
                }
            }
            else if (hasEnergy.length > 1) { // If we have more than one source link ready, unload storage link
                const amount = storageLink.store.getUsedCapacity(RESOURCE_ENERGY);
                if (amount) {
                    const dropoff = getStructureForResource(base, RESOURCE_ENERGY);
                    if (dropoff) {
                        this.emptyLink(kernel, storageLink, dropoff, amount, trace);
                    }
                }
            }
        }
        // Check if we need to terminate due to being stale
        // TODO change in links or storage should cause termination; then remove ttl
        if (shouldTerminate) {
            trace.info('terminating link process', {});
            trace.end();
            return terminate();
        }
        trace.end();
        return sleeping(TICK_STEP);
    }
    updateStructures(trace, kernel, base, room) {
        var _a, _b;
        this.storageId = (_a = room.storage) === null || _a === void 0 ? void 0 : _a.id;
        if (!this.storageId) {
            throw new Error('cannot create a link manager when room does not have storage');
        }
        // TODO move these to a thread
        if (room.storage) {
            this.storageLink = (_b = room.storage.pos.findInRange(FIND_STRUCTURES, 3, {
                filter: (structure) => {
                    return structure.structureType === STRUCTURE_LINK && structure.isActive();
                },
            })[0]) === null || _b === void 0 ? void 0 : _b.id;
        }
        const sources = room.find(FIND_SOURCES);
        this.sourceLinks = sources.map((source) => {
            var _a;
            return (_a = source.pos.findInRange(FIND_STRUCTURES, 2, {
                filter: (structure) => {
                    return structure.structureType === STRUCTURE_LINK && structure.isActive();
                },
            })[0]) === null || _a === void 0 ? void 0 : _a.id;
        }).filter((value) => value);
        const controller = room.controller;
        if (controller) {
            this.sinkLinks = controller.pos.findInRange(FIND_STRUCTURES, 4, {
                filter: (structure) => {
                    return structure.structureType === STRUCTURE_LINK && structure.isActive();
                },
            }).map((link) => {
                return link.id;
            });
        }
    }
    emptyLink(kernel, link, dropoff, amount, trace) {
        this.haulTTL = HAUL_TTL;
        const details = {
            [TASK_ID]: `lu-${link.id}-${Game.time}`,
            [MEMORY_TASK_TYPE]: TASK_HAUL,
            [MEMORY_HAUL_PICKUP]: link.id,
            [MEMORY_HAUL_AMOUNT]: amount,
            [MEMORY_HAUL_RESOURCE]: RESOURCE_ENERGY,
            [MEMORY_HAUL_DROPOFF]: dropoff.id,
        };
        kernel.getTopics().addRequest(getBaseDistributorTopic(this.baseId), UNLOAD_LINK, details, HAUL_TTL + Game.time);
        trace.info('haul energy from storage link', { request: details });
    }
    fillLink(kernel, link, pickup, amount, trace) {
        this.haulTTL = HAUL_TTL;
        const details = {
            [TASK_ID]: `ll-${link.id}-${Game.time}`,
            [MEMORY_TASK_TYPE]: TASK_HAUL,
            [MEMORY_HAUL_PICKUP]: pickup.id,
            [MEMORY_HAUL_AMOUNT]: amount,
            [MEMORY_HAUL_RESOURCE]: RESOURCE_ENERGY,
            [MEMORY_HAUL_DROPOFF]: link.id,
        };
        trace.info('haul energy to storage link', { request: details });
        kernel.getTopics().addRequest(getBaseDistributorTopic(this.baseId), LOAD_LINK, details, HAUL_TTL + Game.time);
    }
}
const transferEnergy = (source, sink, trace) => {
    const amount = _.min([
        source.store.getUsedCapacity(RESOURCE_ENERGY),
        sink.store.getFreeCapacity(RESOURCE_ENERGY),
    ]);
    const result = source.transferEnergy(sink, amount);
    trace.info('transfer energy', {
        source: source.id,
        sink: sink.id,
        result,
    });
};
const canTransfer = (link) => {
    return (link === null || link === void 0 ? void 0 : link.store.getUsedCapacity(RESOURCE_ENERGY)) >= ENERGY_READY_AMOUNT && link.cooldown < 1;
};
const canReceive = (link) => {
    return (link === null || link === void 0 ? void 0 : link.store.getFreeCapacity(RESOURCE_ENERGY)) >= ENERGY_READY_AMOUNT;
};
const notEmpty = (link) => {
    return (link === null || link === void 0 ? void 0 : link.store.getUsedCapacity(RESOURCE_ENERGY)) > 0;
};

const RUN_INTERVAL$2 = 100;
const MAX_NEIGHBORS = 2;
class NeighborsRunnable {
    constructor(baseId) {
        this.baseId = baseId;
    }
    run(kernel, trace) {
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.error('base not found', { baseId: this.baseId });
            return terminate();
        }
        this.updateNeighbors(kernel, base, trace);
        return sleeping(RUN_INTERVAL$2);
    }
    updateNeighbors(kernel, currentBase, trace) {
        // Narrow bases to ones that are nearby
        let nearbyBases = _.filter(kernel.getPlanner().getBases(), (base) => {
            if (currentBase.id === base.id) {
                return false;
            }
            const distance = Game.map.getRoomLinearDistance(currentBase.primary, base.primary);
            if (distance > 6) {
                return false;
            }
            // RAKE calculate path check number of rooms in path, factoring in enemy rooms
            return true;
        });
        // Sort by distance
        nearbyBases = _.sortBy(nearbyBases, (base) => {
            return Game.map.getRoomLinearDistance(currentBase.primary, base.primary);
        });
        // Pick at most nearest 3
        nearbyBases = _.take(nearbyBases, MAX_NEIGHBORS);
        // Set bases neighbors
        currentBase.neighbors = nearbyBases.map((base) => base.id);
        trace.info('updated neighbors', { base: currentBase });
    }
}

const REQUEST_RESOURCES_TTL = 25;
class NukerRunnable {
    constructor(baseId, nuker) {
        this.id = nuker.id;
        this.baseId = baseId;
        this.haulTTL = 0;
        this.prevTime = Game.time;
    }
    run(kernel, trace) {
        trace = trace.begin('nuker_run');
        const ticks = Game.time - this.prevTime;
        this.prevTime = Game.time;
        this.haulTTL -= ticks;
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.error('no base config for room', { baseId: this.baseId });
            return terminate();
        }
        const room = getBasePrimaryRoom(base);
        if (!room) {
            trace.end();
            return terminate();
        }
        const nuker = Game.getObjectById(this.id);
        if (!nuker) {
            trace.end();
            return terminate();
        }
        if (!nuker.isActive()) {
            trace.end();
            return sleeping(100);
        }
        let readyToFire = !nuker.cooldown;
        const neededEnergy = nuker.store.getFreeCapacity(RESOURCE_ENERGY);
        if (neededEnergy > 0) {
            trace.info('need energy', { neededEnergy });
            this.requestResource(kernel, base, RESOURCE_ENERGY, neededEnergy, trace);
            readyToFire = false;
        }
        const neededGhodium = nuker.store.getFreeCapacity(RESOURCE_GHODIUM);
        if (neededGhodium > 0) {
            trace.info('need ghodium', { neededGhodium });
            this.requestResource(kernel, base, RESOURCE_GHODIUM, neededGhodium, trace);
            readyToFire = false;
        }
        if (readyToFire) {
            trace.info('lets play global thermonuclear war');
            const request = kernel.getTopics().getNextRequest(NUKER_TARGETS);
            if (request) {
                const positionStr = request.details.position;
                const posArray = positionStr.split(',');
                let position = null;
                if (posArray && posArray.length === 3) {
                    position = new RoomPosition(posArray[0], posArray[1], posArray[2]);
                }
                else {
                    trace.info('problem with position string', { positionStr });
                }
                if (position !== null) {
                    trace.info('would nuke', { position });
                    const result = nuker.launchNuke(position);
                    trace.notice('nuker launch result', { result, position });
                }
            }
        }
        trace.end();
        return sleeping(REQUEST_RESOURCES_TTL);
    }
    requestResource(kernel, base, resource, amount, trace) {
        const pickup = getStructureWithResource(base, resource);
        if (!pickup) {
            trace.info('requesting resource from governor', { resource, amount });
            const resourceGovernor = kernel.getResourceManager();
            const requested = resourceGovernor.requestResource(base, resource, amount, REQUEST_RESOURCES_TTL, trace);
            if (!requested) {
                resourceGovernor.buyResource(base, resource, amount, REQUEST_RESOURCES_TTL, trace);
            }
            return;
        }
        trace.info('requesting load', {
            nuker: this.id,
            resource: resource,
            amount: amount,
            pickup: pickup.id,
            ttl: REQUEST_RESOURCES_TTL,
        });
        const request = {
            [TASK_ID]: `load-${this.id}-${Game.time}`,
            [MEMORY_TASK_TYPE]: TASK_HAUL,
            [MEMORY_HAUL_PICKUP]: pickup.id,
            [MEMORY_HAUL_RESOURCE]: resource,
            [MEMORY_HAUL_AMOUNT]: amount,
            [MEMORY_HAUL_DROPOFF]: this.id,
        };
        kernel.getTopics().addRequest(getBaseDistributorTopic(this.baseId), HAUL_NUKER, request, REQUEST_RESOURCES_TTL + Game.time);
    }
}

class ObserverRunnable {
    constructor(baseId, observer) {
        this.id = observer.id;
        this.baseId = baseId;
        this.inRangeRooms = inRangeRoomNames(observer.pos.roomName);
    }
    run(kernel, trace) {
        trace = trace.as('observer_run');
        const observer = Game.getObjectById(this.id);
        if (!observer) {
            trace.error('missing structure', { id: this.id });
            trace.end();
            return terminate();
        }
        if (this.justObserved && Game.rooms[this.justObserved]) {
            kernel.getScribe().updateRoom(kernel, Game.rooms[this.justObserved], trace);
        }
        trace.info('in range rooms', { inRange: this.inRangeRooms });
        const getOldestRoomTrace = trace.begin('get_oldest_room');
        const nextRoom = kernel.getScribe().getOldestRoomInList(this.inRangeRooms);
        getOldestRoomTrace.end();
        trace.info('next room', { nextRoom });
        if (!nextRoom) {
            trace.end();
            return running();
        }
        const observeRoomTrace = trace.begin('observe_room');
        const result = observer.observeRoom(nextRoom);
        observeRoomTrace.end();
        if (result === OK) {
            this.justObserved = nextRoom;
        }
        else {
            this.justObserved = null;
        }
        trace.end();
        return running();
    }
}
const inRangeRoomNames = (centerRoomName) => {
    const roomsInRange = [];
    const centerRoomXY = roomNameToXY(centerRoomName);
    const topLeft = [centerRoomXY[0] - 10, centerRoomXY[1] - 10];
    for (let i = 0; i < 21; i++) {
        for (let j = 0; j < 21; j++) {
            const roomName = roomNameFromXY(topLeft[0] + i, topLeft[1] + j);
            if (Game.map.getRoomStatus(roomName).status !== 'closed') {
                roomsInRange.push(roomName);
            }
        }
    }
    return roomsInRange;
};
// https://github.com/screeps/engine/blob/master/src/utils.js
const roomNameFromXY = (x, y) => {
    if (x < 0) {
        x = 'W' + (-x - 1);
    }
    else {
        x = 'E' + (x);
    }
    if (y < 0) {
        y = 'N' + (-y - 1);
    }
    else {
        y = 'S' + (y);
    }
    return ('' + x + y);
};
// https://github.com/screeps/engine/blob/master/src/utils.js
const roomNameToXY = (name) => {
    let xx = parseInt(name.substr(1), 10);
    let verticalPos = 2;
    if (xx >= 100) {
        verticalPos = 4;
    }
    else if (xx >= 10) {
        verticalPos = 3;
    }
    let yy = parseInt(name.substr(verticalPos + 1), 10);
    const horizontalDir = name.charAt(0);
    const verticalDir = name.charAt(verticalPos);
    if (horizontalDir === 'W' || horizontalDir === 'w') {
        xx = -xx - 1;
    }
    if (verticalDir === 'N' || verticalDir === 'n') {
        yy = -yy - 1;
    }
    return [xx, yy];
};

const RUN_INTERVAL$1 = 50;
class RemotesManager {
    constructor(baseId) {
        this.baseId = baseId;
        this.spawnUtilizationConsumer = null;
        this.spawnUtilization = -1;
    }
    run(kernel, trace) {
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.error('base not found', { baseId: this.baseId });
            return terminate();
        }
        if (!this.spawnUtilizationConsumer) {
            const topic = getBaseSpawnUtilizationTopic(base.id);
            this.spawnUtilizationConsumer = kernel.getBroker().getStream(topic).addConsumer(this.baseId);
        }
        // Update
        this.processEvents(kernel, base, trace);
        this.updateRemotes(kernel, base, trace);
        return sleeping(RUN_INTERVAL$1);
    }
    // Process events from the spawn utilization topic
    processEvents(kernel, base, trace) {
        const events = this.spawnUtilizationConsumer.getEvents();
        for (const event of events) {
            this.spawnUtilization = event.data.utilization;
            trace.info('spawn utilization', { spawnUtilization: this.spawnUtilization });
        }
    }
    // Update base rooms/remotes based on the current situation
    updateRemotes(kernel, base, trace) {
        var _a;
        if (this.spawnUtilization === -1) {
            trace.warn('spawn utilization not set');
            //return;
        }
        const primaryRoom = getBasePrimaryRoom(base);
        if (!primaryRoom) {
            trace.warn('primary room not found', { roomName: base.primary });
            return;
        }
        const level = ((_a = primaryRoom === null || primaryRoom === void 0 ? void 0 : primaryRoom.controller) === null || _a === void 0 ? void 0 : _a.level) || 0;
        const numDesired = desiredRemotes(base, level, this.spawnUtilization, trace);
        if (numDesired === 0) {
            trace.info('no remotes desired', { base: base.id, level: level });
            return;
        }
        resetRemotes(base, trace);
        const [remotes, debug] = findRemotes(kernel, base, trace);
        if (remotes.length === 0) {
            trace.info('no remotes found', { base: base.id, level: level });
            return;
        }
        trace.info('found remotes', { base: base.id, level: level, numDesired, remotes: remotes, debug: debug });
        for (let i = 0; i < Math.min(numDesired, remotes.length); i++) {
            addRoom(base, remotes[i], trace);
        }
    }
}

const UPDATE_DAMAGED_STRUCTURES_TTL = 40;
const UPDATE_DAMAGED_SECONDARY_TTL = 15;
const UPGRADER_BUFFER = 25000;
// TODO increase this later, we should be able to sustain at least one nuke
// before the walls break
const MAX_WALL_HITS = 11000000;
class RepairRunnable {
    constructor(baseId) {
        this.baseId = baseId;
        this.threadUpdateDamagedStructure = threadBaseRoom('damaged_structures_thread', UPDATE_DAMAGED_STRUCTURES_TTL)(this.updateDamagedStructures.bind(this));
        this.threadUpdateDamagedSecondaryStructures = threadBaseRoom('secondary_structures_thread', UPDATE_DAMAGED_SECONDARY_TTL)(this.updateDamagedSecondaryStructures.bind(this));
    }
    run(kernel, trace) {
        trace = trace.as('repair_run');
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.error('no base config for room', { baseId: this.baseId });
            return terminate();
        }
        const room = getBasePrimaryRoom(base);
        if (!room) {
            trace.error('base primary room not visible', { baseId: this.baseId });
            trace.end();
            return sleeping(10);
        }
        this.threadUpdateDamagedStructure(trace, kernel, base, room);
        this.threadUpdateDamagedSecondaryStructures(trace, kernel, base, room);
        trace.end();
        return running();
    }
    updateDamagedStructures(trace, kernel, base, room) {
        const damagedStructures = room.find(FIND_STRUCTURES, {
            filter: (s) => {
                return s.hits < s.hitsMax && (s.structureType != STRUCTURE_WALL && s.structureType != STRUCTURE_RAMPART &&
                    s.structureType != STRUCTURE_ROAD);
            },
        });
        base.damagedStructures = _.map(damagedStructures, 'id');
        trace.info('damaged structures', {
            room: this.baseId,
            damagedStructures: base.damagedStructures,
        });
    }
    updateDamagedSecondaryStructures(trace, kernel, base, room) {
        const rcLevel = room.controller.level.toString();
        const rcLevelHitsMax = RAMPART_HITS_MAX[rcLevel] || 10000;
        const energyFullness = getEnergyFullness(base) * 10;
        // was .45 but changed to to .4 to increase the HP of walls early levels
        base.defenseHitsLimit = rcLevelHitsMax * Math.pow(0.40, (10 - energyFullness));
        if (room.storage && room.storage.store.getUsedCapacity(RESOURCE_ENERGY) < 50000) {
            base.defenseHitsLimit = 10000;
        }
        // If energy in reserve is less then we need to sustain a max ugprader,
        // then limit the amount our defense hits
        const reserveEnergy = getStoredResourceAmount(base, RESOURCE_ENERGY);
        const reserveBuffer = getReserveBuffer(base);
        if (reserveEnergy < reserveBuffer + UPGRADER_BUFFER) {
            base.defenseHitsLimit = _.min([base.defenseHitsLimit, MAX_WALL_HITS]);
        }
        let damagedSecondaryStructures = room.find(FIND_STRUCTURES, {
            filter: (s) => {
                return s.hits < s.hitsMax && (s.structureType == STRUCTURE_RAMPART ||
                    s.structureType == STRUCTURE_WALL) &&
                    s.hits < base.defenseHitsLimit;
            },
        });
        damagedSecondaryStructures = _.sortBy(damagedSecondaryStructures, (structure) => {
            return structure.hits;
        });
        base.damagedSecondaryStructures = _.map(damagedSecondaryStructures, 'id');
        trace.info('damaged secondary structures', {
            room: this.baseId,
            defenseHitsLimit: base.defenseHitsLimit,
            damagedSecondaryStructures: base.damagedSecondaryStructures,
        });
    }
}

const DROPOFF_TTL$1 = 200;
const REQUEST_WORKER_TTL = 50;
const PRODUCE_EVENTS_TTL = 50;
const RED_ALERT_TTL = 200;
const CONTAINER_TTL$1 = 250;
class MineralRunnable extends PersistentMemory {
    constructor(mineral) {
        super(mineral.id);
        this.id = mineral.id;
        this.position = mineral.pos;
        this.creepPosition = null;
        this.threadProduceEvents = threadBase('consume_events', PRODUCE_EVENTS_TTL)(this.produceEvents.bind(this));
        this.threadUpdateDropoff = threadBase('update_dropoff', DROPOFF_TTL$1)(this.updateDropoff.bind(this));
        this.threadRequestHarvesters = threadBase('request_miners', REQUEST_WORKER_TTL)(this.requestHarvesters.bind(this));
        this.threadBuildExtractor = threadBase('build_extractor', CONTAINER_TTL$1)(this.buildExtractor.bind(this));
    }
    run(kernel, trace) {
        trace = trace.begin('mineral_run');
        trace.info('mineral run', {
            mineralId: this.id,
            position: this.position,
            creepPosition: this.creepPosition,
        });
        const mineral = Game.getObjectById(this.id);
        if (!mineral) {
            trace.error('mineral not found', { id: this.id });
            trace.end();
            return terminate();
        }
        const base = kernel.getPlanner().getBaseByRoom(mineral.room.name);
        if (!base) {
            trace.error('no base config', { room: mineral.room.name });
            trace.end();
            return terminate();
        }
        // If red alert, don't do anything
        if (base.alertLevel === AlertLevel.RED) {
            trace.error('red alert', { room: mineral.room.name });
            trace.end();
            return sleeping(RED_ALERT_TTL);
        }
        if (!this.creepPosition) {
            this.populatePositions(trace, kernel, base, mineral);
        }
        this.threadProduceEvents(trace, kernel, base, mineral);
        this.threadUpdateDropoff(trace, kernel, base, mineral);
        this.threadRequestHarvesters(trace, kernel, base, mineral);
        this.threadBuildExtractor(trace, kernel, base, mineral);
        trace.end();
        let sleepFor = 100;
        if (mineral.mineralAmount === 0) {
            sleepFor = mineral.ticksToRegeneration;
        }
        return sleeping(sleepFor);
    }
    produceEvents(trace, kernel, base, mineral) {
        const data = {
            id: mineral.id,
            position: mineral.pos,
        };
        kernel.getBroker().getStream(getLogisticsTopic(base.id)).
            publish(new Event(this.id, Game.time, LogisticsEventType.RequestRoad, data));
        const hudLine = {
            key: `${this.id}`,
            room: mineral.room.name,
            text: `mineral(${mineral.id}) - `,
            time: Game.time,
            order: 4,
        };
        kernel.getBroker().getStream(getLinesStream()).publish(new Event(this.id, Game.time, HudEventSet, hudLine));
    }
    populatePositions(trace, kernel, base, mineral) {
        trace.info('populate positions', { room: mineral.room.name });
        const memory = this.getMemory(trace);
        // Check memory for creep position
        const creepPosition = memory.creepPosition;
        if (creepPosition) {
            trace.info('creep position in memory', { room: mineral.room.name });
            this.creepPosition = new RoomPosition(creepPosition.x, creepPosition.y, creepPosition.roomName);
        }
        const basePos = new RoomPosition(base.origin.x, base.origin.y - 1, base.origin.roomName);
        const [pathResult, details] = getPath(kernel, mineral.pos, basePos, roadPolicy, trace);
        if (!pathResult || !pathResult.path.length) {
            trace.error('path not found', { basePos, mineral: mineral.pos, details });
            return;
        }
        trace.info('creep position set', { creepPosition: this.creepPosition });
        this.creepPosition = pathResult.path[0];
        // Update memory
        memory.creepPosition = this.creepPosition;
        this.setMemory(memory, false);
    }
    updateDropoff(_trace, _kernel, base, _mineral) {
        var _a;
        this.dropoffId = (_a = getStructureForResource(base, RESOURCE_ENERGY)) === null || _a === void 0 ? void 0 : _a.id;
    }
    requestHarvesters(trace, kernel, base, mineral) {
        var _a, _b, _c, _d;
        if (mineral.mineralAmount === 0) {
            trace.info('no minerals to harvest');
            return;
        }
        if (!this.creepPosition) {
            trace.error('creep position not set', { creepPosition: this.creepPosition });
            return;
        }
        if (!mineral.room) {
            trace.error('mineral room not visible', { room: (_a = mineral.room) === null || _a === void 0 ? void 0 : _a.name });
            return;
        }
        const room = mineral.room;
        const username = kernel.getPlanner().getUsername();
        if (((_b = room === null || room === void 0 ? void 0 : room.controller) === null || _b === void 0 ? void 0 : _b.owner) && room.controller.owner.username !== username) {
            trace.info('room owned by someone else', { roomId: room.name, owner: (_d = (_c = room.controller) === null || _c === void 0 ? void 0 : _c.owner) === null || _d === void 0 ? void 0 : _d.username });
            return;
        }
        const baseCreeps = kernel.getCreepsManager().getCreepsByBase(base.id);
        const numHarvesters = baseCreeps.filter((creep) => {
            const role = creep.memory[MEMORY_ROLE];
            return role === WORKER_HARVESTER &&
                creep.memory[MEMORY_SOURCE] === this.id &&
                creepIsFresh(creep);
        }).length;
        trace.info('num harvesters', { numHarvesters });
        if (numHarvesters < 1) {
            const positionStr = [this.creepPosition.x, this.creepPosition.y, this.creepPosition.roomName].join(',');
            const memory = {
                [MEMORY_SOURCE]: this.id,
                [MEMORY_SOURCE_POSITION]: positionStr,
                [MEMORY_ASSIGN_ROOM]: mineral.room.name,
                [MEMORY_BASE]: base.id,
            };
            trace.info('requesting harvester', { mineralId: this.id, memory });
            const request = createSpawnRequest(PRIORITY_MINER_PRIMARY, REQUEST_WORKER_TTL + Game.time, WORKER_HARVESTER, memory, null, 0);
            kernel.getTopics().addRequestV2(getBaseSpawnTopic(base.id), request);
        }
    }
    buildExtractor(trace, kernel, base, mineral) {
        var _a;
        if (((_a = mineral.room.controller) === null || _a === void 0 ? void 0 : _a.level) < 6) {
            trace.info('room too low for extractor', { id: this.id });
            return;
        }
        const extractor = mineral.pos.lookFor(LOOK_STRUCTURES).find((structure) => {
            return structure.structureType === STRUCTURE_EXTRACTOR;
        });
        if (!extractor) {
            const site = mineral.pos.lookFor(LOOK_CONSTRUCTION_SITES).find((site) => {
                return site.structureType === STRUCTURE_EXTRACTOR;
            });
            if (!site) {
                mineral.room.createConstructionSite(mineral.pos, STRUCTURE_EXTRACTOR);
                trace.warn('building extractor', { id: this.id });
            }
        }
    }
}

const RUN_TTL$4 = 20;
const STRUCTURE_TTL = 200;
const DROPOFF_TTL = 200;
const BUILD_LINK_TTL = 200;
const CONTAINER_TTL = 250;
class SourceRunnable extends PersistentMemory {
    constructor(source) {
        super(source.id);
        this.id = source.id;
        this.creepPosition = null;
        this.linkPosition = null;
        this.openPositions = null;
        this.threadProduceEvents = threadBase('consume_events', RUN_TTL$4)(this.produceEvents.bind(this));
        this.threadUpdateStructures = threadBase('update_structures', STRUCTURE_TTL)(this.updateStructures.bind(this));
        this.threadUpdateDropoff = threadBase('update_dropoff', DROPOFF_TTL)(this.updateDropoff.bind(this));
        this.threadBuildContainer = threadBase('build_container', CONTAINER_TTL)(this.buildContainer.bind(this));
        this.threadBuildLink = threadBase('build_link', BUILD_LINK_TTL)(this.buildLink.bind(this));
        this.threadRequestMiners = threadBase('request_miners', RUN_TTL$4)(this.requestMiners.bind(this));
        this.threadRequestHauling = threadBase('reqeust_hauling', RUN_TTL$4)(this.requestHauling.bind(this));
    }
    run(kernel, trace) {
        trace = trace.begin('source_run');
        trace.info('source run', {
            sourceId: this.id,
            containerId: this.containerId,
            linkId: this.linkId,
            creepPosition: this.creepPosition,
            linkPosition: this.linkPosition,
            openPositions: this.openPositions,
        });
        const source = Game.getObjectById(this.id);
        if (!source) {
            trace.error('source not found', { id: this.id });
            trace.end();
            return terminate();
        }
        if (!source.room) {
            trace.error('source room not visible', { id: this.id });
            trace.end();
            return terminate();
        }
        const base = kernel.getPlanner().getBaseByRoom(source.room.name);
        if (!base) {
            trace.error('no base config', { room: source.room.name });
            trace.end();
            return terminate();
        }
        if (!this.creepPosition || !this.linkPosition || !this.openPositions) {
            trace.info('creep or link position not set');
            this.populatePositions(trace, kernel, base, source);
        }
        this.threadProduceEvents(trace, kernel, base, source);
        this.threadUpdateStructures(trace, kernel, base, source);
        this.threadUpdateDropoff(trace, kernel, base, source);
        // If green, then build stuff
        if (base.alertLevel === AlertLevel.GREEN) {
            this.threadBuildContainer(trace, kernel, base, source);
            this.threadBuildLink(trace, kernel, base, source);
        }
        this.threadRequestMiners(trace, kernel, base, source);
        this.threadRequestHauling(trace, kernel, base, source);
        trace.getMetricsCollector().gauge('source_energy_remaining', source.energy, { source: this.id, base: base.id, room: source.room.name });
        trace.end();
        return sleeping(RUN_TTL$4);
    }
    produceEvents(trace, kernel, base, source) {
        const creepPosition = this.creepPosition;
        if (!creepPosition) {
            trace.error('no creep position', { room: source.room.name });
            return;
        }
        const data = {
            id: source.id,
            position: creepPosition,
        };
        kernel.getBroker().getStream(getLogisticsTopic(base.id)).
            publish(new Event(this.id, Game.time, LogisticsEventType.RequestRoad, data));
        const hudLine = {
            key: `${this.id}`,
            room: source.room.name,
            text: `source (${source.id}) - ` +
                `container: ${this.creepPosition} (${this.containerId}), ` +
                `link: ${this.linkPosition} (${this.linkId})`,
            time: Game.time,
            order: 4,
        };
        kernel.getBroker().getStream(getLinesStream()).publish(new Event(this.id, Game.time, HudEventSet, hudLine));
    }
    populatePositions(trace, kernel, base, source) {
        const memory = this.getMemory(trace);
        // Check memory for creep position
        const creepPosition = memory.creepPosition;
        if (creepPosition) {
            trace.info('creep position in memory', { room: source.room.name });
            this.creepPosition = new RoomPosition(creepPosition.x, creepPosition.y, creepPosition.roomName);
        }
        const linkPosition = memory.linkPosition;
        if (linkPosition) {
            trace.info('link position in memory', { room: source.room.name });
            this.linkPosition = new RoomPosition(linkPosition.x, linkPosition.y, linkPosition.roomName);
        }
        // Determine number of open spaces around the source
        let openPositions = getNearbyPositions(source.pos, 1);
        const terrain = source.room.getTerrain();
        openPositions = openPositions.filter(pos => {
            // If wall, blocked
            if (terrain.get(pos.x, pos.y) === TERRAIN_MASK_WALL) {
                return false;
            }
            const blocker = source.room.lookForAt(LOOK_STRUCTURES, pos).find((s) => {
                return OBSTACLE_OBJECT_TYPES.indexOf(s.structureType) !== -1;
            });
            if (blocker) {
                return false;
            }
            return true;
        });
        this.openPositions = openPositions;
        if (this.creepPosition && this.linkPosition && this.openPositions) {
            trace.info('creep and link positions in memory', { room: source.room.name, openPositions: this.openPositions });
            return;
        }
        const basePos = new RoomPosition(base.origin.x, base.origin.y - 1, base.origin.roomName);
        const [pathResult, details] = getPath(kernel, source.pos, basePos, roadPolicy, trace);
        if (!pathResult || !pathResult.path.length) {
            trace.error('path not found', { basePos: basePos, source: source.pos, details });
            return;
        }
        trace.info('creep position set', { creepPosition: this.creepPosition });
        this.creepPosition = pathResult.path[0];
        const availableLinkPos = getNearbyPositions(this.creepPosition, 1);
        trace.info('available link positions', { availableLinkPos });
        const filtered = availableLinkPos.filter((pos) => {
            // Remove creep position
            if (this.creepPosition.isEqualTo(pos)) {
                return false;
            }
            // Don't use the next pos in the road
            if (pathResult.path[1].isEqualTo(pos)) {
                return false;
            }
            // Dont use any positions with structures, construction sites, or walls
            const blocker = pos.look().find((result) => {
                if (result.type === 'structure' || result.type === 'constructionSite') {
                    return true;
                }
                if (result.type === 'terrain' && result.terrain === 'wall') {
                    return true;
                }
                return false;
            });
            if (blocker) {
                return false;
            }
            return true;
        });
        if (filtered.length === 0) {
            trace.error('no available link position', { creepPosition: this.creepPosition, filtered, availableLinkPos });
        }
        else {
            this.linkPosition = filtered[0];
        }
        trace.warn('creep and link position was not set: setting', {
            sourceId: this.id,
            creepPosition: this.creepPosition,
            linkPosition: this.linkPosition,
            openPositions: this.openPositions,
        });
        // Update memory
        memory.creepPosition = this.creepPosition;
        memory.linkPosition = this.linkPosition;
        this.setMemory(memory, false);
    }
    updateStructures(trace, _kernel, _base, _source) {
        if (!this.creepPosition) {
            trace.error('creep position not set', { creepPosition: this.creepPosition });
            return;
        }
        // Pick container
        const container = this.creepPosition.lookFor(LOOK_STRUCTURES).find((s) => {
            return s.structureType === STRUCTURE_CONTAINER;
        });
        this.containerId = container === null || container === void 0 ? void 0 : container.id;
        // Pick link
        const link = this.linkPosition.lookFor(LOOK_STRUCTURES).find((s) => {
            return s.structureType === STRUCTURE_LINK;
        });
        this.linkId = link === null || link === void 0 ? void 0 : link.id;
    }
    updateDropoff(_trace, _kernel, base, _source) {
        var _a;
        this.dropoffId = (_a = getStructureForResource(base, RESOURCE_ENERGY)) === null || _a === void 0 ? void 0 : _a.id;
    }
    requestMiners(trace, kernel, base, source) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q, _r, _s, _t, _u, _v;
        if (!source.room) {
            trace.error('source room not visible', { source: source });
            return;
        }
        const room = source.room;
        const username = kernel.getPlanner().getUsername();
        if (((_b = (_a = room === null || room === void 0 ? void 0 : room.controller) === null || _a === void 0 ? void 0 : _a.owner) === null || _b === void 0 ? void 0 : _b.username) && ((_d = (_c = room === null || room === void 0 ? void 0 : room.controller) === null || _c === void 0 ? void 0 : _c.owner) === null || _d === void 0 ? void 0 : _d.username) !== username) {
            trace.info('room owned by someone else', {
                roomId: room === null || room === void 0 ? void 0 : room.name,
                owner: (_f = (_e = room === null || room === void 0 ? void 0 : room.controller) === null || _e === void 0 ? void 0 : _e.owner) === null || _f === void 0 ? void 0 : _f.username
            });
            return;
        }
        if (((_h = (_g = room === null || room === void 0 ? void 0 : room.controller) === null || _g === void 0 ? void 0 : _g.reservation) === null || _h === void 0 ? void 0 : _h.username) && ((_k = (_j = room === null || room === void 0 ? void 0 : room.controller) === null || _j === void 0 ? void 0 : _j.reservation) === null || _k === void 0 ? void 0 : _k.username) !== username) {
            trace.info('room reserved by someone else', {
                roomId: room === null || room === void 0 ? void 0 : room.name,
                username: (_m = (_l = room === null || room === void 0 ? void 0 : room.controller) === null || _l === void 0 ? void 0 : _l.reservation) === null || _m === void 0 ? void 0 : _m.username
            });
            return;
        }
        let desiredParts = 6;
        if ((((_p = (_o = room === null || room === void 0 ? void 0 : room.controller) === null || _o === void 0 ? void 0 : _o.owner) === null || _p === void 0 ? void 0 : _p.username) && ((_r = (_q = room === null || room === void 0 ? void 0 : room.controller) === null || _q === void 0 ? void 0 : _q.owner) === null || _r === void 0 ? void 0 : _r.username) !== username) &&
            (((_t = (_s = room === null || room === void 0 ? void 0 : room.controller) === null || _s === void 0 ? void 0 : _s.reservation) === null || _t === void 0 ? void 0 : _t.username) && ((_v = (_u = room === null || room === void 0 ? void 0 : room.controller) === null || _u === void 0 ? void 0 : _u.reservation) === null || _v === void 0 ? void 0 : _v.username) !== username)) {
            desiredParts = 3; // Sources have half as much energy
        }
        const miners = kernel.getCreepsManager().getCreepsByBaseAndRole(base.id, WORKER_MINER).
            filter((creep) => {
            return creep.memory[MEMORY_SOURCE] === this.id;
        });
        const numMiners = miners.length;
        const workPartCount = miners.reduce((sum, creep) => {
            return sum + creep.getActiveBodyparts(WORK);
        }, 0);
        trace.info('num miners', {
            numMiners,
            workPartCount,
            desiredParts,
            numPositions: this.openPositions.length
        });
        if (numMiners >= this.openPositions.length) {
            trace.info('no more open positions', { numMiners, numPositions: this.openPositions.length });
            return;
        }
        if (workPartCount >= desiredParts) {
            trace.info('enough miners', { workPartCount });
            return;
        }
        // We do not have enough miners and there is room
        let positionStr = [this.creepPosition.x, this.creepPosition.y, this.creepPosition.roomName].join(',');
        if (numMiners > 0) {
            const positions = this.openPositions.filter((pos) => {
                // If miner is already assigned to this position, don't use it
                const str = [pos.x, pos.y, pos.roomName].join(',');
                if (miners.find(miner => miner.memory[MEMORY_SOURCE_POSITION] === str)) {
                    return false;
                }
                return true;
            });
            if (!positions.length) {
                trace.error('no available positions', { numMiners, positions, openPositions: this.openPositions });
                return;
            }
            positionStr = [positions[0].x, positions[0].y, positions[0].roomName].join(',');
        }
        let priority = PRIORITY_MINER_PRIMARY;
        // if remote, use lower priority
        if (room.name !== base.primary) {
            priority = PRIORITY_MINER_REMOTE;
        }
        // Reduce priority as a source has more miners
        // Early RCL, there can be multiple miners per source we need to ensure
        // that workers/haulers are also spawned.
        priority -= numMiners * 3;
        const role = WORKER_MINER;
        const memory = {
            [MEMORY_SOURCE]: this.id,
            [MEMORY_SOURCE_CONTAINER]: this.containerId,
            [MEMORY_SOURCE_POSITION]: positionStr,
            [MEMORY_ASSIGN_ROOM]: room.name,
            [MEMORY_BASE]: base.id,
        };
        trace.info('requesting miner', { sourceId: this.id, priority, memory });
        const request = createSpawnRequest(priority, RUN_TTL$4 + Game.time, role, memory, null, 0);
        kernel.getTopics().addRequestV2(getBaseSpawnTopic(base.id), request);
    }
    requestHauling(trace, kernel, base, source) {
        var _a;
        const container = Game.getObjectById(this.containerId);
        if (!container) {
            trace.info('no container');
            return;
        }
        const haulers = kernel.getCreepsManager().getCreepsByBaseAndRole(base.id, WORKER_HAULER);
        const workers = kernel.getCreepsManager().getCreepsByBaseAndRole(base.id, ROLE_WORKER);
        const creeps = haulers.concat(workers);
        const avgHaulerCapacity = _.sum(creeps, (creep) => {
            return creep.store.getCapacity(RESOURCE_ENERGY);
        });
        const creepsWithTask = creeps.filter((creep) => {
            const task = creep.memory[MEMORY_TASK_TYPE];
            const pickup = creep.memory[MEMORY_HAUL_PICKUP];
            return task === TASK_HAUL && pickup === this.containerId;
        });
        const creepWithTaskCapacity = creepsWithTask.reduce((total, hauler) => {
            return total += hauler.store.getFreeCapacity();
        }, 0);
        const averageLoad = avgHaulerCapacity / creeps.length;
        let loadSize = 300;
        if (averageLoad >= 50) {
            loadSize = averageLoad;
        }
        const storeUsedCapacity = container.store.getUsedCapacity();
        const untaskedUsedCapacity = storeUsedCapacity - creepWithTaskCapacity;
        const loadsToHaul = Math.floor(untaskedUsedCapacity / loadSize);
        let priority = HAUL_CONTAINER;
        // prioritize hauling primary room
        if (base.primary === ((_a = source.room) === null || _a === void 0 ? void 0 : _a.name)) {
            priority += HAUL_BASE_ROOM;
        }
        trace.info('requesting hauling', {
            sourceId: this.id,
            priority,
            loadsToHaul,
            averageLoad,
            loadSize,
            storeUsedCapacity,
            untaskedUsedCapacity,
            creepWithTaskCapacity,
            creepsWithTask: creepsWithTask.length
        });
        for (let i = 0; i < loadsToHaul; i++) {
            // Reduce priority for each load after first
            const loadPriority = priority - LOAD_FACTOR * i;
            const details = {
                [TASK_ID]: `sch-${this.id}-${Game.time}`,
                [MEMORY_TASK_TYPE]: TASK_HAUL,
                [MEMORY_HAUL_PICKUP]: this.containerId,
                [MEMORY_HAUL_DROPOFF]: this.dropoffId || undefined,
                [MEMORY_HAUL_RESOURCE]: RESOURCE_ENERGY,
            };
            trace.info('requesting hauling', { sourceId: this.id, i, loadPriority, details });
            kernel.getTopics().addRequest(getBaseHaulerTopic(base.id), loadPriority, details, RUN_TTL$4 + Game.time);
        }
    }
    buildContainer(trace, _kernel, _base, source) {
        if (source.energy) {
            trace.info('only build container if exhausting source', { id: this.id });
            return;
        }
        if (!this.creepPosition) {
            trace.error('no creep position', { id: this.id });
            return;
        }
        if (this.containerId) {
            const container = Game.getObjectById(this.containerId);
            if (container) {
                trace.info('container already built', { container });
                return;
            }
        }
        const sites = this.creepPosition.lookFor(LOOK_CONSTRUCTION_SITES).filter((site) => {
            return site.structureType === STRUCTURE_CONTAINER;
        });
        if (sites.length) {
            trace.info('container site found', { sites });
            return;
        }
        const result = this.creepPosition.createConstructionSite(STRUCTURE_CONTAINER);
        if (result !== OK) {
            trace.error('failed to build container', { result });
            return;
        }
        trace.info('container created', { id: this.id });
    }
    buildLink(trace, kernel, base, source) {
        var _a, _b;
        if (!this.linkPosition) {
            trace.error('no link position', { room: source.room.name, id: this.id });
            return;
        }
        if (this.linkId) {
            const link = Game.getObjectById(this.linkId);
            if (link) {
                trace.info('link found', { link });
                return;
            }
        }
        if (!((_a = source.room) === null || _a === void 0 ? void 0 : _a.controller)) {
            trace.error('no visibility or controller', { baseId: base.id, roomName: source.room.name, id: this.id });
            return;
        }
        const roomLevel = ((_b = source.room.controller) === null || _b === void 0 ? void 0 : _b.level) || 0;
        if (roomLevel < 6) {
            trace.info('room level too low', { roomLevel });
            return;
        }
        const linkSites = this.linkPosition.lookFor(LOOK_CONSTRUCTION_SITES).find((s) => {
            return s.structureType === STRUCTURE_LINK;
        });
        if (linkSites) {
            trace.info('link sites found', { linkSites });
            return;
        }
        const linksInRoom = source.room.find(FIND_STRUCTURES, {
            filter: (s) => {
                return s.structureType === STRUCTURE_LINK;
            },
        });
        const linkSitesInRoom = source.room.find(FIND_CONSTRUCTION_SITES, {
            filter: (s) => {
                return s.structureType === STRUCTURE_LINK;
            },
        });
        const maxLinks = CONTROLLER_STRUCTURES['link'][roomLevel];
        if (maxLinks <= linksInRoom.length + linkSitesInRoom.length) {
            trace.info('too many links', { maxLinks, linksInRoom });
            return;
        }
        const result = this.linkPosition.createConstructionSite(STRUCTURE_LINK);
        if (result !== OK) {
            trace.error('failed to build link', { result });
            return;
        }
        trace.notice('link created', { id: this.id });
    }
}

const MIN_RESERVATION_TICKS = 4000;
const NO_VISION_TTL$1 = 20;
const MIN_TTL$1 = 10;
const REQUEST_RESERVER_TTL = 25;
const UPDATE_PROCESSES_TTL$1 = 50;
class RoomRunnable {
    constructor(id, scheduler) {
        this.id = id;
        this.scheduler = scheduler;
        // Threads
        this.threadUpdateProcessSpawning = threadBaseRoom('spawn_room_processes_thread', UPDATE_PROCESSES_TTL$1)(this.handleProcessSpawning.bind(this));
        this.threadRequestReserver = threadBase('request_reserver_thread', REQUEST_RESERVER_TTL)(this.requestReserver.bind(this));
    }
    run(kernel, trace) {
        trace = trace.begin('room_run');
        trace.info('room run', {
            id: this.id,
        });
        const base = kernel.getPlanner().getBaseByRoom(this.id);
        if (!base) {
            trace.error('no base config, terminating', { id: this.id });
            trace.end();
            return terminate();
        }
        // Request reservers if needed
        if (this.id != base.primary) {
            this.threadRequestReserver(trace, kernel, base, this.id);
        }
        const room = Game.rooms[this.id];
        if (!room) {
            trace.notice('cannot find room in game', { id: this.id });
            return sleeping(NO_VISION_TTL$1);
        }
        this.threadUpdateProcessSpawning(trace, kernel, base, room);
        trace.end();
        return sleeping(MIN_TTL$1);
    }
    handleProcessSpawning(trace, kernel, base, room) {
        var _a, _b, _c, _d;
        if (((_a = room.controller) === null || _a === void 0 ? void 0 : _a.my) || !((_c = (_b = room.controller) === null || _b === void 0 ? void 0 : _b.owner) === null || _c === void 0 ? void 0 : _c.username)) {
            // Sources
            room.find(FIND_SOURCES).forEach((source) => {
                const sourceId = `${source.id}`;
                if (!this.scheduler.hasProcess(sourceId)) {
                    trace.info('found source without process, starting', { sourceId: source.id, room: this.id });
                    this.scheduler.registerProcess(new Process(sourceId, 'sources', Priorities.RESOURCES, new SourceRunnable(source)));
                }
            });
            if (base.primary === room.name && ((_d = room.controller) === null || _d === void 0 ? void 0 : _d.level) >= 6) {
                // Mineral
                room.find(FIND_MINERALS).forEach((mineral) => {
                    const mineralId = `${mineral.id}`;
                    if (!this.scheduler.hasProcess(mineralId)) {
                        this.scheduler.registerProcess(new Process(mineralId, 'mineral', Priorities.RESOURCES, new MineralRunnable(mineral)));
                    }
                });
            }
        }
    }
    requestReserver(trace, kernel, base, roomName) {
        var _a, _b, _c;
        const reserver = _.find(Game.creeps, (creep) => {
            const role = creep.memory[MEMORY_ROLE];
            return (role === WORKER_RESERVER) &&
                creep.memory[MEMORY_ASSIGN_ROOM] === roomName && creepIsFresh(creep);
        });
        if (reserver) {
            return;
        }
        // If base is under threat, don't worry about remotes
        if (base.alertLevel !== AlertLevel.GREEN) {
            trace.notice('not requesting reserver, base alert level is not green', {
                roomName: roomName,
                baseId: base.id,
                alertLevel: base.alertLevel,
            });
            return;
        }
        const roomEntry = kernel.getScribe().getRoomById(roomName);
        if (!roomEntry) {
            trace.error('cannot find room entry', { roomName: roomName });
            return;
        }
        // If owned by me we don't need reserver
        if (((_a = roomEntry.controller) === null || _a === void 0 ? void 0 : _a.owner) === kernel.getPlanner().getUsername()) {
            if (((_b = roomEntry.controller) === null || _b === void 0 ? void 0 : _b.level) > 0) {
                trace.notice('not requesting reserver, room is owned by me', {
                    roomName: roomName,
                    baseId: base.id,
                });
                return;
            }
            // Reservation, check min reservation ticks
            if (((_c = roomEntry.controller) === null || _c === void 0 ? void 0 : _c.downgrade) > MIN_RESERVATION_TICKS) {
                trace.notice('not requesting reserver, room is owned by me and reservation is long enough', {
                    roomName: roomName,
                    baseId: base.id,
                });
                return;
            }
        }
        trace.notice('sending reserve request to base', { room: roomName, base: base.id });
        const priority = PRIORITY_RESERVER;
        const ttl = REQUEST_RESERVER_TTL + Game.time;
        const role = WORKER_RESERVER;
        const memory = {
            [MEMORY_ASSIGN_ROOM]: this.id,
            [MEMORY_BASE]: base.id,
        };
        let topic = getShardSpawnTopic();
        if (baseEnergyStorageCapacity(base) >= 650) {
            topic = getBaseSpawnTopic(base.id);
        }
        const request = createSpawnRequest(priority, ttl, role, memory, null, 0);
        kernel.getTopics().addRequestV2(topic, request);
    }
}

const PRICES = new Map([
    [RESOURCE_ENERGY, { max: 10, min: 1.0 }],
    [RESOURCE_POWER, { max: 100, min: 10 }],
    [RESOURCE_OPS, { max: 100, min: 10 }],
    [RESOURCE_HYDROGEN, { max: 5, min: 0.6 }],
    [RESOURCE_OXYGEN, { max: 5, min: 0.3 }],
    [RESOURCE_HYDROXIDE, { max: 10, min: 0.5 }],
    [RESOURCE_LEMERGIUM, { max: 10, min: 0.2 }],
    [RESOURCE_UTRIUM, { max: 10, min: 0.4 }],
    [RESOURCE_UTRIUM_LEMERGITE, { max: 10, min: 1.0 }],
    [RESOURCE_KEANIUM, { max: 10, min: 0.2 }],
    [RESOURCE_ZYNTHIUM, { max: 10, min: 0.1 }],
    [RESOURCE_ZYNTHIUM_KEANITE, { max: 10, min: 0.5 }],
    [RESOURCE_GHODIUM, { max: 10, min: 2.0 }],
    [RESOURCE_CATALYST, { max: 20, min: 0.5 }],
    [RESOURCE_GHODIUM_HYDRIDE, { max: 10, min: 0.5 }],
    [RESOURCE_GHODIUM_OXIDE, { max: 10, min: 0.5 }],
    [RESOURCE_GHODIUM_ACID, { max: 10, min: 0.5 }],
    [RESOURCE_GHODIUM_ALKALIDE, { max: 10, min: 0.5 }],
    [RESOURCE_CATALYZED_GHODIUM_ACID, { max: 30.0, min: 17 }],
    [RESOURCE_CATALYZED_GHODIUM_ALKALIDE, { max: 30.0, min: 12.0 }],
    [RESOURCE_KEANIUM_HYDRIDE, { max: 10, min: 0.5 }],
    [RESOURCE_KEANIUM_OXIDE, { max: 10, min: 0.5 }],
    [RESOURCE_KEANIUM_ACID, { max: 10, min: 0.5 }],
    [RESOURCE_KEANIUM_ALKALIDE, { max: 10, min: 0.5 }],
    [RESOURCE_CATALYZED_KEANIUM_ACID, { max: 30, min: 0.5 }],
    [RESOURCE_CATALYZED_KEANIUM_ALKALIDE, { max: 30, min: 10.0 }],
    [RESOURCE_LEMERGIUM_HYDRIDE, { max: 10, min: 0.5 }],
    [RESOURCE_LEMERGIUM_OXIDE, { max: 10, min: 0.5 }],
    [RESOURCE_LEMERGIUM_ACID, { max: 10, min: 0.5 }],
    [RESOURCE_LEMERGIUM_ALKALIDE, { max: 10, min: 0.5 }],
    [RESOURCE_CATALYZED_LEMERGIUM_ACID, { max: 30, min: 0.5 }],
    [RESOURCE_CATALYZED_LEMERGIUM_ALKALIDE, { max: 30, min: 8.0 }],
    [RESOURCE_UTRIUM_HYDRIDE, { max: 10, min: 0.5 }],
    [RESOURCE_UTRIUM_OXIDE, { max: 10, min: 0.5 }],
    [RESOURCE_UTRIUM_ACID, { max: 10, min: 0.5 }],
    [RESOURCE_UTRIUM_ALKALIDE, { max: 10, min: 0.5 }],
    [RESOURCE_CATALYZED_UTRIUM_ACID, { max: 30, min: 2.000 }],
    [RESOURCE_CATALYZED_UTRIUM_ALKALIDE, { max: 30, min: 5 }],
    [RESOURCE_ZYNTHIUM_HYDRIDE, { max: 10, min: 0.5 }],
    [RESOURCE_ZYNTHIUM_OXIDE, { max: 10, min: 0.5 }],
    [RESOURCE_ZYNTHIUM_ACID, { max: 10, min: 0.5 }],
    [RESOURCE_ZYNTHIUM_ALKALIDE, { max: 10, min: 0.5 }],
    [RESOURCE_CATALYZED_ZYNTHIUM_ACID, { max: 30, min: 0.5 }],
    [RESOURCE_CATALYZED_ZYNTHIUM_ALKALIDE, { max: 30, min: 5 }],
    [RESOURCE_SILICON, { max: 100, min: 10 }],
    [RESOURCE_METAL, { max: 100, min: 10 }],
    [RESOURCE_BIOMASS, { max: 100, min: 10 }],
    [RESOURCE_MIST, { max: 100, min: 10 }],
    [RESOURCE_UTRIUM_BAR, { max: 100, min: 10 }],
    [RESOURCE_LEMERGIUM_BAR, { max: 100, min: 10 }],
    [RESOURCE_ZYNTHIUM_BAR, { max: 100, min: 10 }],
    [RESOURCE_KEANIUM_BAR, { max: 100, min: 10 }],
    [RESOURCE_GHODIUM_MELT, { max: 100, min: 10 }],
    [RESOURCE_OXIDANT, { max: 100, min: 10 }],
    [RESOURCE_REDUCTANT, { max: 100, min: 10 }],
    [RESOURCE_PURIFIER, { max: 100, min: 10 }],
    [RESOURCE_BATTERY, { max: 100, min: 10 }],
    [RESOURCE_COMPOSITE, { max: 100, min: 10 }],
    [RESOURCE_CRYSTAL, { max: 100, min: 10 }],
    [RESOURCE_LIQUID, { max: 100, min: 10 }],
    [RESOURCE_WIRE, { max: 100, min: 10 }],
    [RESOURCE_SWITCH, { max: 100, min: 10 }],
    [RESOURCE_TRANSISTOR, { max: 100, min: 10 }],
    [RESOURCE_MICROCHIP, { max: 100, min: 10 }],
    [RESOURCE_CIRCUIT, { max: 100, min: 10 }],
    [RESOURCE_DEVICE, { max: 100, min: 10 }],
    [RESOURCE_CELL, { max: 100, min: 10 }],
    [RESOURCE_PHLEGM, { max: 100, min: 10 }],
    [RESOURCE_TISSUE, { max: 100, min: 10 }],
    [RESOURCE_MUSCLE, { max: 100, min: 10 }],
    [RESOURCE_ORGANOID, { max: 100, min: 10 }],
    [RESOURCE_ORGANISM, { max: 100, min: 10 }],
    [RESOURCE_ALLOY, { max: 100, min: 10 }],
    [RESOURCE_TUBE, { max: 100, min: 10 }],
    [RESOURCE_FIXTURES, { max: 100, min: 10 }],
    [RESOURCE_FRAME, { max: 100, min: 10 }],
    [RESOURCE_HYDRAULICS, { max: 100, min: 10 }],
    [RESOURCE_MACHINE, { max: 100, min: 10 }],
    [RESOURCE_CONDENSATE, { max: 100, min: 10 }],
    [RESOURCE_CONCENTRATE, { max: 100, min: 10 }],
    [RESOURCE_EXTRACT, { max: 100, min: 10 }],
    [RESOURCE_SPIRIT, { max: 100, min: 10 }],
    [RESOURCE_EMANATION, { max: 100, min: 10 }],
    [RESOURCE_ESSENCE, { max: 100, min: 10 }],
]);

const MIDPOINT = 10000;
const SLOPE = 1 / 10000;
class SigmoidPricing {
    constructor(prices) {
        this.prices = prices;
    }
    getMarketPrice(orderType, resource) {
        let orders = Game.market.getAllOrders({ type: orderType, resourceType: resource });
        const myOrderIds = Object.keys(Game.market.orders);
        orders = _.filter(orders, (order) => {
            return myOrderIds.indexOf(order.id) === -1;
        });
        if (!orders.length) {
            return orderType === ORDER_BUY ? 0.001 : null;
        }
        let price = null;
        if (orderType === ORDER_BUY) {
            price = _.max(orders.map((order) => order.price));
        }
        else {
            price = _.min(orders.map((order) => order.price));
        }
        return price;
    }
    getPrice(orderType, resource, amount) {
        const range = this.prices.get(resource);
        if (!range) {
            throw new Error(`invalid resource: ${resource}`);
        }
        let price = sigmoid(amount, range.max);
        const marketPrice = this.getMarketPrice(orderType, resource);
        if (marketPrice !== null) {
            if (orderType === ORDER_BUY) {
                price = _.min([marketPrice, price]);
            }
            else {
                price = _.max([marketPrice, price]);
            }
        }
        if (orderType === ORDER_BUY && price > range.max) {
            price = range.max;
        }
        if (orderType === ORDER_SELL && price < range.min) {
            price = range.min;
        }
        return price;
    }
}
function sigmoid(x, max, mid = MIDPOINT, slope = SLOPE) {
    return max / (1 + Math.pow(Math.E, -slope * (-x + mid)));
}

const TASK_PHASE_HAUL_RESOURCE = 'phase_transfer_resource';
const TASK_PHASE_TRANSACT = 'phase_transact';
const TASK_PHASE_TRANSFER = 'phase_transfer';
const TASK_TTL = 50;
const MAX_TERMINAL_ENERGY = 1000;
const PROCESS_TASK_TTL = 10;
const REQUEST_RETURN_ENERGY_TTL = 20;
const ORDER_MGMT_TTL = 1000;
const HAUL_OLD_SELL_ORDER_TTL = 20;
const UPDATE_ENERGY_VALUE_TTL = 2500;
function getBaseTerminalTopic(base) {
    return `base_${base.id}_${TOPIC_TERMINAL_TASK}`;
}
class TerminalRunnable {
    constructor(baseId, terminal) {
        this.baseId = baseId;
        this.terminalId = terminal.id;
        this.prevTime = Game.time;
        this.processTaskTTL = 0;
        this.returnEnergyTTL = REQUEST_RETURN_ENERGY_TTL;
        this.updateOrdersTTL = ORDER_MGMT_TTL;
        this.pricer = new SigmoidPricing(PRICES);
        this.threadHaulOldSellOrders = threadBase('haul_old_sell_orders_thread', HAUL_OLD_SELL_ORDER_TTL)(this.haulOldSellOrders.bind(this));
        this.threadUpdateEnergyValue = threadKernel('update_energy_thread', UPDATE_ENERGY_VALUE_TTL)(this.updateEnergyValue.bind(this));
    }
    run(kernel, trace) {
        trace = trace.begin('terminal_run');
        const ticks = Game.time - this.prevTime;
        this.prevTime = Game.time;
        this.processTaskTTL -= ticks;
        this.updateOrdersTTL -= ticks;
        this.returnEnergyTTL -= ticks;
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.error('no base config for room', { baseId: this.baseId });
            return terminate();
        }
        const terminal = Game.getObjectById(this.terminalId);
        // If terminal no longer exists, terminate
        if (!terminal) {
            trace.info('terminal not found - terminating', {});
            trace.end();
            return terminate();
        }
        if (!terminal.isActive()) {
            trace.end();
            return sleeping(100);
        }
        this.threadHaulOldSellOrders(trace, kernel, base, terminal);
        this.threadUpdateEnergyValue(trace, kernel);
        const task = base.terminalTask || null;
        if (!task) {
            this.processTaskTTL = -1;
            const task = kernel.getTopics().getNextRequest(getBaseTerminalTopic(base));
            if (task) {
                base.terminalTask = task.details;
            }
        }
        trace.info('terminal run', {
            ticks,
            processTaskTTL: this.processTaskTTL,
            returnEnergyTTL: this.returnEnergyTTL,
            updateOrdersTTL: this.updateOrdersTTL,
            task,
        });
        if (task && this.processTaskTTL < 0) {
            this.processTaskTTL = PROCESS_TASK_TTL;
            this.processTask(kernel, base, terminal, task, ticks, trace);
        }
        else if (!task) {
            const terminalAmount = terminal.store.getUsedCapacity(RESOURCE_ENERGY);
            if (terminalAmount > MAX_TERMINAL_ENERGY && this.returnEnergyTTL < 0) {
                this.returnEnergyTTL = REQUEST_RETURN_ENERGY_TTL;
                const amountToTransfer = terminalAmount - MAX_TERMINAL_ENERGY;
                trace.info('send energy to storage', { amountToTransfer });
                this.sendEnergyToStorage(kernel, base, terminal, amountToTransfer, REQUEST_RETURN_ENERGY_TTL, trace);
            }
        }
        if (this.updateOrdersTTL < 0) {
            this.updateOrdersTTL = ORDER_MGMT_TTL;
            this.updateOrders(kernel, terminal, trace);
        }
        trace.end();
        return running();
    }
    isIdle(base) {
        return !!base.terminalTask;
    }
    setTask(base, task) {
        base.terminalTask = task;
    }
    getTask(base) {
        return base.terminalTask;
    }
    clearTask(base) {
        base.terminalTask = null;
    }
    processTask(kernel, base, terminal, task, ticks, trace) {
        const details = task.details;
        const taskType = details[TERMINAL_TASK_TYPE];
        // Maintain task TTL. We want to abort hard to perform tasks
        let ttl = details[TASK_TTL$2];
        if (ttl === undefined) {
            ttl = TASK_TTL;
        }
        if (ttl < 0) {
            this.clearTask(base);
            return;
        }
        else {
            terminal.room.memory[TERMINAL_TASK].details[TASK_TTL$2] = ttl - ticks;
        }
        trace.info('processTask', { task });
        switch (taskType) {
            case TASK_TRANSFER:
                this.transferResource(kernel, base, terminal, details, trace);
                break;
            case TASK_MARKET_ORDER: {
                // Perform market order
                const orderType = details[MEMORY_ORDER_TYPE];
                if (orderType === ORDER_SELL) {
                    this.sell(kernel, base, terminal, details, trace);
                }
                else if (orderType === ORDER_BUY) {
                    this.buy(kernel, base, terminal, details, trace);
                }
                else {
                    this.clearTask(base);
                }
                break;
            }
            default:
                this.clearTask(base);
        }
        return;
    }
    transferResource(kernel, base, terminal, task, trace) {
        const resource = task[TRANSFER_RESOURCE];
        const amount = task[TRANSFER_AMOUNT];
        const roomName = task[TRANSFER_ROOM];
        const phase = task[TASK_PHASE] || TASK_PHASE_HAUL_RESOURCE;
        trace.info('transfer resource', { resource, amount, roomName, phase });
        switch (phase) {
            case TASK_PHASE_HAUL_RESOURCE: {
                // Check if we should move to next phase
                const terminalAmount = terminal.store.getUsedCapacity(resource);
                if (terminalAmount >= amount) {
                    trace.info('terminal amount gte desired amount', { terminalAmount, amount });
                    terminal.room.memory[TERMINAL_TASK].details[TASK_PHASE] = TASK_PHASE_TRANSFER;
                    break;
                }
                const pickup = getStructureWithResource(base, resource);
                if (!pickup) {
                    if (!terminalAmount) {
                        trace.info('no pickup and no resources in terminal', {});
                        this.clearTask(base);
                        break;
                    }
                    trace.info('no pickup, but resources in terminal', { terminalAmount });
                    terminal.room.memory[TERMINAL_TASK].details[TASK_PHASE] = TASK_PHASE_TRANSFER;
                    terminal.room.memory[TERMINAL_TASK].details[TRANSFER_AMOUNT] = terminalAmount;
                    break;
                }
                trace.info('requesting resource transfer to terminal', { pickup: pickup.id, resource, amount });
                this.haulResourceToTerminal(kernel, base, terminal, pickup, resource, amount);
                break;
            }
            case TASK_PHASE_TRANSFER: {
                let haulAmount = amount;
                const energyRequired = Game.market.calcTransactionCost(amount, terminal.room.name, roomName);
                // If we are transfering energy we need energy in addition to what we want to transfer
                if (resource === RESOURCE_ENERGY) {
                    haulAmount += energyRequired;
                    trace.info('padded energy', { amount, added: energyRequired });
                }
                const energyReady = this.haulTransferEnergyToTerminal(kernel, base, terminal, resource, haulAmount, roomName, trace);
                if (!energyReady) {
                    trace.info('energy not ready', { amount, roomName, haulAmount });
                    break;
                }
                const result = terminal.send(resource, amount, roomName);
                trace.info('send resource', { resource, amount, roomName, result });
                if (result !== OK) {
                    trace.info('send resource failed', { result });
                    break;
                }
                this.clearTask(base);
                break;
            }
            default:
                this.clearTask(base);
        }
    }
    buy(kernel, base, terminal, task, trace) {
        const resource = task[MEMORY_ORDER_RESOURCE];
        const amount = task[MEMORY_ORDER_AMOUNT];
        const currentAmount = terminal.store.getUsedCapacity(resource);
        let missingAmount = amount - currentAmount;
        trace.info('buy order', {
            resource,
            amount,
            currentAmount,
            missingAmount,
        });
        // Buy in at least blocks of 1000, to avoid stupid small orders
        missingAmount = Math.max(1000, missingAmount);
        if (currentAmount >= amount) {
            trace.info('buy order satisfied');
            this.clearTask(base);
            return;
        }
        if (terminal.cooldown) {
            trace.info('buy order failed: cooling down');
            return;
        }
        const orders = Game.market.getAllOrders({ type: ORDER_SELL, resourceType: resource });
        const sellOrder = _.sortBy(orders, (order) => {
            const transferEnergy = Game.market.calcTransactionCost(missingAmount, terminal.room.name, order.roomName);
            return order.price + (transferEnergy * this.energyValue);
        })[0];
        const resources = kernel.getResourceManager().getSharedResources();
        const reserveAmount = resources.get(resource) || 0;
        const maxBuyPrice = this.pricer.getPrice(ORDER_BUY, resource, reserveAmount);
        if (!sellOrder || sellOrder.price > maxBuyPrice) {
            trace.info('sell orders too expensive: creating buy order', { resource, orderPrice: sellOrder === null || sellOrder === void 0 ? void 0 : sellOrder.price, maxBuyPrice });
            this.createBuyOrder(base, terminal, resource, amount, trace);
            this.clearTask(base);
            return;
        }
        const dealAmount = Math.min(missingAmount, sellOrder.remainingAmount);
        const energyReady = this.prepareTransferEnergy(kernel, base, terminal, sellOrder, dealAmount, trace);
        if (!energyReady) {
            trace.info('deal energy not ready', { dealAmount });
            return;
        }
        const result = Game.market.deal(sellOrder.id, dealAmount, terminal.room.name);
        trace.info('buy deal result', {
            orderId: sellOrder.id,
            dealAmount,
            price: sellOrder.price,
            destRoom: terminal.room.name,
            result,
        });
    }
    sell(kernel, base, terminal, task, trace) {
        const resource = task[MEMORY_ORDER_RESOURCE];
        const amount = task[MEMORY_ORDER_AMOUNT];
        const phase = task[TASK_PHASE] || TASK_PHASE_HAUL_RESOURCE;
        switch (phase) {
            case TASK_PHASE_HAUL_RESOURCE: {
                // Check if we should move to next phase
                const terminalAmount = terminal.store.getUsedCapacity(resource);
                if (terminalAmount >= amount) {
                    terminal.room.memory[TERMINAL_TASK].details[TASK_PHASE] = TASK_PHASE_TRANSACT;
                    break;
                }
                const pickup = getStructureWithResource(base, resource);
                if (!pickup) {
                    if (!terminalAmount) {
                        this.clearTask(base);
                        break;
                    }
                    terminal.room.memory[TERMINAL_TASK].details[TASK_PHASE] = TASK_PHASE_TRANSACT;
                    terminal.room.memory[TERMINAL_TASK].details[MEMORY_ORDER_AMOUNT] = terminalAmount;
                    break;
                }
                this.haulResourceToTerminal(kernel, base, terminal, pickup, resource, amount);
                break;
            }
            case TASK_PHASE_TRANSACT: {
                // Check if we are done selling
                if (terminal.store.getUsedCapacity(resource) === 0 || amount < 1) {
                    this.clearTask(base);
                    break;
                }
                // Get lowest buy order
                let buyOrders = Game.market.getAllOrders({ type: ORDER_BUY, resourceType: resource });
                buyOrders = buyOrders.filter((order) => {
                    return order.remainingAmount > 0;
                });
                const buyOrder = _.sortBy(buyOrders, (order) => {
                    const transferEnergy = Game.market.calcTransactionCost(amount, terminal.room.name, order.roomName);
                    return order.price + (transferEnergy * this.energyValue);
                }).reverse()[0];
                // Get desired purchase price based on current stockpile
                const resources = kernel.getResourceManager().getSharedResources();
                const currentAmount = resources.get(resource) || 0;
                const minSellPrice = this.pricer.getPrice(ORDER_SELL, resource, currentAmount);
                // If no buy orders or price is too low, create a sell order
                if (!buyOrder || buyOrder.price < minSellPrice) {
                    trace.info('no orders or sell prices too low, creating sell order');
                    this.createSellOrder(base, terminal, resource, amount, trace);
                    this.clearTask(base);
                    return;
                }
                // Make sure we have enough energy to perform the deal
                const dealAmount = _.min([amount, buyOrder.remainingAmount]);
                const energyReady = this.prepareTransferEnergy(kernel, base, terminal, buyOrder, dealAmount, trace);
                if (!energyReady) {
                    trace.info('deal energy not ready', { dealAmount });
                    return;
                }
                // If terminal is on cool down, try again in a bit
                if (terminal.cooldown) {
                    return;
                }
                // Transact the deal
                const result = Game.market.deal(buyOrder.id, dealAmount, terminal.room.name);
                if (result == OK) {
                    trace.info('sold resources', {
                        orderId: buyOrder.id,
                        dealAmount,
                        price: buyOrder.price,
                        destRoom: terminal.room.name,
                        result,
                    });
                    // deduct amount transacted from total amount we want to buy
                    terminal.room.memory[TERMINAL_TASK].details[MEMORY_ORDER_AMOUNT] -= dealAmount;
                }
                break;
            }
            default:
                trace.error('BROKEN MARKET LOGIC', phase);
                this.clearTask(base);
        }
    }
    prepareTransferEnergy(kernel, base, terminal, order, amount, trace) {
        const energyRequired = Game.market.calcTransactionCost(amount, terminal.room.name, order.roomName);
        // If we are transfering energy we need energy in addition to what we want to transfer
        let haulAmount = energyRequired;
        if (order.resourceType === RESOURCE_ENERGY) {
            haulAmount = amount + energyRequired;
            trace.info('padded haul amount', { haulAmount, added: energyRequired });
        }
        return this.haulTransferEnergyToTerminal(kernel, base, terminal, order.resourceType, haulAmount, order.roomName, trace);
    }
    createBuyOrder(base, terminal, resource, amount, trace) {
        // Check if we already have a sell order for the room and resource
        const duplicateBuyOrders = Object.values(Game.market.orders).filter((order) => {
            return order.type === ORDER_BUY && order.resourceType === resource &&
                order.roomName === terminal.room.name && order.remainingAmount > 0;
        });
        if (duplicateBuyOrders.length) {
            trace.info('duplicate buy orders found', { duplicateBuyOrders });
            this.clearTask(base);
            return;
        }
        const buyPrice = this.pricer.getPrice(ORDER_BUY, resource, amount);
        // Create buy order
        const buyOrder = {
            type: ORDER_BUY,
            resourceType: resource,
            price: buyPrice,
            totalAmount: amount,
            roomName: terminal.room.name,
        };
        const result = Game.market.createOrder(buyOrder);
        trace.info('create buy order result', { result, buyOrder });
    }
    createSellOrder(base, terminal, resource, amount, trace) {
        // Check if we already have a sell order for the room and resource
        const duplicateSellOrders = Object.values(Game.market.orders).filter((order) => {
            return order.type === ORDER_SELL && order.resourceType === resource &&
                order.roomName === terminal.room.name && order.remainingAmount > 0;
        });
        if (duplicateSellOrders.length) {
            trace.info('duplicate sell orders found', { duplicateSellOrders });
            this.clearTask(base);
            return;
        }
        const sellPrice = this.pricer.getPrice(ORDER_SELL, resource, amount);
        // Create buy order
        const order = {
            type: ORDER_SELL,
            resourceType: resource,
            price: sellPrice,
            totalAmount: amount,
            roomName: terminal.room.name,
        };
        const result = Game.market.createOrder(order);
        trace.info('create sell order result', { result, order });
    }
    updateOrders(kernel, terminal, trace) {
        trace.info('updating prices on buy/sell orders');
        Object.values(Game.market.orders).filter((order) => {
            return order.roomName === terminal.room.name;
        }).forEach((order) => {
            if (order.remainingAmount === 0) {
                const result = Game.market.cancelOrder(order.id);
                trace.info('order is complete; cancelled', { result, orderId: order.id });
                return;
            }
            const resources = kernel.getResourceManager().getSharedResources();
            const currentAmount = resources.get(order.resourceType) || 0;
            const price = this.pricer.getPrice(order.type, order.resourceType, currentAmount);
            if (order.price !== price) {
                Game.market.changeOrderPrice(order.id, price);
                trace.info('updating order price', {
                    orderId: order.id,
                    previousPrice: order.price, newPrice: price, resource: order.resourceType,
                });
            }
        });
    }
    haulOldSellOrders(trace, kernel, base, terminal) {
        trace.info('checking for incomplete sell orders');
        const ordersWithMissingResources = Object.values(Game.market.orders).filter((order) => {
            return order.type === ORDER_SELL && order.roomName === terminal.room.name &&
                order.remainingAmount - order.amount > 0;
        });
        trace.info('incomplete orders', { ordersWithMissingResources });
        ordersWithMissingResources.forEach((order) => {
            const resourceType = order.resourceType;
            const pickup = getStructureWithResource(base, resourceType);
            if (!pickup) {
                const result = Game.market.cancelOrder(order.id);
                trace.info('cancelling order: resource not available', { result, order });
                return;
            }
            const missingAmount = order.remainingAmount - order.amount;
            if (pickup.store.getUsedCapacity(order.resourceType) < missingAmount) {
                const result = Game.market.cancelOrder(order.id);
                trace.info('cancelling order: not enough of the resource available', { result, order, missingAmount });
                return;
            }
            trace.info('requesting hauling', { resource: order.resourceType, missingAmount });
            this.sendHaulRequest(kernel, base, terminal, pickup, order.resourceType, missingAmount, HAUL_TERMINAL);
        });
    }
    haulResourceToTerminal(kernel, base, terminal, pickup, resource, amount) {
        const baseCreeps = kernel.getCreepsManager().getCreepsByBase(this.baseId);
        const numHaulers = baseCreeps.filter((creep) => {
            return creep.memory[MEMORY_TASK_TYPE] === TASK_HAUL &&
                creep.memory[MEMORY_HAUL_RESOURCE] === resource &&
                creep.memory[MEMORY_HAUL_DROPOFF] === terminal.id;
        }).length;
        // If we already have a hauler assigned, don't assign more
        if (numHaulers) {
            return;
        }
        this.sendHaulRequest(kernel, base, terminal, pickup, resource, amount, HAUL_TERMINAL);
    }
    haulTransferEnergyToTerminal(kernel, base, terminal, resource, amount, destinationRoom, trace) {
        const currentEnergy = terminal.store.getUsedCapacity(RESOURCE_ENERGY);
        if (currentEnergy < amount) {
            // If we are low on energy don't take any from reserve
            if (getStoredResourceAmount(base, RESOURCE_ENERGY) > 20000) {
                const pickup = getStructureWithResource(base, RESOURCE_ENERGY);
                if (!pickup) {
                    return false;
                }
                const requestAmount = amount - currentEnergy;
                trace.info('requesting', { resource, amount: requestAmount });
                this.sendHaulRequest(kernel, base, terminal, pickup, RESOURCE_ENERGY, requestAmount, HAUL_TERMINAL);
                return false;
            }
            return false;
        }
        return true;
    }
    sendHaulRequest(kernel, base, terminal, pickup, resource, amount, priority) {
        amount = _.min([amount, pickup.store.getUsedCapacity(resource)]);
        const details = {
            [TASK_ID]: `mrl-${terminal.id}-${Game.time}`,
            [MEMORY_TASK_TYPE]: TASK_HAUL,
            [MEMORY_HAUL_PICKUP]: pickup.id,
            [MEMORY_HAUL_RESOURCE]: resource,
            [MEMORY_HAUL_AMOUNT]: amount,
            [MEMORY_HAUL_DROPOFF]: terminal.id,
        };
        kernel.getTopics().addRequest(getBaseDistributorTopic(base.id), priority, details, PROCESS_TASK_TTL + Game.time);
    }
    sendEnergyToStorage(kernel, base, terminal, amount, ttl, trace) {
        const reserve = getStructureForResource(base, RESOURCE_ENERGY);
        if (!reserve) {
            trace.info('could not find dropoff for energy', { amount });
            return;
        }
        trace.info('sending request to haul energy from terminal', { amount, dropoff: reserve.id });
        const details = {
            [TASK_ID]: `meu-${terminal.id}-${Game.time}`,
            [MEMORY_TASK_TYPE]: TASK_HAUL,
            [MEMORY_HAUL_PICKUP]: terminal.id,
            [MEMORY_HAUL_RESOURCE]: RESOURCE_ENERGY,
            [MEMORY_HAUL_AMOUNT]: amount,
            [MEMORY_HAUL_DROPOFF]: reserve.id,
        };
        kernel.getTopics().addRequest(getBaseDistributorTopic(base.id), HAUL_TERMINAL, details, ttl + Game.time);
    }
    updateEnergyValue(trace) {
        const energyHistory = Game.market.getHistory(RESOURCE_ENERGY);
        trace.info('updating energy value', { energyHistory });
        // private servers can return energyHistory as an empty object
        if (!energyHistory || !Array.isArray(energyHistory) || !energyHistory.length) {
            this.energyValue = 1;
            return;
        }
        const dailyAvgs = energyHistory.map((order) => order.avgPrice);
        if (!dailyAvgs.length) {
            this.energyValue = 1;
            return;
        }
        this.energyValue = _.sum(dailyAvgs) / dailyAvgs.length;
    }
}

const REQUEST_ENERGY_TTL = 10;
const REQUEST_ENERGY_THRESHOLD = 500;
const EMERGENCY_RESERVE = 250;
class TowerRunnable {
    constructor(baseId, tower) {
        this.baseId = baseId;
        this.towerId = tower.id;
        this.haulTTL = 0;
        this.repairTTL = 0;
        this.prevTime = Game.time;
    }
    run(kernel, trace) {
        trace = trace.begin('tower_run');
        const ticks = Game.time - this.prevTime;
        this.prevTime = Game.time;
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.error('no base config, terminating', { id: this.baseId });
            trace.end();
            return terminate();
        }
        const tower = Game.getObjectById(this.towerId);
        if (!tower) {
            trace.error('no tower, terminating', { id: this.towerId });
            trace.end();
            return terminate();
        }
        if (!tower.isActive()) {
            trace.error('tower is inactive, sleeping', { id: this.towerId });
            trace.end();
            return sleeping(100);
        }
        // Count towers based on which have energy
        const numPoweredTowers = tower.room.find(FIND_MY_STRUCTURES, {
            filter: (s) => {
                return s.structureType === STRUCTURE_TOWER && s.store.getUsedCapacity(RESOURCE_ENERGY) > 10;
            },
        }).length;
        this.haulTTL -= ticks;
        this.repairTTL -= ticks;
        const towerUsed = tower.store.getUsedCapacity(RESOURCE_ENERGY);
        trace.info('tower runnable', {
            room: tower.room.name,
            id: this.towerId,
            haulTTL: this.haulTTL,
            repairTTL: this.repairTTL,
            repairTarget: this.repairTarget,
            energy: towerUsed,
            numPoweredTowers: numPoweredTowers,
        });
        // Request energy
        if (towerUsed < REQUEST_ENERGY_THRESHOLD && this.haulTTL < 0) {
            this.haulTTL = REQUEST_ENERGY_TTL;
            trace.info('requesting energy', {});
            this.requestEnergy(kernel, base, tower, REQUEST_ENERGY_TTL, trace);
        }
        // Attack hostiles
        const roomName = tower.room.name;
        let targets = kernel.getTopics().getFilteredRequests(getBasePriorityTargetsTopic(this.baseId), (target) => {
            return target.details.roomName === roomName;
        });
        // Remove targets that can heal too much
        targets = targets.filter((target) => {
            return numPoweredTowers * 600 > target.details.healingPower;
        });
        trace.info('targets', {
            targets: targets.map((t) => {
                return {
                    id: t.details.id,
                    healing: t.details.healingPower,
                };
            }),
        });
        if (targets.length) {
            const target = Game.getObjectById(targets[0].details.id);
            const result = tower.attack(target);
            trace.info('attacking', { target: targets[0].details.id, result });
            trace.end();
            return running();
        }
        // let healTargets = kernel.getTopics().getFilteredRequests(getBasePriorityHealsTopic(this.baseId),
        //   (target) => {
        //     return target.details.roomName === roomName;
        //   },
        // );
        // Heal damaged creeps
        //if (!this.damagedCreep && this.orgRoom.damagedCreeps.length) {
        //  this.damagedCreep = this.orgRoom.damagedCreeps.shift();
        //}
        if (this.damagedCreep) {
            const creep = Game.creeps[this.damagedCreep];
            if (!creep || creep.hits >= creep.hitsMax) {
                this.damagedCreep = null;
            }
            else {
                const result = tower.heal(creep);
                trace.info('healing', { target: creep.id, result });
                trace.end();
                return running();
            }
        }
        // Not above attack/heal reserve, skip repair logic
        if (towerUsed < EMERGENCY_RESERVE) {
            trace.info('skipping repairs low energy', { towerUsed });
            trace.end();
            return running();
        }
        // If low on CPU bucket, stop repairing
        if (Game.cpu.bucket < 1000) {
            trace.info('skipping repairs low bucket', { bucket: Game.cpu.bucket });
            trace.end();
            return running();
        }
        // Repair focus TTL that spreads repairs out
        if (this.repairTarget && this.repairTTL < 0) {
            trace.info('repair target ttl hit', {});
            this.repairTarget = null;
            this.repairTTL = 0;
        }
        // If target is repaired, picking another target
        if (this.repairTarget) {
            const target = Game.getObjectById(this.repairTarget);
            if (!target || target.hits >= target.hitsMax) {
                trace.info('repair target done/missing', { target });
                this.repairTarget = null;
                this.repairTTL = 0;
            }
        }
        // Repair damaged structure
        if (!this.repairTarget && base.damagedStructures.length) {
            this.repairTarget = base.damagedStructures.shift();
            this.repairTTL = 10;
        }
        // Do not repair secondary structures or roads if room is low on energy
        const baseEnergy = getStoredResourceAmount(base, RESOURCE_ENERGY);
        if (baseEnergy < 10000) {
            trace.info('skipping repairs low energy', { energy: baseEnergy });
            this.repairTarget = null;
            this.repairTTL = 0;
            trace.end();
            return running();
        }
        // Repair damaged secondary structures
        if (!this.repairTarget && base.damagedSecondaryStructures.length) {
            let nextRepairTarget = null;
            let nextReapirTargetId = null;
            for (let i = 0; i < base.damagedSecondaryStructures.length; i++) {
                nextReapirTargetId = base.damagedSecondaryStructures[0];
                trace.info('damaged secondary structure', {
                    nextReapirTargetId,
                    length: base.damagedSecondaryStructures.length
                });
                nextRepairTarget = Game.getObjectById(nextReapirTargetId);
                if (!nextRepairTarget) {
                    trace.info('damaged secondary structure not found', { nextReapirTargetId });
                    continue;
                }
                trace.info('damaged secondary structure', { nextReapirTargetId, nextRepairTarget });
                break;
            }
            if (nextRepairTarget) {
                if (nextRepairTarget.hitsMax - nextRepairTarget.hits < 100000 &&
                    base.damagedSecondaryStructures.length) {
                    this.repairTarget = base.damagedSecondaryStructures.shift();
                }
                else {
                    this.repairTarget = nextReapirTargetId;
                }
                trace.info('repair damaged secondary target', { target: this.repairTarget });
                this.repairTTL = 10;
            }
        }
        // If no repair target sleep for a bit
        if (!this.repairTarget) {
            trace.info('no repair repair', {});
            trace.end();
            return sleeping(5);
        }
        const target = Game.getObjectById(this.repairTarget);
        if (!target) {
            trace.error('repair target missing', { target });
            this.repairTarget = null;
            this.repairTTL = 0;
            trace.end();
            return running();
        }
        const result = tower.repair(target);
        trace.info('repair', { target, result, ttl: this.repairTTL });
        trace.end();
        return running();
    }
    requestEnergy(kernel, base, tower, ttl, trace) {
        const towerUsed = tower.store.getUsedCapacity(RESOURCE_ENERGY);
        const towerFree = tower.store.getFreeCapacity(RESOURCE_ENERGY);
        const towerTotal = tower.store.getCapacity(RESOURCE_ENERGY);
        const pickup = getStructuresWithResource(base, RESOURCE_ENERGY).shift();
        if (!pickup) {
            trace.info('no pickup', {});
            return;
        }
        const priority = (base.alertLevel !== AlertLevel.GREEN ?
            HAUL_TOWER_HOSTILES : HAUL_TOWER) - (towerUsed / towerTotal);
        const details = {
            [TASK_ID]: `tel-${tower.id}-${Game.time}`,
            [MEMORY_TASK_TYPE]: TASK_HAUL,
            [MEMORY_HAUL_PICKUP]: pickup.id,
            [MEMORY_HAUL_RESOURCE]: RESOURCE_ENERGY,
            [MEMORY_HAUL_AMOUNT]: towerFree,
            [MEMORY_HAUL_DROPOFF]: tower.id,
        };
        kernel.getTopics().addRequest(getBaseDistributorTopic(this.baseId), priority, details, ttl + Game.time);
        trace.info('request energy', { priority, details, towerUsed, towerTotal });
    }
}

const RUN_INTERVAL = 100;
const MAX_WALL_SITES = 5;
class WallsRunnable {
    constructor(baseId) {
        this.baseId = baseId;
    }
    run(kernel, trace) {
        const base = kernel.getPlanner().getBaseById(this.baseId);
        if (!base) {
            trace.error('base not found', { baseId: this.baseId });
            return terminate();
        }
        if (!base.walls.length) {
            this.updateBaseWalls(kernel, base, trace);
        }
        const baseLevel = getBaseLevel(base);
        if (baseLevel >= 3) {
            this.buildWalls(kernel, base, trace);
        }
        return sleeping(RUN_INTERVAL);
    }
    updateBaseWalls(kernel, base, trace) {
        const baseBounds = {
            x1: base.origin.x - 9, y1: base.origin.y - 9,
            x2: base.origin.x + 9, y2: base.origin.y + 9,
        };
        const [walls] = getCutTiles(base.primary, [baseBounds], ENTIRE_ROOM_BOUNDS);
        base.walls = walls;
        trace.info('updated walls', { base: base });
    }
    buildWalls(kernel, base, trace) {
        if (!base.walls) {
            return;
        }
        const room = getBasePrimaryRoom(base);
        if (!room) {
            trace.error('no primary room');
            return;
        }
        if (!room.storage) {
            trace.info('no storage');
            return;
        }
        trace.info('building walls', { roomId: room.name });
        let numWallSites = 0;
        base.walls.forEach((wall) => {
            if (numWallSites >= MAX_WALL_SITES) {
                return;
            }
            const position = new RoomPosition(wall.x, wall.y, room.name);
            const road = position.lookFor(LOOK_STRUCTURES).find((structure) => {
                return structure.structureType === STRUCTURE_ROAD;
            });
            const roadSite = position.lookFor(LOOK_CONSTRUCTION_SITES).find((site) => {
                return site.structureType === STRUCTURE_ROAD;
            });
            const passage = _.find(base.passages, { x: position.x, y: position.y });
            let expectedStructure = STRUCTURE_WALL;
            if (road || roadSite || passage) {
                expectedStructure = STRUCTURE_RAMPART;
            }
            // Every other spot is a rampart
            if (position.x % 2 != position.y % 2) {
                expectedStructure = STRUCTURE_RAMPART;
            }
            const structures = position.lookFor(LOOK_STRUCTURES);
            // Remove undesired structures
            const undesiredStructures = structures.filter((structure) => {
                return structure.structureType !== expectedStructure && structure.structureType !== STRUCTURE_ROAD;
            });
            undesiredStructures.forEach((structure) => {
                structure.destroy();
            });
            // If desired structure is present, we are done
            const desiredStructure = structures.find((structure) => {
                return structure.structureType === expectedStructure;
            });
            if (desiredStructure) {
                trace.info('structure present', { structure: desiredStructure.structureType, position: position });
                return;
            }
            let foundSite = false;
            // Check for sites
            const sites = position.lookFor(LOOK_CONSTRUCTION_SITES);
            if (sites) {
                const expectedSite = sites.find((site) => {
                    return site.structureType === expectedStructure;
                });
                if (expectedSite) {
                    trace.info('site present', { site: expectedSite.structureType });
                    numWallSites++;
                    foundSite = true;
                }
                else {
                    sites.forEach((site) => {
                        // dont remove road site
                        if (site.structureType === STRUCTURE_ROAD) {
                            return;
                        }
                        trace.info('wrong site, remove', { existing: site.structureType, expected: expectedStructure });
                        site.remove();
                    });
                }
            }
            if (!foundSite) {
                const result = room.createConstructionSite(wall.x, wall.y, expectedStructure);
                if (result !== OK) {
                    trace.error('failed to build structure', { result, pos: position, structureType: expectedStructure });
                    return;
                }
                trace.notice('building site', { wall, structureType: expectedStructure });
            }
        });
    }
}

const MIN_ENERGY = 100000;
const MIN_TICKS_TO_DOWNGRADE = 150000;
const MIN_UPGRADERS = 1;
const MAX_UPGRADERS = 10;
const UPGRADER_ENERGY = 25000;
const MAX_EXPLORERS = 3;
const NO_VISION_TTL = 20;
const MIN_TTL = 10;
const ENERGY_REQUEST_TTL = 50;
const REQUEST_CLAIMER_TTL = 50;
const REQUEST_REPAIRER_TTL = 30;
const REQUEST_BUILDER_TTL = 30;
const REQUEST_UPGRADER_TTL = 30;
const CHECK_SAFE_MODE_TTL = 10;
const UPDATE_PROCESSES_TTL = 20;
const PRODUCE_STATUS_TTL = 30;
const ABANDON_BASE_TTL = 50;
const REQUEST_EXPLORER_TTL = 100;
const UPDATE_BOOSTER_TTL = 5;
const ABANDON_BASE_AFTER = 10000;
const HOSTILE_DAMAGE_THRESHOLD = 0;
const HOSTILE_HEALING_THRESHOLD = 600;
var DEFENSE_POSTURE;
(function (DEFENSE_POSTURE) {
    DEFENSE_POSTURE["OPEN"] = "open";
    DEFENSE_POSTURE["CLOSED"] = "closed";
    DEFENSE_POSTURE["UNKNOWN"] = "unknown";
})(DEFENSE_POSTURE || (DEFENSE_POSTURE = {}));
const importantStructures = [
    STRUCTURE_SPAWN,
    STRUCTURE_STORAGE,
    STRUCTURE_TERMINAL,
    STRUCTURE_TOWER,
];
class BaseRunnable {
    constructor(id, scheduler) {
        this.id = id;
        this.scheduler = scheduler;
        this.defensePosture = DEFENSE_POSTURE.UNKNOWN;
        // Metrics
        this.missingProcesses = 0;
        // Threads
        this.threadProduceHudStatus = threadBase('produce_status_thread', PRODUCE_STATUS_TTL)(this.produceHudStatus.bind(this));
        this.threadAbandonBase = threadBase('abandon_base_check', ABANDON_BASE_TTL)(this.abandonBase.bind(this));
        this.threadUpdateProcessSpawning = threadBaseRoom('spawn_room_processes_thread', UPDATE_PROCESSES_TTL)(this.handleProcessSpawning.bind(this));
        this.threadRequestRepairer = threadBaseRoom('request_repairs_thread', REQUEST_REPAIRER_TTL)(this.requestRepairer.bind(this));
        this.threadRequestBuilder = threadBaseRoom('request_builder_thead', REQUEST_BUILDER_TTL)(this.requestBuilder.bind(this));
        this.threadRequestUpgrader = threadBaseRoom('request_upgrader_thread', REQUEST_UPGRADER_TTL)(this.requestUpgrader.bind(this));
        this.threadRequestExplorer = threadBaseRoom('request_explorers_thread', REQUEST_EXPLORER_TTL)(this.requestExplorer.bind(this));
        this.threadCheckSafeMode = threadBaseRoom('check_safe_mode_thread', CHECK_SAFE_MODE_TTL)(this.checkSafeMode.bind(this));
        this.threadRequestEnergy = threadBaseRoom('request_energy_thread', ENERGY_REQUEST_TTL)(this.requestEnergy.bind(this));
        this.threadUpdateAlertLevel = threadBaseRoom('update_alert_level_thread', UPDATE_PROCESSES_TTL)(this.updateAlertLevel.bind(this));
        // Pump events from booster runnable and set booster state on the Base
        this.threadUpdateBoosters = threadBase('update_booster_thread', UPDATE_BOOSTER_TTL)(this.updateBoosters.bind(this));
    }
    run(kernel, trace) {
        var _a;
        trace = trace.begin('base_run');
        // Check if base process should terminate
        const base = kernel.getPlanner().getBaseById(this.id);
        if (!base) {
            trace.error('no base config, terminating', { base: this.id });
            trace.end();
            return terminate();
        }
        // Inform other processes of room status
        this.threadProduceHudStatus(trace, kernel, base);
        // Base life cycle
        this.threadAbandonBase(trace, kernel, base);
        // if room not visible or controller is unclaimed
        const room = getBasePrimaryRoom(base);
        if (!room || ((_a = room.controller) === null || _a === void 0 ? void 0 : _a.level) === 0) {
            trace.info('room not visible or level 0', { room: room === null || room === void 0 ? void 0 : room.name });
            this.requestClaimer(kernel, trace);
            trace.end();
            return sleeping(NO_VISION_TTL);
        }
        this.threadUpdateProcessSpawning(trace, kernel, base, room);
        // Defense
        // this.threadUpdateRampartAccess(trace, base, room);
        this.threadCheckSafeMode(trace, kernel, base, room);
        // Logistics
        this.threadRequestEnergy(trace, kernel, base, room);
        // Creeps
        this.threadRequestUpgrader(trace, kernel, base, room);
        if (base.alertLevel === AlertLevel.GREEN) {
            this.threadRequestBuilder(trace, kernel, base, room);
            this.threadRequestRepairer(trace, kernel, base, room);
            this.threadRequestExplorer(trace, kernel, base, room);
        }
        // Alert level
        this.threadUpdateAlertLevel(trace, kernel, base);
        const roomVisual = new RoomVisual(base.primary);
        roomVisual.text('O', base.origin.x, base.origin.y, { color: '#FFFFFF' });
        roomVisual.text('P', base.parking.x, base.parking.y, { color: '#FFFFFF' });
        trace.end();
        return sleeping(MIN_TTL);
    }
    abandonBase(trace, kernel, base) {
        var _a, _b, _c, _d;
        trace = trace.begin('abandon_base');
        const room = getBasePrimaryRoom(base);
        if (!room) {
            if (base.addedAt < Game.time - ABANDON_BASE_AFTER) {
                trace.info('cannot see room significant time, abandoning base', { base: base.id });
                kernel.getPlanner().removeBase(base.id, trace);
            }
            trace.end();
            return;
        }
        trace.notice('abandoning base check', {
            base: this.id,
            added: base.addedAt,
            abandonTtl: Game.time - base.addedAt,
            roomLevel: (_a = room.controller) === null || _a === void 0 ? void 0 : _a.level,
            roomOwner: (_b = room.controller) === null || _b === void 0 ? void 0 : _b.owner,
            reservation: (_c = room.controller) === null || _c === void 0 ? void 0 : _c.reservation,
        });
        const spawns = room.find(FIND_MY_SPAWNS);
        if (spawns.length > 0) {
            trace.end();
            return;
        }
        const towers = room.find(FIND_MY_STRUCTURES, {
            filter: (structure) => {
                return structure.structureType === STRUCTURE_TOWER && structure.isActive();
            },
        });
        if (towers.length > 0) {
            trace.end();
            return;
        }
        if (((_d = room.controller) === null || _d === void 0 ? void 0 : _d.level) > 4) {
            trace.end();
            return;
        }
        if (base.addedAt < Game.time - ABANDON_BASE_AFTER) {
            trace.warn('base lacks spawners, towers, and is low level, abandon', { base: this.id });
            kernel.getPlanner().removeBase(base.id, trace);
            trace.end();
            return;
        }
        trace.end();
    }
    requestClaimer(kernel, trace) {
        const enroute = _.find(Game.creeps, {
            memory: {
                [MEMORY_ROLE]: WORKER_RESERVER,
                [MEMORY_ASSIGN_SHARD]: Game.shard.name,
                [MEMORY_ASSIGN_ROOM]: this.id,
                [MEMORY_BASE]: this.id,
            },
        });
        if (enroute) {
            trace.notice('claimer already enroute', { id: this.id, name: enroute.name, pos: enroute.pos });
            return;
        }
        const detail = {
            role: WORKER_RESERVER,
            memory: {
                [MEMORY_ROLE]: WORKER_RESERVER,
                [MEMORY_ASSIGN_SHARD]: Game.shard.name,
                [MEMORY_ASSIGN_ROOM]: this.id,
                [MEMORY_BASE]: this.id,
            },
        };
        trace.notice('requesting claimer', { id: this.id, detail });
        kernel.getTopics().addRequest(getShardSpawnTopic(), PRIORITY_RESERVER, detail, REQUEST_CLAIMER_TTL + Game.time);
    }
    handleProcessSpawning(trace, kernel, base, primaryRoom) {
        var _a;
        let missingProcesses = 0;
        // Spawn Manager
        const spawnManagerId = `spawns_${base.id}`;
        if (!this.scheduler.hasProcess(spawnManagerId)) {
            trace.info('starting spawn manager', { id: this.id });
            missingProcesses++;
            this.scheduler.registerProcess(new Process(spawnManagerId, 'spawns', Priorities.CORE_LOGISTICS, new SpawnManager(spawnManagerId, base.id)));
        }
        // Towers
        primaryRoom.find(FIND_MY_STRUCTURES, {
            filter: (structure) => {
                return structure.structureType === STRUCTURE_TOWER && structure.isActive();
            },
        }).forEach((tower) => {
            const towerId = `${tower.id}`;
            if (!this.scheduler.hasProcess(towerId)) {
                trace.info('starting tower', { id: tower.id });
                missingProcesses++;
                const process = new Process(towerId, 'towers', Priorities.DEFENCE, new TowerRunnable(this.id, tower));
                process.setSkippable(false);
                this.scheduler.registerProcess(process);
            }
        });
        // Nukes
        primaryRoom.find(FIND_MY_STRUCTURES, {
            filter: (structure) => structure.structureType === STRUCTURE_NUKER && structure.isActive(),
        }).forEach((nuker) => {
            const nukeId = `${nuker.id}`;
            if (!this.scheduler.hasProcess(nukeId)) {
                trace.info('starting nuke', { id: nukeId });
                missingProcesses++;
                this.scheduler.registerProcess(new Process(nukeId, 'nukes', Priorities.OFFENSE, new NukerRunnable(this.id, nuker)));
            }
        });
        if ((_a = primaryRoom.terminal) === null || _a === void 0 ? void 0 : _a.isActive()) {
            // Terminal runnable
            const terminalId = primaryRoom.terminal.id;
            if (!this.scheduler.hasProcess(terminalId)) {
                trace.info('starting terminal', { id: terminalId });
                missingProcesses++;
                this.scheduler.registerProcess(new Process(terminalId, 'terminals', Priorities.DEFENCE, new TerminalRunnable(this.id, primaryRoom.terminal)));
            }
        }
        // Link Manager
        const linkManagerId = `links_${this.id}`;
        if (!this.scheduler.hasProcess(linkManagerId) && primaryRoom.storage) {
            trace.info('starting link manager', { id: linkManagerId });
            missingProcesses++;
            this.scheduler.registerProcess(new Process(linkManagerId, 'links', Priorities.RESOURCES, new LinkManager(linkManagerId, this.id)));
        }
        // Labs Manager
        const labsManagerId = `labs_${this.id}`;
        if (!this.scheduler.hasProcess(labsManagerId)) {
            trace.info('starting labs manager', { id: labsManagerId });
            missingProcesses++;
            this.scheduler.registerProcess(new Process(labsManagerId, 'labs', Priorities.LOGISTICS, new LabsManager(base.id, labsManagerId, this.scheduler, trace)));
            missingProcesses++;
        }
        // Construction
        const constructionId = `construction_${this.id}`;
        if (!this.scheduler.hasProcess(constructionId)) {
            trace.info('starting construction', { id: constructionId });
            missingProcesses++;
            this.scheduler.registerProcess(new Process(constructionId, 'construction', Priorities.CORE_LOGISTICS, new BaseConstructionRunnable(base.id, constructionId)));
            missingProcesses++;
        }
        // Remotes
        const remotesManagerId = `remotes_${this.id}`;
        if (!this.scheduler.hasProcess(remotesManagerId)) {
            trace.info('starting remote manager', { id: remotesManagerId });
            missingProcesses++;
            this.scheduler.registerProcess(new Process(remotesManagerId, 'remotes', Priorities.CORE_LOGISTICS, new RemotesManager(this.id)));
        }
        // Rooms
        base.rooms.forEach((room) => {
            const roomId = `room_${room}`;
            const hasRoomProcess = this.scheduler.hasProcess(roomId);
            if (!hasRoomProcess) {
                trace.info('starting room', { id: roomId });
                missingProcesses++;
                this.scheduler.registerProcess(new Process(roomId, 'rooms', Priorities.CRITICAL, new RoomRunnable(room, this.scheduler)));
            }
        });
        // Controller
        const controllerProcessId = primaryRoom.controller.id;
        if (!this.scheduler.hasProcess(controllerProcessId)) {
            trace.info('starting controller', { id: primaryRoom.controller.id });
            missingProcesses++;
            const controllerRunnable = new ControllerRunnable(primaryRoom.controller.id);
            this.scheduler.registerProcess(new Process(controllerProcessId, 'controller', Priorities.CRITICAL, controllerRunnable));
        }
        // Road network and hauling
        const logisticsId = `logistics_${this.id}`;
        const hasLogisticsProcess = this.scheduler.hasProcess(logisticsId);
        if (!hasLogisticsProcess) {
            trace.info('starting logistics', { id: logisticsId });
            missingProcesses++;
            this.scheduler.registerProcess(new Process(logisticsId, 'logistics', Priorities.LOGISTICS, new LogisticsRunnable(this.id)));
        }
        // Walls
        const wallsId = `walls_${this.id}`;
        const hasWallsProcess = this.scheduler.hasProcess(wallsId);
        if (!hasWallsProcess) {
            trace.info('starting walls', { id: wallsId });
            missingProcesses++;
            this.scheduler.registerProcess(new Process(wallsId, 'walls', Priorities.DEFENCE, new WallsRunnable(this.id)));
        }
        // Neighbors
        const neighborsId = `neighbors_${this.id}`;
        if (!this.scheduler.hasProcess(neighborsId)) {
            trace.info('starting neighbors', { id: neighborsId });
            missingProcesses++;
            this.scheduler.registerProcess(new Process(neighborsId, 'neighbors', Priorities.CRITICAL, new NeighborsRunnable(this.id)));
        }
        const defenseManagerId = `defense_manager_${this.id}`;
        if (!this.scheduler.hasProcess(defenseManagerId)) {
            // Defense manager, must run before towers and defenders
            const defenseManager = new DefenseManager(kernel, base, trace);
            this.scheduler.registerProcess(new Process(defenseManagerId, 'defense_manager', Priorities.CRITICAL, defenseManager));
        }
        // Repairs
        const repairId = `repair_${this.id}`;
        if (!this.scheduler.hasProcess(repairId)) {
            trace.info('starting repair', { id: repairId });
            missingProcesses++;
            const repairRunnable = new RepairRunnable(this.id);
            this.scheduler.registerProcess(new Process(repairId, 'repair', Priorities.CRITICAL, repairRunnable));
        }
        // Observer runnable
        const observerStructures = primaryRoom.find(FIND_MY_STRUCTURES, {
            filter: (structure) => {
                return structure.structureType === STRUCTURE_OBSERVER;
            },
        });
        if (observerStructures.length) {
            const observerId = observerStructures[0].id;
            const hasProcess = this.scheduler.hasProcess(observerId);
            if (!hasProcess) {
                trace.info('starting observer', { id: observerId });
                missingProcesses++;
                this.scheduler.registerProcess(new Process(observerId, 'observer', Priorities.EXPLORATION, new ObserverRunnable(base.id, observerStructures[0])));
            }
        }
        this.missingProcesses = missingProcesses;
    }
    requestRepairer(trace, kernel, base, room) {
        var _a;
        let maxHits = 0;
        let hits = 0;
        // Don't request repairers until we have at least upgraded the room
        if (((_a = room.controller) === null || _a === void 0 ? void 0 : _a.level) < 2) {
            return;
        }
        const roomStructures = room.find(FIND_STRUCTURES);
        roomStructures.forEach((s) => {
            if (s.structureType === STRUCTURE_WALL || s.structureType === STRUCTURE_RAMPART) {
                return;
            }
            if (s.hitsMax > 0 && s.hits > 0) {
                maxHits += s.hitsMax;
                hits += s.hits;
            }
        });
        let hitsPercentage = 1;
        if (maxHits > 0) {
            hitsPercentage = hits / maxHits;
        }
        const numRepairers = kernel.getCreepsManager().getCreepsByBaseAndRole(this.id, WORKER_REPAIRER).length;
        trace.info('need repairers?', { id: this.id, hitsPercentage, numRepairers });
        // Repairer requests
        let desiredRepairers = 0;
        let repairerPriority = PRIORITY_REPAIRER;
        if (hitsPercentage < 0.8) {
            trace.info('need more repairers', { id: this.id, hitsPercentage });
            desiredRepairers = 1;
        }
        if (hitsPercentage < 0.6) {
            trace.info('need more repairers', { id: this.id, hitsPercentage });
            desiredRepairers = 2;
            repairerPriority = PRIORITY_REPAIRER_URGENT;
        }
        if (Game.cpu.bucket < 1000) {
            trace.info('bucket low', { bucket: Game.cpu.bucket });
            desiredRepairers = 0;
        }
        if (numRepairers >= desiredRepairers) {
            trace.info('already have enough repairers', { id: this.id, numRepairers, desiredRepairers });
            return;
        }
        trace.info('request repairers', { id: this.id, desiredRepairers, numRepairers });
        // @TODO create role base creation methods
        const memory = {
            [MEMORY_ASSIGN_ROOM]: this.id,
            [MEMORY_BASE]: base.id,
        };
        const request = createSpawnRequest(repairerPriority, REQUEST_REPAIRER_TTL + Game.time, WORKER_REPAIRER, memory, null, 0);
        kernel.getTopics().addRequestV2(getBaseSpawnTopic(base.id), request);
    }
    requestBuilder(trace, kernel, base, room) {
        if (!Object.values(Game.spawns).length) {
            trace.info('no spawns, dont spawn builders');
            return;
        }
        if (!room.storage) {
            trace.info('no storage, dont spawn builders');
            return;
        }
        const builders = kernel.getCreepsManager().getCreepsByBaseAndRole(this.id, WORKER_BUILDER);
        const numConstructionSites = room.find(FIND_MY_CONSTRUCTION_SITES).length;
        trace.info('num constructions sites', { numConstructionSites });
        let desiredBuilders = 0;
        if (numConstructionSites) {
            desiredBuilders = desiredBuilders = Math.ceil(numConstructionSites / 10);
        }
        if (desiredBuilders > 3) {
            desiredBuilders = 3;
        }
        trace.info('checking builders', { id: this.id, numConstructionSites, desiredBuilders });
        if (builders.length >= desiredBuilders) {
            trace.info('already have enough builders', { id: this.id, numConstructionSites, desiredBuilders });
            return;
        }
        const priority = PRIORITY_BUILDER - (builders.length * 2);
        const ttl = REQUEST_BUILDER_TTL + Game.time;
        const memory = {
            [MEMORY_ASSIGN_ROOM]: this.id,
            [MEMORY_ASSIGN_SHARD]: Game.shard.name,
            [MEMORY_BASE]: base.id,
        };
        const request = createSpawnRequest(priority, ttl, WORKER_BUILDER, memory, null, 0);
        kernel.getTopics().addRequestV2(getBaseSpawnTopic(base.id), request);
        // @CONFIRM builders are being spawned
    }
    requestUpgrader(trace, kernel, base, room) {
        var _a, _b;
        if (!room.storage) {
            trace.info('no storage, dont spawn upgraders');
            return;
        }
        const numUpgraders = kernel.getCreepsManager().getCreepsByBaseAndRole(this.id, WORKER_UPGRADER).length;
        let parts = 1;
        let desiredUpgraders = MIN_UPGRADERS;
        let maxParts = 15;
        const reserveEnergy = getStoredResourceAmount(base, RESOURCE_ENERGY);
        const reserveBuffer = getReserveBuffer(base);
        if (!((_a = room.controller) === null || _a === void 0 ? void 0 : _a.my)) {
            trace.error('not my room');
            desiredUpgraders = 0;
        }
        else if (room.controller.level === 8) {
            parts = (reserveEnergy - reserveBuffer) / 1500;
            desiredUpgraders = 1;
            // In an effort to reduce CPU usage, we only spawn upgraders if they have many parts or
            // if we are close go being downgraded In theory, we will build bigger creeps instead of
            // smaller ones and rate won't be a big reduction
            if (parts < 15 && room.controller.ticksToDowngrade > MIN_TICKS_TO_DOWNGRADE) {
                desiredUpgraders = 0;
            }
            trace.info('max level room', {
                parts, desiredUpgraders, ticksToDowngrade: room.controller.ticksToDowngrade,
                reserveEnergy, reserveBuffer,
            });
        }
        else if (room.storage) { // @ORG-REFACTOR replace has storage with base phase
            const roomCapacity = room.energyCapacityAvailable;
            maxParts = Math.floor(roomCapacity / 200);
            if (maxParts > 15) {
                maxParts = 15;
            }
            if (((_b = room.storage) === null || _b === void 0 ? void 0 : _b.isActive()) && reserveEnergy > reserveBuffer) {
                parts = (reserveEnergy - reserveBuffer) / 1500;
            }
            else if (!room.storage && reserveEnergy > 1000) {
                parts = reserveEnergy - 1000 / 1500;
            }
            desiredUpgraders = Math.ceil(parts / maxParts);
            trace.info('has storage', { desiredUpgraders, maxParts, parts, reserveEnergy, reserveBuffer });
        }
        const energyLimit = ((parts - 1) * 150) + 200;
        // Don't let it create a ton of upgraders
        if (desiredUpgraders > MAX_UPGRADERS) {
            desiredUpgraders = MAX_UPGRADERS;
        }
        trace.info('request upgraders', {
            desiredUpgraders,
            numUpgraders,
            parts,
            energyLimit,
        });
        for (let i = 0; i < desiredUpgraders - numUpgraders; i++) {
            // Reduce priority by number of existing and requested upgraders
            const upgraderPriority = PRIORITY_UPGRADER - ((numUpgraders + i) * 2);
            const memory = {
                [MEMORY_ASSIGN_ROOM]: this.id,
                [MEMORY_BASE]: base.id,
            };
            const request = createSpawnRequest(upgraderPriority, REQUEST_UPGRADER_TTL + Game.time, WORKER_UPGRADER, memory, null, energyLimit);
            kernel.getTopics().addRequestV2(getBaseSpawnTopic(base.id), request);
            // @CONFIRM that upgraders are being created
        }
    }
    requestExplorer(trace, kernel, base) {
        const shardConfig = kernel.getConfig();
        if (!shardConfig.explorers) {
            trace.info('shard does not allow explorers');
            return;
        }
        const explorers = kernel.getCreepsManager().
            getCreepsByBaseAndRole(this.id, WORKER_EXPLORER);
        if (explorers.length < MAX_EXPLORERS) {
            trace.info('requesting explorer');
            const priority = EXPLORER;
            const ttl = REQUEST_EXPLORER_TTL + Game.time;
            const role = WORKER_EXPLORER;
            const memory = {
                [MEMORY_BASE]: base.id,
            };
            const request = createSpawnRequest(priority, ttl, role, memory, null, 0);
            kernel.getTopics().addRequestV2(getBaseSpawnTopic(base.id), request);
        }
        else {
            trace.info('not requesting explorer', { numExplorers: explorers.length });
        }
    }
    checkSafeMode(trace, kernel, base, room) {
        const controller = room.controller;
        if (!controller) {
            trace.info('controller not found');
            return;
        }
        let enableSafeMode = false;
        let hostiles = room.find(FIND_HOSTILE_CREEPS);
        // filter out hostiles without attack, ranged attack, or work parts
        hostiles = hostiles.filter((hostile) => {
            return hostile.getActiveBodyparts(ATTACK) ||
                hostile.getActiveBodyparts(RANGED_ATTACK) ||
                hostile.getActiveBodyparts(WORK);
        });
        // Filter friendly creeps
        const friends = kernel.getFriends();
        hostiles = hostiles.filter((creep) => friends.indexOf(creep.owner.username) === -1);
        trace.info('hostiles', { hostiles: hostiles.length });
        if (hostiles) {
            // BUG: Proximity to a link has triggered safe mode
            // find may not work as expected
            const infrastructure = room.find(FIND_MY_STRUCTURES, {
                filter: (structure) => {
                    return _.find(importantStructures, structure.structureType);
                },
            });
            trace.info('infrastructure', { infrastructure: infrastructure.length });
            // Iterate through critical infrastructure and check if any are under attack
            for (const structure of infrastructure) {
                if (structure.pos.findInRange(hostiles, 3).length) {
                    trace.notice('critical infrastructure under attack');
                    enableSafeMode = true;
                    break;
                }
            }
        }
        if (enableSafeMode) {
            if (controller.safeMode) {
                trace.info('safe mode already active');
                return;
            }
            // If hostiles present spawn defenders and/or activate safe mode
            if (controller.safeModeAvailable && !controller.safeMode && !controller.safeModeCooldown) {
                controller.activateSafeMode();
                trace.warn('activating safe mode');
                return;
            }
        }
        else {
            trace.info('do not enable safe mode');
        }
    }
    requestEnergy(trace, kernel, base, room) {
        var _a, _b;
        const terminalEnergy = ((_a = room.terminal) === null || _a === void 0 ? void 0 : _a.store.getUsedCapacity(RESOURCE_ENERGY)) || 0;
        const storageEnergy = ((_b = room.storage) === null || _b === void 0 ? void 0 : _b.store.getUsedCapacity(RESOURCE_ENERGY)) || 0;
        trace.info('room energy', {
            terminalEnergy,
            storageEnergy,
            roomLevel: getBaseLevel(base),
            desiredBuffer: getReserveBuffer(base),
            UPGRADER_ENERGY,
            MIN_ENERGY,
        });
        let requestEnergy = false;
        // if we are below minimum energy, request more
        if (storageEnergy + terminalEnergy < MIN_ENERGY) {
            requestEnergy = true;
        }
        /*
        // If not level 8, request more energy then buffer for upgrades
        if (Game.market.credits > CREDIT_RESERVE && orgRoom.getRoomLevel() < 8 &&
          storageEnergy + terminalEnergy < orgRoom.getReserveBuffer() + UPGRADER_ENERGY) {
          requestEnergy = true;
        }
        */
        if (requestEnergy) {
            const amount = 5000;
            trace.info('requesting energy from governor', { amount, resource: RESOURCE_ENERGY });
            const resourceGovernor = kernel.getResourceManager();
            const requested = resourceGovernor.requestResource(base, RESOURCE_ENERGY, amount, ENERGY_REQUEST_TTL, trace);
            if (!requested) {
                resourceGovernor.buyResource(base, RESOURCE_ENERGY, amount, ENERGY_REQUEST_TTL, trace);
            }
        }
    }
    produceHudStatus(trace, kernel, base) {
        const resources = getStoredResources(base);
        const roomStatus = {
            [BASE_STATUS_NAME]: base.primary,
            [BASE_STATUS_LEVEL]: getBaseLevel(base),
            [BASE_STATUS_LEVEL_COMPLETED]: getBaseLevelCompleted(base),
            [BASE_STATUS_ENERGY]: resources.get(RESOURCE_ENERGY) || 0,
            [BASE_STATUS_ALERT_LEVEL]: base.alertLevel,
        };
        const status = {
            details: roomStatus,
        };
        trace.info('producing base status', { status });
        kernel.getTopics().addRequest(ROOM_STATUES, 1, status, PRODUCE_STATUS_TTL + Game.time);
        const line = {
            key: `base_${base.id}`,
            room: base.primary,
            order: 0,
            text: `Base: ${base.id} - status: ${base.alertLevel}, level: ${getBaseLevel(base)}, ` +
                `Rooms: ${base.rooms.join(',')}, Neighbors: ${base.neighbors.join(',')}`,
            time: Game.time,
        };
        const event = new Event(base.primary, Game.time, HudEventSet, line);
        kernel.getBroker().getStream(getLinesStream()).publish(event);
        const reserveEnergy = getStoredResourceAmount(base, RESOURCE_ENERGY);
        const reserveBuffer = getReserveBuffer(base);
        const parts = (reserveEnergy - reserveBuffer) / 1500;
        const upgraderLine = {
            key: `base_${base.id}_upgrader`,
            room: base.primary,
            order: 1,
            text: `Energy: ${reserveEnergy}, Buffer: ${reserveBuffer}, Parts: ${parts}`,
            time: Game.time,
        };
        const upgraderEvent = new Event(base.primary, Game.time, HudEventSet, upgraderLine);
        kernel.getBroker().getStream(getLinesStream()).publish(upgraderEvent);
        const indicatorStream = kernel.getBroker().getStream(getDashboardStream());
        base.rooms.forEach((roomName) => {
            // Alert indicator
            let alertLevelStatus = HudIndicatorStatus.Green;
            if (base.alertLevel === AlertLevel.RED) {
                alertLevelStatus = HudIndicatorStatus.Red;
            }
            else if (base.alertLevel === AlertLevel.YELLOW) {
                alertLevelStatus = HudIndicatorStatus.Yellow;
            }
            const alertLevelIndicator = {
                room: roomName, key: 'alert', display: 'A',
                status: alertLevelStatus,
            };
            indicatorStream.publish(new Event(roomName, Game.time, HudEventSet, alertLevelIndicator));
        });
        // Processes
        let processStatus = HudIndicatorStatus.Green;
        if (this.missingProcesses > 1) {
            processStatus = HudIndicatorStatus.Red;
        }
        else if (this.missingProcesses === 1) {
            processStatus = HudIndicatorStatus.Yellow;
        }
        const keyProcessesIndicator = {
            room: base.primary, key: 'processes', display: 'P',
            status: processStatus,
        };
        indicatorStream.publish(new Event(base.id, Game.time, HudEventSet, keyProcessesIndicator));
    }
    updateAlertLevel(trace, kernel, base, _room) {
        // check if strong enemies are present in base
        const roomEntry = kernel.getScribe().getRoomById(base.primary);
        if (!roomEntry) {
            trace.warn('room not found, assuming hostile presence', { room: base.primary });
            base.alertLevel = AlertLevel.YELLOW;
            return;
        }
        if (beingSieged(roomEntry)) {
            trace.warn('room being sieged', {
                room: base.primary,
                hostileDamage: roomEntry.hostilesDmg,
                hostileHealing: roomEntry.hostilesHealing,
            });
            base.alertLevel = AlertLevel.RED;
            return;
        }
        // check if strong enemies are present in rooms
        const rooms = base.rooms;
        const hostileRoom = rooms.find((roomName) => {
            const roomEntry = kernel.getScribe().getRoomById(roomName);
            if (!roomEntry) {
                trace.info('room not found', { room: roomName });
                return false;
            }
            if (beingSieged(roomEntry)) {
                return true;
            }
            return false;
        });
        if (hostileRoom) {
            trace.warn('hostile presence detected', {
                room: hostileRoom,
            });
            base.alertLevel = AlertLevel.YELLOW;
            return;
        }
        // check if neighbor is under red alert
        const neighbors = base.neighbors;
        const redNeighbor = neighbors.find((id) => {
            const neighborBase = kernel.getPlanner().getBaseById(id);
            if (!neighborBase) {
                trace.warn('neighbor base not found, should not happen', { id });
                return false;
            }
            if (neighborBase.alertLevel === AlertLevel.RED) {
                return true;
            }
            return false;
        });
        if (redNeighbor) {
            trace.warn('red neighbor detected', {
                redNeighbor,
            });
            base.alertLevel = AlertLevel.YELLOW;
            return;
        }
        trace.info('no significant hostile presence', { level: base.alertLevel, baseId: base.id });
        base.alertLevel = AlertLevel.GREEN;
    }
    updateBoosters(trace, kernel, base) {
        const topic = kernel.getTopics().getTopic(getBaseBoostTopic(base));
        if (!topic) {
            trace.info('no topic', { room: this.id });
            return;
        }
        topic.forEach((message) => {
            const details = message.details;
            trace.info('booster position', { room: this.id, details });
            setBoostPosition(base, details.position);
            setLabsByAction(base, details.labsByAction);
            base.storedEffects = details.storedEffects;
            base.labsByAction = details.labsByAction;
        });
    }
}
function beingSieged(roomEntry) {
    if (roomEntry.hostilesDmg > HOSTILE_DAMAGE_THRESHOLD ||
        roomEntry.hostilesHealing > HOSTILE_HEALING_THRESHOLD) {
        return true;
    }
    return false;
}

/**
 * Scribe
 *
 * Journal room state so that we have some object permanence when we lose visibility
 * into rooms. When a room has not been visited in a a while remote the stale journal
 * entry.
 *
 * TODO: Store room state and restore on restart - bug?
 */
const RUN_TTL$3 = 5;
const WRITE_MEMORY_INTERVAL = 50;
const REMOVE_STALE_ENTRIES_INTERVAL = 100;
const UPDATE_BASE_COUNT = 50;
const PRODUCE_EVENTS_INTERVAL = 10;
const JOURNAL_ENTRY_TTL = 10;
const YELLOW_JOURNAL_AGE = 500;
const RED_JOURNAL_AGE = 1000;
const MAX_JOURNAL_AGE = 2500;
class Scribe {
    constructor(trace) {
        const journal = {
            rooms: new Map(),
            creeps: new Map(),
        };
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        if (Memory.scribe) {
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            trace.info('Loading scribe journal', { memory: JSON.stringify(Memory.scribe) });
            try {
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                journal.rooms = new Map(Memory.scribe.rooms);
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                journal.creeps = new Map(Memory.scribe.creeps);
                trace.notice('Loaded scribe journal', {
                    rooms: journal.rooms.size,
                    creeps: journal.creeps.size,
                });
            }
            catch (e) {
                trace.error('Error loading scribe journal', e);
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                delete Memory.scribe;
            }
        }
        this.journal = journal;
        this.globalBaseCount = -2;
        this.threadRemoveStaleJournalEntries = threadKernel('remove_stale', REMOVE_STALE_ENTRIES_INTERVAL)(this.removeStaleJournalEntries.bind(this));
        this.threadWriteMemory = threadKernel('write_memory', WRITE_MEMORY_INTERVAL)(this.writeMemory.bind(this));
        this.threadUpdateBaseCount = threadKernel('update_base_count', UPDATE_BASE_COUNT)(this.updateBaseCount.bind(this));
        this.threadProduceEvents = threadKernel('produce_events', PRODUCE_EVENTS_INTERVAL)(this.produceEvents.bind(this));
    }
    run(kernel, trace) {
        trace = trace.begin('run');
        const updateRoomsTrace = trace.begin('update_rooms');
        // Iterate rooms and update if stale
        Object.values(Game.rooms).forEach((room) => {
            const entry = this.getRoomById(room.name);
            if (!entry || Game.time - entry.lastUpdated > JOURNAL_ENTRY_TTL) {
                this.updateRoom(kernel, room, updateRoomsTrace);
            }
        });
        updateRoomsTrace.end();
        this.threadRemoveStaleJournalEntries(trace, kernel);
        this.threadWriteMemory(trace, kernel);
        this.threadUpdateBaseCount(trace, kernel);
        this.threadProduceEvents(trace, kernel);
        /*
        const username = this.getKingdom().config.username;
        const friends = this.getKingdom().config.friends;
        const visual = Game.map.visual;
        this.getRooms().forEach((room) => {
          const age = Game.time - room.lastUpdated;
          const owner = room.controller?.owner || null;
    
          visual.text(age.toString(), new RoomPosition(49, 47, room.id), {
            align: 'right',
            fontSize: 4,
          });
    
          let roomPosture = '';
          if (owner && owner !== username) {
            roomPosture += '⚔️';
          }
          if (owner === username) {
            roomPosture += '🟢';
          }
          if (room.controller?.safeMode > 0) {
            roomPosture += '💢';
          }
    
          visual.text(roomPosture, new RoomPosition(0, 4, room.id), {
            align: 'left',
            fontSize: 6,
          });
        });
        */
        const metrics = trace.getMetricsCollector();
        metrics.gauge('scribe_rooms_total', this.getRooms().length, {});
        const oldestRoom = _.min(this.getRooms(), (room) => room.lastUpdated);
        metrics.gauge('scribe_rooms_max_age', Game.time - oldestRoom.lastUpdated, {});
        trace.info('Scribe run complete', {
            rooms: this.getRooms().length,
            oldest: Game.time - oldestRoom.lastUpdated,
        });
        trace.end();
        return sleeping(RUN_TTL$3);
    }
    writeMemory(trace, _kernel) {
        trace.info('write_memory', { cpu: Game.cpu });
        if (Game.cpu.bucket < 1000) {
            trace.info('clearing journal from memory to reduce CPU load');
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            Memory.scribe = null;
            return;
        }
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        Memory.scribe = {
            rooms: Array.from(this.journal.rooms.entries()),
            creeps: Array.from(this.journal.creeps.entries()),
        };
    }
    updateBaseCount(trace, kernel) {
        if (this.globalBaseCount === -2) {
            // skip the first time, we want to give shards time to update their remote memory
            trace.info('skipping updateBaseCount');
            this.globalBaseCount = -1;
            return;
        }
        const bases = kernel.getPlanner().getBases();
        let baseCount = bases.length;
        // Iterate shards and get their base counts
        this.getShardList().forEach((shard) => {
            var _a;
            if (shard === Game.shard.name) {
                return;
            }
            const shardMemory = this.getRemoteShardMemory(shard);
            baseCount += ((_a = shardMemory === null || shardMemory === void 0 ? void 0 : shardMemory.status) === null || _a === void 0 ? void 0 : _a.numColonies) || 0;
        });
        trace.info('update_base_count', { baseCount: baseCount });
        this.globalBaseCount = baseCount;
    }
    produceEvents(trace, kernel) {
        const indicatorStream = kernel.getBroker().getStream(getDashboardStream());
        const rooms = this.getRooms();
        trace.info('produce_events', { rooms: rooms.length });
        rooms.forEach((roomEntry) => {
            const age = Game.time - roomEntry.lastUpdated;
            // Processes
            let processStatus = HudIndicatorStatus.Green;
            if (age > RED_JOURNAL_AGE) {
                processStatus = HudIndicatorStatus.Red;
            }
            else if (age > YELLOW_JOURNAL_AGE) {
                processStatus = HudIndicatorStatus.Yellow;
            }
            const spawnLengthIndicator = { room: roomEntry.id, key: 'journal', display: '👁️', status: processStatus };
            indicatorStream.publish(new Event(`journal_${roomEntry.id}`, Game.time, HudEventSet, spawnLengthIndicator));
        });
    }
    getShardList() {
        const name = Game.shard.name;
        if (name.startsWith('shard') && name !== 'shardSeason') {
            return ['shard3', 'shard2', 'shard1', 'shard0'];
        }
        return [Game.shard.name];
    }
    getGlobalBaseCount() {
        if (this.globalBaseCount < 0) {
            return null;
        }
        return this.globalBaseCount;
    }
    removeStaleJournalEntries(trace) {
        const numBefore = this.journal.rooms.size;
        // Create new map of fresh rooms
        const rooms = new Map();
        for (const [id, entry] of this.journal.rooms) {
            if (Game.time - entry.lastUpdated > MAX_JOURNAL_AGE) {
                trace.info('removing stale journal entry', { id: id });
                continue;
            }
            rooms.set(id, entry);
        }
        this.journal.rooms = rooms;
        const numAfter = this.journal.rooms.size;
        trace.info('remove_stale', { numBefore, numAfter });
        trace.getMetricsCollector().counter('scribe_stale_total', numBefore - numAfter);
    }
    getRoomById(roomId) {
        return this.journal.rooms.get(roomId) || null;
    }
    getRooms() {
        return Array.from(this.journal.rooms.values());
    }
    getOldestRoomInList(rooms) {
        const knownRooms = Array.from(this.journal.rooms.keys());
        const missingRooms = _.difference(rooms, knownRooms);
        if (missingRooms.length) {
            return _.shuffle(missingRooms)[0];
        }
        // Create array and narrow to only rooms in the provided list
        const inRangeRooms = _.pick(Array.from(this.journal.rooms.values()), (room) => {
            return rooms.indexOf(room.id) > -1;
        });
        // Sort in ascending order so we pull the oldest room first
        const sortedRooms = _.sortBy(inRangeRooms, 'lastUpdated');
        if (!sortedRooms.length) {
            return null;
        }
        return sortedRooms[0].id;
    }
    getRoomsUpdatedRecently() {
        return Array.from(this.journal.rooms.values()).filter((room) => {
            return Game.time - room.lastUpdated < 500;
        }).map((room) => {
            return room.id;
        });
    }
    getRoomsWithPowerBanks() {
        return Array.from(this.journal.rooms.values()).filter((room) => {
            if (!room.powerBanks) {
                return false;
            }
            return room.powerBanks.length > 0;
        }).map((room) => {
            return [room.id, Game.time - room.lastUpdated, room.powerBanks[0].ttl];
        });
    }
    getRoomsWithInvaderBases() {
        const found = [];
        for (const [_key, room] of this.journal.rooms) {
            if (room.invaderCoreLevel > 0) {
                found.push([room.id, Game.time - room.lastUpdated]);
            }
        }
        return found;
    }
    getRoomsWithHostileTowers(kernel) {
        return Array.from(this.journal.rooms.values()).filter((room) => {
            if (!room.controller || room.controller.owner === kernel.getPlanner().getUsername()) {
                return false;
            }
            if (!room.numTowers) {
                return false;
            }
            return true;
        }).map((room) => {
            return [room.id, room.numTowers, room.controller.owner];
        });
    }
    getRoomsWithPortals() {
        return Array.from(this.journal.rooms.values()).filter((room) => {
            if (!room.portals) {
                return false;
            }
            return room.portals.length > 0;
        }).map((room) => {
            return [room.id, Game.time - room.lastUpdated, room.portals.map((portal) => {
                    return portal.destinationShard;
                })];
        });
    }
    getRoomsWithMyConstructionSites() {
        let rooms = Array.from(this.journal.rooms.values()).filter((room) => {
            return room.myConstructionSites > 0;
        });
        rooms = _.sortByOrder(rooms, 'myConstructionSites', 'desc');
        return rooms.map((room) => {
            return { id: room.id, sites: room.myConstructionSites };
        });
    }
    getHostileRooms(kernel, trace) {
        return Array.from(this.journal.rooms.values()).filter((room) => {
            var _a, _b;
            trace.info("room entry", {
                room: room.id,
                controller: room.controller,
                owner: (_a = room.controller) === null || _a === void 0 ? void 0 : _a.owner
            });
            if (!room.controller || room.controller.owner === kernel.getPlanner().getUsername()) {
                return false;
            }
            const base = kernel.getPlanner().getBases()[0];
            if (!base) {
                trace.error("no bases found");
                return false;
            }
            const baseStatus = (_b = Game.map.getRoomStatus(base.primary)) === null || _b === void 0 ? void 0 : _b.status;
            if (room.specialRoom && baseStatus === 'normal') {
                return false;
            }
            if (room.controller.level < 1) {
                return false;
            }
            if (room.controller.safeMode) {
                return false;
            }
            return true;
        }).map((room) => {
            return {
                id: room.id,
                numTowers: room.numTowers,
                numKeyStructures: room.numKeyStructures,
                owner: room.controller.owner,
                level: room.controller.level,
                controllerPos: room.controller.pos,
            };
        });
    }
    updateRoom(kernel, roomObject, trace) {
        var _a, _b, _c, _d, _e, _f;
        trace = trace.begin('update_room');
        trace = trace.withFields(new Map([['room', roomObject.name]]));
        const end = trace.startTimer('update_room');
        const room = {
            id: roomObject.name,
            lastUpdated: Game.time,
            specialRoom: false,
            status: null,
            controller: null,
            storage: null,
            hasSpawns: false,
            spawnLocation: null,
            numSources: 0,
            hasHostiles: false,
            hostilesByOwner: new Map(),
            hostilesDmgByOwner: new Map(),
            hostilesDmg: 0,
            hostilesHealing: 0,
            hasKeepers: false,
            keepersDmg: 0,
            keepersHealing: 0,
            invaderCorePos: null,
            invaderCoreLevel: null,
            invaderCoreTime: null,
            numTowers: 0,
            numKeyStructures: 0,
            myConstructionSites: 0,
            mineral: null,
            powerBanks: [],
            portals: [],
            deposits: [],
        };
        const controllerEnd = trace.startTimer('controller');
        if (roomObject.controller) {
            let owner = null;
            if (roomObject.controller.owner) {
                owner = roomObject.controller.owner.username;
            }
            if ((_b = (_a = roomObject.controller) === null || _a === void 0 ? void 0 : _a.reservation) === null || _b === void 0 ? void 0 : _b.username) {
                owner = roomObject.controller.reservation.username;
            }
            let downgrade = 0;
            if ((_c = roomObject.controller) === null || _c === void 0 ? void 0 : _c.reservation) {
                downgrade = roomObject.controller.reservation.ticksToEnd;
            }
            else if (((_d = roomObject.controller) === null || _d === void 0 ? void 0 : _d.level) >= 1) {
                downgrade = roomObject.controller.ticksToDowngrade;
            }
            room.controller = {
                owner: owner,
                level: roomObject.controller.level,
                safeMode: roomObject.controller.safeMode || 0,
                safeModeAvailable: roomObject.controller.safeModeAvailable,
                pos: roomObject.controller.pos,
                downgrade,
            };
            const spawns = roomObject.find(FIND_STRUCTURES, {
                filter: (structure) => {
                    return structure.structureType === STRUCTURE_SPAWN;
                },
            });
            room.hasSpawns = spawns.length > 0;
            room.spawnLocation = (_e = spawns[0]) === null || _e === void 0 ? void 0 : _e.pos;
        }
        if (roomObject.storage) {
            room.storage = {
                energy: roomObject.storage.store.getUsedCapacity(RESOURCE_ENERGY),
            };
        }
        controllerEnd();
        const roomStatusEnd = trace.startTimer('roomStatus');
        let status = null;
        try {
            status = Game.map.getRoomStatus(roomObject.name);
        }
        catch (e) {
            trace.warn('problem getting room status from game, assuming normal', e);
        }
        room.status = (status === null || status === void 0 ? void 0 : status.status) || 'normal';
        room.specialRoom = room.status !== 'normal';
        roomStatusEnd();
        const sourcesEnd = trace.startTimer('sources');
        room.numSources = roomObject.find(FIND_SOURCES).length;
        sourcesEnd();
        const hostilesEnd = trace.startTimer('hostiles');
        const hostiles = roomObject.find(FIND_HOSTILE_CREEPS);
        // Filter friendly creeps
        const friends = kernel.getFriends();
        const hostileCreeps = hostiles.filter((creep) => {
            const owner = creep.owner.username;
            return friends.indexOf(owner) === -1 && owner !== 'Source Keeper';
        });
        room.hasHostiles = hostileCreeps.length > 0;
        if (room.hasHostiles) {
            room.hostilesByOwner = hostileCreeps.reduce((map, creep) => {
                const owner = creep.owner.username;
                const count = map.get(owner) || 0;
                map.set(owner, count + 1);
                return map;
            }, new Map());
            room.hostilesDmgByOwner = hostileCreeps.reduce((map, creep) => {
                const owner = creep.owner.username;
                const count = map.get(owner) || 0;
                map.set(owner, count + scoreAttacking(creep));
                return map;
            }, new Map());
            room.hostilesDmg = _.sum(hostileCreeps, (creep) => {
                return scoreAttacking(creep);
            });
            room.hostilesHealing = _.sum(hostileCreeps, (creep) => {
                return scoreHealing(creep);
            });
        }
        // keepers
        const keepers = hostiles.filter((creep) => {
            const owner = creep.owner.username;
            return owner === 'Source Keeper';
        });
        room.hasKeepers = keepers.length > 0;
        if (room.hasKeepers) {
            room.keepersDmg = _.sum(keepers, (creep) => {
                return scoreAttacking(creep);
            });
            room.keepersHealing = _.sum(keepers, (creep) => {
                return scoreHealing(creep);
            });
        }
        const invaderCores = roomObject.find(FIND_HOSTILE_STRUCTURES, {
            filter: (structure) => {
                return structure.structureType === STRUCTURE_INVADER_CORE && structure.isActive();
            },
        });
        if (invaderCores.length) {
            room.invaderCorePos = invaderCores[0].pos;
            room.invaderCoreLevel = invaderCores[0].level;
            room.invaderCoreTime = (_f = invaderCores[0].effects[EFFECT_COLLAPSE_TIMER]) === null || _f === void 0 ? void 0 : _f.ticksRemaining;
        }
        hostilesEnd();
        const towersEnd = trace.startTimer('towers');
        room.numTowers = roomObject.find(FIND_STRUCTURES, {
            filter: (structure) => {
                var _a;
                return structure.structureType === STRUCTURE_TOWER &&
                    ((_a = structure.owner) === null || _a === void 0 ? void 0 : _a.username) !== kernel.getPlanner().getUsername();
            },
        }).length;
        towersEnd();
        const keyStructuresEnd = trace.startTimer('keyStructures');
        room.numKeyStructures = roomObject.find(FIND_HOSTILE_STRUCTURES, {
            filter: (structure) => {
                return structure.isActive() && (structure.structureType === STRUCTURE_TOWER ||
                    structure.structureType === STRUCTURE_SPAWN ||
                    structure.structureType === STRUCTURE_TERMINAL ||
                    structure.structureType === STRUCTURE_NUKER ||
                    structure.structureType === STRUCTURE_INVADER_CORE);
            },
        }).length;
        keyStructuresEnd();
        const constructionSitesEnd = trace.startTimer('constructionSites');
        room.myConstructionSites = roomObject.find(FIND_MY_CONSTRUCTION_SITES).length;
        constructionSitesEnd();
        const mineralEnd = trace.startTimer('mineral');
        room.mineral = null;
        const minerals = roomObject.find(FIND_MINERALS);
        if (minerals.length) {
            room.mineral = minerals[0].mineralType;
        }
        mineralEnd();
        const portalsEnd = trace.startTimer('portals');
        room.portals = [];
        const portals = roomObject.find(FIND_STRUCTURES, {
            filter: (structure) => {
                return structure.structureType === STRUCTURE_PORTAL;
            },
        });
        room.portals = portals.map((portal) => {
            return {
                id: portal.id,
                pos: portal.pos,
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                destinationShard: portal.destination.shard,
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                destinationRoom: portal.destination.room,
            };
        });
        portalsEnd();
        const powerBanksEnd = trace.startTimer('powerBanks');
        room.powerBanks = roomObject.find(FIND_STRUCTURES, {
            filter: (structure) => {
                return structure.structureType === STRUCTURE_POWER_BANK;
            },
        }).map((powerBank) => {
            return {
                id: powerBank.id,
                hits: powerBank.hits,
                ttl: powerBank.ticksToDecay,
                power: powerBank.power,
                pos: powerBank.pos,
            };
        });
        powerBanksEnd();
        const depositsEnd = trace.startTimer('deposits');
        room.deposits = roomObject.find(FIND_DEPOSITS).map((deposit) => {
            return {
                type: deposit.depositType,
                cooldown: deposit.cooldown,
                ttl: deposit.ticksToDecay,
            };
        });
        depositsEnd();
        this.journal.rooms.set(room.id, room);
        const duration = end();
        trace.info('updated room', { duration, room: room.id });
        trace.end();
    }
    clearRoom(roomId) {
        this.journal.rooms.delete(roomId);
    }
    getLocalShardMemory() {
        if (typeof (InterShardMemory) === 'undefined') {
            return {};
        }
        return JSON.parse(InterShardMemory.getLocal() || '{}');
    }
    setLocalShardMemory(memory) {
        if (typeof (InterShardMemory) === 'undefined') {
            return;
        }
        return InterShardMemory.setLocal(JSON.stringify(memory));
    }
    getRemoteShardMemory(shardName) {
        if (typeof (InterShardMemory) === 'undefined') {
            return {};
        }
        return JSON.parse(InterShardMemory.getRemote(shardName) || '{}');
    }
    getPortals(shardName) {
        const roomsWithPortals = [];
        for (const [_, entry] of this.journal.rooms) {
            if (entry.portals) {
                const shardPortals = entry.portals.filter(portal => portal.destinationShard === shardName);
                roomsWithPortals.push(...shardPortals);
            }
        }
        return roomsWithPortals;
    }
    setCreepBackup(creep) {
        const localMemory = this.getLocalShardMemory();
        if (!localMemory.creep_backups) {
            localMemory.creep_backups = new Map();
        }
        localMemory.creep_backups.set(creep.name, {
            name: creep.name,
            memory: creep.memory,
            ttl: Game.time,
        });
        this.setLocalShardMemory(localMemory);
    }
    getCreepBackup(shardName, creepName) {
        var _a;
        const remoteMemory = this.getRemoteShardMemory(shardName);
        return ((_a = remoteMemory.creep_backups) === null || _a === void 0 ? void 0 : _a.get(creepName)) || null;
    }
}

const BASE_SECTOR_RADIUS = 4;
function getNextRoomToScout(kernel, creep) {
    const base = getCreepBase(kernel, creep);
    if (!base) {
        return null;
    }
    // Get rooms considered with the bases range/sector
    const nearbyRooms = getNearbyRooms(base.primary, BASE_SECTOR_RADIUS);
    if (nearbyRooms.length === 0) {
        return null;
    }
    const explorers = kernel.getCreepsManager().getCreeps().
        filter(c => c.memory[MEMORY_ROLE] === WORKER_EXPLORER);
    const assignedRooms = explorers.map((explorer) => explorer.memory[MEMORY_ASSIGN_ROOM]);
    // Filter rooms down to stale entries and entries already being scouted
    const staleEntries = nearbyRooms.filter((room) => {
        // filter out rooms that are already being scouted
        if (assignedRooms.includes(room)) {
            return false;
        }
        const roomEntry = kernel.getScribe().getRoomById(room);
        if (!roomEntry) {
            return true;
        }
        return Game.time - roomEntry.lastUpdated > YELLOW_JOURNAL_AGE;
    });
    // If no stale entries, return null
    if (staleEntries.length === 0) {
        return null;
    }
    // Sort by distance from creep
    const sortedBy = _.sortBy(staleEntries, (room) => {
        return Game.map.getRoomLinearDistance(creep.pos.roomName, room);
    });
    return sortedBy[0];
}
function getNearbyRooms(center, distance) {
    const rooms = [];
    const centerStatus = Game.map.getRoomStatus(center);
    const baseCord = getRoomCordFromRoomName(center);
    for (let x = -distance + baseCord.x; x <= distance + baseCord.x; x++) {
        for (let y = -distance + baseCord.y; y <= distance + baseCord.y; y++) {
            const roomName = getRoomNameFromRoomCord({ x: x, y: y });
            if (roomName == center) {
                continue;
            }
            // Room must have same status as center/base
            const status = Game.map.getRoomStatus(roomName);
            if (status.status !== centerStatus.status) {
                continue;
            }
            rooms.push(roomName);
        }
    }
    return rooms;
}
function getRoomNameFromRoomCord(roomCord) {
    const horizontal = roomCord.x < 0 ? 'W' : 'E';
    const vertical = roomCord.y < 0 ? 'N' : 'S';
    if (horizontal === 'W') {
        roomCord.x += 1;
    }
    if (vertical === 'N') {
        roomCord.y += 1;
    }
    const absX = Math.abs(roomCord.x);
    const absY = Math.abs(roomCord.y);
    return horizontal + absX + vertical + absY;
}
function getRoomCordFromRoomName(roomName) {
    // get index of E/W
    const eIndex = roomName.indexOf('E');
    const wIndex = roomName.indexOf('W');
    const horizontal = eIndex > -1 ? eIndex : wIndex;
    // get index of N/S
    const nIndex = roomName.indexOf('N');
    const sIndex = roomName.indexOf('S');
    const vertical = sIndex > -1 ? sIndex : nIndex;
    // get x
    let x = parseInt(roomName.substring(horizontal + 1, vertical), 10);
    if (wIndex > -1) {
        x = -x;
        x -= 1;
    }
    // get y
    let y = parseInt(roomName.substring(vertical + 1, roomName.length), 10);
    if (nIndex > -1) {
        y = -y;
        y -= 1;
    }
    return { x, y };
}

const RUN_TTL$2 = 10;
const BASE_PROCESSES_TTL = 50;
const EXPAND_TTL = 500;
const MAX_SPAWNLESS_BASES = 1;
class BaseManager {
    constructor(config, scheduler, trace) {
        this.config = config;
        this.scheduler = scheduler;
        this.shards = [];
        this.bases = new Map();
        this.shards.push(Game.shard.name);
        let bases = new Map();
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        if (Memory.bases) {
            try {
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                trace.warn('found shard memory', { bases: Memory.bases.length });
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                bases = new Map(Memory.bases);
            }
            catch (e) {
                trace.error('failed to load bases', { e });
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                delete Memory.bases;
            }
        }
        else {
            trace.warn('no shard config found, bootstraping?');
        }
        // Setup known bases
        Array.from(bases.values()).forEach((base) => {
            trace.notice('setting up base', { base });
            const origin = new RoomPosition(base.origin.x, base.origin.y, base.origin.roomName);
            const parking = new RoomPosition(base.parking.x, base.parking.y, base.parking.roomName);
            this.addBase(base.id, base.addedAt || Game.time, base.isPublic, origin, parking, base.rooms, base.walls || [], base.passages || [], base.neighbors || [], base.alertLevel || AlertLevel.GREEN, trace);
        });
        // Check for spawns without bases
        Object.values(Game.spawns).forEach((spawn) => {
            const roomName = spawn.room.name;
            const origin = new RoomPosition(spawn.pos.x, spawn.pos.y + 4, spawn.pos.roomName);
            trace.info('checking spawn', { roomName, origin });
            const parking = new RoomPosition(origin.x + 5, origin.y, origin.roomName);
            if (!this.bases.has(roomName)) {
                trace.warn('found unknown base', { roomName });
                this.addBase(roomName, Game.time, false, origin, parking, [roomName], [], [], [], AlertLevel.GREEN, trace);
            }
        });
        trace.notice('bases configs', { bases: this.bases });
        this.threadBaseProcesses = threadKernel('base_processes', BASE_PROCESSES_TTL)(this.baseProcesses.bind(this));
        // TODO make this an iterator
        this.expandBasesThread = threadKernel('expand', EXPAND_TTL)(this.expandBases.bind(this));
    }
    run(kernel, trace) {
        this.threadBaseProcesses(trace, kernel);
        this.expandBasesThread(trace, kernel);
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        Memory.bases = Array.from(this.bases.entries());
        return sleeping(RUN_TTL$2);
    }
    getShards() {
        return this.shards;
    }
    getBase(baseId) {
        return this.bases.get(baseId);
    }
    getBases() {
        return Array.from(this.bases.values());
    }
    getBaseMap() {
        return this.bases;
    }
    getBaseById(baseId) {
        return this.bases.get(baseId);
    }
    getBaseByRoom(roomName) {
        for (const base of this.bases.values()) {
            if (base.rooms.includes(roomName)) {
                return base;
            }
        }
        return null;
    }
    getUsername() {
        if (!this.username) {
            const spawn = _.first(_.values(Game.spawns));
            if (!spawn) {
                throw new Error('no spawns found');
            }
            this.username = spawn.owner.username;
        }
        return this.username;
    }
    getClosestBaseInRange(roomName, range = 5) {
        let selectedBase = null;
        let selectedBaseDistance = 99999;
        Object.values(this.getBases()).forEach((base) => {
            const distance = Game.map.getRoomLinearDistance(base.primary, roomName);
            if (distance <= range && selectedBaseDistance > distance) {
                selectedBase = base;
                selectedBaseDistance = distance;
            }
        });
        return selectedBase;
    }
    addBase(primaryRoom, createTime, isPublic, origin, parking, rooms, walls, passages, neighbors, alertLevel, trace) {
        if (this.bases.has(primaryRoom)) {
            trace.error('base already exists', { primaryRoom });
            return;
        }
        if (rooms.indexOf(primaryRoom) === -1) {
            trace.warn('primary room not in rooms', { primaryRoom, rooms });
            rooms.unshift(primaryRoom);
        }
        this.bases.set(primaryRoom, {
            id: primaryRoom,
            addedAt: createTime,
            isPublic: isPublic,
            primary: primaryRoom,
            rooms: rooms,
            origin: origin,
            parking: parking,
            walls: walls,
            passages: passages,
            neighbors: neighbors,
            alertLevel: alertLevel,
            boostPosition: null,
            boosts: new Map(),
            // @REFACTOR check these defaults
            storedEffects: new Map(),
            labsByAction: new Map(),
            terminalTask: null,
            defenseHitsLimit: 0,
            damagedStructures: [],
            damagedSecondaryStructures: [],
        });
        return this.bases.get(primaryRoom);
    }
    removeBase(baseId, trace) {
        const base = this.getBase(baseId);
        if (!base) {
            trace.error('base does not exist', { baseId });
            return;
        }
        this.bases.delete(baseId);
    }
    baseProcesses(trace, kernel) {
        // If any defined colonies don't exist, run it
        const bases = kernel.getPlanner().getBases();
        bases.forEach((base) => {
            const baseProcessId = `base_${base.id}`;
            const hasProcess = this.scheduler.hasProcess(baseProcessId);
            if (hasProcess) {
                return;
            }
            trace.warn('starting base process');
            this.scheduler.registerProcess(new Process(baseProcessId, 'base', Priorities.CRITICAL, new BaseRunnable(base.id, this.scheduler)));
        });
    }
    expandBases(trace, kernel) {
        if (!this.config.autoExpand) {
            trace.warn('auto expand disabled');
            return;
        }
        const spawnlessBases = this.getBases().filter((base) => {
            const roomEntry = kernel.getScribe().getRoomById(base.primary);
            if (!roomEntry) {
                return false;
            }
            return !roomEntry.hasSpawns;
        });
        if (spawnlessBases.length >= MAX_SPAWNLESS_BASES) {
            trace.notice('too many spawnless bases', { lowLevelBases: spawnlessBases.length });
            return;
        }
        const scribe = kernel.getScribe();
        const globalBaseCount = scribe.getGlobalBaseCount();
        if (!globalBaseCount) {
            trace.notice('do not know global base count yet');
            return;
        }
        const allowedColonies = Game.gcl.level;
        if (globalBaseCount >= allowedColonies) {
            trace.notice('max GCL colonies reached', { globalBaseCount, allowedColonies });
            return;
        }
        const bases = this.getBases();
        const numColonies = bases.length;
        const shardBaseMax = (this.config.maxColonies || 9999);
        if (numColonies >= shardBaseMax) {
            trace.notice('max config colonies reached', { numColonies, shardBaseMax });
            return;
        }
        const results = pickExpansion(kernel, trace);
        if (results.selected) {
            const roomName = results.selected;
            const distance = results.distance;
            const origin = results.origin;
            const parking = new RoomPosition(origin.x + 5, origin.y + 5, origin.roomName);
            trace.notice('selected room, adding base', { roomName, distance, origin, parking });
            this.addBase(roomName, Game.time, false, origin, parking, [roomName], [], [], [], AlertLevel.GREEN, trace);
            return;
        }
        trace.notice('no expansion selected');
    }
    reportMetrics(kernel, metrics, trace) {
        this.getBases().forEach((base) => {
            var _a, _b, _c, _d;
            const primaryRoom = getBasePrimaryRoom(base);
            if (!primaryRoom) {
                trace.info('no primary room for base, not reporting metrics', { base });
                return;
            }
            const roomLevel = ((_a = primaryRoom.controller) === null || _a === void 0 ? void 0 : _a.level) || 0;
            metrics.gauge("base_level", roomLevel, { base: base.id });
            const roomProgress = ((_b = primaryRoom.controller) === null || _b === void 0 ? void 0 : _b.progress) || 0;
            const roomProgressTotal = ((_c = primaryRoom.controller) === null || _c === void 0 ? void 0 : _c.progressTotal) || 0;
            let roomProgressPercent = 0;
            if (roomProgressTotal > 0 && roomProgress > 0) {
                roomProgressPercent = roomProgress / roomProgressTotal;
            }
            metrics.gauge("base_level_progress", roomProgressPercent, { base: base.id });
            const energyAvailable = primaryRoom.energyAvailable;
            metrics.gauge("base_energy_available", energyAvailable, { base: base.id });
            const storedEnergy = ((_d = primaryRoom.storage) === null || _d === void 0 ? void 0 : _d.store.getUsedCapacity(RESOURCE_ENERGY)) || 0;
            metrics.gauge("base_energy_stored", storedEnergy, { base: base.id });
            // Base creeps by role
            const baseCreeps = kernel.getCreepsManager().getCreepsByBase(base.id);
            const roleCounts = _.countBy(baseCreeps, (creep) => {
                return creep.memory[MEMORY_ROLE];
            });
            _.each(roleCounts, (count, role) => {
                metrics.gauge(`base_creep_role_total`, count, { base: base.id, role });
            });
            metrics.gauge("base_rooms_total", base.rooms.length, { base: base.id });
            let missingRoomEntries = 0;
            // Report room level metrics
            base.rooms.forEach((roomName) => {
                // Remote Room Entry age
                const roomEntry = kernel.getScribe().getRoomById(roomName);
                if (roomEntry) {
                    metrics.gauge('scribe_room_entry_age', Game.time - roomEntry.lastUpdated, { room: roomName, base: base.id });
                    metrics.gauge('scribe_room_entry_hostile_attack_total', roomEntry.hostilesDmg, { room: roomName, base: base.id });
                    metrics.gauge('scribe_room_entry_hostile_healing_total', roomEntry.hostilesHealing, { room: roomName, base: base.id });
                }
                else {
                    missingRoomEntries++;
                }
            });
            metrics.gauge('scribe_room_entry_missing', missingRoomEntries, { base: base.id });
            // get nearby rooms and report if fresh room entries
            const nearbyRoomNames = getNearbyRooms(base.primary, BASE_SECTOR_RADIUS);
            const freshNearbyRoomEntries = nearbyRoomNames.filter((roomName) => {
                // if no room entry, not fresh
                const roomEntry = kernel.getScribe().getRoomById(roomName);
                if (!roomEntry) {
                    return false;
                }
                // If last updated is older then "yellow" threshold, not fresh
                if (roomEntry.lastUpdated < Game.time - YELLOW_JOURNAL_AGE) {
                    return false;
                }
                return true;
            });
            metrics.gauge('base_fresh_sector_room_entries', freshNearbyRoomEntries.length, { base: base.id });
        });
    }
}

const behavior$c = sequenceNode('attacker_root', [
    leafNode('move_node', (creep, trace) => {
        const x = _.min([_.max([creep.memory[MEMORY_POSITION_X], 0]), 49]);
        const y = _.min([_.max([creep.memory[MEMORY_POSITION_Y], 0]), 49]);
        const roomId = creep.memory[MEMORY_POSITION_ROOM];
        if (x === undefined || y === undefined || !roomId) {
            trace.info('missing position memory', { x, y, roomId });
            return SUCCESS;
        }
        const position = new RoomPosition(x, y, roomId);
        trace.info('creep status', { current: creep.pos, desired: position, fatigue: creep.fatigue });
        if (creep.pos.isEqualTo(position)) {
            trace.info('creep in position', { position });
            return SUCCESS;
        }
        const ignoreCreeps = creep.pos.inRangeTo(position, 1);
        const result = creep.moveTo(position, { reusePath: 5, ignoreCreeps });
        trace.info('move to', { result, creepPos: creep.pos, position });
        return SUCCESS;
    }),
    leafNode('attack_heal_node', (creep, trace) => {
        let didHeal = false;
        const heal = creep.memory[MEMORY_HEAL];
        if (heal) {
            const healTarget = Game.getObjectById(heal);
            if (healTarget) {
                const healTargetDistance = creep.pos.getRangeTo(healTarget);
                if (healTarget && healTargetDistance <= 1) {
                    const healResult = creep.heal(healTarget);
                    trace.info('heal result', { healResult });
                    didHeal = true;
                }
                else if (healTarget && healTargetDistance <= 3) {
                    const rangedHealResult = creep.rangedHeal(healTarget);
                    trace.info('ranged heal result', { rangedHealResult });
                    didHeal = true;
                }
            }
        }
        const attack = creep.memory[MEMORY_ATTACK];
        if (attack) {
            const attackTarget = Game.getObjectById(attack);
            if (attackTarget) {
                const attackDistance = creep.pos.getRangeTo(attackTarget);
                let didDismantle = false;
                if (!didHeal && attackDistance <= 1 && creep.getActiveBodyparts(WORK) > 0) {
                    const dismantleResult = creep.dismantle(attackTarget);
                    trace.info('dismantle result', { dismantleResult });
                    didDismantle = true;
                }
                if (!didHeal && !didDismantle && attackDistance <= 1 && creep.getActiveBodyparts(ATTACK) > 0) {
                    const attackResult = creep.attack(attackTarget);
                    trace.info('attack result', { attackResult });
                }
                if (creep.getActiveBodyparts(RANGED_ATTACK) > 0) {
                    const rangedResult = creep.rangedAttack(attackTarget);
                    trace.info('ranged attack result', { rangedResult });
                }
            }
        }
        return SUCCESS;
    }),
]);
const roleAttacker = {
    run: rootNode('attacker', behaviorBoosts(behavior$c)),
};

const selectInfrastructureSites = leafNode('selectSite', (creep, trace, kernel) => {
    const sites = getInfrastructureSites(creep.room);
    if (sites.length === 0) {
        return FAILURE;
    }
    const base = getCreepBase(kernel, creep);
    if (!base) {
        trace.error('No base config for creep');
        return FAILURE;
    }
    // Sort sites by distance from base origin, this ensures the base is built first
    _.sortBy(sites, (site) => {
        return site.pos.getRangeTo(base.origin);
    });
    const site = sites[0];
    setDestination(creep, site.id, site.room.name);
    return SUCCESS;
});
const selectSite = leafNode('selectSite', (creep, trace, kernel) => {
    const sites = getPrioritizedSites(creep.room);
    if (sites.length === 0) {
        return FAILURE;
    }
    const base = getCreepBase(kernel, creep);
    if (!base) {
        trace.error('No base config for creep');
        return FAILURE;
    }
    // Sort sites by distance from base origin, this ensures the base is built first
    _.sortBy(sites, (site) => {
        return site.pos.getRangeTo(base.origin);
    });
    const site = sites[0];
    setDestination(creep, site.id, site.room.name);
    return SUCCESS;
});
const build = leafNode('build', (creep) => {
    const destination = Game.getObjectById(creep.memory[MEMORY_DESTINATION]);
    if (!destination) {
        return SUCCESS;
    }
    const result = creep.build(destination);
    if (result === ERR_NOT_ENOUGH_RESOURCES) {
        return SUCCESS;
    }
    if (result === ERR_INVALID_TARGET) {
        return FAILURE;
    }
    if (result != OK) {
        return FAILURE;
    }
    if (creep.store.getUsedCapacity() === 0) {
        return SUCCESS;
    }
    return RUNNING;
});

const selectNearbyLink = leafNode('select_nearby_link', (creep, trace, _kernal) => {
    // Favor near by stores
    let nearByLinks = creep.pos.findInRange(FIND_STRUCTURES, 8, {
        filter: (structure) => {
            if (structure.structureType !== STRUCTURE_LINK) {
                return false;
            }
            return structure.store.getUsedCapacity(RESOURCE_ENERGY) > 0;
        },
    });
    nearByLinks = _.sortBy(nearByLinks, (structure) => {
        return creep.pos.getRangeTo(structure);
    });
    // If we have a target set the destination
    if (nearByLinks.length) {
        trace.info('selecting nearby link', { linkId: nearByLinks[0].id });
        setDestination(creep, nearByLinks[0].id);
        return SUCCESS;
    }
    setDestination(creep, null);
    trace.info('did not find a nearby link');
    return FAILURE;
});
const selectStorageForDeposit = leafNode('select_storage', (creep, trace, kernel) => {
    const base = getCreepBase(kernel, creep);
    if (!base) {
        trace.error('No base config for creep');
        return FAILURE;
    }
    const energyReserve = getStructureWithResource(base, RESOURCE_ENERGY);
    if (energyReserve) {
        trace.info('selecting reserve', { id: energyReserve.id });
        setDestination(creep, energyReserve.id);
        return SUCCESS;
    }
    setDestination(creep, null);
    trace.info('did not find reserve with energy', { roomName: creep.room.name });
    return FAILURE;
});
const selectContainer = leafNode('select_container', (creep, trace, _kernel) => {
    // If no nearby stores or the room lacks storage, try to get energy from the nearest container
    const containerInRoom = creep.pos.findClosestByRange(FIND_STRUCTURES, {
        filter: (structure) => {
            if (structure.structureType !== STRUCTURE_CONTAINER) {
                return false;
            }
            return structure.store.getUsedCapacity(RESOURCE_ENERGY) > 0;
        },
    });
    if (containerInRoom) {
        trace.info('selecting container in room', { id: containerInRoom.id });
        setDestination(creep, containerInRoom.id);
        return SUCCESS;
    }
    setDestination(creep, null);
    trace.info('did not find a container with energy');
    return FAILURE;
});
const selectDroppedEnergy = leafNode('select_dropped_energy', (creep, trace, _kernel) => {
    const droppedEnergy = creep.pos.findClosestByRange(FIND_DROPPED_RESOURCES, {
        filter: (resource) => {
            if (resource.resourceType !== RESOURCE_ENERGY) {
                return false;
            }
            return resource.amount > 0;
        },
    });
    if (droppedEnergy) {
        trace.info('selecting dropped energy in room', { id: droppedEnergy.id });
        setDestination(creep, droppedEnergy.id);
        return SUCCESS;
    }
    setDestination(creep, null);
    trace.info('did not find any dropped energy');
    return FAILURE;
});
const selectMoveFill = (selector) => {
    return sequenceNode('fill_from_selector', [
        selector,
        cachedMoveToMemoryObjectId(MEMORY_DESTINATION, 1, commonPolicy),
        leafNode('fill_creep', (creep, trace) => {
            return fillCreepFromDestination(creep, trace);
        }),
    ]);
};
const getSomeEnergy = runUntilConditionMet('get_some_energy_until_success', (creep, trace, _kernel) => {
    const freeCapacity = creep.store.getFreeCapacity(RESOURCE_ENERGY);
    trace.info('creep free capacity', { freeCapacity });
    return creep.store.getUsedCapacity(RESOURCE_ENERGY) > 0;
}, selectorNode('select_and_fill_with_energy', [
    selectMoveFill(selectNearbyLink),
    selectMoveFill(selectStorageForDeposit),
    selectMoveFill(selectContainer),
    selectMoveFill(selectDroppedEnergy),
]));
const getEnergy = repeatUntilConditionMet('get_energy_until_success', (creep, trace, _kernel) => {
    const freeCapacity = creep.store.getFreeCapacity(RESOURCE_ENERGY);
    trace.info('creep free capacity', { freeCapacity });
    return creep.store.getFreeCapacity(RESOURCE_ENERGY) <= 0;
}, selectorNode('select_and_fill_with_energy', [
    selectMoveFill(selectStorageForDeposit),
    selectMoveFill(selectNearbyLink),
    selectMoveFill(selectContainer),
    selectMoveFill(selectDroppedEnergy),
]));
const parkingLot = leafNode('parking_lot', (creep, trace, kernel) => {
    const base = getCreepBase(kernel, creep);
    if (!(base === null || base === void 0 ? void 0 : base.parking)) {
        trace.error('no parking config for creep', { creepName: creep.name });
        return FAILURE;
    }
    let idle = creep.memory[MEMORY_IDLE];
    if (!idle) {
        idle = 1;
    }
    else {
        idle++;
    }
    creep.memory[MEMORY_IDLE] = idle;
    if (creep.pos.inRangeTo(base.parking, 1)) {
        trace.info('in range of parking lot');
        // TODO may move this to something cheaper
        const hasNuker = base.parking.lookFor(LOOK_STRUCTURES).find((structure) => {
            return structure.structureType === STRUCTURE_NUKER;
        });
        // Pray to the God of Nukes
        if (idle % 20 === 0) {
            if (hasNuker) {
                creep.say('🙏');
            }
            else {
                creep.say('🥪');
            }
        }
        return FAILURE;
    }
    trace.info('moving to parking lot', { parkingLot: base.parking });
    const result = creep.moveTo(base.parking, {
        reusePath: 50,
        maxOps: 1500,
        ignoreCreeps: false,
    });
    if (result !== OK && result !== ERR_TIRED) {
        trace.info('could not move to parking lot', {
            result,
            parkingLot: base.parking,
            creepName: creep.name,
            pos: creep.pos,
        });
    }
    return FAILURE;
});
const sign = `Hi!`;
const updateSign = repeatUntilConditionMet('check_sign', (creep, _trace, _kernel) => {
    if (!creep.room || !creep.room.controller || !creep.room.controller.sign) {
        return true;
    }
    const current = creep.room.controller.sign.text;
    if (current.indexOf('Respawn Area') !== -1) {
        return true;
    }
    return current === sign;
}, sequenceNode('update_sign', [
    leafNode('pick_room_controller', (creep) => {
        setDestination(creep, creep.room.controller.id);
        return SUCCESS;
    }),
    moveToCreepMemory(MEMORY_DESTINATION, 1, false, 25, 1500),
    leafNode('set_sign', (creep, trace, _kernel) => {
        const result = creep.signController(creep.room.controller, sign);
        trace.info('set sign', { result });
        if (result === OK) {
            return SUCCESS;
        }
        return SUCCESS;
    }),
]));

const behavior$b = sequenceNode('builder_root', [
    moveToShard(MEMORY_ASSIGN_SHARD),
    repeatUntilConditionMet('move_to_room', (creep, _trace, _kingdom) => {
        const assignedRoom = creep.memory[MEMORY_ASSIGN_ROOM];
        // If creep doesn't have a room assigned, we are done
        if (!assignedRoom) {
            return true;
        }
        // If the creep reaches the room we are done
        if (creep.room.name === assignedRoom) {
            return true;
        }
        return false;
    }, sequenceNode('move_to_room_controller', [
        leafNode('set_controller_location', (creep, trace, kingdom) => {
            var _a, _b;
            const assignedRoom = creep.memory[MEMORY_ASSIGN_ROOM];
            let posStr = [25, 25, assignedRoom].join(',');
            const roomEntry = kingdom.getScribe().getRoomById(assignedRoom);
            if ((_a = roomEntry === null || roomEntry === void 0 ? void 0 : roomEntry.controller) === null || _a === void 0 ? void 0 : _a.pos) {
                const pos = (_b = roomEntry.controller) === null || _b === void 0 ? void 0 : _b.pos;
                posStr = [pos.x, pos.y, pos.roomName].join(',');
            }
            creep.memory[MEMORY_ASSIGN_ROOM_POS] = posStr;
            return SUCCESS;
        }),
        cachedMoveToMemoryPos(MEMORY_ASSIGN_ROOM_POS, 3, commonPolicy),
    ])),
    setCommuteDuration,
    getEnergy,
    repeatUntilConditionMet('build_sites', (creep, _trace, _kingdom) => {
        return creep.store.getUsedCapacity(RESOURCE_ENERGY) === 0;
    }, sequenceNode('build_construction_site', [
        selectorNode('pick_something', [
            selectSite,
            parkingLot,
        ]),
        cachedMoveToMemoryObjectId(MEMORY_DESTINATION, 3, singleRoomPolicy),
        build,
    ])),
]);
const roleBuilder = {
    run: rootNode('builder', behaviorBoosts(behavior$b)),
};

const behavior$a = sequenceNode('defender_root', [
    cachedMoveToRoom(MEMORY_ASSIGN_ROOM, warPartySingleFilePolicy),
    leafNode('attack_hostiles', (creep, trace, kernel) => {
        const base = getCreepBase(kernel, creep);
        if (!base) {
            trace.error('creep has no base', { memory: creep.memory });
        }
        // Heal self or adjacent creep if one has lower HP
        let healTarget = null;
        if (creep.hits < creep.hitsMax) {
            healTarget = creep;
        }
        // heal adjacent creeps with lowest HP
        let friendlyCreepsNearby = creep.pos.findInRange(FIND_MY_CREEPS, 1, {
            filter: (c) => c.hits < c.hitsMax,
        });
        friendlyCreepsNearby = _.sortBy(friendlyCreepsNearby, (a) => {
            return a.hits / a.hitsMax;
        });
        const first = friendlyCreepsNearby[0];
        if (first && (!healTarget || first.hits < healTarget.hits)) {
            healTarget = first;
        }
        // if we have a heal target, heal it
        if (healTarget) {
            const result = creep.heal(healTarget);
            trace.info('healing self', { result });
        }
        // Get targets in the room
        const roomId = creep.room.name;
        const targets = kernel.getTopics().
            getFilteredRequests(getBasePriorityTargetsTopic(base.id), (target) => {
            return target.details.roomName === roomId;
        });
        trace.info('room targets', { targets });
        // We move to the room target and attack the highest priority target in range,
        // which could also be the room target or a target of opportunity
        let moveTarget = null;
        let attackTarget = null;
        if (targets.length) {
            moveTarget = Game.getObjectById(targets[0].details.id);
            const inRangeHostiles = _.find(targets, (target) => {
                const hostile = Game.getObjectById(target.details.id);
                return hostile && creep.pos.inRangeTo(hostile, 3);
            });
            if (inRangeHostiles) {
                attackTarget = Game.getObjectById(inRangeHostiles.details.id);
            }
        }
        trace.info('target', { moveTarget, attackTarget });
        if (attackTarget) {
            const result = creep.rangedAttack(attackTarget);
            trace.info('ranged attack result', { result, targetId: attackTarget.id });
        }
        // If not in rampart, move to target
        if (!moveTarget) {
            return moveToAssignedPosition(creep, trace, kernel);
        }
        // If we are less then 2/3 health, move back and heal
        if (creep.hits < creep.hitsMax * 0.666) {
            trace.info('too damaged, flee');
            const base = getCreepBase(kernel, creep);
            if (base) {
                const result = creep.moveTo(base.origin);
                trace.info('moving to base origin to heal', { result });
                return RUNNING;
            }
        }
        if (creep.pos.getRangeTo(moveTarget) <= 3) {
            trace.info('target in range');
            return RUNNING;
        }
        const result = move(creep, moveTarget, 3);
        trace.info('move to target', { result, moveTarget });
        return RUNNING;
    }),
]);
const moveToAssignedPosition = (creep, trace, kernel) => {
    let position = null;
    // Check if creep knows last known position
    const positionString = creep.memory[MEMORY_ASSIGN_ROOM_POS] || null;
    if (positionString) {
        const posArray = positionString.split(',');
        if (posArray && posArray.length === 3) {
            position = new RoomPosition(posArray[0], posArray[1], posArray[2]);
        }
        else {
            trace.info('invalid position string', { positionString });
        }
    }
    else {
        trace.info('failed to get position string');
    }
    // If don't have a last known position, go to parking lot
    if (!position) {
        const base = getCreepBase(kernel, creep);
        if (base) {
            position = base.parking;
        }
        else {
            trace.info('could not get creep base config');
        }
    }
    if (!position) {
        trace.info('not able to determine destination, failing');
        return FAILURE;
    }
    // Check if we are at the destination
    if (creep.pos.getRangeTo(position) < 1) {
        trace.info('reached last known position or parking lot, waiting...');
        return SUCCESS;
    }
    // Move to destination
    const result = move(creep, position, 1);
    trace.info('move to last known hostile position or parking lot', { result, position });
    return RUNNING;
};
const move = (creep, target, range = 3) => {
    const result = creep.moveTo(target, { visualizePathStyle: { stroke: '#ffffff' }, range });
    if (result === ERR_NO_BODYPART) {
        return FAILURE;
    }
    if (result === ERR_INVALID_TARGET) {
        return FAILURE;
    }
    return SUCCESS;
};
const roleDefender = {
    run: rootNode('defender', behaviorBoosts(behavior$a)),
};

const behavior$9 = sequenceNode('defender_drone_root', [
    leafNode('move_node', (creep, trace) => {
        const x = creep.memory[MEMORY_POSITION_X];
        const y = creep.memory[MEMORY_POSITION_Y];
        const roomId = creep.memory[MEMORY_POSITION_ROOM];
        if (!x || !y || !roomId) {
            trace.info('missing position data', { x, y, roomId });
            return SUCCESS;
        }
        const position = new RoomPosition(x, y, roomId);
        if (creep.pos.isEqualTo(position)) {
            trace.info('creep at position', { position });
            return SUCCESS;
        }
        let result = null;
        if (creep.room.name != roomId) {
            result = creep.moveTo(position, {
                reusePath: 50,
                visualizePathStyle: { stroke: '#ffffff' }
            });
        }
        else {
            let ignoreCreeps = false;
            if (creep.memory[MEMORY_DEFENSE_IN_POSITION]) {
                ignoreCreeps = true;
            }
            result = creep.moveTo(position, {
                reusePath: 20,
                ignoreCreeps,
                visualizePathStyle: { stroke: '#ffffff' }
            });
        }
        trace.info('move to', { result, position });
        return SUCCESS;
    }),
    leafNode('attack_node', (creep, trace) => {
        const attack = creep.memory[MEMORY_ATTACK];
        if (!attack) {
            trace.info('no attack target');
            return SUCCESS;
        }
        const target = Game.getObjectById(attack);
        if (!target) {
            trace.info('target gone, did it die?', { attack });
            return SUCCESS;
        }
        if (!creep.pos.inRangeTo(target, 3)) {
            trace.info('target not in range attack range', { target });
            return SUCCESS;
        }
        const result = creep.rangedAttack(target);
        trace.info('range attack', { result, target });
        return SUCCESS;
    }),
    leafNode('heal_node', (creep, trace) => {
        const heal = creep.memory[MEMORY_HEAL];
        if (!heal) {
            trace.info('no heal target');
            return SUCCESS;
        }
        const target = Game.getObjectById(heal);
        if (!creep.pos.inRangeTo(target, 1)) {
            trace.info('target not in heal range', { target });
            return SUCCESS;
        }
        const result = creep.heal(target);
        trace.info('heal', { result, target });
        return SUCCESS;
    }),
]);
const roleDefenderDrone = {
    run: rootNode('defender_drone', behaviorBoosts(behavior$9)),
};

const getHaulTaskFromBaseTopic = leafNode('pick_haul_task', (creep, trace, kernel) => {
    // lookup base from kingdom
    const baseId = creep.memory[MEMORY_BASE];
    const base = kernel.getPlanner().getBaseById(baseId);
    if (!base) {
        trace.error('could not find base', { name: creep.name, memory: creep.memory });
        creep.suicide();
        return FAILURE;
    }
    // get next haul task
    const task = kernel.getTopics().getNextRequest(getBaseHaulerTopic(baseId));
    if (!task) {
        trace.info('no haul task');
        return FAILURE;
    }
    storeHaulTask(creep, task, trace);
    return SUCCESS;
});
const storeHaulTask = (creep, task, trace) => {
    trace.log('store haul task', { task });
    // set task details
    creep.memory[TASK_ID] = task.details[TASK_ID];
    creep.memory[MEMORY_TASK_TYPE] = TASK_HAUL;
    creep.memory[MEMORY_HAUL_PICKUP] = task.details[MEMORY_HAUL_PICKUP];
    creep.memory[MEMORY_HAUL_RESOURCE] = task.details[MEMORY_HAUL_RESOURCE];
    if (task.details[MEMORY_HAUL_AMOUNT]) {
        creep.memory[MEMORY_HAUL_AMOUNT] = task.details[MEMORY_HAUL_AMOUNT];
    }
    else {
        // Clear this, "needs energy" task was limiting regular haul tasks
        delete creep.memory[MEMORY_HAUL_AMOUNT];
    }
    creep.memory[MEMORY_HAUL_DROPOFF] = task.details[MEMORY_HAUL_DROPOFF];
    // const taskId = creep.memory[MEMORY.TASK_ID] || '?';
    // creep.say(taskId);
};
const clearTask = leafNode('clear_haul_task', (creep, _trace, _kingdom) => {
    delete creep.memory[TASK_ID];
    delete creep.memory[MEMORY_TASK_TYPE];
    delete creep.memory[MEMORY_HAUL_PICKUP];
    delete creep.memory[MEMORY_HAUL_RESOURCE];
    delete creep.memory[MEMORY_HAUL_AMOUNT];
    delete creep.memory[MEMORY_HAUL_DROPOFF];
    delete creep.memory[MEMORY_DESTINATION];
    return SUCCESS;
});
const loadCreep = leafNode('load_resource', (creep, trace, _kingdom) => {
    if (creep.store.getFreeCapacity() === 0) {
        trace.info('creep is full');
        return SUCCESS;
    }
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const pickup = Game.getObjectById(creep.memory[MEMORY_HAUL_PICKUP]);
    if (!pickup) {
        creep.say('⬆️❌');
        trace.error('could not find pickup', { id: creep.memory[MEMORY_HAUL_PICKUP] });
        return FAILURE;
    }
    const resource = creep.memory[MEMORY_HAUL_RESOURCE] || undefined;
    let amount = creep.memory[MEMORY_HAUL_AMOUNT] || undefined;
    let result = null;
    if (pickup instanceof Resource) {
        result = creep.pickup(pickup);
        trace.info('pickup resource', {
            pickup: pickup.id,
        });
    }
    else {
        const structure = pickup;
        if (amount > creep.store.getFreeCapacity(resource)) {
            amount = creep.store.getFreeCapacity(resource);
        }
        if (amount > structure.store.getUsedCapacity(resource)) {
            amount = structure.store.getUsedCapacity(resource);
        }
        // If we are seeing a specific amount, we are done when we have that amount in the hold
        if (creep.store.getUsedCapacity(resource) >= amount) {
            return SUCCESS;
        }
        if (amount === 0) {
            trace.error('zero amount', { resource, amount, creep, pickup });
            return FAILURE;
        }
        result = creep.withdraw(structure, resource, amount);
        trace.info('withdraw resource', {
            structure: structure.id,
            resource,
            amount,
            result,
        });
    }
    if (result === ERR_INVALID_ARGS) {
        trace.error('invalid args', { resource, amount, pickup });
        return FAILURE;
    }
    if (result === ERR_FULL) {
        trace.notice('full', { resource, amount, pickup });
        return SUCCESS;
    }
    if (result === ERR_NOT_ENOUGH_RESOURCES) {
        trace.notice('not enough resources', { resource, amount, pickup });
        return SUCCESS;
    }
    if (result !== OK) {
        trace.error('could not load resource', { result, resource, amount, pickup });
        return FAILURE;
    }
    // If we do not wait until next tick, the creep will not
    // know it's full
    return RUNNING;
});
// TODO marge these to into a generic one that takes a memory key to an object id
const emptyCreep$1 = leafNode('empty_creep', (creep, trace, _kingdom) => {
    const destination = Game.getObjectById(creep.memory[MEMORY_HAUL_DROPOFF]);
    if (!destination) {
        creep.say('⬇️❌');
        trace.error('no dump destination', { name: creep.name, memory: creep.memory });
        return FAILURE;
    }
    const resources = Object.keys(creep.store);
    const resource = resources.pop();
    const result = creep.transfer(destination, resource);
    trace.info('transfer result', { result, resource, resources });
    if (result === ERR_FULL) {
        trace.warn('transfer error: full', { result, resource, resources });
        const dropResult = creep.drop(resource);
        trace.info('drop result', { result: dropResult, resource, resources });
    }
    else if (result !== OK) {
        trace.error('transfer error', { result, resource, resources });
        return FAILURE;
    }
    // We have more resources to unload
    if (resources.length > 0) {
        trace.info('more do dump', { resources });
        return RUNNING;
    }
    return SUCCESS;
});
const emptyToDestination = leafNode('empty_creep_to_destination', (creep, trace, _kingdom) => {
    const destinationId = creep.memory[MEMORY_DESTINATION];
    const destination = Game.getObjectById(destinationId);
    if (!destination) {
        creep.say('⬇️❌');
        trace.error('no dump destination', { name: creep.name, memory: creep.memory });
        return FAILURE;
    }
    const resources = Object.keys(creep.store);
    const resource = resources.pop();
    const result = creep.transfer(destination, resource);
    trace.info('transfer result', { result, resource, resources });
    if (result === ERR_FULL) {
        trace.info('transfer error: full', { result, resource, resources });
    }
    else if (result !== OK) {
        trace.error('transfer error', { result, resource, resources });
        return FAILURE;
    }
    // We have more resources to unload
    if (resources.length > 0) {
        trace.info('more do dump', { resources });
        return RUNNING;
    }
    return SUCCESS;
});

const roadWorker = (behaviorNode) => {
    return sequenceAlwaysNode('build_repair_road_and_do_behavior', [
        leafNode('build_repair_road', (creep, trace, _kernel) => {
            // If energy too low, do not build/repair roads
            if (creep.store.getUsedCapacity(RESOURCE_ENERGY) <= creep.store.getCapacity(RESOURCE_ENERGY) * 0.8) {
                return SUCCESS;
            }
            // If work parts, do not build/repair roads
            if (creep.getActiveBodyparts(WORK) === 0) {
                return SUCCESS;
            }
            // Check for road to repair
            const road = creep.pos.lookFor(LOOK_STRUCTURES).find((structure) => {
                return structure.structureType === STRUCTURE_ROAD;
            });
            if (road && road.hitsMax - road.hits > 100) {
                const result = creep.repair(road);
                if (result !== OK) {
                    trace.error('repair error', {
                        result,
                        roadPosition: [road.pos.x, road.pos.y, road.pos.roomName].join(','),
                    });
                }
                return SUCCESS;
            }
            // Check for road to build
            const constructionSite = creep.pos.lookFor(LOOK_CONSTRUCTION_SITES).find((structure) => {
                return structure.structureType === STRUCTURE_ROAD;
            });
            if (constructionSite) {
                const result = creep.build(constructionSite);
                if (result !== OK) {
                    trace.error('build error', { result });
                }
                return SUCCESS;
            }
            return SUCCESS;
        }),
        behaviorNode,
    ]);
};

const selectNextTaskOrPark = selectorNode('pick_something', [
    leafNode('pick_haul_task', (creep, trace, kernel) => {
        // lookup base from kernel
        const base = getCreepBase(kernel, creep);
        if (!base) {
            trace.error('could not find creep base', { name: creep.name, memory: creep.memory });
            creep.suicide();
            return;
        }
        // get next haul task
        const task = kernel.getTopics().getMessageOfMyChoice(getBaseDistributorTopic(base.id), (messages) => {
            const sorted = _.sortByOrder(messages, [
                'priority',
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                (message) => {
                    const dropoff = Game.getObjectById(message.details[MEMORY_HAUL_DROPOFF]);
                    if (!dropoff) {
                        return -1;
                    }
                    return creep.pos.getRangeTo(dropoff);
                },
            ], ['desc', 'asc']);
            trace.info('sorted core haul tasks', { sorted });
            if (sorted.length) {
                return sorted[0];
            }
            return null;
        });
        if (!task) {
            trace.info('no haul task');
            return FAILURE;
        }
        storeHaulTask(creep, task, trace);
        return SUCCESS;
    }),
    parkingLot,
]);
const emptyCreep = leafNode('empty_creep', (creep, trace, _kernel) => {
    if (creep.store.getUsedCapacity() === 0) {
        return SUCCESS;
    }
    const destinationId = creep.memory[MEMORY_DESTINATION];
    const destination = Game.getObjectById(destinationId);
    if (!destination) {
        trace.error('could not find destination', { destinationId });
        return FAILURE;
    }
    const desiredResource = creep.memory[MEMORY_HAUL_RESOURCE];
    if (!desiredResource) {
        trace.error('no desired resource', { creep });
        return FAILURE;
    }
    const taskId = creep.memory[TASK_ID];
    let resources = Object.keys(creep.store);
    if (!taskId.startsWith('lu-')) {
        // If not a link unload, we should unload what isnt needed
        resources = _.difference(resources, [desiredResource]);
    }
    if (!resources.length) {
        return SUCCESS;
    }
    const resource = resources.pop();
    const result = creep.transfer(destination, resource);
    if (result !== OK) {
        trace.error('transfer error', { result, resource, resources });
        return FAILURE;
    }
    trace.info('transfer result', { result, resource, resources });
    // We have more resources to unload
    if (resources.length > 0) {
        return RUNNING;
    }
    return SUCCESS;
});
const unloadIfNeeded = selectorNode('unload_creep_if_needed', [
    leafNode('check_if_unload_needed', (creep, trace, kernel) => {
        const desiredResource = creep.memory[MEMORY_HAUL_RESOURCE];
        if (!desiredResource) {
            throw new Error('Hauler task missing desired resource');
        }
        const taskId = creep.memory[TASK_ID];
        const loadedResources = Object.keys(creep.store);
        let toUnload = loadedResources;
        if (!taskId.startsWith('lu-')) {
            // If not a link unload, we should unload what isnt needed
            toUnload = _.difference(loadedResources, [desiredResource]);
        }
        trace.info('to unload', { loadedResources, desiredResource, toUnload });
        if (toUnload.length) {
            const base = getCreepBase(kernel, creep);
            if (!base) {
                throw new Error('Unable to get room for creep');
            }
            const reserve = getStructureForResource(base, toUnload[0]);
            creep.memory[MEMORY_DESTINATION] = reserve.id;
            trace.info('unloading at', {
                loaded: JSON.stringify(loadedResources),
                toUnload: JSON.stringify(toUnload),
                desired: JSON.stringify([desiredResource]),
                dropoff: reserve.id,
            });
            return FAILURE;
        }
        return SUCCESS;
    }),
    sequenceNode('move_and_unload', [
        moveToCreepMemory(MEMORY_DESTINATION, 1, false, 25, 500),
        emptyCreep,
    ]),
]);
const loadIfNeeded = selectorNode('load_creep_if_needed', [
    leafNode('has_resource', (creep, trace, _kernel) => {
        const dropoffId = creep.memory[MEMORY_HAUL_DROPOFF];
        if (!dropoffId) {
            throw new Error('Hauler task missing dropoff');
        }
        const dropoff = Game.getObjectById(dropoffId);
        if (!dropoff) {
            trace.error('no dropoff', { dropoffId });
            return FAILURE;
        }
        const resource = creep.memory[MEMORY_HAUL_RESOURCE];
        if (!resource) {
            trace.error('no resource', { creep });
            return FAILURE;
        }
        let amount = creep.memory[MEMORY_HAUL_AMOUNT];
        if (!amount) {
            const taskId = creep.memory[TASK_ID] || 'unknown task id';
            throw new Error(`Hauler task missing amount: ${taskId}`);
        }
        if (creep.store.getUsedCapacity(resource) > amount * 0.5) {
            return SUCCESS;
        }
        if (amount >= creep.store.getCapacity(resource)) {
            amount = creep.store.getCapacity(resource);
            creep.memory[MEMORY_HAUL_AMOUNT] = amount;
        }
        if (amount < creep.store.getCapacity(resource) && resource === RESOURCE_ENERGY) {
            creep.memory[MEMORY_HAUL_AMOUNT] = creep.store.getCapacity(resource);
        }
        trace.info('has resource', {
            resource,
            amount,
            creepAmount: creep.store.getUsedCapacity(resource),
        });
        if (creep.store.getUsedCapacity(resource) >= amount) {
            return SUCCESS;
        }
        if (dropoff instanceof StructureExtension) {
            // Update amount to be a full creep so that we can fill multiple extensions
            creep.memory[MEMORY_HAUL_AMOUNT] = creep.store.getCapacity(RESOURCE_ENERGY);
        }
        return FAILURE;
    }),
    sequenceNode('get_resource', [
        moveToCreepMemory(MEMORY_HAUL_PICKUP, 1, false, 25, 1000),
        loadCreep,
    ]),
]);
const deliver = sequenceNode('deliver', [
    leafNode('use_memory_dropoff', (creep, _trace, _kernel) => {
        const dropoff = creep.memory[MEMORY_HAUL_DROPOFF];
        if (dropoff) {
            setDestination(creep, dropoff);
            return SUCCESS;
        }
        return FAILURE;
    }),
    moveToCreepMemory(MEMORY_DESTINATION, 1, false, 25, 250),
    leafNode('empty_creep', (creep, trace, _kernel) => {
        if (creep.store.getUsedCapacity() === 0) {
            return SUCCESS;
        }
        const destination = Game.getObjectById(creep.memory[MEMORY_DESTINATION]);
        if (!destination) {
            return FAILURE;
        }
        const resource = Object.keys(creep.store).pop();
        const result = creep.transfer(destination, resource);
        trace.info('transfer', { resource, result });
        if (result === ERR_FULL) {
            return SUCCESS;
        }
        if (result === ERR_NOT_ENOUGH_RESOURCES) {
            return SUCCESS;
        }
        if (result === ERR_INVALID_TARGET) {
            return SUCCESS;
        }
        if (result != OK) {
            return FAILURE;
        }
        return SUCCESS;
    }),
]);
const behavior$8 = sequenceNode('core_task_or_extensions', [
    clearTask,
    selectNextTaskOrPark,
    unloadIfNeeded,
    loadIfNeeded,
    deliver,
]);
const roleDistributor = {
    run: rootNode('distributor', behaviorBoosts(roadWorker(behavior$8))),
};

const explorerPolicy = {
    room: {
        avoidHostileRooms: false,
        avoidFriendlyRooms: false,
        avoidRoomsWithKeepers: false,
        avoidRoomsWithTowers: false,
        avoidUnloggedRooms: false,
        sameRoomStatus: true,
        costMatrixType: AllowedCostMatrixTypes.COMMON,
    },
    destination: {
        range: 1,
    },
    path: {
        allowIncomplete: true,
        maxSearchRooms: 12,
        maxOps: 5000,
        maxPathRooms: 6,
        ignoreCreeps: true,
        sourceKeeperBuffer: 3,
        hostileCreepBuffer: 8,
    },
};
const behavior$7 = sequenceNode('explorer_root', [
    leafNode('select_next_room', (creep, trace, kingdom) => {
        // Don't notify me when creep wonders into hostile room
        creep.notifyWhenAttacked(false);
        // If creep is assigned room it is not in, then move to that room
        if (creep.memory[MEMORY_ASSIGN_ROOM] &&
            creep.pos.roomName !== creep.memory[MEMORY_ASSIGN_ROOM]) {
            trace.info('creep is assigned room already, moving to it');
            return SUCCESS;
        }
        const nextRoom = getNextRoomToScout(kingdom, creep);
        if (nextRoom) {
            trace.notice(`found next room to scout: ${nextRoom}`);
            creep.memory[MEMORY_ASSIGN_ROOM] = nextRoom;
            return SUCCESS;
        }
        trace.info('no next room to scout');
        const currentRoomStatus = Game.map.getRoomStatus(creep.room.name);
        let exits = Object.values(Game.map.describeExits(creep.room.name));
        exits = exits.filter((exit) => {
            const exitStatus = Game.map.getRoomStatus(exit);
            return currentRoomStatus.status === exitStatus.status;
        });
        // Unvisited exists first
        const unvisited = exits.filter((roomName) => {
            const entry = kingdom.getScribe().getRoomById(roomName);
            return !entry;
        });
        if (unvisited.length > 0) {
            trace.info('unvisited exits', { unvisited });
            creep.memory[MEMORY_ASSIGN_ROOM] = unvisited[0];
            return SUCCESS;
        }
        // Try to pick oldest room first if it hasn't been visited recently
        const ageSorted = _.sortBy(exits, (roomName) => {
            const entry = kingdom.getScribe().getRoomById(roomName);
            return entry.lastUpdated;
        });
        const oldestRoom = kingdom.getScribe().getRoomById(ageSorted[0]);
        if (Game.time - oldestRoom.lastUpdated > 50) {
            trace.info('age sorted exits', { ageSorted });
            creep.memory[MEMORY_ASSIGN_ROOM] = ageSorted[0];
            return SUCCESS;
        }
        // All exits are recently visited, pick random exit
        const shuffled = _.shuffle(exits);
        trace.info('next room', { next: shuffled[0] });
        creep.memory[MEMORY_ASSIGN_ROOM] = shuffled[0];
        return SUCCESS;
    }),
    cachedMoveToRoom(MEMORY_ASSIGN_ROOM, explorerPolicy),
    leafNode('move_into_room', (creep, trace, kingdom) => {
        // Record room
        kingdom.getScribe().updateRoom(kingdom, creep.room, trace);
        // Move one step into the room
        creep.moveTo(new RoomPosition(25, 25, creep.room.name), { maxOps: 100 });
        return SUCCESS;
    }),
]);
const roleExplorer = {
    run: rootNode('explorer', behavior$7),
};

const harvest$1 = leafNode('fill_creep', (creep, trace, _kingdom) => {
    const destination = Game.getObjectById(creep.memory[MEMORY_SOURCE]);
    if (!destination) {
        return FAILURE;
    }
    if (destination.energy === 0) {
        return RUNNING;
    }
    const result = creep.harvest(destination);
    trace.info('harvest', { result });
    if (creep.store.getFreeCapacity() === 0) {
        return SUCCESS;
    }
    // NOTICE switched this to FAILURE from RUNNING, this may cause problems
    if (result === ERR_NOT_ENOUGH_RESOURCES) {
        return RUNNING;
    }
    if (result === OK) {
        return RUNNING;
    }
    return FAILURE;
});

const selectRoomDropoff = selectorNode('selectRoomDropoff', [
    leafNode('use_memory_dropoff', (creep) => {
        const dropoff = creep.memory[MEMORY_HAUL_DROPOFF];
        if (dropoff) {
            setDestination(creep, dropoff);
            return SUCCESS;
        }
        return FAILURE;
    }),
    leafNode('pick_adjacent_container', (creep, _trace, _kernel) => {
        const role = creep.memory[MEMORY_ROLE];
        // haulers should pick containers near the spawner
        // TODO this is hacky and feels bad
        if (role && (role === WORKER_DISTRIBUTOR || role === WORKER_HAULER)) {
            return FAILURE;
        }
        const targets = creep.pos.findInRange(FIND_STRUCTURES, 2, {
            filter: (structure) => {
                return structure.structureType == STRUCTURE_CONTAINER &&
                    structure.store.getFreeCapacity(RESOURCE_ENERGY) > 0 &&
                    structure.isActive();
            },
        });
        if (!targets || !targets.length) {
            return FAILURE;
        }
        setDestination(creep, targets[0].id);
        return SUCCESS;
    }),
    leafNode('pick_adjacent_link', (creep, _trace, _kernel) => {
        const role = creep.memory[MEMORY_ROLE];
        if (role && role === WORKER_DISTRIBUTOR) {
            return FAILURE;
        }
        const targets = creep.pos.findInRange(FIND_STRUCTURES, 2, {
            filter: (structure) => {
                // TODO things seeking to gain energy should use another function
                return structure.structureType == STRUCTURE_LINK &&
                    structure.store.getFreeCapacity(RESOURCE_ENERGY) > 0 &&
                    structure.isActive();
            },
        });
        if (!targets || !targets.length) {
            return FAILURE;
        }
        setDestination(creep, targets[0].id);
        return SUCCESS;
    }),
    leafNode('pick_storage', (creep, trace, kernel) => {
        var _a;
        const role = creep.memory[MEMORY_ROLE];
        if (role && role === WORKER_DISTRIBUTOR) {
            return FAILURE;
        }
        const base = getCreepBase(kernel, creep);
        if (!base) {
            trace.error('could not find creep base', { name: creep.name, memory: creep.memory });
            creep.suicide();
            return FAILURE;
        }
        const room = getBasePrimaryRoom(base);
        if (!room) {
            return FAILURE;
        }
        if (!room.storage || !((_a = room.storage) === null || _a === void 0 ? void 0 : _a.isActive())) {
            return FAILURE;
        }
        const baseCreeps = kernel.getCreepsManager().getCreepsByBase(base.id);
        const distributors = _.filter(baseCreeps, (creep) => {
            return creep.memory[MEMORY_ROLE] === WORKER_DISTRIBUTOR;
        });
        if (!distributors.length) {
            return FAILURE;
        }
        setDestination(creep, room.storage.id);
        return SUCCESS;
    }),
    leafNode('pick_tower', (creep, _trace, _kernel) => {
        const role = creep.memory[MEMORY_ROLE];
        if (role && role === WORKER_DISTRIBUTOR) {
            return FAILURE;
        }
        const targets = creep.room.find(FIND_STRUCTURES, {
            filter: (structure) => {
                return structure.structureType == STRUCTURE_TOWER &&
                    structure.store.getFreeCapacity(RESOURCE_ENERGY) > 0 &&
                    structure.isActive();
            },
        });
        if (!targets || !targets.length) {
            return FAILURE;
        }
        setDestination(creep, targets[0].id);
        return SUCCESS;
    }),
    leafNode('pick_container', (creep, trace, kernel) => {
        const role = creep.memory[MEMORY_ROLE];
        if (role && role === WORKER_DISTRIBUTOR) {
            return FAILURE;
        }
        const base = getCreepBase(kernel, creep);
        if (!base) {
            trace.error('could not find creep base', { name: creep.name, memory: creep.memory });
            creep.suicide();
            return FAILURE;
        }
        const room = getBasePrimaryRoom(base);
        if (!room) {
            trace.error('could not find primary room', { baseId: base.id, roomName: base.primary });
            creep.suicide();
            return FAILURE;
        }
        const distributors = room.find(FIND_MY_CREEPS, {
            filter: (creep) => {
                return creep.memory[MEMORY_ROLE] === WORKER_DISTRIBUTOR;
            },
        });
        if (!distributors.length) {
            return FAILURE;
        }
        const target = getStructuresWithResource(base, RESOURCE_ENERGY);
        if (!target.length) {
            return FAILURE;
        }
        setDestination(creep, target[0].id);
        return SUCCESS;
    }),
    leafNode('pick_spawner_extension', (creep, trace, kernel) => {
        const base = getCreepBase(kernel, creep);
        if (!base) {
            trace.error('could not find creep base', { name: creep.name, memory: creep.memory });
            creep.suicide();
            return FAILURE;
        }
        const room = getBasePrimaryRoom(base);
        if (!room) {
            trace.error('could not find primary room', { baseId: base.id, roomName: base.primary });
            creep.suicide();
            return FAILURE;
        }
        const targets = room.find(FIND_STRUCTURES, {
            filter: (structure) => {
                return (structure.structureType == STRUCTURE_EXTENSION ||
                    structure.structureType == STRUCTURE_SPAWN) &&
                    structure.store.getFreeCapacity(RESOURCE_ENERGY) > 0;
            },
        });
        if (!targets.length) {
            return FAILURE;
        }
        setDestination(creep, targets[0].id);
        return SUCCESS;
    }),
]);

/**
 * Harvester creep
 *
 * Harvests mineral nodes.
 */
const behavior$6 = sequenceNode('haul_energy', [
    cachedMoveToMemoryPos(MEMORY_SOURCE_POSITION, 0, commonPolicy),
    setCommuteDuration,
    harvest$1,
    sequenceNode('dump', [
        selectRoomDropoff,
        cachedMoveToMemoryObjectId(MEMORY_DESTINATION, 1, commonPolicy),
        leafNode('empty_creep', (creep, trace, _kingdom) => {
            if (creep.store.getUsedCapacity() === 0) {
                return SUCCESS;
            }
            const destination = Game.getObjectById(creep.memory[MEMORY_DESTINATION]);
            if (!destination) {
                trace.info('no destination', { destination: creep.memory[MEMORY_DESTINATION] });
                return SUCCESS;
            }
            const resource = Object.keys(creep.store).pop();
            const result = creep.transfer(destination, resource);
            trace.info('transfer', { result, resource });
            if (result !== OK) {
                trace.error('transfer failed', { result });
                return FAILURE;
            }
            return RUNNING;
        }),
    ]),
]);
const roleHarvester = {
    run: rootNode('hauler', behaviorBoosts(roadWorker(behavior$6))),
};

const haulerPolicy = {
    room: {
        avoidHostileRooms: true,
        avoidFriendlyRooms: false,
        avoidRoomsWithKeepers: true,
        avoidRoomsWithTowers: false,
        avoidUnloggedRooms: false,
        sameRoomStatus: true,
        costMatrixType: AllowedCostMatrixTypes.HAULER,
    },
    destination: {
        range: 1,
    },
    path: {
        allowIncomplete: true,
        maxSearchRooms: 12,
        maxOps: 6000,
        maxPathRooms: 6,
        ignoreCreeps: true,
        sourceKeeperBuffer: 3,
        preferRoadSites: true,
    },
};
const behavior$5 = sequenceNode('haul_task', [
    clearTask,
    selectorNode('pick_something', [
        getHaulTaskFromBaseTopic,
        parkingLot,
    ]),
    cachedMoveToMemoryObjectId(MEMORY_HAUL_PICKUP, 1, haulerPolicy),
    loadCreep,
    cachedMoveToMemoryObjectId(MEMORY_HAUL_DROPOFF, 1, haulerPolicy),
    emptyCreep$1,
]);
const roleHauler = {
    run: rootNode('hauler', behaviorBoosts(roadWorker(behavior$5))),
};

const harvest = leafNode('harvest', (creep, trace, _kingdom) => {
    // If miner is full, then dump
    if (!creep.store.getFreeCapacity(RESOURCE_ENERGY)) {
        const link = creep.pos.findInRange(FIND_MY_STRUCTURES, 1, {
            filter: (structure) => {
                return structure.structureType === STRUCTURE_LINK;
            },
        })[0];
        if (link) {
            const amount = _.min([
                link.store.getFreeCapacity(RESOURCE_ENERGY),
                creep.store.getUsedCapacity(RESOURCE_ENERGY),
            ]);
            if (amount) {
                const result = creep.transfer(link, RESOURCE_ENERGY, amount);
                trace.info('creep transfer to link', { result, amount });
                return RUNNING;
            }
        }
    }
    const destinationId = creep.memory[MEMORY_SOURCE];
    const destination = Game.getObjectById(destinationId);
    if (!destination) {
        trace.info('destination not found', { destinationId });
        return FAILURE;
    }
    if (destination.energy === 0) {
        return FAILURE;
    }
    const result = creep.harvest(destination);
    trace.info('harvest result', { result });
    if (result === ERR_NOT_IN_RANGE) {
        trace.info('not in range result', { result, destinationId });
        return FAILURE;
    }
    if (creep.store.getFreeCapacity() === 0) {
        trace.info('creep has no free capacity', {});
        return SUCCESS;
    }
    if (result === ERR_NOT_ENOUGH_RESOURCES) {
        trace.info('not enough resources', { result });
        return FAILURE;
    }
    if (result === OK) {
        trace.info('ok result', { result });
        return RUNNING;
    }
    trace.info('harvest no ok', { result });
    return FAILURE;
});
const buildNearbySites = leafNode('build_nearby_sites', (creep, trace, _kingdom) => {
    // if creep out of energy, refill if some available
    if (creep.store.getUsedCapacity(RESOURCE_ENERGY) === 0) {
        return refillMiner(creep, trace);
    }
    const sites = creep.pos.findInRange(FIND_CONSTRUCTION_SITES, 1);
    if (!sites.length) {
        trace.info('no construction site', {});
        return FAILURE;
    }
    const result = creep.build(sites[0]);
    trace.info('build result', { result });
    return RUNNING;
});
const repairContainer = leafNode('repair_container', (creep, trace, _kingdom) => {
    // if creep out of energy, refill if some available
    if (creep.store.getUsedCapacity(RESOURCE_ENERGY) === 0) {
        return refillMiner(creep, trace);
    }
    const container = creep.pos.lookFor(LOOK_STRUCTURES).find((site) => {
        return site.structureType === STRUCTURE_CONTAINER;
    });
    if (!container) {
        trace.info('no container', { pos: creep.pos });
        return FAILURE;
    }
    if (container.hits === container.hitsMax) {
        trace.info('container full health', { hits: container.hits });
        return FAILURE;
    }
    const result = creep.repair(container);
    trace.info('repair result', { result });
    return RUNNING;
});
const moveEnergyToLink = leafNode('move_energy_to_link', (creep, trace, _kingdom) => {
    const source = Game.getObjectById(creep.memory[MEMORY_SOURCE]);
    if (!source) {
        trace.info('source not found');
        return FAILURE;
    }
    if (source.energy > 0) {
        trace.info('source has energy, stop trying to load link');
        return SUCCESS;
    }
    const link = creep.pos.findInRange(FIND_MY_STRUCTURES, 1, {
        filter: (structure) => {
            return structure.structureType === STRUCTURE_LINK;
        },
    })[0];
    if (!link) {
        trace.info('no link');
        return FAILURE;
    }
    let target = null;
    const freeCapacity = creep.store.getFreeCapacity(RESOURCE_ENERGY);
    trace.info('creep free capacity', { freeCapacity });
    if (freeCapacity > 0) {
        target = creep.pos.findInRange(FIND_STRUCTURES, 1, {
            filter: (structure) => {
                return structure.structureType === STRUCTURE_CONTAINER &&
                    structure.store.getUsedCapacity(RESOURCE_ENERGY) > 0;
            },
        })[0];
        if (target) {
            const result = creep.withdraw(target, RESOURCE_ENERGY);
            trace.info('withdraw energy', { result, targetId: target.id });
            return RUNNING;
        }
        target = creep.pos.findInRange(FIND_DROPPED_RESOURCES, 1, {
            filter: (resource) => {
                return resource.resourceType === RESOURCE_ENERGY;
            },
        })[0];
        if (target) {
            const result = creep.pickup(target);
            trace.info('pickup energy', { result, targetId: target.id });
            return RUNNING;
        }
        trace.info('found no energy to pickup');
        return FAILURE;
    }
    if (link.store.getFreeCapacity(RESOURCE_ENERGY) > 0) {
        const result = creep.transfer(link, RESOURCE_ENERGY);
        trace.info('transfer energy to link', { result });
    }
    return RUNNING;
});
const waitUntilSourceReady = leafNode('wait_until_ready', (creep) => {
    const source = Game.getObjectById(creep.memory[MEMORY_SOURCE]);
    if (!source) {
        return FAILURE;
    }
    if (creep.pos.getRangeTo(source) > 1) {
        return FAILURE;
    }
    if (source.energy < 1) {
        return RUNNING;
    }
    return SUCCESS;
});
// Refill creep with nearby dropped resources or container
function refillMiner(creep, trace) {
    const dropped = creep.pos.findInRange(FIND_DROPPED_RESOURCES, 1, {
        filter: (resource) => {
            return resource.resourceType === RESOURCE_ENERGY;
        }
    })[0];
    if (dropped) {
        const result = creep.pickup(dropped);
        if (result !== OK) {
            trace.info('pickup not ok', { result });
            return FAILURE;
        }
        trace.info('pickup ok', { result });
        return RUNNING;
    }
    const container = creep.pos.findInRange(FIND_STRUCTURES, 1, {
        filter: (structure) => {
            return structure.structureType === STRUCTURE_CONTAINER;
        },
    })[0];
    if (container) {
        if (container.store.getUsedCapacity(RESOURCE_ENERGY) === 0) {
            trace.info('no energy in container', { store: container.store });
            return FAILURE;
        }
        const result = creep.withdraw(container, RESOURCE_ENERGY);
        if (result !== OK) {
            trace.info('withdraw not ok', { result });
            return FAILURE;
        }
        trace.info('withdraw ok', { result });
        return RUNNING;
    }
    trace.info('no energy to pick up', { pos: creep.pos });
    return FAILURE;
}
const behavior$4 = sequenceNode('mine_energy', [
    cachedMoveToMemoryPos(MEMORY_SOURCE_POSITION, 0, commonPolicy),
    setCommuteDuration,
    repeatUntilFailure('mine_until_failure', sequenceNode('get_energy_and_dump', [
        selectorNode('get_energy', [
            harvest,
            buildNearbySites,
            repairContainer,
            moveEnergyToLink,
            waitUntilSourceReady,
        ]),
    ])),
]);
const roleMiner = {
    run: rootNode('miner', behaviorBoosts(behavior$4)),
};

const selectStructureToRepair = leafNode('selectStructureToRepair', (creep, trace, kernel) => {
    const base = getCreepBase(kernel, creep);
    if (!base) {
        trace.error('no room on creep', { name: creep.name, memory: creep.memory });
        creep.suicide();
        return FAILURE;
    }
    const target = getNextDamagedStructure(base);
    if (!target) {
        return FAILURE;
    }
    trace.info('selected target', { target: target.id, pos: target.pos });
    setDestination(creep, target.id);
    return SUCCESS;
});
const repair = leafNode('repair_structure', (creep, trace, _kingdom) => {
    const destination = Game.getObjectById(creep.memory[MEMORY_DESTINATION]);
    if (!destination) {
        return FAILURE;
    }
    // TODO this should not be a failure, I need to makea RepeatCondition node
    if (destination.hits >= destination.hitsMax) {
        return SUCCESS;
    }
    const result = creep.repair(destination);
    trace.info('repairing', { destination: destination.id, result });
    if (result != OK) {
        return FAILURE;
    }
    return RUNNING;
});
const behavior$3 = sequenceNode('repair', [
    cachedMoveToRoom(MEMORY_ASSIGN_ROOM, commonPolicy),
    setCommuteDuration,
    getEnergy,
    repeatUntilConditionMet('repair_until_empty', (creep, _trace, _kingdom) => {
        if (creep.store.getUsedCapacity() === 0) {
            return true;
        }
        return false;
    }, sequenceNode('select_and_repair', [
        selectStructureToRepair,
        moveToCreepMemory(MEMORY_DESTINATION, 1, false, 50, 1000),
        repair,
    ])),
]);
const roleRepairer = {
    run: rootNode('repairer', behaviorBoosts(behavior$3)),
};

const behavior$2 = sequenceNode('reserver_root', [
    moveToShard(MEMORY_ASSIGN_SHARD),
    leafNode('set_controller_location', (creep, trace, kernel) => {
        var _a, _b;
        const assignedRoom = creep.memory[MEMORY_ASSIGN_ROOM];
        let posStr = [25, 25, assignedRoom].join(',');
        const roomEntry = kernel.getScribe().getRoomById(assignedRoom);
        if ((_a = roomEntry === null || roomEntry === void 0 ? void 0 : roomEntry.controller) === null || _a === void 0 ? void 0 : _a.pos) {
            const pos = (_b = roomEntry.controller) === null || _b === void 0 ? void 0 : _b.pos;
            posStr = [pos.x, pos.y, pos.roomName].join(',');
        }
        creep.memory[MEMORY_ASSIGN_ROOM_POS] = posStr;
        return SUCCESS;
    }),
    cachedMoveToMemoryPos(MEMORY_ASSIGN_ROOM_POS, 1, commonPolicy),
    repeatUntilSuccess('move_to_rc', leafNode('move', (creep) => {
        const room = creep.memory[MEMORY_ASSIGN_ROOM];
        // If creep doesn't have a harvest room assigned, we are done
        if (!room) {
            return SUCCESS;
        }
        if (!creep.room.controller) {
            return SUCCESS;
        }
        return moveTo(creep, creep.room.controller.pos, 1, false, 25, 1500);
    })),
    setCommuteDuration,
    repeatUntilSuccess('reserve', leafNode('claim', (creep, trace, kernel) => {
        var _a, _b, _c;
        const roomId = creep.memory[MEMORY_ASSIGN_ROOM];
        // If reserver doesn't have a room assigned, we are done
        if (!roomId) {
            trace.warn('no room assigned', { creep: creep.name });
            return SUCCESS;
        }
        // If the creep is not in the right room we are done
        if (creep.room.name !== roomId) {
            trace.warn('wrong room', { creep: creep.name, room: creep.room.name, assignedRoom: roomId });
            return SUCCESS;
        }
        const base = getCreepBase(kernel, creep);
        if (!base) {
            trace.error('no base config', { memory: creep.memory });
            creep.suicide();
            return FAILURE;
        }
        const room = creep.room;
        if (!room) {
            trace.info('unable to get creep room');
            return FAILURE;
        }
        if (!room.controller) {
            trace.info('no controller in room');
            return FAILURE;
        }
        const unowned = !((_a = room.controller) === null || _a === void 0 ? void 0 : _a.owner) && !((_b = room.controller) === null || _b === void 0 ? void 0 : _b.reservation);
        const claimedByMe = ((_c = room.controller) === null || _c === void 0 ? void 0 : _c.my) || false;
        const username = kernel.getPlanner().getUsername();
        const reservedByMe = room.controller && room.controller.reservation &&
            room.controller.reservation.username === username;
        const isPrimary = room.name === base.primary;
        const controller = room.controller;
        trace.info('reserver', {
            room: room.name,
            unowned,
            claimedByMe,
            reservedByMe,
            isPrimary,
            controller,
        });
        if (controller && controller.upgradeBlocked) {
            trace.info('upgrade/attack blocked', { ttl: room.controller.upgradeBlocked });
            return FAILURE;
        }
        if (!unowned && !claimedByMe && !reservedByMe) {
            const result = creep.attackController(creep.room.controller);
            trace.info('attackController', { result });
            if (result != OK) {
                return FAILURE;
            }
        }
        else {
            if (isPrimary) {
                const result = creep.claimController(creep.room.controller);
                trace.info('claimController', { result });
                if (result != OK) {
                    return FAILURE;
                }
            }
            else {
                const result = creep.reserveController(creep.room.controller);
                trace.info('reserveController', { result });
                if (result != OK) {
                    return FAILURE;
                }
            }
        }
        return RUNNING;
    })),
    updateSign,
]);
const roleReserver = {
    run: rootNode('reserver', behaviorBoosts(behavior$2)),
};

const controllerDumpPolicy = {
    room: {
        avoidHostileRooms: true,
        avoidFriendlyRooms: false,
        avoidRoomsWithKeepers: true,
        avoidRoomsWithTowers: false,
        avoidUnloggedRooms: false,
        sameRoomStatus: true,
        costMatrixType: AllowedCostMatrixTypes.COMMON,
    },
    destination: {
        range: 1,
    },
    path: {
        allowIncomplete: true,
        maxSearchRooms: 12,
        maxOps: 5000,
        maxPathRooms: 6,
        ignoreCreeps: true,
    },
};
const behavior$1 = sequenceNode('upgrader_root', [
    getSomeEnergy,
    moveToShard(MEMORY_ASSIGN_SHARD),
    leafNode('set_controller_location', (creep, trace, kingdom) => {
        var _a;
        const assignedRoom = creep.memory[MEMORY_ASSIGN_ROOM];
        const roomEntry = kingdom.getScribe().getRoomById(assignedRoom);
        if (!((_a = roomEntry === null || roomEntry === void 0 ? void 0 : roomEntry.controller) === null || _a === void 0 ? void 0 : _a.pos)) {
            trace.error('no controller pos', { assignedRoom });
            return FAILURE;
        }
        const pos = roomEntry.controller.pos;
        const posStr = [pos.x, pos.y, pos.roomName].join(',');
        creep.memory[MEMORY_ASSIGN_ROOM_POS] = posStr;
        return SUCCESS;
    }),
    cachedMoveToMemoryPos(MEMORY_ASSIGN_ROOM_POS, 3, controllerDumpPolicy),
    setCommuteDuration,
    repeatUntilSuccess('upgrade_until_empty', leafNode('upgrade_controller', (creep, trace, _kingdom) => {
        const result = creep.upgradeController(creep.room.controller);
        trace.info('upgrade result', { result });
        if (result == ERR_NOT_ENOUGH_RESOURCES) {
            return SUCCESS;
        }
        if (result != OK) {
            return FAILURE;
        }
        return RUNNING;
    })),
    updateSign,
]);
const roleUpgrader = {
    run: rootNode('upgrader', behaviorBoosts(roadWorker(behavior$1))),
};

/**
 * Worker creep
 *
 * Early game creep. Used when there is no storage. Replaced by more specialized
 * creeps when storage is built.
 *
 * Requirements:
 * - Process hauling tasks
 * - Move and pick up energy
 * - Move and drop off energy at spawn, extension
 * - If no drop offs with capacity, then build structures or upgrade controller
 *
 */
const selectDropoff = module.exports.selectRoomDropoff = selectorNode('selectRoomDropoff', [
    leafNode('use_memory_dropoff', (creep) => {
        const dropoff = creep.memory[MEMORY_HAUL_DROPOFF];
        if (dropoff) {
            setDestination(creep, dropoff);
            return SUCCESS;
        }
        return FAILURE;
    }),
    leafNode('pick_storage', (creep, trace, kernel) => {
        var _a;
        const base = getCreepBase(kernel, creep);
        if (!base) {
            trace.error('could not find creep base', { name: creep.name, memory: creep.memory });
            // creep.suicide();
            return FAILURE;
        }
        const room = getBasePrimaryRoom(base);
        if (!room) {
            trace.error('could not find base primary room', { base });
            return FAILURE;
        }
        if (!room.storage) {
            return FAILURE;
        }
        if (!((_a = room.storage) === null || _a === void 0 ? void 0 : _a.isActive())) {
            return FAILURE;
        }
        const distributors = _.filter(kernel.getCreepsManager().getCreepsByBase(base.id), (creep) => {
            return creep.memory[MEMORY_ROLE] === WORKER_DISTRIBUTOR;
        });
        if (!distributors.length) {
            return FAILURE;
        }
        setDestination(creep, room.storage.id);
        return SUCCESS;
    }),
    leafNode('pick_tower', (creep, _trace, _kernel) => {
        const targets = creep.room.find(FIND_STRUCTURES, {
            filter: (structure) => {
                return structure.structureType == STRUCTURE_TOWER &&
                    structure.store.getFreeCapacity(RESOURCE_ENERGY) > 0 &&
                    structure.isActive();
            },
        });
        if (!targets || !targets.length) {
            return FAILURE;
        }
        setDestination(creep, targets[0].id);
        return SUCCESS;
    }),
    leafNode('pick_spawner_extension', (creep, trace, kernel) => {
        const base = getCreepBase(kernel, creep);
        if (!base) {
            trace.error('could not find creep base', { name: creep.name, memory: creep.memory });
            creep.suicide();
            return FAILURE;
        }
        const room = getBasePrimaryRoom(base);
        if (!room) {
            return FAILURE;
        }
        const targets = room.find(FIND_STRUCTURES, {
            filter: (structure) => {
                return (structure.structureType == STRUCTURE_EXTENSION ||
                    structure.structureType == STRUCTURE_SPAWN) &&
                    structure.store.getFreeCapacity(RESOURCE_ENERGY) > 0;
            },
        });
        if (!targets.length) {
            return FAILURE;
        }
        setDestination(creep, targets[0].id);
        return SUCCESS;
    }),
]);
const behavior = sequenceNode('haul_energy', [
    clearTask,
    cachedMoveToRoom(MEMORY_BASE, commonPolicy),
    selectorNode('pick_something', [
        getHaulTaskFromBaseTopic,
        parkingLot,
    ]),
    cachedMoveToMemoryObjectId(MEMORY_HAUL_PICKUP, 1, haulerPolicy),
    loadCreep,
    clearTask,
    selectorNode('dump_or_build_or_upgrade', [
        repeatUntilFailure('dump_until_no_dropoff', sequenceNode('dump_energy', [
            leafNode('fail_when_empty', (creep, _trace, _kernel) => {
                if (creep.store.getUsedCapacity(RESOURCE_ENERGY) > 0) {
                    return SUCCESS;
                }
                return FAILURE;
            }),
            selectDropoff,
            cachedMoveToMemoryObjectId(MEMORY_DESTINATION, 1, haulerPolicy),
            emptyToDestination,
        ])),
        leafNode('succeed_when_empty', (creep, _trace, _kernel) => {
            if (creep.store.getUsedCapacity(RESOURCE_ENERGY) === 0) {
                return SUCCESS;
            }
            return FAILURE;
        }),
        sequenceNode('build_construction_site', [
            selectInfrastructureSites,
            cachedMoveToMemoryObjectId(MEMORY_DESTINATION, 3, haulerPolicy),
            build,
        ]),
        leafNode('succeed_when_empty', (creep, _trace, _kernel) => {
            if (creep.store.getUsedCapacity(RESOURCE_ENERGY) === 0) {
                return SUCCESS;
            }
            return FAILURE;
        }),
        sequenceNode('upgrade_controller', [
            leafNode('pick_room_controller', (creep, _trace, kernel) => {
                const base = getCreepBase(kernel, creep);
                if (!base) {
                    return FAILURE;
                }
                const room = Game.rooms[base.primary];
                if (!room) {
                    return FAILURE;
                }
                if (!room.controller) {
                    return FAILURE;
                }
                setDestination(creep, room.controller.id);
                return SUCCESS;
            }),
            cachedMoveToMemoryObjectId(MEMORY_DESTINATION, 3, haulerPolicy),
            setCommuteDuration,
            repeatUntilSuccess('upgrade_until_empty', leafNode('upgrade_controller', (creep, trace, _kernel) => {
                const result = creep.upgradeController(creep.room.controller);
                trace.info('upgrade result', { result });
                if (result == ERR_NOT_ENOUGH_RESOURCES) {
                    return SUCCESS;
                }
                if (result != OK) {
                    return SUCCESS;
                }
                return RUNNING;
            })),
        ]),
    ]),
]);
const roleWorker = {
    run: rootNode('worker', behaviorBoosts(roadWorker(behavior))),
};

class CreepManager {
    constructor(scheduler) {
        this.roleToCreepLogic = {
            [WORKER_ATTACKER]: roleAttacker,
            [WORKER_DEFENDER]: roleDefender,
            [WORKER_DEFENDER_BOOSTED]: roleDefender,
            [WORKER_DEFENDER_DRONE]: roleDefenderDrone,
            [WORKER_DISTRIBUTOR]: roleDistributor,
            [WORKER_MINER]: roleMiner,
            [ROLE_WORKER]: roleWorker,
            [WORKER_HARVESTER]: roleHarvester,
            [WORKER_UPGRADER]: roleUpgrader,
            [WORKER_BUILDER]: roleBuilder,
            [WORKER_REPAIRER]: roleRepairer,
            [WORKER_HAULER]: roleHauler,
            [WORKER_RESERVER]: roleReserver,
            [WORKER_EXPLORER]: roleExplorer,
            [ROLE_HARASSER]: roleHarasser,
        };
        this.id = 'creep_manager';
        this.scheduler = scheduler;
        this.creepsByBase = new Map();
    }
    run(kernel, trace) {
        trace = trace.begin('creep_manager_run');
        this.creeps = Object.values(Game.creeps);
        this.creepsByBase = _.reduce(this.creeps, (acc, creep) => {
            const base = creep.memory[MEMORY_BASE];
            if (!base) {
                trace.warn(`Creep ${creep.name} has no base assigned`);
                return acc;
            }
            const creeps = acc.get(base) || [];
            acc.set(base, creeps.concat([creep]));
            return acc;
        }, new Map());
        this.creepsByBaseAndRole = _.reduce(this.creeps, (bases, creep) => {
            const base = creep.memory[MEMORY_BASE];
            if (!base) {
                return bases;
            }
            if (!bases.has(base)) {
                bases.set(base, new Map());
            }
            const role = creep.memory[MEMORY_ROLE];
            if (!role) {
                return bases;
            }
            if (!bases.get(base).get(role)) {
                bases.get(base).set(role, []);
            }
            bases.get(base).get(role).push(creep);
            return bases;
        }, new Map);
        trace.info(`Found ${this.creeps.length} creeps`);
        // TODO, make this more efficient.
        this.creeps.forEach((creep) => {
            if (creep.spawning) {
                return;
            }
            const hasProcess = this.scheduler.hasProcess(creep.name);
            if (hasProcess) {
                return;
            }
            // Check if creep has role, if not then check if near portal and get stored memory
            if (!creep.memory[MEMORY_ROLE]) {
                creep.pos.findInRange(FIND_STRUCTURES, 10, {
                    filter: (structure) => {
                        return structure.structureType === STRUCTURE_PORTAL;
                    },
                }).forEach((portal) => {
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any
                    const destination = portal.destination;
                    if (destination.shard) {
                        const backup = kernel.getScribe().getCreepBackup(destination.shard, creep.name);
                        if (backup) {
                            creep.memory = backup.memory;
                        }
                    }
                });
            }
            // Create processes for any creeps that do not have a process
            // registered with the scheduler
            const process = this.getCreepProcess(creep.name, creep);
            if (!process) {
                trace.error('creep has no process', { creep: creep.name, memory: creep.memory });
                const result = creep.suicide();
                if (result !== OK) {
                    trace.error('suicide failed', { result, creep: creep.name, memory: creep.memory });
                }
                return;
            }
            trace.info(`Creating process for ${creep.name}`);
            this.scheduler.registerProcess(process);
        });
        if (Game.time % 100 === 0) {
            // Cleanup old creep memory
            for (const i in Memory.creeps) {
                if (!Game.creeps[i]) {
                    delete Memory.creeps[i];
                }
            }
        }
        trace.end();
        return running();
    }
    getCreeps() {
        return this.creeps;
    }
    getCreepsByBase(baseName) {
        return this.creepsByBase.get(baseName) || [];
    }
    getCreepsByBaseAndRole(base, role) {
        var _a;
        return ((_a = this.creepsByBaseAndRole.get(base)) === null || _a === void 0 ? void 0 : _a.get(role)) || [];
    }
    getCreepProcess(name, creep) {
        const role = creep.memory[MEMORY_ROLE] || null;
        if (!role) {
            throw new Error(`Creep ${creep.name} has no role`);
        }
        const roleDefinition = DEFINITIONS.get(role);
        if (!roleDefinition) {
            throw new Error(`Creep ${name} has ${role} which does not have role definition`);
        }
        const behavior = this.getCreepBehavior(name, role);
        if (!behavior) {
            throw new Error(`Creep ${name} has ${role} which does not have behavior defined`);
        }
        // add 1 to the priority because creeps need to be lower priority than their manager
        const priority = roleDefinition.processPriority + 1;
        const process = new Process(name, role, priority, behavior);
        process.setSkippable(roleDefinition.skippable);
        return process;
    }
    getCreepBehavior(name, role) {
        const behavior = this.getBehaviorByRole(role);
        return {
            run: (kernel, trace) => {
                const creep = Game.creeps[name];
                if (!creep) {
                    trace.info('creep not found; terminating process', { name });
                    return terminate();
                }
                const fields = new Map();
                fields.set('creepPos', [creep.pos.x, creep.pos.y, creep.pos.roomName].join(','));
                trace = trace.withFields(fields);
                if (creep.spawning) {
                    // TODO sleep for whoever mich longer it will take to spawn
                    return running();
                }
                // On first tick, set the start tick
                const startTick = creep.memory[MEMORY_START_TICK];
                if (!startTick) {
                    creep.memory[MEMORY_START_TICK] = Game.time;
                }
                behavior.run(creep, trace, kernel);
                return running();
            },
        };
    }
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    getBehaviorByRole(role) {
        const behavior = this.roleToCreepLogic[role];
        if (!behavior) {
            throw new Error(`Unable to get behavior for ${role}`);
        }
        return behavior;
    }
    reportMetrics(metrics) {
        // Report creep counts by role
        const creepsByBase = _.countBy(Game.creeps, (c) => c.memory[MEMORY_BASE]);
        _.forEach(creepsByBase, (value, key) => {
            metrics.gauge(`creeps_base_total`, value, { base: key });
        });
    }
}

const RUN_TTL$1 = 50;
const MAX_BASE_LEVEL = 2;
class InvaderManager {
    constructor(id, scheduler) {
        this.id = id;
        this.scheduler = scheduler;
    }
    run(kernel, trace) {
        trace = trace.begin('invader_manager_run');
        const rooms = getRoomEntriesWithInvaderBases(kernel, trace);
        trace.notice('found defeatable invader bases', {
            rooms: rooms.map((roomEntry) => {
                return { id: roomEntry.id, pos: roomEntry.invaderCorePos };
            }),
        });
        rooms.forEach((roomEntry) => {
            trace.info('requesting attack', { roomId: roomEntry.id });
            const attackRequest = {
                status: AttackStatus.REQUESTED,
                roomId: roomEntry.id,
            };
            kernel.getTopics().addRequest(ATTACK_ROOM, 1, attackRequest, ATTACK_ROOM_TTL + Game.time);
        });
        trace.end();
        return sleeping(RUN_TTL$1);
    }
}
const getRoomEntriesWithInvaderBases = (kernel, trace) => {
    const end = trace.startTimer('getRoomEntriesWithInvaderBases');
    const weakRooms = kernel.getScribe().getRooms().filter((roomEntry) => {
        if (!roomEntry.invaderCoreLevel) {
            return false;
        }
        if (roomEntry.invaderCoreLevel <= 0) {
            return false;
        }
        if (roomEntry.invaderCoreLevel > MAX_BASE_LEVEL) {
            trace.info('invader base too strong', { id: roomEntry.id, level: roomEntry.invaderCoreLevel });
            return false;
        }
        // TODO remove when we dont have any room entries without the pos
        if (!roomEntry.invaderCorePos) {
            trace.error('no position for invader base', { id: roomEntry.id });
            return false;
        }
        return true;
    });
    end();
    return weakRooms;
};

const RESERVE_LIMIT = 20000;
const REACTION_BATCH_SIZE = 1000;
const MIN_CREDITS = 200000;
const MIN_BOOST_CREDITS = 1000000;
const MIN_SELL_ORDER_SIZE = 1000;
const MAX_SELL_AMOUNT = 25000;
const MIN_ROOM_ENERGY = 100000;
const ENERGY_BALANCE_AMOUNT = 5000;
const UPDATE_RESOURCES_TTL = 50;
const REQUEST_REACTION_TTL = 250;
const REQUEST_SELL_TTL = 500;
const REQUEST_DISTRIBUTE_BOOSTS = 250;
const CONSUME_STATUS_TTL = 25;
const BALANCE_ENERGY_TTL = 50;
// Try to ensure that all colonies are ready to
// boost creeps with these effects
const MIN_CRITICAL_COMPOUND = 1000;
const CRITICAL_EFFECTS = {
    'upgradeController': ['XGH2O', 'GH2O', 'GH'],
    // 'capacity': ['XKH2O', 'KH2O', 'KH'],
    'heal': ['XLHO2', 'LHO2', 'LO'],
    'attack': ['XUH2O', 'UH2O', 'UH'],
    'rangedAttack': ['XKHO2', 'KHO2', 'KO'],
    'damage': ['XGHO2', 'GHO2', 'GO'],
    'dismantle': ['XZH2O', 'ZH2O', 'ZH'],
};
class ResourceManager {
    constructor(kernel) {
        this.availableReactions = {};
        this.reactorStatuses = [];
        this.roomStatuses = [];
        this.reactionStats = {};
        this.reactionStatuses = {};
        this.kernel = kernel;
        this.pricer = new SigmoidPricing(PRICES);
        this.resources = new Map();
        this.sharedResources = new Map();
        this.availableReactions = {};
        this.reactorStatuses = [];
        this.roomStatuses = [];
        this.reactionStats = {};
        this.reactionStatuses = {};
        this.reactionStatusStreamConsumer = kernel.getBroker().
            getStream(REACTION_STATUS_STREAM).addConsumer('resource_governor');
        this.threadUpdateResources = threadKernel('update_resources_thread', UPDATE_RESOURCES_TTL)((_trace, _kernel) => {
            this.resources = this.getBaseResources();
            this.sharedResources = this.getSharedResources();
        });
        this.threadRequestReactions = threadKernel('request_reactions_thread', REQUEST_REACTION_TTL)((trace, kernel) => {
            this.availableReactions = this.getReactions(kernel, trace);
            this.requestReactions(trace, kernel);
        });
        this.threadRequestSellExtraResources = threadKernel('request_sell_resources_tread', REQUEST_SELL_TTL)((trace, kernel) => {
            this.requestSellResource(trace, kernel);
        });
        this.threadDistributeBoosts = threadKernel('distribute_boosts_thread', REQUEST_DISTRIBUTE_BOOSTS)((trace, kernel) => {
            this.distributeBoosts(trace, kernel);
        });
        this.threadConsumeStatuses = threadKernel('statuses_thread', CONSUME_STATUS_TTL)(this.consumeStatuses.bind(this));
        this.threadConsumeReactionStatusStream = threadKernel('reaction_stream', CONSUME_STATUS_TTL)(this.consumeReactionStatusStream.bind(this));
        this.threadBalanceEnergy = threadKernel('balance_energy_thread', BALANCE_ENERGY_TTL)(this.balanceEnergy.bind(this));
    }
    run(kernel, trace) {
        this.threadUpdateResources(trace, kernel);
        this.threadRequestReactions(trace, kernel);
        this.threadRequestSellExtraResources(trace, kernel);
        this.threadDistributeBoosts(trace, kernel);
        this.threadConsumeStatuses(trace, kernel);
        this.threadConsumeReactionStatusStream(trace, kernel);
        this.threadBalanceEnergy(trace, kernel);
        return running();
    }
    getBaseWithTerminalWithResource(resource, destBase = null) {
        const basesWithTerminal = this.kernel.getPlanner().getBases().reduce((acc, base) => {
            const room = getBasePrimaryRoom(base);
            if (!room) {
                return acc;
            }
            // Don't return the room needing the resources
            if (destBase && base.id && room.name === destBase.id) {
                return acc;
            }
            // If base doesn't have a terminal don't include it
            if (!room.terminal) {
                return acc;
            }
            const isCritical = Object.values(CRITICAL_EFFECTS).reduce((acc, compounds) => {
                if (acc) {
                    return acc;
                }
                if (compounds.indexOf(resource) != -1) {
                    return true;
                }
                return false;
            }, false);
            let amount = getStoredResourceAmount(base, resource);
            if (isCritical) {
                amount -= MIN_CRITICAL_COMPOUND;
            }
            if (resource === RESOURCE_ENERGY && amount < MIN_ROOM_ENERGY) {
                return acc;
            }
            if (amount <= 0) {
                return acc;
            }
            return acc.concat({ room: room, amount });
        }, []);
        return _.sortBy(basesWithTerminal, 'amount').reverse().shift();
    }
    getTerminals() {
        return this.kernel.getPlanner().getBases().reduce((acc, base) => {
            const room = getBasePrimaryRoom(base);
            if (!room) {
                return acc;
            }
            // If base doesn't have a terminal don't include it
            if (!room.terminal) {
                return acc;
            }
            return acc.concat(room.terminal);
        }, []);
    }
    getSharedResources() {
        return this.kernel.getPlanner().getBases().reduce((acc, base) => {
            const primaryRoom = getBasePrimaryRoom(base);
            if (!primaryRoom) {
                return acc;
            }
            if (!primaryRoom.terminal) {
                return acc;
            }
            const baseResources = getStoredResources(base);
            Array.from(baseResources.keys()).forEach((resource) => {
                // Find first effect for that resource
                const isCritical = Object.values(CRITICAL_EFFECTS).find((compounds) => {
                    if (compounds.indexOf(resource) != -1) {
                        return true;
                    }
                    return false;
                }, false);
                // If critical, subtract min compound amount
                let amount = baseResources.get(resource);
                if (isCritical) {
                    amount -= MIN_CRITICAL_COMPOUND;
                }
                if (amount < 0) {
                    amount = 0;
                }
                // Update the resource counts
                const current = acc.get(resource) || 0;
                acc.set(resource, amount + current);
            });
            return acc;
        }, new Map());
    }
    getBaseResources() {
        return this.kernel.getPlanner().getBases().reduce((acc, base) => {
            const primaryRoom = getBasePrimaryRoom(base);
            if (!primaryRoom) {
                return acc;
            }
            if (!primaryRoom.terminal) {
                return acc;
            }
            const baseResources = getStoredResources(base);
            Array.from(baseResources.keys()).forEach((resource) => {
                const current = acc.get(resource) || 0;
                acc.set(resource, baseResources.get(resource) + current);
            });
            return acc;
        }, new Map());
    }
    getAmountInReserve(kernel, resource) {
        return kernel.getPlanner().getBases().reduce((acc, base) => {
            return acc + getStoredResourceAmount(base, resource);
        }, 0);
    }
    getReactions(kernel, trace) {
        const availableReactions = new Map();
        const missingOneInput = new Map();
        const overReserve = new Map();
        const firstInputs = Object.keys(REACTIONS);
        firstInputs.forEach((inputA) => {
            // If we don't have a full batch, move onto next
            if (!this.resources.has(inputA) || this.resources.get(inputA) < REACTION_BATCH_SIZE) {
                trace.info('dont have enough of first resource', {
                    inputA,
                    amountA: this.resources.get(inputA) || 0,
                    REACTION_BATCH_SIZE,
                });
                return;
            }
            const secondInputs = Object.keys(REACTIONS[inputA]);
            secondInputs.forEach((inputB) => {
                const output = REACTIONS[inputA][inputB];
                // If we don't have a full batch if input mark missing one and go to next
                if (!this.resources.has(inputB) || this.resources.get(inputB) < REACTION_BATCH_SIZE) {
                    if (!missingOneInput.has(output)) {
                        trace.info('dont have enough of second resource', {
                            inputA,
                            inputB,
                            amountA: this.resources.get(inputA) || 0,
                            amountB: this.resources.get(inputB) || 0,
                            output,
                            REACTION_BATCH_SIZE,
                        });
                        missingOneInput.set(output, { inputA, inputB, output });
                    }
                    return;
                }
                // Check if we need more of the output
                if (this.resources.get(output) > RESERVE_LIMIT && !overReserve.has(output)) {
                    overReserve.set(output, { inputA, inputB, output });
                    return;
                }
                // If reaction isn't already present add it
                if (!availableReactions.has(output)) {
                    trace.info('adding available reaction', {
                        inputA,
                        inputB,
                        amountA: this.resources.get(inputA) || 0,
                        amountB: this.resources.get(inputB) || 0,
                        output,
                    });
                    availableReactions.set(output, { inputA, inputB, output });
                }
            });
        });
        const sortedAvailableReactions = this.prioritizeReactions(availableReactions, 0);
        const sortedMissingOneInput = this.prioritizeReactions(missingOneInput, 5);
        // overReserve = this.prioritizeReactions(overReserve, 10);
        let nextReactions = [].concat(sortedAvailableReactions);
        if (missingOneInput.size && Game.market.credits > MIN_CREDITS) {
            nextReactions = nextReactions.concat(sortedMissingOneInput);
        }
        // nextReactions = nextReactions.concat(overReserve.reverse());
        trace.info('available reactions', { nextReactions, availableReactions, missingOneInput, overReserve });
        return nextReactions;
    }
    prioritizeReactions(reactions, penalty) {
        return _.sortBy(Array.from(reactions.values()), (reaction) => {
            let priority = REACTION_PRIORITIES[reaction['output']];
            // Reduce priority linearly based on amount of resource (more = lower priority)
            const amount = this.resources.get(reaction['output']) || 0;
            priority = priority * _.max([0, 1 - (amount / RESERVE_LIMIT)]);
            priority -= penalty;
            reaction['priority'] = priority;
            return priority;
        });
    }
    buyResource(base, resource, amount, ttl, trace) {
        trace.info('requesting resource purchase', { baseId: base.id, resource, amount, ttl });
        const primaryRoom = getBasePrimaryRoom(base);
        if (!primaryRoom) {
            trace.info('no primary room', { baseId: base.id });
            return;
        }
        // We can't request a transfer if room lacks a terminal
        if (!primaryRoom.terminal) {
            trace.info('room does not have a terminal', { baseId: base.id });
            return false;
        }
        const details = {
            [TERMINAL_TASK_TYPE]: TASK_MARKET_ORDER,
            [MEMORY_ORDER_TYPE]: ORDER_BUY,
            [MEMORY_ORDER_RESOURCE]: resource,
            [MEMORY_ORDER_AMOUNT]: amount,
        };
        if (Game.market.credits < MIN_CREDITS) {
            trace.info('below min credits, not purchasing resource', { resource, amount });
            return false;
        }
        trace.info('purchase resource', { baseId: base.id, resource, amount });
        this.kernel.getTopics().addRequest(getBaseTerminalTopic(base), TERMINAL_BUY, details, ttl + Game.time);
        return true;
    }
    requestResource(base, resource, amount, ttl, trace) {
        trace.log('requesting resource transfer', { baseId: base.id, resource, amount, ttl });
        const primaryRoom = getBasePrimaryRoom(base);
        if (!primaryRoom) {
            trace.error('no primary room', { baseId: base.id });
            return;
        }
        // We can't request a transfer if room lacks a terminal
        if (!primaryRoom.terminal) {
            trace.log('room does not have a terminal', { baseId: base.id });
            return false;
        }
        // Don't send transfer request if a terminal already has the task
        const inProgress = this.kernel.getPlanner().getBases().filter((base) => {
            if (!base.terminalTask) {
                return false;
            }
            return base.terminalTask[TRANSFER_RESOURCE] === resource &&
                base.terminalTask[TRANSFER_BASE] === base.id;
        }).length > 0;
        if (inProgress) {
            trace.log('task already in progress', { base: base.id, resource, amount, ttl });
            return true;
        }
        const sourceBase = this.getBaseWithTerminalWithResource(resource, base);
        if (!sourceBase) {
            trace.log('no rooms with resource', { resource });
            return false;
        }
        const sourceResources = getStoredResourceAmount(sourceBase, resource);
        amount = _.min([sourceResources, amount]);
        trace.log('requesting resource from other room', { source: sourceBase.id, resource, amount });
        const request = {
            [TERMINAL_TASK_TYPE]: TASK_TRANSFER,
            [TRANSFER_RESOURCE]: resource,
            [TRANSFER_AMOUNT]: amount,
            [TRANSFER_BASE]: base.id,
        };
        this.kernel.getTopics().addRequest(getBaseTerminalTopic(sourceBase), TERMINAL_TRANSFER, request, ttl + Game.time);
        return true;
    }
    createBuyOrder(room, resource, amount, trace) {
        if (!room.hasTerminal()) {
            trace.log('room does not have terminal', { roomName: room.name });
            return;
        }
        // Check if we already have a buy order for the room and resource
        const duplicateBuyOrders = Object.values(Game.market.orders).filter((order) => {
            return order.type === ORDER_BUY && order.roomName === room.id &&
                order.resourceType === resource;
        });
        if (duplicateBuyOrders.length) {
            trace.log('already have an order for resource', { resource });
            return;
        }
        const currentAmount = this.resources.get(resource) || 0;
        const price = this.pricer.getPrice(ORDER_BUY, resource, currentAmount);
        // Create buy order
        const order = {
            type: ORDER_BUY,
            resourceType: resource,
            price: price,
            totalAmount: amount,
            roomName: room.id,
        };
        const result = Game.market.createOrder(order);
        trace.log('create order result', { order, result });
    }
    requestReactions(trace, kernel) {
        Object.values(this.availableReactions).forEach((reaction) => {
            const details = {
                [REACTOR_TASK_TYPE]: REACTION,
                [REACTOR_INPUT_A]: reaction['inputA'],
                [REACTOR_INPUT_B]: reaction['inputB'],
                [REACTOR_OUTPUT]: reaction['output'],
                [REACTOR_AMOUNT]: REACTION_BATCH_SIZE,
            };
            kernel.getTopics().addRequest(TASK_REACTION, reaction['priority'], details, REQUEST_REACTION_TTL + Game.time);
        });
        const reactions = kernel.getTopics().getTopic(TASK_REACTION);
        if (!reactions || !reactions.length) {
            trace.info('no reactions to request');
            return;
        }
        trace.info('requested reactions', {
            reactions: reactions.map((r) => {
                return {
                    output: r.details[REACTOR_OUTPUT],
                    priority: r.priority,
                };
            }),
        });
    }
    requestSellResource(trace, kernel) {
        this.sharedResources.forEach((amount, resource) => {
            if (resource === RESOURCE_ENERGY) {
                return;
            }
            const excess = amount - RESERVE_LIMIT;
            if (excess < MIN_SELL_ORDER_SIZE) {
                return;
            }
            // Check if we already have a buy order for the room and resource
            const duplicateOrders = Object.values(Game.market.orders).filter((order) => {
                return order.type === ORDER_SELL && order.resourceType === resource;
            });
            if (duplicateOrders.length && Game.market.credits > MIN_CREDITS) {
                return;
            }
            const sourceBase = this.getBaseWithTerminalWithResource(resource);
            if (!sourceBase) {
                return;
            }
            const sourceResources = getStoredResourceAmount(sourceBase, resource);
            const sellAmount = _.min([sourceResources, excess, MAX_SELL_AMOUNT]);
            const details = {
                [TERMINAL_TASK_TYPE]: TASK_MARKET_ORDER,
                [MEMORY_ORDER_TYPE]: ORDER_SELL,
                [MEMORY_ORDER_RESOURCE]: resource,
                [MEMORY_ORDER_AMOUNT]: sellAmount,
            };
            kernel.getTopics().addRequest(getBaseTerminalTopic(sourceBase), TERMINAL_BUY, details, REQUEST_SELL_TTL + Game.time);
        });
    }
    distributeBoosts(trace, kernel) {
        trace.info('balancing boosts');
        kernel.getPlanner().getBases().forEach((base) => {
            const baseTrace = trace.withFields(new Map([['base', base.id]]));
            const baseEnd = baseTrace.startTimer('base');
            if (!base.boostPosition) {
                baseEnd();
                return;
            }
            const primaryRoom = getBasePrimaryRoom(base);
            if (!primaryRoom) {
                baseEnd();
                return;
            }
            //const allEffects = getStoredBoosterEffects(base);
            const availableEffects = getStoredEffects(base);
            Object.entries(CRITICAL_EFFECTS).forEach(([effectName, compounds]) => {
                const effectTrace = baseTrace.withFields(new Map([
                    ['effect', effectName],
                    ['compounds', compounds.length.toString()],
                ]));
                const effectsEnd = effectTrace.startTimer('effect');
                /*
                const effect = allEffects[effectName];
                if (!effect) {
                  effectTrace.log('missing effect', {effectName});
                  effectsEnd();
                  return;
                }
                */
                const bestCompound = compounds[0];
                const baseResources = getStoredResources(base);
                const currentAmount = baseResources.get(bestCompound) || 0;
                const availableEffect = availableEffects.get(effectName);
                if (!availableEffect || currentAmount < MIN_CRITICAL_COMPOUND) {
                    effectTrace.info('maybe request/buy best compound', {
                        baseId: base.id,
                        bestCompound,
                        currentAmount,
                        credits: Game.market.credits,
                        MIN_CRITICAL_COMPOUND,
                    });
                    const minimumCritical = MIN_CRITICAL_COMPOUND;
                    // TODO if base at a war, increase minimum critical resources
                    const requested = this.requestResource(base, bestCompound, minimumCritical - currentAmount, REQUEST_DISTRIBUTE_BOOSTS, effectTrace);
                    // If we couldnt request resource, try buying
                    if (!requested && Game.market.credits > MIN_BOOST_CREDITS) {
                        this.buyResource(base, bestCompound, minimumCritical - currentAmount, REQUEST_DISTRIBUTE_BOOSTS, effectTrace);
                    }
                }
                else {
                    effectTrace.info('have booster', { baseId: base.id, effectName, bestCompound, currentAmount });
                }
                effectsEnd();
            });
            baseEnd();
        });
    }
    consumeReactionStatusStream(trace) {
        const events = this.reactionStatusStreamConsumer.getEvents();
        trace.log('reaction status events', { events });
        events.forEach((event) => {
            switch (event.type) {
                case REACTION_STATUS_START:
                case REACTION_STATUS_UPDATE:
                    this.reactionStatuses[event.key] = event.data;
                    break;
                case REACTION_STATUS_STOP:
                    delete this.reactionStatuses[event.key];
                    break;
                default:
                    throw new Error(`Unknown reaction status event type: ${event.type}`);
            }
        });
    }
    consumeStatuses(trace, kernel) {
        trace.info('consuming statues');
        const reactorStatuses = kernel.getTopics().getTopic(ACTIVE_REACTIONS) || [];
        this.reactorStatuses = reactorStatuses;
        trace.info('reactor statuses', { length: reactorStatuses.length });
        this.reactionStats = reactorStatuses.reduce((acc, status) => {
            const resource = status.details[REACTION_STATUS_RESOURCE];
            if (!acc[resource]) {
                acc[resource] = 0;
            }
            const phase = status.details[REACTION_STATUS_PHASE];
            if (phase === TASK_PHASE_REACT) {
                acc[resource] += 1;
            }
            return acc;
        }, {});
        const roomStatuses = kernel.getTopics().getTopic(ROOM_STATUES) || [];
        if (!roomStatuses.length) {
            trace.warn('no room statues in topic, skipping update');
            return;
        }
        this.roomStatuses = _.pluck(roomStatuses, 'details');
        trace.info('room statuses', { length: roomStatuses.length });
    }
    balanceEnergy(trace, kernel) {
        if (this.roomStatuses.length < 2) {
            trace.info('not enough rooms to balance');
            return;
        }
        const hasTerminals = _.filter(this.roomStatuses, { [BASE_STATUS_TERMINAL]: true });
        if (hasTerminals.length < 2) {
            trace.info('not enough terminals to balance');
            return;
        }
        const energySorted = _.sortByAll(hasTerminals, [
            ['details', BASE_STATUS_ENERGY].join('.'),
        ]);
        const levelAndEnergySorted = _.sortByAll(hasTerminals, [
            ['details', BASE_STATUS_LEVEL].join('.'),
            ['details', BASE_STATUS_LEVEL_COMPLETED].join('.'),
            ['details', BASE_STATUS_ENERGY].join('.'),
        ]);
        trace.info('sorted', { energySorted, levelAndEnergySorted });
        const sinkRoom = levelAndEnergySorted[0];
        const baseStatus = energySorted[energySorted.length - 1];
        if (sinkRoom === baseStatus) {
            trace.info('sink and source are same');
            return;
        }
        const energyDiff = baseStatus[BASE_STATUS_ENERGY] - sinkRoom[BASE_STATUS_ENERGY];
        if (energyDiff < ENERGY_BALANCE_AMOUNT * 2) {
            trace.info('energy different too small, no need to send energy', { energyDiff });
            return;
        }
        const sourceRoomName = baseStatus[BASE_STATUS_NAME];
        const sourceBase = kernel.getPlanner().getBaseByRoom(sourceRoomName);
        if (!sourceBase) {
            trace.warn('source base not found', { sourceRoomName });
            return;
        }
        const sinkRoomName = sinkRoom[BASE_STATUS_NAME];
        const request = {
            [TERMINAL_TASK_TYPE]: TASK_TRANSFER,
            [TRANSFER_RESOURCE]: RESOURCE_ENERGY,
            [TRANSFER_AMOUNT]: ENERGY_BALANCE_AMOUNT,
            [TRANSFER_BASE]: sinkRoomName,
        };
        trace.notice('send energy request', { request });
        kernel.getTopics().addRequest(getBaseTerminalTopic(sourceBase), TERMINAL_ENERGY_BALANCE, request, BALANCE_ENERGY_TTL + Game.time);
    }
}

const SHARD_MEMORY_TTL = 50;
class ShardManager {
    constructor() {
        this.threadUpdateShardMemory = threadKernel('update_shard_memory', SHARD_MEMORY_TTL)(this.updateShardMemory.bind(this));
    }
    run(kernel, trace) {
        trace = trace.begin('kernel_governor');
        trace.info('kernel governor run', {});
        this.threadUpdateShardMemory(trace, kernel);
        trace.end();
        return sleeping(SHARD_MEMORY_TTL);
    }
    updateShardMemory(trace, kernel) {
        trace.info('update_shard_memory');
        const scribe = kernel.getScribe();
        let shardMemory = scribe.getLocalShardMemory();
        const bases = kernel.getPlanner().getBases();
        shardMemory.status = {
            numColonies: bases.length,
        };
        shardMemory.time = Game.time;
        shardMemory = this.requestClaimersFromOtherShards(kernel, shardMemory, trace);
        shardMemory = this.requestBuildersFromOtherShards(kernel, shardMemory, trace);
        kernel.getPlanner().getShards().forEach((shardName) => {
            if (shardName === Game.shard.name) {
                return;
            }
            const shardMemory = kernel.getScribe().getRemoteShardMemory(shardName);
            trace.info('shard memory', { shardName, shardMemory });
            this.handleClaimerRequests(kernel, shardMemory.request_claimer || new Map(), trace);
            this.handleBuilderRequests(kernel, shardMemory.request_builder || new Map(), trace);
            /* TODO remove if not used Jan 2022
            const shardConfig: ShardConfig = kernel.getPlanner().getShardConfig(shardName);
            if (!shardConfig) {
              return;
            }
      
            trace.log('kernel governor shard', {shardName, shardConfig})
      
            const primaryBase: Base = Object.values(shardConfig)[0];
            if (!primaryBase || !primaryBase.primary) {
              return;
            }
      
            trace.log('kernel governor base', {shardName, primaryBase})
           */
            /* TODO
            if (!shardMemory.ttl) {
              shardMemory = this.sendClaimer(shardName, primaryBase.primary, shardMemory, trace);
            }
            */
        });
        trace.info('setting local memory', { shardMemory });
        scribe.setLocalShardMemory(shardMemory);
    }
    // TODO update to check memory specifically
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    findCreeps(needle) {
        return _.filter(Game.creeps, needle);
    }
    requestClaimersFromOtherShards(kernel, localMemory, trace) {
        localMemory.request_claimer = new Map();
        // Check if we need to request reservers
        const claimedRooms = Object.values(Game.rooms).filter((room) => {
            var _a;
            return (_a = room.controller) === null || _a === void 0 ? void 0 : _a.my;
        });
        const bases = kernel.getPlanner().getBases();
        if (!bases.length) {
            return localMemory;
        }
        if (bases.length && !claimedRooms.length) {
            const request = {
                baseId: bases[0].id,
                base: bases[0].id,
                shard: Game.shard.name,
                room: bases[0].primary,
                ttl: Game.time,
            };
            const enroute = this.findCreeps({
                memory: {
                    [MEMORY_ROLE]: WORKER_RESERVER,
                    [MEMORY_ASSIGN_SHARD]: request.shard,
                    [MEMORY_ASSIGN_ROOM]: request.room,
                    [MEMORY_BASE]: request.baseId,
                },
            });
            if (!enroute.length) {
                localMemory.request_claimer.set(bases[0].primary, request);
                trace.info('requesting claimer from another shard', { request });
            }
        }
        return localMemory;
    }
    handleClaimerRequests(kernel, requests, trace) {
        Array.from(requests.values()).forEach((creepRequest) => {
            const memory = {
                [MEMORY_ASSIGN_SHARD]: creepRequest.shard,
                [MEMORY_ASSIGN_ROOM]: creepRequest.room,
                [MEMORY_BASE]: creepRequest.baseId,
            };
            const enroute = this.findCreeps({
                [MEMORY_ROLE]: WORKER_RESERVER,
                memory,
            });
            trace.info('checking if claimer in-flight', {
                shardName: creepRequest.shard,
                enroute: enroute.map((creep) => creep.id),
            });
            if (enroute.length) {
                return;
            }
            const priorities = PRIORITY_RESERVER;
            const ttl = SHARD_MEMORY_TTL + Game.time;
            const role = WORKER_RESERVER;
            const request = createSpawnRequest(priorities, ttl, role, memory, null, 0);
            trace.info('relaying claimer request from remote shard', { request });
            kernel.getTopics().addRequestV2(getShardSpawnTopic(), request);
        });
    }
    sendClaimer(shardName, roomName, memory, trace) {
        const request = {
            shard: shardName,
            room: roomName,
            ttl: Game.time,
        };
        if (!memory.request_claimer) {
            memory.request_claimer = new Map();
        }
        memory.request_claimer.set(roomName, request);
        trace.info('adding request for claimer', { request });
        return memory;
    }
    requestBuildersFromOtherShards(kernel, localMemory, trace) {
        localMemory.request_builder = new Map();
        // Check if we need to request builders
        const claimedRooms = Object.values(Game.rooms).filter((room) => {
            var _a;
            return (_a = room.controller) === null || _a === void 0 ? void 0 : _a.my;
        });
        const bases = kernel.getPlanner().getBases();
        if (!bases.length) {
            return localMemory;
        }
        if (!Object.values(Game.spawns).length && claimedRooms.length) {
            const request = {
                baseId: bases[0].id,
                base: bases[0].id,
                shard: Game.shard.name,
                room: bases[0].primary,
                ttl: Game.time,
            };
            const enroute = this.findCreeps({
                memory: {
                    [MEMORY_ROLE]: WORKER_BUILDER,
                    [MEMORY_ASSIGN_SHARD]: request.shard,
                    [MEMORY_ASSIGN_ROOM]: request.room,
                    [MEMORY_BASE]: request.baseId,
                },
            });
            if (enroute.length < 6) {
                localMemory.request_builder.set(bases[0].primary, request);
                trace.info('requesting builder from another shard', { request });
            }
        }
        return localMemory;
    }
    handleBuilderRequests(kernel, requests, trace) {
        Array.from(requests.values()).forEach((creepRequest) => {
            const memory = {
                [MEMORY_ROLE]: WORKER_BUILDER,
                [MEMORY_ASSIGN_SHARD]: creepRequest.shard,
                [MEMORY_ASSIGN_ROOM]: creepRequest.room,
                [MEMORY_BASE]: creepRequest.baseId,
            };
            const enroute = this.findCreeps({ memory });
            if (enroute.length) {
                trace.info('builder in-flight', { shardName: creepRequest.shard, enroute: enroute.map((creep) => creep.id) });
                return;
            }
            const priority = PRIORITY_BUILDER;
            const ttl = SHARD_MEMORY_TTL + Game.time;
            const role = WORKER_BUILDER;
            const request = createSpawnRequest(priority, ttl, role, memory, null, 0);
            trace.info('relaying builder request from remote shard', { request });
            kernel.getTopics().addRequestV2(getShardSpawnTopic(), request);
        });
    }
}

const CLEANUP_INTERVAL = 100;
const SITE_TTL = 10000;
class SiteJanitor {
    constructor() {
        this.sites = new Map();
    }
    run(kernel, trace) {
        trace.warn('running site janitor');
        let purged = 0;
        let created = 0;
        let updated = 0;
        let removed = 0;
        for (const [_, entry] of this.sites) {
            // remove entry if site does not exist in game
            if (!Game.constructionSites[entry.id]) {
                purged++;
                this.sites.delete(entry.id);
            }
        }
        _.each(Game.constructionSites, (site) => {
            // if we have not seen the site, add an entry
            if (!this.sites.has(site.id)) {
                this.sites.set(site.id, {
                    id: site.id,
                    completeness: site.progress,
                    observed: Game.time,
                });
                created++;
                return;
            }
            // if we have seen the site and the progress is different, update the observed time
            if (this.sites.get(site.id).completeness !== site.progress) {
                this.sites.get(site.id).completeness = site.progress;
                this.sites.get(site.id).observed = Game.time;
                updated++;
                return;
            }
            // if the site has been observed for too long, remove it
            if (Game.time - this.sites.get(site.id).observed > SITE_TTL) {
                site.remove();
                this.sites.delete(site.id);
                removed++;
                return;
            }
        });
        trace.info('site janitor stats', {
            purged,
            created,
            updated,
            removed,
        });
        return sleeping(CLEANUP_INTERVAL);
    }
}

/**
 * Managers War Parties, including the creation and recreation on startup.
 *
 *
 */
const WAR_PARTY_RUN_TTL = 100;
const BASE_ATTACK_RANGE = 3;
const MAX_BASES_PER_TARGET = 3;
const MAX_WAR_PARTIES_PER_BASE = 1;
const MAX_HARASSERS_PER_BASE = 1;
const CONSUME_EVENTS_TTL = 20;
const RUN_TTL = 10;
function newWarParty(id, baseId, target, muster, createdAt) {
    return {
        id: id,
        baseId: baseId,
        target: target,
        muster: muster,
        createdAt: createdAt,
    };
}
class WarManager {
    constructor(id, scheduler) {
        // TODO make a target struct
        this.targets = [];
        this.id = id;
        this.scheduler = scheduler;
        this.warParties = null;
        this.targets = [];
        this.consumeEventsThread = threadKernel('events_thread', CONSUME_EVENTS_TTL)(this.consumeEvents.bind(this));
        this.updateWarPartiesThread = threadKernel('update_warparties', WAR_PARTY_RUN_TTL)(this.updateWarParties.bind(this));
        this.mapUpdateThread = threadKernel('map_thread', 1)(this.mapUpdate.bind(this));
    }
    run(kernel, trace) {
        trace = trace.begin('war_manager_run');
        if (this.warParties === null) {
            trace.info('restoring war parites');
            this.restoreFromMemory(kernel, trace);
        }
        this.updateWarPartiesThread(trace, kernel);
        this.consumeEventsThread(trace, kernel);
        this.mapUpdateThread(trace, kernel);
        // Write post event status
        trace.info('war manager state', {
            targets: this.targets,
            warPartyIds: this.warParties.map((warParty) => warParty.id),
            autoAttack: kernel.getConfig().autoAttack,
        });
        trace.end();
        return sleeping(RUN_TTL);
    }
    getTargets() {
        return this.targets;
    }
    getWarParties() {
        return this.warParties;
    }
    consumeEvents(trace, kernel) {
        // Process events
        let targets = this.targets;
        const topic = kernel.getTopics().getTopic(ATTACK_ROOM);
        if (!topic) {
            trace.warn('no attack room topic');
            return;
        }
        trace.info('consuming events', { length: topic.length });
        let event = null;
        // eslint-disable-next-line no-cond-assign
        while (event = topic.shift()) {
            switch (event.details.status) {
                case AttackStatus.REQUESTED:
                    trace.info('requested attack', { target: event.details.roomId });
                    if (targets.indexOf(event.details.roomId) === -1) {
                        targets.push(event.details.roomId);
                    }
                    break;
                case AttackStatus.COMPLETED:
                    trace.info('attack completed', { roomId: event.details.roomId });
                    // clear target room so we don't try to pick it before it's journal entry is updated
                    kernel.getScribe().clearRoom(event.details.roomId);
                    // remove room from targets when completed
                    targets = targets.filter((target) => target !== event.details.roomId);
                    break;
                default:
                    throw new Error(`invalid status ${event.details.status}`);
            }
        }
        // address bug with duplicate entries
        targets = _.uniq(targets);
        // filter targets if room entry is stale/missing
        targets = targets.filter((target) => {
            const roomEntry = kernel.getScribe().getRoomById(target);
            if (!roomEntry) {
                return false;
            }
            return true;
        });
        // prioritize targets by strength
        targets = targets.sort((a, b) => {
            const aEntry = kernel.getScribe().getRoomById(a);
            const bEntry = kernel.getScribe().getRoomById(b);
            return scoreRoomDamage(aEntry) - scoreRoomDamage(bEntry);
        });
        /*
        let bases = kernel.getPlanner().getBases();
        trace.info('allowed bases', {bases: bases.map((base) => base.primary)});
        // Sort targets so closest to other bases is prioritized
        targets = _.sortBy(targets, (target) => {
          bases = _.sortBy(bases, (base: Base) => {
            return Game.map.getRoomLinearDistance(base.primary, target);
          });
    
          if (bases.length === 0) {
            return 999;
          }
    
          return Game.map.getRoomLinearDistance(bases[0].primary, target);
        });
        */
        trace.notice('targets', { targets });
        this.targets = targets;
    }
    updateWarParties(trace, kernel) {
        // Update list of war parties
        this.warParties = this.warParties.filter((party) => {
            return this.scheduler.hasProcess(party.id);
        });
        if (!kernel.getConfig().autoAttack) {
            trace.info('auto attack disabled');
            return;
        }
        if (!this.targets || this.targets.length === 0) {
            trace.info('no target rooms');
            return;
        }
        const bases = kernel.getPlanner().getBases();
        const targetNumBasesAssigned = {};
        const baseAssignments = {};
        this.targets.forEach((target) => {
            var _a;
            const roomEntry = kernel.getScribe().getRoomById(target);
            if (!roomEntry) {
                trace.error('no room entry', { target });
                return;
            }
            if (targetNumBasesAssigned[target] === undefined) {
                targetNumBasesAssigned[target] = 0;
            }
            if (roomEntry.controller) {
                if (roomEntry.controller.safeMode > 150) {
                    trace.info('controller is in safe mode', { target });
                    return;
                }
                // Send reservers to block if no towers and not in safe mode
                if (roomEntry.numTowers === 0 && ((_a = roomEntry.controller) === null || _a === void 0 ? void 0 : _a.level) > 0) {
                    trace.info('no towers and still claimed, send reserver', { target });
                    this.sendReserver(kernel, target, trace);
                }
            }
            // Determine if we need to send attackers
            if (roomEntry.numTowers > 0 || roomEntry.numKeyStructures > 0) {
                // Locate nearby colonies and spawn war parties
                bases.forEach((base) => {
                    // If we have assigned enough rooms, move on
                    if (targetNumBasesAssigned[target] >= MAX_BASES_PER_TARGET) {
                        return;
                    }
                    // Check if base already assigned
                    if (baseAssignments[base.id]) {
                        return;
                    }
                    const baseTrace = trace.withFields(new Map([['base', base.primary]]));
                    // TODO check for path to target
                    const linearDistance = Game.map.getRoomLinearDistance(base.primary, target);
                    if (linearDistance > BASE_ATTACK_RANGE) {
                        baseTrace.info('linear distance too far', {
                            base: base.primary, target, linearDistance, BASE_ATTACK_RANGE,
                        });
                        return;
                    }
                    const baseRoom = Game.rooms[base.primary];
                    if (!baseRoom) {
                        baseTrace.warn('no base room', { target, base });
                        return;
                    }
                    trace.info('assigning base', { base: base.primary, target });
                    targetNumBasesAssigned[target]++;
                    baseAssignments[base.id] = target;
                    // If passive, don't attack targets unless base is flagged aggressive
                    const flagId = `aggressive_${base.id}`;
                    if (kernel.getConfig().temperament === Temperaments.Passive && !Game.flags[flagId]) {
                        trace.info('temperament is passive, dont attack');
                        this.writeMemory(trace);
                        return;
                    }
                    this.attack(kernel, base, baseRoom, roomEntry, trace);
                    return;
                });
            }
        });
        this.writeMemory(trace);
    }
    mapUpdate(trace, kernel) {
        if (this.warParties) {
            this.warParties.forEach((party) => {
                const base = kernel.getPlanner().getBaseById(party.baseId);
                if (!base) {
                    trace.error('no base', { baseId: party.baseId });
                    return;
                }
                Game.map.visual.line(new RoomPosition(25, 25, base.primary), new RoomPosition(25, 25, party.target), {});
            });
        }
        this.targets.forEach((target) => {
            Game.map.visual.text('❌', new RoomPosition(25, 25, target), {
                align: 'center',
                fontSize: 20,
            });
        });
    }
    attack(kernel, base, baseRoom, targetRoomEntry, trace) {
        const boosts = newMultipliers();
        const baseStorage = baseRoom.storage;
        if (baseStorage) {
            const availableHealingBoost = scoreStorageHealing(baseStorage);
            boosts[HEAL] = availableHealingBoost;
        }
        const availableEnergyCapacity = baseRoom.energyCapacityAvailable;
        // We want at least one creep attacking per quad, so ensure tank works with 3 creeps
        const roomDamage = scoreRoomDamage(targetRoomEntry) / 4;
        const [parts, ok] = buildAttacker(roomDamage, availableEnergyCapacity, boosts, trace);
        if (ok) {
            trace.warn('building attacker', { parts });
            this.sendWarParty(kernel, base, targetRoomEntry, parts, trace);
        }
        else if (targetRoomEntry.invaderCoreLevel < 1) { // don't harass invaders bases
            // check if an adjacent room has units, if not don't send harasser
            const enemy = targetRoomEntry.controller.owner;
            const adjacentRooms = Game.map.describeExits(targetRoomEntry.id);
            const remote = Object.values(adjacentRooms).find((roomName) => {
                var _a, _b;
                const entry = kernel.getScribe().getRoomById(roomName);
                if (!entry) {
                    trace.info('no entry', { roomName });
                    return false;
                }
                // make sure room is owned by the target enemy
                if (((_a = entry.controller) === null || _a === void 0 ? void 0 : _a.owner) !== enemy) {
                    trace.info('room not owned by enemy', { roomName, enemy, owner: (_b = entry.controller) === null || _b === void 0 ? void 0 : _b.owner });
                    return false;
                }
                // make sure room does not have units that can do significant damage
                if (entry.hostilesDmgByOwner.get(enemy) || 0 > 50) {
                    trace.info('room has hostile damage', { roomName, enemy, dmg: entry.hostilesDmgByOwner.get(enemy) });
                    return false;
                }
                if (entry.hostilesByOwner.get(enemy) || 0 > 0) {
                    trace.info('room has hostile units', { roomName, enemy, hostiles: entry.hostilesByOwner.get(enemy) });
                    return false;
                }
                trace.info('room is undefended remote', { roomName, enemy, entry });
                return true;
            });
            if (remote) {
                trace.warn('could not build attacker, harass', {
                    base: base.primary,
                    targetRoom: targetRoomEntry.id,
                    roomDamage, availableEnergyCapacity, boosts,
                });
                this.sendHarassers(kernel, base, targetRoomEntry, trace);
            }
            else {
                trace.info('no remote rooms to harass', { base: base.primary, targetRoom: targetRoomEntry.id });
            }
        }
        else {
            trace.warn('could not build attacker, do nothing', {
                base: base.primary,
                targetRoom: targetRoomEntry.id,
                roomDamage, availableEnergyCapacity, boosts,
            });
        }
    }
    sendReserver(kernel, target, trace) {
        const numReservers = _.filter(Game.creeps, (creep) => {
            const role = creep.memory[MEMORY_ROLE];
            return (role === WORKER_RESERVER) &&
                creep.memory[MEMORY_ASSIGN_ROOM] === target && creepIsFresh(creep);
        }).length;
        if (numReservers < 1) {
            const priority = PRIORITY_RESERVER;
            const ttl = WAR_PARTY_RUN_TTL + Game.time;
            const role = WORKER_RESERVER;
            const memory = {
                [MEMORY_ASSIGN_ROOM]: target,
            };
            const request = createSpawnRequest(priority, ttl, role, memory, null, 0);
            trace.notice('requesting reserver', { request });
            kernel.getTopics().addRequestV2(getShardSpawnTopic(), request);
        }
        else {
            trace.info('reserver already exists', { numReservers });
        }
    }
    sendHarassers(kernel, base, targetRoom, trace) {
        // get number of harassers
        const numHarassers = _.filter(Game.creeps, (creep) => {
            return creep.memory[MEMORY_ROLE] === ROLE_HARASSER &&
                creep.memory[MEMORY_BASE] === base.id &&
                creep.memory[MEMORY_HARASS_BASE] === targetRoom.id &&
                creepIsFresh(creep);
        });
        // if we have too many, don't bother
        if (numHarassers.length >= MAX_HARASSERS_PER_BASE) {
            trace.info('too many harassers', { numHarassers });
            return;
        }
        /*
        if (this.lastHarassTime + HARASS_COOLDOWN < Game.time) {
          trace.info('delaying next harasser', {
            lastHarassTime: this.lastHarassTime,
            cooldown: this.lastHarassTime + HARASS_COOLDOWN - Game.time
          });
          return;
        }
        */
        // request more
        const priorities = PRIORITY_HARASSER;
        const ttl = WAR_PARTY_RUN_TTL + Game.time;
        const role = ROLE_HARASSER;
        const memory = {
            [MEMORY_BASE]: base.id,
            [MEMORY_HARASS_BASE]: targetRoom.id,
        };
        const request = createSpawnRequest(priorities, ttl, role, memory, null, 0);
        trace.notice('requesting harasser', { request });
        kernel.getTopics().addRequestV2(getBaseSpawnTopic(base.id), request);
    }
    sendWarParty(kernel, base, targetRoom, _parts, trace) {
        const numBaseWarParties = this.warParties.filter((party) => {
            return party.baseId === base.id;
        }).length;
        if (numBaseWarParties >= MAX_WAR_PARTIES_PER_BASE) {
            trace.info('too many war parties', { numBaseWarParties });
            return;
        }
        trace.info('base parties', {
            baseId: base.id,
            numBaseWarParties,
            max: MAX_WAR_PARTIES_PER_BASE,
        });
        const partyId = `war_party_${targetRoom.id}_${base.primary}_${Game.time}`;
        const muster = getMusterVector(base);
        trace.info('creating war party', { target: targetRoom.id, partyId });
        const warPartyRunnable = this.createAndScheduleWarParty(base, partyId, targetRoom.id, muster, Game.time, trace);
        if (warPartyRunnable) {
            this.warParties.push(newWarParty(partyId, base.id, targetRoom.id, muster, Game.time));
        }
    }
    createAndScheduleWarParty(base, id, target, muster, createdAt, _trace) {
        const party = new WarPartyRunnable(id, base.id, target, muster, createdAt);
        const process = new Process(id, 'war_party', Priorities.OFFENSE, party);
        process.setSkippable(false);
        this.scheduler.registerProcess(process);
        return party;
    }
    writeMemory(_trace) {
        // Update memory
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        Memory.war = {
            targets: this.targets,
            parties: this.warParties,
        };
    }
    restoreFromMemory(kernel, trace) {
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const memory = Memory;
        trace.info('restore memory', { war: memory.war || null });
        // If there is no memory, setup defaults
        if (!memory.war) {
            memory.war = {
                targets: [],
                parties: [],
            };
        }
        this.targets = memory.war.targets || [];
        this.warParties = memory.war.parties.map((party) => {
            trace.info('restoring party', { party });
            // Validate the war party details
            if (!party.id || !party.target || !party.baseId || !party.muster || !party.createdAt) {
                trace.error('invalid party', { party });
                return null;
            }
            // Convert object to RoomPosition
            party.muster.pos = new RoomPosition(party.muster.pos.x, party.muster.pos.y, party.muster.pos.roomName);
            const base = kernel.getPlanner().getBaseById(party.baseId);
            if (!base) {
                trace.warn('not create war party, cannot find base config', { baseId: party.baseId });
                return null;
            }
            const runnable = this.createAndScheduleWarParty(base, party.id, party.target, party.muster, party.createdAt, trace);
            if (!runnable) {
                trace.error('failed to create war party', { party });
                return null;
            }
            return newWarParty(party.id, party.baseId, party.target, party.muster, party.createdAt);
        }).filter((party) => {
            return party;
        });
        trace.info('restore complete', {
            targets: this.targets,
            warParties: this.warParties.map((warParty) => warParty.id),
        });
    }
}

class AI {
    constructor(config, scheduler, trace) {
        var _a;
        trace = trace.begin('ai_constructor');
        this.config = config;
        this.scheduler = scheduler;
        // ========= IPC ==========
        this.broker = new EventBroker();
        this.topics = new Topics();
        // ========= Caches =======
        this.pathCache = new PathCache(250, getPath);
        this.costMatrixCache = new CostMatrixCache();
        // ========= Core =========
        // Memory manager process, prepare memory cache and cleans old caches
        // Things that extend PersistentMemory must be after this
        const memoryManager = new MemoryManager();
        scheduler.registerProcess(new Process('memory_manager', 'memory_manager', Priorities.CRITICAL, memoryManager));
        // Journal game state, object permanence
        this.scribe = new Scribe(trace);
        trace.info('scribe created', { numRooms: this.scribe.getRooms().length });
        scheduler.registerProcess(new Process('scribe', 'scribe', Priorities.CRITICAL, this.scribe));
        // Base manager, decides when to expand
        this.planning = new BaseManager(config, this.scheduler, trace);
        scheduler.registerProcess(new Process('bases', 'planning', Priorities.CRITICAL, this.planning));
        // Creep manager
        this.creeps = new CreepManager(scheduler);
        scheduler.registerProcess(new Process(this.creeps.id, 'creeps_manager', Priorities.CRITICAL, this.creeps));
        // TODO rename this
        // Kingdom Governor (Inter-shard communication)
        scheduler.registerProcess(new Process('kingdom_governor', 'kingdom_governor', Priorities.CRITICAL, new ShardManager()));
        // ========= High-level managers
        // Resource Manager
        this.resourceManager = new ResourceManager(this);
        scheduler.registerProcess(new Process('resource_manager', 'resource_manager', Priorities.LOGISTICS, this.resourceManager));
        // Site Janitor
        const siteJanitorId = 'site_janitor';
        const siteJanitor = new SiteJanitor();
        scheduler.registerProcess(new Process(siteJanitorId, 'site_janitor', Priorities.CRITICAL, siteJanitor));
        // Buffer manager
        const bufferManagerId = 'buffer_manager';
        const bufferManager = new BufferManager(bufferManagerId, trace);
        scheduler.registerProcess(new Process(bufferManagerId, 'buffer_manager', Priorities.DEFENCE, bufferManager));
        // Invader manager
        if (((_a = Game.shard) === null || _a === void 0 ? void 0 : _a.name) !== 'shard3') {
            const invaderManagerId = 'invader_manager';
            const invaderManager = new InvaderManager(invaderManagerId, scheduler);
            scheduler.registerProcess(new Process(invaderManagerId, 'invader_manager', Priorities.ATTACK, invaderManager));
        }
        // War manager
        const warManagerId = 'war_manager';
        const warManager = new WarManager(warManagerId, scheduler);
        scheduler.registerProcess(new Process(warManagerId, 'war_manager', Priorities.ATTACK, warManager));
        this.warManager = warManager;
        // ======= Debugging tools ========
        // Path debugger
        const hudRunnable = new HUDRunnable();
        scheduler.registerProcess(new Process('hud', 'hud', Priorities.DEBUG, hudRunnable));
        // Path debugger
        const pathDebuggerId = 'path_debugger';
        const pathDebugger = new PathDebugger(pathDebuggerId, this);
        scheduler.registerProcess(new Process(pathDebuggerId, 'path_debugger', Priorities.DEBUG, pathDebugger));
        // CostMatrix debugger
        const costMatrixDebuggerId = 'costmatrix_debugger';
        const costMatrixDebugger = new CostMatrixDebugger(costMatrixDebuggerId);
        scheduler.registerProcess(new Process(costMatrixDebuggerId, 'costmatrix_debugger', Priorities.DEBUG, costMatrixDebugger));
        // Expansion debugger
        const expandDebuggerId = 'expand_debugger';
        const expandDebugger = new BasesDebugger(expandDebuggerId, this);
        scheduler.registerProcess(new Process(expandDebuggerId, 'expand_debugger', Priorities.DEBUG, expandDebugger));
        // Min cut debugger
        const minCutDebuggerId = 'mincut_debugger';
        const minCutDebugger = new MinCutDebugger(minCutDebuggerId);
        scheduler.registerProcess(new Process(minCutDebuggerId, 'mincut_debugger', Priorities.DEBUG, minCutDebugger));
        // Muster debugger
        const musterDebuggerId = 'muster_debugger';
        const musterDebugger = new MusterDebugger(musterDebuggerId);
        scheduler.registerProcess(new Process(musterDebuggerId, 'muster_debugger', Priorities.DEBUG, musterDebugger));
        // ======= ========================
        trace.end();
    }
    tick(trace) {
        // Remove old messages from broker
        if (Game.time % 25 === 0) {
            this.broker.removeConsumed();
        }
        // Run the scheduler
        this.scheduler.tick(this, trace);
        if (Game.time % 5 === 0) {
            const end = trace.startTimer('update_stats');
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            Memory.stats = {};
            end();
        }
        trace.end();
    }
    getConfig() {
        return this.config;
    }
    getTopics() {
        return this.topics;
    }
    getBroker() {
        return this.broker;
    }
    getScheduler() {
        return this.scheduler;
    }
    getCreepsManager() {
        return this.creeps;
    }
    getPlanner() {
        return this.planning;
    }
    getScribe() {
        return this.scribe;
    }
    getPathCache() {
        return this.pathCache;
    }
    getCostMatrixCache() {
        return this.costMatrixCache;
    }
    getResourceManager() {
        return this.resourceManager;
    }
    getWarManager() {
        return this.warManager;
    }
    getFriends() {
        return this.config.friends;
    }
    getNeutrals() {
        return this.config.neutral;
    }
    getDontAttack() {
        return this.getFriends().concat(this.getNeutrals());
    }
    getAvoid() {
        return this.config.avoid;
    }
    getKOS() {
        return this.config.kos;
    }
    getNewTracer() {
        return new Tracer('tracer', new Map(), new Metrics());
    }
    debugRemotes(baseId) {
        const trace = new Tracer('remote_debugger', new Map([['pid', 'remote_debugger']]), new Metrics());
        const base = this.planning.getBase(baseId);
        const [rooms, debug] = findRemotes(this, base, trace);
        trace.notice('remote rooms', { rooms, debug });
    }
    getPathDebugger() {
        return this.scheduler.getProcess('path_debugger').runnable;
    }
    getCostMatrixDebugger() {
        return this.scheduler.getProcess('costmatrix_debugger').runnable;
    }
    getBasesDebugger() {
        return this.scheduler.getProcess('expand_debugger').runnable;
    }
    getMinCutDebugger() {
        return this.scheduler.getProcess('mincut_debugger').runnable;
    }
    getMusterDebugger() {
        return this.scheduler.getProcess('muster_debugger').runnable;
    }
}

const DEFAULT_LOG_WHEN_PID = null;
const DEFAULT_LOG_COLOR = false;
const DEFAULT_METRIC_REPORT = false;
const DEFAULT_METRIC_CONSOLE = false;
const DEFAULT_METRIC_FILTER = null;
const DEFAULT_METRIC_MIN = 0.1;
const DEFAULT_CPU_THROTTLE = 300;
const DEFAULT_SLOW_PROCESS = 10;
// On start copy memory values to debugging control flags
// eslint-disable-next-line @typescript-eslint/no-explicit-any
global.LOG_WHEN_PID = Memory.LOG_WHEN_PID || DEFAULT_LOG_WHEN_PID;
// eslint-disable-next-line @typescript-eslint/no-explicit-any
global.LOG_COLOR = Memory.LOG_COLOR || false;
// eslint-disable-next-line @typescript-eslint/no-explicit-any
global.METRIC_REPORT = Memory.METRIC_REPORT || DEFAULT_METRIC_REPORT;
// eslint-disable-next-line @typescript-eslint/no-explicit-any
global.METRIC_CONSOLE = Memory.METRIC_CONSOLE || DEFAULT_METRIC_CONSOLE;
// eslint-disable-next-line @typescript-eslint/no-explicit-any
global.METRIC_FILTER = Memory.METRIC_FILTER || DEFAULT_METRIC_FILTER;
// eslint-disable-next-line @typescript-eslint/no-explicit-any
global.METRIC_MIN = Memory.METRIC_MIN || DEFAULT_METRIC_MIN;
// eslint-disable-next-line @typescript-eslint/no-explicit-any
global.CPU_THROTTLE = Memory.CPU_THROTTLE || DEFAULT_CPU_THROTTLE;
// eslint-disable-next-line @typescript-eslint/no-explicit-any
global.SLOW_PROCESS = Memory.SLOW_PROCESS || DEFAULT_SLOW_PROCESS;
// Memory hack variables
let lastMemoryTick = 0;
let lastMemory = null;
// AI CPU usage tracking
let previousTick = 0; // Track previous tick time for display
const previousSkipped = 0;
console.log('***** STARTING AI *****');
const bootTick = Game.time;
const shardName = Game.shard.name;
let shardConfig = shards[shardName];
if (!shardConfig) {
    console.log('no shard config found for shard', shardName);
    shardConfig = shards.default;
}
console.log('selected shard config', JSON.stringify(shardConfig));
const metrics = new Metrics({ shard: Game.shard.name });
const scheduler = new Scheduler();
scheduler.setCPUThrottle(global.CPU_THROTTLE);
scheduler.setSlowProcessThreshold(global.SLOW_PROCESS);
const trace = new Tracer('boot', new Map([['shard', Game.shard.name]]), new Metrics());
const ai = new AI(shardConfig, scheduler, trace);
global.AI = ai; // So we can access it from the console
const loop = function () {
    metrics.gauge('time', Game.time);
    const trace = new Tracer('loop', new Map([['shard', Game.shard.name]]), metrics);
    // Set process id filter
    trace.setLogFilter(global.LOG_WHEN_PID);
    trace.setLogColors(global.LOG_COLOR);
    const end = trace.startTimer('memory_hack');
    // memory hack from Dissi
    if (lastMemoryTick && lastMemory && Game.time === (lastMemoryTick + 1)) {
        delete global.Memory;
        global.Memory = lastMemory;
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        RawMemory._parsed = lastMemory;
    }
    else {
        Memory;
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        lastMemory = RawMemory._parsed;
    }
    lastMemoryTick = Game.time;
    end();
    // Update memory for debugging controls control flags
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    Memory.LOG_WHEN_PID = global.LOG_WHEN_PID || DEFAULT_LOG_WHEN_PID;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    Memory.LOG_COLOR = global.LOG_COLOR || DEFAULT_LOG_COLOR;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    Memory.METRIC_REPORT = global.METRIC_REPORT || DEFAULT_METRIC_REPORT;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    Memory.METRIC_CONSOLE = global.METRIC_CONSOLE || DEFAULT_METRIC_CONSOLE;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    Memory.METRIC_FILTER = global.METRIC_FILTER || DEFAULT_METRIC_FILTER;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    Memory.METRIC_MIN = global.METRIC_MIN || DEFAULT_METRIC_MIN;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    Memory.CPU_THROTTLE = global.CPU_THROTTLE || DEFAULT_CPU_THROTTLE;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    Memory.SLOW_PROCESS = global.SLOW_PROCESS || DEFAULT_SLOW_PROCESS;
    // Enable metric collection
    if (global.METRIC_REPORT === true || global.METRIC_CONSOLE) {
        trace.setCollectMetrics(true);
    }
    // Filter metrics by name
    if (global.METRIC_FILTER) {
        trace.setMetricFilter(global.METRIC_FILTER);
    }
    // Filter metrics by minimum value
    if (global.METRIC_MIN >= 0) {
        trace.setMetricMin(global.METRIC_MIN);
    }
    scheduler.setCPUThrottle(global.CPU_THROTTLE);
    scheduler.setSlowProcessThreshold(global.SLOW_PROCESS);
    const tickSinceStart = Game.time - bootTick;
    metrics.gauge('ticks_since_start', tickSinceStart);
    console.log('======== TICK', Game.time, Game.shard.name, '==== prev cpu:', previousTick, previousSkipped, Game.cpu.bucket, tickSinceStart);
    // Tick the AI
    ai.tick(trace);
    // Output metrics to console that match the filter
    if (global.METRIC_CONSOLE) {
        trace.outputMetrics();
    }
    // Output metric aggregations to console
    if (global.METRIC_REPORT === true) {
        addTraceMetrics(trace);
        setActive();
        reportMetrics();
    }
    else {
        setInactive();
    }
    // TODO report event broker topics
    ai.getTopics().reportMetrics(metrics);
    ai.getCreepsManager().reportMetrics(metrics);
    ai.getPlanner().reportMetrics(ai, metrics, trace);
    // Global control level related metrics
    metrics.gauge('global_control_level', Game.gcl.level);
    metrics.gauge('global_control_progress', Game.gcl.progress);
    metrics.gauge('global_control_progress_total', Game.gcl.progressTotal);
    // Get CPU spent on AI
    previousTick = Game.cpu.getUsed();
    metrics.counter('cpu_used_total', previousTick);
    // Write metrics to Memory to be fetched by screeps-scraper
    metrics.write();
};

exports.loop = loop;
